// *F* Developed 2011-2014 rejecht <rejecht at outlook dot com>

class AwardPRI extends xPlayerReplicationInfo
	HideDropDown
	;

var(Replicated) int ReplicatedKills;

replication
{
	reliable if (bNetDirty && (Role == ROLE_Authority))
		ReplicatedKills
		;
}


var protected const class<Award>
	AwardClass_Event,
	AwardClass_Death,
	AwardClass_WeaponDeath,
	AwardClass_Kill,
	AwardClass_WeaponKill,
	AwardClass_Suicide,
	AwardClass_WeaponSuicide;


struct SStatistics_Damage
{
	var class<DamageType> Class_DamageType;

	var int
		Count_Deaths,
		Count_Kills,
		Count_Suicides,
		AwardDeath_Index,
		AwardDeath_Trigger,
		AwardDeath_Multiplier,
		AwardKill_Index,
		AwardKill_Trigger,
		AwardKill_Multiplier,
		AwardSuicide_Index,
		AwardSuicide_Trigger,
		AwardSuicide_Multiplier;

	var bool
		bAwardDeath,
		bAwardKill,
		bAwardSuicide;
};


struct SStatistics_Weapon
{
	var class<Weapon> Class_Weapon;

	var int
		Count_Deaths,
		Count_Kills,
		Count_Suicides,
		AwardDeath_Index,
		AwardDeath_Trigger,
		AwardDeath_Multiplier,
		AwardKill_Index,
		AwardKill_Trigger,
		AwardKill_Multiplier,
		AwardSuicide_Index,
		AwardSuicide_Trigger,
		AwardSuicide_Multiplier;

	var bool
		bAwardDeath,
		bAwardKill,
		bAwardSuicide;
};


struct SStatistics_WeaponDamage
{
	var class<WeaponDamageType> Class_WeaponDamageType; // Primary Key

	var class<Weapon> Class_Weapon; // Foreign Key

	var float Accuracy;

	var int
		Fired,
		FireModeIndex,
		Projectiles,
		WeaponDamageDone_Enemies,
		WeaponDamageDone_Team,
		WeaponHitsDone_Enemies,
		WeaponHitsDone_Team,
		WeaponDamageTaken_Enemies,
		WeaponDamageTaken_Team,
		WeaponHitsTaken_Enemies,
		WeaponHitsTaken_Team,
		Count_Deaths,
		Count_Kills,
		Count_Suicides,
		AwardDeath_Index,
		AwardDeath_Trigger,
		AwardDeath_Multiplier,
		AwardKill_Index,
		AwardKill_Trigger,
		AwardKill_Multiplier,
		AwardSuicide_Index,
		AwardSuicide_Trigger,
		AwardSuicide_Multiplier;

	var bool
		bAwardDeath, // Bound to a death award.
		bAwardKill, // Bound to a kill award.
		bAwardSuicide; // Bound to a suicide award.
};


struct SStatistics_Event
{
	var name Name_Event;

	var int
		Count_Triggered,
		Award_Index,
		Award_Trigger,
		Award_Multiplier;
};


struct SStatistics_AdrenalineCombo
{
	var class<Combo> Class_Combo;
	var int Count_Activations;
};

// #

var array<SStatistics_Damage> Statistics_Damage;
var array<SStatistics_Weapon> Statistics_Weapon;
var array<SStatistics_WeaponDamage> Statistics_WeaponDamage;
var array<SStatistics_Event> Statistics_Event;
var array<SStatistics_AdrenalineCombo> Statistics_AdrenalineCombo;

// #Records

struct SMetrics
{
	var float Accuracy; // Client-Side Calculation

	var int
		TotalWeaponDamageDone_Enemies,
		TotalWeaponDamageDone_Self,
		TotalWeaponDamageDone_Team,
		TotalWeaponDamageTaken_Enemies,
		TotalWeaponDamageTaken_Team,
		TotalWeaponHitsTaken_Enemies,
		TotalWeaponHitsTaken_Self,
		TotalWeaponHitsTaken_Team,
		TotalHeadShotHits,
		TotalHeadShotKills,
		LongestHeadShotDistance,
		Kills,
		Deaths,
		Suicides,
		Teamkills,
		TeamDeaths,
		AdrenalineGained,
		AdrenalineLost;

	var byte Played;
};

struct SAwardRecord
{
	var class<Award> Class_Award;
	var int Award_Index, Award_Multiplier;
	var Color CaptionColor;
};

// #

struct SWeaponDamageRecord
{
	var class<WeaponDamageType> Class_WeaponDamageType;
	var string Title;
	var int Kills, Deaths, Suicides, DamageDone, Fired;
	var byte Accuracy;
	var bool bHasData;
};

struct SWeaponRecord
{
	var class<Weapon> Class_Weapon;
	var array<SWeaponDamageRecord> WeaponDamageRecord;
	var int Kills, Deaths, Suicides, DamageDone, Fired;
	var byte Accuracy;
	var bool bHasData;
};

// #

var SMetrics Metrics; // Totals

var array<SAwardRecord> AwardRecord; // Compiled Array

var array<SWeaponRecord> ClientWeaponRecord;

// #AwardTrigger

// - <Trigger> If zero, then the award triggers each time.
// - <Trigger> If greater than zero, then the award triggers once the trigger
// is reached.
// - <Trigger> Any value less than zero has no function, but could be used
// to trigger every N time.
// - <Index> The index associated with the award.
// - Constants may only be modified through default properties.

struct SAwardTrigger_WeaponDeath
{
	var const class<Weapon> Class_Weapon;
	var int Trigger, Index;
};

struct SAwardTrigger_WeaponKill
{
	var const class<Weapon> Class_Weapon;
	var int Trigger, Index;
};

struct SAwardTrigger_WeaponSuicide
{
	var const class<Weapon> Class_Weapon;
	var int Trigger, Index;
};

struct SAwardTrigger_WeaponDamageDeath
{
	var const class<WeaponDamageType> Class_WeaponDamageType;
	var int Trigger, Index;
};

struct SAwardTrigger_WeaponDamageKill
{
	var const class<WeaponDamageType> Class_WeaponDamageType;
	var int Trigger, Index;
};

struct SAwardTrigger_WeaponDamageSuicide
{
	var const class<WeaponDamageType> Class_WeaponDamageType;
	var int Trigger, Index;
};

struct SAwardTrigger_DamageDeath
{
	var const class<DamageType> Class_DamageType;
	var int Trigger, Index;
};

struct SAwardTrigger_DamageKill
{
	var const class<DamageType> Class_DamageType;
	var int Trigger, Index;
};

struct SAwardTrigger_DamageSuicide
{
	var const class<DamageType> Class_DamageType;
	var int Trigger, Index;
};

struct SAwardTrigger_Event
{
	var const name Name_Event;
	var int Trigger, Index;
};

// #Declarations

var protected const array<SAwardTrigger_WeaponDeath> AwardTrigger_WeaponDeath;
var protected const array<SAwardTrigger_DamageDeath> AwardTrigger_DamageDeath;

var protected const array<SAwardTrigger_WeaponKill> AwardTrigger_WeaponKill;
var protected const array<SAwardTrigger_DamageKill> AwardTrigger_DamageKill;
var protected const array<SAwardTrigger_WeaponDamageKill> AwardTrigger_WeaponDamageKill;

var protected const array<SAwardTrigger_WeaponSuicide> AwardTrigger_WeaponSuicide;
var protected const array<SAwardTrigger_DamageSuicide> AwardTrigger_DamageSuicide;

var protected const array<SAwardTrigger_WeaponDamageDeath> AwardTrigger_WeaponDamageDeath;
var protected const array<SAwardTrigger_WeaponDamageSuicide> AwardTrigger_WeaponDamageSuicide;

var protected const array<SAwardTrigger_Event> AwardTrigger_Event;

// #Lifetime

simulated event PostBeginPlay ()
{
	Super.PostBeginPlay ();


	if (Role == ROLE_Authority)
	{
		Statistics_PostBeginPlay ();
	}
}

// #

protected function Statistics_PostBeginPlay ()
{
}

// #

function int Statistics_RegisterWeaponDamageType (Class<WeaponDamageType> WeaponDamageTypeClass, int FireModeIndex, optional int Projectiles)
{
	local int Index;

	Projectiles = Clamp (Projectiles, 1, 999);

	for (Index = 0; Index < Statistics_WeaponDamage.Length; ++Index)
	{
		if (Statistics_WeaponDamage[Index].Class_WeaponDamageType != WeaponDamageTypeClass)
		{
			continue;
		}
		else
		{
			return Index;
		}
	}


	Index = Statistics_WeaponDamage.Length;
	Statistics_WeaponDamage.Length = Index + 1;

	Statistics_WeaponDamage[Index].Class_WeaponDamageType = WeaponDamageTypeClass;
	Statistics_WeaponDamage[Index].Class_Weapon = WeaponDamageTypeClass.Default.WeaponClass;
	Statistics_WeaponDamage[Index].FireModeIndex = FireModeIndex;
	Statistics_WeaponDamage[Index].Projectiles = Projectiles;

	Award_BindByIndex_WeaponDamageDeath (Index);
	Award_BindByIndex_WeaponDamageKill (Index);
	Award_BindByIndex_WeaponDamageSuicide (Index);

	return Index;
}

function Statistics_LogWeaponFiredByIndex (int Index, int Load)
{
	if (Statistics_ShouldLog ())
	{
		Statistics_WeaponDamage[Index].Fired += Load;
	}
}

// #

simulated function Statistics_UpdateMetrics_Award
(
	class<Award> Class_Award,
	int Award_Index,
	int Award_Multiplier
)
{
	local int Index;

	for (Index = 0; Index < AwardRecord.Length; ++Index)
	{
		if
		(
			(Class_Award == AwardRecord[Index].Class_Award)
			&&
			(Award_Index == AwardRecord[Index].Award_Index)
		)
		{
			AwardRecord[Index].Award_Multiplier = Award_Multiplier;

			return;
		}
	}

	// New Award

	Index = AwardRecord.Length;
	AwardRecord.Length = Index + 1;

	AwardRecord[Index].Class_Award = Class_Award;
	AwardRecord[Index].Award_Index = Award_Index;
	AwardRecord[Index].Award_Multiplier = Award_Multiplier;
	AwardRecord[Index].CaptionColor = Class_Award.Static.GetCaptionColor (Award_Index);
}


simulated function Statistics_UpdateMetrics_AdrenalineCombo
(
	class<Combo> Class_Combo,
	int Count_Activations
)
{
	local int Index;
	local int Count;

	Count = Statistics_AdrenalineCombo.Length;

	for (Index = 0; Index < Count; ++Index)
	{
		if (Statistics_AdrenalineCombo[Index].Class_Combo != Class_Combo)
		{
			continue;
		}
		else
		{
			Statistics_AdrenalineCombo[Index].Count_Activations = Count_Activations;

			return;
		}
	}


	Index = Statistics_AdrenalineCombo.Length;
	Statistics_AdrenalineCombo.Length = Index + 1;

	Statistics_AdrenalineCombo[Index].Class_Combo = Class_Combo;
	Statistics_AdrenalineCombo[Index].Count_Activations = Count_Activations;
}

// #

protected simulated function bool Statistics_IsWeaponRelevant (out SWeaponRecord WR)
{
	return
	(
		(WR.DamageDone > 0)
		||
		(WR.Kills > 0)
		||
		(WR.Accuracy > 0)
		||
		(WR.Fired > 0)
		||
		(WR.Deaths > 0)
		||
		(WR.Suicides > 0)
	);
}

protected simulated function bool Statistics_IsWeaponDamageRelevant (out AwardPRI.SWeaponDamageRecord WDR)
{
	return
	(
		(WDR.DamageDone > 0)
		||
		(WDR.Kills > 0)
		||
		(WDR.Accuracy > 0)
		||
		(WDR.Fired > 0)
		||
		(WDR.Deaths > 0)
		||
		(WDR.Suicides > 0)
	);
}

// In case a future weapon damage type has no WeaponClass set.

protected simulated function Class<Weapon> Statistics_GetAssociatedWeaponClass (Class<WeaponDamageType> WeaponDamageTypeClass)
{
	return WeaponDamageTypeClass.Default.WeaponClass;
}

// In case a new lookup method is used.

protected simulated function string Statistics_GetText_WeaponDamageTypeName (Class<WeaponDamageType> WeaponDamageTypeClass)
{
	return Class'WeaponDamageTypeMap'.Static.GetText_WeaponDamageType (WeaponDamageTypeClass);
}

// - RePlayer calls when it receives new data.
// - Receives a WeaponDamageType.
// - Groups WeaponDamageType under associated WeaponClass.

simulated function Statistics_UpdateMetrics_Weapon
(
	class<WeaponDamageType> WeaponDamageTypeClass,
	int DamageDone,
	byte Accuracy,
	int Fired,
	int Kills,
	int Deaths,
	int Suicides
)
{
	local class<Weapon> AssociatedWeaponClass;
	local int IndexWeapon, IndexWeaponDamageType, Count;

	AssociatedWeaponClass = Statistics_GetAssociatedWeaponClass (WeaponDamageTypeClass);

	// Find the weapon to update this weapon damage type for..

	Count = ClientWeaponRecord.Length;

	for (IndexWeapon = 0; IndexWeapon < Count; ++IndexWeapon)
	{
		if (AssociatedWeaponClass != ClientWeaponRecord[IndexWeapon].Class_Weapon)
		{
			continue;
		}
		else
		{
			IndexWeaponDamageType = Statistics_GetIndex_WeaponDamageRecord (IndexWeapon, WeaponDamageTypeClass);


			ClientWeaponRecord[IndexWeapon].WeaponDamageRecord[IndexWeaponDamageType].Class_WeaponDamageType = WeaponDamageTypeClass;
			ClientWeaponRecord[IndexWeapon].WeaponDamageRecord[IndexWeaponDamageType].Accuracy = Accuracy;
			ClientWeaponRecord[IndexWeapon].WeaponDamageRecord[IndexWeaponDamageType].DamageDone = DamageDone;
			ClientWeaponRecord[IndexWeapon].WeaponDamageRecord[IndexWeaponDamageType].Fired = Fired;
			ClientWeaponRecord[IndexWeapon].WeaponDamageRecord[IndexWeaponDamageType].Kills = Kills;
			ClientWeaponRecord[IndexWeapon].WeaponDamageRecord[IndexWeaponDamageType].Deaths = Deaths;
			ClientWeaponRecord[IndexWeapon].WeaponDamageRecord[IndexWeaponDamageType].Suicides = Suicides;

			ClientWeaponRecord[IndexWeapon].WeaponDamageRecord[IndexWeaponDamageType].bHasData = Statistics_IsWeaponDamageRelevant (ClientWeaponRecord[IndexWeapon].WeaponDamageRecord[IndexWeaponDamageType]);

			ClientWeaponRecord[IndexWeapon].WeaponDamageRecord[IndexWeaponDamageType].Title = Statistics_GetText_WeaponDamageTypeName (WeaponDamageTypeClass);


			Statistics_UpdateMetrics_WeaponTotals ();

			ClientWeaponRecord[IndexWeapon].bHasData = Statistics_IsWeaponRelevant (ClientWeaponRecord[IndexWeapon]);

			return;
		}
	}

	// - The AssociatedWeaponClass was not found thus implying the weapon
	// damage type has not been added either.
	// - Add the weapon first.

	IndexWeapon = ClientWeaponRecord.Length;
	ClientWeaponRecord.Length = IndexWeapon + 1;

	ClientWeaponRecord[IndexWeapon].Class_Weapon = AssociatedWeaponClass;
	ClientWeaponRecord[IndexWeapon].Accuracy = Accuracy;
	ClientWeaponRecord[IndexWeapon].DamageDone = DamageDone;
	ClientWeaponRecord[IndexWeapon].Fired = Fired;
	ClientWeaponRecord[IndexWeapon].Kills = Kills;
	ClientWeaponRecord[IndexWeapon].Deaths = Deaths;
	ClientWeaponRecord[IndexWeapon].Suicides = Suicides;

	// Add the weapon damage type.

	IndexWeaponDamageType = ClientWeaponRecord[IndexWeapon].WeaponDamageRecord.Length;
	ClientWeaponRecord[IndexWeapon].WeaponDamageRecord.Length = IndexWeaponDamageType + 1;

	ClientWeaponRecord[IndexWeapon].WeaponDamageRecord[IndexWeaponDamageType].Class_WeaponDamageType = WeaponDamageTypeClass;
	ClientWeaponRecord[IndexWeapon].WeaponDamageRecord[IndexWeaponDamageType].Accuracy = Accuracy;
	ClientWeaponRecord[IndexWeapon].WeaponDamageRecord[IndexWeaponDamageType].DamageDone = DamageDone;
	ClientWeaponRecord[IndexWeapon].WeaponDamageRecord[IndexWeaponDamageType].Fired = Fired;
	ClientWeaponRecord[IndexWeapon].WeaponDamageRecord[IndexWeaponDamageType].Kills = Kills;
	ClientWeaponRecord[IndexWeapon].WeaponDamageRecord[IndexWeaponDamageType].Deaths = Deaths;
	ClientWeaponRecord[IndexWeapon].WeaponDamageRecord[IndexWeaponDamageType].Suicides = Suicides;
	ClientWeaponRecord[IndexWeapon].WeaponDamageRecord[IndexWeaponDamageType].Title = Statistics_GetText_WeaponDamageTypeName (WeaponDamageTypeClass);

	// Find out whether it has received useful data.

	ClientWeaponRecord[IndexWeapon].WeaponDamageRecord[IndexWeaponDamageType].bHasData = Statistics_IsWeaponDamageRelevant (ClientWeaponRecord[IndexWeapon].WeaponDamageRecord[IndexWeaponDamageType]);

	//

	Statistics_UpdateMetrics_WeaponTotals ();

	//

	ClientWeaponRecord[IndexWeapon].bHasData = Statistics_IsWeaponRelevant (ClientWeaponRecord[IndexWeapon]);
}

// Runs through all weapons and accumulates data from its associated weapon damage types.

protected simulated function Statistics_UpdateMetrics_WeaponTotals ()
{
	local int IndexWeapon;
	local int IndexWeaponDamageType;
	local int Count;
	local int TotalWeaponAccuracy;

	for
	(
		IndexWeapon = 0;
		IndexWeapon < ClientWeaponRecord.Length;
		++IndexWeapon
	)
	{
		Count = 0;
		TotalWeaponAccuracy = 0;

		ClientWeaponRecord[IndexWeapon].Kills = 0;
		ClientWeaponRecord[IndexWeapon].Deaths = 0;
		ClientWeaponRecord[IndexWeapon].Suicides = 0;
		ClientWeaponRecord[IndexWeapon].DamageDone = 0;
		ClientWeaponRecord[IndexWeapon].Fired = 0;
		ClientWeaponRecord[IndexWeapon].Accuracy = 0;

		for
		(
			IndexWeaponDamageType = 0;
			IndexWeaponDamageType < ClientWeaponRecord[IndexWeapon].WeaponDamageRecord.Length;
			++IndexWeaponDamageType
		)
		{
			ClientWeaponRecord[IndexWeapon].Kills += ClientWeaponRecord[IndexWeapon].WeaponDamageRecord[IndexWeaponDamageType].Kills;
			ClientWeaponRecord[IndexWeapon].Deaths += ClientWeaponRecord[IndexWeapon].WeaponDamageRecord[IndexWeaponDamageType].Deaths;
			ClientWeaponRecord[IndexWeapon].Suicides += ClientWeaponRecord[IndexWeapon].WeaponDamageRecord[IndexWeaponDamageType].Suicides;
			ClientWeaponRecord[IndexWeapon].Fired += ClientWeaponRecord[IndexWeapon].WeaponDamageRecord[IndexWeaponDamageType].Fired;
			ClientWeaponRecord[IndexWeapon].DamageDone += ClientWeaponRecord[IndexWeapon].WeaponDamageRecord[IndexWeaponDamageType].DamageDone;


			if (ClientWeaponRecord[IndexWeapon].WeaponDamageRecord[IndexWeaponDamageType].Accuracy > 0)
			{
				++Count;

				TotalWeaponAccuracy += ClientWeaponRecord[IndexWeapon].WeaponDamageRecord[IndexWeaponDamageType].Accuracy;
			}
		}


		if (Count > 1)
		{
			ClientWeaponRecord[IndexWeapon].Accuracy = TotalWeaponAccuracy / Count;
		}
		else
		{
			ClientWeaponRecord[IndexWeapon].Accuracy = TotalWeaponAccuracy;
		}
	}
}

// #

protected function int Statistics_GetIndex_Weapon
(
	Class<Weapon> WeaponClass
)
{
	local int Index;

	for (Index = 0; Index < Statistics_Weapon.Length; ++Index)
	{
		if (Statistics_Weapon[Index].Class_Weapon != WeaponClass)
		{
			continue;
		}
		else
		{
			return Index;
		}
	}


	Index = Statistics_Weapon.Length;
	Statistics_Weapon.Length = Index + 1;

	Statistics_Weapon[Index].Class_Weapon = WeaponClass;

	Award_BindByIndex_WeaponDeath (Index);
	Award_BindByIndex_WeaponKill (Index);
	Award_BindByIndex_WeaponSuicide (Index);


	return Index;
}

// Only used client-side when collecting stats.

protected simulated function int Statistics_GetIndex_WeaponDamageRecord
(
	int IndexWeapon,
	Class<WeaponDamageType> WeaponDamageTypeClass
)
{
	local int IndexWeaponDamageType;

	for
	(
		IndexWeaponDamageType = 0;
		IndexWeaponDamageType < ClientWeaponRecord[IndexWeapon].WeaponDamageRecord.Length;
		++IndexWeaponDamageType
	)
	{
		if (ClientWeaponRecord[IndexWeapon].WeaponDamageRecord[IndexWeaponDamageType].Class_WeaponDamageType != WeaponDamageTypeClass)
		{
			continue;
		}
		else
		{
			return IndexWeaponDamageType;
		}
	}

	//

	IndexWeaponDamageType = ClientWeaponRecord[IndexWeapon].WeaponDamageRecord.Length;
	ClientWeaponRecord[IndexWeapon].WeaponDamageRecord.Length = IndexWeaponDamageType + 1;

	ClientWeaponRecord[IndexWeapon].Accuracy += ClientWeaponRecord[IndexWeapon].WeaponDamageRecord[IndexWeaponDamageType].Accuracy;
	ClientWeaponRecord[IndexWeapon].DamageDone += ClientWeaponRecord[IndexWeapon].WeaponDamageRecord[IndexWeaponDamageType].DamageDone;
	ClientWeaponRecord[IndexWeapon].Fired += ClientWeaponRecord[IndexWeapon].WeaponDamageRecord[IndexWeaponDamageType].Fired;
	ClientWeaponRecord[IndexWeapon].Kills += ClientWeaponRecord[IndexWeapon].WeaponDamageRecord[IndexWeaponDamageType].Kills;
	ClientWeaponRecord[IndexWeapon].Deaths += ClientWeaponRecord[IndexWeapon].WeaponDamageRecord[IndexWeaponDamageType].Deaths;
	ClientWeaponRecord[IndexWeapon].Suicides += ClientWeaponRecord[IndexWeapon].WeaponDamageRecord[IndexWeaponDamageType].Suicides;

	return IndexWeaponDamageType;
}

protected function int Statistics_GetIndex_Damage
(
	class<DamageType> DamageTypeClass
)
{
	local int Index;

	for (Index = 0; Index < Statistics_Damage.Length; ++Index)
	{
		if (Statistics_Damage[Index].Class_DamageType != DamageTypeClass)
		{
			continue;
		}
		else
		{
			return Index;
		}
	}


	Index = Statistics_Damage.Length;
	Statistics_Damage.Length = Index + 1;

	Statistics_Damage[Index].Class_DamageType = DamageTypeClass;

	Award_BindByIndex_DamageDeath (Index);
	Award_BindByIndex_DamageKill (Index);
	Award_BindByIndex_DamageSuicide (Index);


	return Index;
}

// #

function int Statistics_GetAmount_WeaponDamageDone ()
{
	return Metrics.TotalWeaponDamageDone_Enemies;
}

// #

protected function Award_BindByIndex_DamageDeath (int IndexDamage)
{
	local int IndexAward;

	for (IndexAward = 0; IndexAward < AwardTrigger_DamageDeath.Length; ++IndexAward)
	{
		if (Statistics_Damage[IndexDamage].Class_DamageType != AwardTrigger_DamageDeath[IndexAward].Class_DamageType)
		{
			continue;
		}
		else
		{
			Statistics_Damage[IndexDamage].AwardDeath_Index = AwardTrigger_DamageDeath[IndexAward].Index;
			Statistics_Damage[IndexDamage].AwardDeath_Trigger = AwardTrigger_DamageDeath[IndexAward].Trigger;
			Statistics_Damage[IndexDamage].bAwardDeath = True;

			return;
		}
	}
}

protected function Award_BindByIndex_DamageKill (int IndexDamage)
{
	local int IndexAward;

	for (IndexAward = 0; IndexAward < AwardTrigger_DamageKill.Length; ++IndexAward)
	{
		if (Statistics_Damage[IndexDamage].Class_DamageType != AwardTrigger_DamageKill[IndexAward].Class_DamageType)
		{
			continue;
		}
		else
		{
			Statistics_Damage[IndexDamage].AwardKill_Index = AwardTrigger_DamageKill[IndexAward].Index;
			Statistics_Damage[IndexDamage].AwardKill_Trigger = AwardTrigger_DamageKill[IndexAward].Trigger;
			Statistics_Damage[IndexDamage].bAwardKill = True;

			return;
		}
	}
}

protected function Award_BindByIndex_DamageSuicide (int IndexDamage)
{
	local int IndexAward;

	for (IndexAward = 0; IndexAward < AwardTrigger_DamageSuicide.Length; ++IndexAward)
	{
		if (Statistics_Damage[IndexDamage].Class_DamageType != AwardTrigger_DamageSuicide[IndexAward].Class_DamageType)
		{
			continue;
		}
		else
		{
			Statistics_Damage[IndexDamage].AwardSuicide_Index = AwardTrigger_DamageSuicide[IndexAward].Index;
			Statistics_Damage[IndexDamage].AwardSuicide_Trigger = AwardTrigger_DamageSuicide[IndexAward].Trigger;
			Statistics_Damage[IndexDamage].bAwardSuicide = True;

			return;
		}
	}
}

// #Queries

function float Statistics_GetWeaponAccuracy ()
{
	local float Average, Hits;
	local int Count, Fired, TotalFired, Index;

	for (Index = 0; Index < Statistics_WeaponDamage.Length; ++Index)
	{
		Fired = Statistics_WeaponDamage[Index].Fired;

		TotalFired += Fired;

		Hits = Statistics_WeaponDamage[Index].WeaponHitsDone_Enemies;

		if (Fired > 1)
		{
			if (Hits > 0)
			{
				Statistics_WeaponDamage[Index].Accuracy = FMin ((Hits / Statistics_WeaponDamage[Index].Fired) * 100.0f, 100.0f); // Recalculate

				if (Statistics_WeaponDamage[Index].Accuracy > 0.0f)
				{
					Average += Statistics_WeaponDamage[Index].Accuracy;

					++Count;
				}
			}
			else
			{
				Statistics_WeaponDamage[Index].Accuracy = 0.0f;
			}
		}
		else if (Fired > 1)
		{
			Statistics_WeaponDamage[Index].Accuracy = FMin ((Hits / Statistics_WeaponDamage[Index].Fired) * 100.0f, 100.0f); // Recalculate
		}
	}

	// Player must have fired at least ten shots before being considered.

	if
	(
		(TotalFired > 9)
		&&
		(Count > 0)
	)
	{
		Average /= Count;
	}
	else
	{
		Average = 0.0f;
	}

	return Average;
}

// #AdrenalineStatistics

// Preconditions:
// - (Amount > 0)

function Statistics_AdrenalineGained (float Amount)
{
	Metrics.AdrenalineGained += Amount;
}

// Preconditions:
// - (Amount < 0)

function Statistics_AdrenalineLost (float Amount)
{
	Metrics.AdrenalineLost += Abs (Amount);
}

function Statistics_DoCombo (class<Combo> ComboClass)
{
	++Statistics_AdrenalineCombo[Statistics_GetIndex_AdrenalineCombo (ComboClass)].Count_Activations;


	if (Controller(Owner) != None)
	{
		ReTeamGame(Level.Game).PlayerUsedCombo (Controller(Owner), ComboClass);
    }
}

// Preconditions:
// - (Class_Combo != None)

protected function int Statistics_GetIndex_AdrenalineCombo
(
	Class<Combo> ComboClass
)
{
	local int Index;

	for (Index = 0; Index < Statistics_AdrenalineCombo.Length; ++Index)
	{
		if (Statistics_AdrenalineCombo[Index].Class_Combo != ComboClass)
		{
			continue;
		}
		else
		{
			return Index;
		}
	}


	Index = Statistics_AdrenalineCombo.Length;
	Statistics_AdrenalineCombo.Length = Index + 1;

	Statistics_AdrenalineCombo[Index].Class_Combo = ComboClass;


	return Index;
}

function Statistics_SelfInflictedDamage
(
	class<DamageType> Class_DamageType,
	int Amount
)
{
	if (Statistics_ShouldLog ())
	{
		if (Class<WeaponDamageType>(Class_DamageType) != None)
		{
			Metrics.TotalWeaponDamageDone_Self += Amount;
		}
	}
}

// #TeamDamage

function Statistics_LogTeamDamageDone
(
	class<DamageType> Class_DamageType,
	int Amount
)
{
	local int Index;

	if (Statistics_ShouldLog ())
	{
		if (Class<WeaponDamageType>(Class_DamageType) != None)
		{
			Metrics.TotalWeaponDamageDone_Team += Amount;

			Index = Statistics_GetIndex_WeaponDamage (Class<WeaponDamageType> (Class_DamageType));

			++Statistics_WeaponDamage[Index].WeaponHitsDone_Team;
			Statistics_WeaponDamage[Index].WeaponDamageDone_Team += Amount;
		}
		else if (class<VehicleDamageType>(Class_DamageType) != None)
		{
		}
		else
		{
		}
	}
}

function Statistics_LogTeamDamageTaken
(
	class<DamageType> Class_DamageType,
	int Amount
)
{
	local int Index;

	if (Statistics_ShouldLog ())
	{
		if (Class<WeaponDamageType>(Class_DamageType) != None)
		{
			++Metrics.TotalWeaponHitsTaken_Team;

			Metrics.TotalWeaponDamageTaken_Team += Amount;

			Index = Statistics_GetIndex_WeaponDamage (class<WeaponDamageType>(Class_DamageType));

			++Statistics_WeaponDamage[Index].WeaponHitsTaken_Team;
			Statistics_WeaponDamage[Index].WeaponDamageTaken_Team += Amount;
		}
		else if (class<VehicleDamageType>(Class_DamageType) != None)
		{
		}
		else
		{
		}
	}
}

// - Always log while live.
// - Allow logging in a short window after ending a live round.

function bool Statistics_ShouldLog ()
{
	local ReGRI ReGRI;

	ReGRI = ReGRI(Level.GRI);

	return
	(
		ReGRI.IsLive ()
		||
		(
			ReGRI.IsRestarting ()
			&&
			((Level.TimeSeconds - ReGRI.GetTimeSeconds_RoundRestarting ()) < 0.33f)
		)
	);
}

// #EnemyDamage

// Example:
// - DamTypeSniperHeadShot gets logged.
// -- DamTypeSniperShot already logged.
// --- Undoes DamTypeSniperShot.

function Statistics_LogEnemyDamageDone
(
	Class<DamageType> DamageTypeClass,
	int Damage,
	Pawn InstigatorPawn,
	Pawn EnemyPawn
)
{
	local Class<WeaponDamageType> WeaponDamageTypeClass;
	local int Index;

	if (Statistics_ShouldLog ())
	{
		WeaponDamageTypeClass = Class<WeaponDamageType>(DamageTypeClass);

		if (WeaponDamageTypeClass != None)
		{
			Metrics.TotalWeaponDamageDone_Enemies += Damage;

			//

			Index = Statistics_GetIndex_WeaponDamage (WeaponDamageTypeClass);

			//

			++Statistics_WeaponDamage[Index].WeaponHitsDone_Enemies;

			Statistics_WeaponDamage[Index].WeaponDamageDone_Enemies += Damage;

			//

			if
			(
				(Statistics_WeaponDamage[Index].FireModeIndex >= 2)
			)
			{
				Statistics_WeaponDamage[Index].Fired = Statistics_WeaponDamage[Index].WeaponHitsDone_Enemies;
			}

			//

			if (Statistics_IsHeadShot (WeaponDamageTypeClass))
			{
				Statistics_EnemyHeadShotWorkaround (WeaponDamageTypeClass, InstigatorPawn, EnemyPawn);
			}
		}
		else if (class<VehicleDamageType> (DamageTypeClass) != None)
		{
		}
		else
		{
		}
	}
}

// #

protected function Statistics_EnemyHeadShotWorkaround
(
	Class<WeaponDamageType> WeaponDamageTypeClass,
	Pawn InstigatorPawn,
	Pawn EnemyPawn
)
{
	local float Distance;

	++Metrics.TotalHeadShotHits;

	if
	(
		(InstigatorPawn != None)
		&&
		(EnemyPawn != None)
	)
	{
		Distance = VSize (InstigatorPawn.Location - EnemyPawn.Location) * 0.01875f;

		if (Distance > Metrics.LongestHeadShotDistance)
		{
			Metrics.LongestHeadShotDistance = Distance;
		}
	}
}


// #

function Statistics_RecordEnemyDamageTaken
(
	class<DamageType> Class_DamageType,
	int Amount
)
{
	local int Index;

	if (Statistics_ShouldLog ())
	{
		if (Class<WeaponDamageType>(Class_DamageType) != None)
		{
			Index = Statistics_GetIndex_WeaponDamage (class<WeaponDamageType>(Class_DamageType));

			++Metrics.TotalWeaponHitsTaken_Enemies;
			Metrics.TotalWeaponDamageTaken_Enemies += Amount;

			++Statistics_WeaponDamage[Index].WeaponHitsTaken_Enemies;
			Statistics_WeaponDamage[Index].WeaponDamageTaken_Enemies += Amount;
		}
		else if (Class<VehicleDamageType> (Class_DamageType) != None)
		{
		}
		else
		{
		}
	}
}

// - To return the index associated with the WeaponDamageType so that the
// caller can modify statistics.

function int Statistics_GetIndex_WeaponDamage
(
	Class<WeaponDamageType> Class_WeaponDamageType
)
{
	local int Index;

	for (Index = 0; Index < Statistics_WeaponDamage.Length; ++Index)
	{
		if (Statistics_WeaponDamage[Index].Class_WeaponDamageType != Class_WeaponDamageType)
		{
			continue; // Most Frequent
		}
		else
		{
			return Index; // *Changed* 2013-12-21 23:17:15
		}
	}


	Index = Statistics_WeaponDamage.Length;
	Statistics_WeaponDamage.Length = Index + 1;

	Statistics_WeaponDamage[Index].Class_WeaponDamageType = Class_WeaponDamageType;
	Statistics_WeaponDamage[Index].Class_Weapon = Class_WeaponDamageType.default.WeaponClass;

	Award_BindByIndex_WeaponDamageDeath (Index);
	Award_BindByIndex_WeaponDamageKill (Index);
	Award_BindByIndex_WeaponDamageSuicide (Index);


	return Index;
}

// #

static function bool Statistics_IsHeadShot (Class<DamageType> DamageTypeClass)
{
	return
	(
		(Class<DamTypeSniperHeadShot>(DamageTypeClass) != None)
		||
		(Class<DamTypeClassicHeadShot>(DamageTypeClass) != None)
		||
		(Class<DamTypeShockBeamHeadShot>(DamageTypeClass) != None)
	);
}

// Compiler
// - When an award is added, it is also put in this stream for
// easier replication.
// - If the award exists, its multiplier gets updated.
// - Missing the last parameter, which could be anything, such as
// progress, distance, other stuff related to the award.

protected function Statistics_RecordAward
(
	class<Award> Class_Award,
	int IndexAward,
	int Multiplier
)
{
	local int Index;

	for (Index = 0; Index < AwardRecord.Length; ++Index)
	{
		if
		(
			(Class_Award == AwardRecord[Index].Class_Award)
			&&
			(AwardRecord[Index].Award_Index == IndexAward)
		)
		{
			AwardRecord[Index].Award_Multiplier = Multiplier;

			ReTeamGame(Level.Game).PlayerAwarded (Controller(Owner), Self, Class_Award, IndexAward, Multiplier);

			return;
		}
	}

	// New Award

	Index = AwardRecord.Length;
	AwardRecord.Length = Index + 1;

	AwardRecord[Index].Class_Award = Class_Award;
	AwardRecord[Index].Award_Index = IndexAward;
	AwardRecord[Index].Award_Multiplier = Multiplier;
	AwardRecord[Index].CaptionColor = Class_Award.Static.GetCaptionColor (IndexAward);

	//

	ReTeamGame(Level.Game).PlayerAwarded (Controller(Owner), Self, Class_Award, IndexAward, Multiplier);
}

// Bindings

protected function Award_BindByIndex_WeaponDamageDeath (int IndexWeapon)
{
	local int IndexAward;

	for (IndexAward = 0; IndexAward < AwardTrigger_WeaponDamageDeath.Length; ++IndexAward)
	{
		if (Statistics_WeaponDamage[IndexWeapon].Class_WeaponDamageType != AwardTrigger_WeaponDamageDeath[IndexAward].Class_WeaponDamageType)
		{
			continue;
		}
		else
		{
			Statistics_WeaponDamage[IndexWeapon].AwardDeath_Index = AwardTrigger_WeaponDamageDeath[IndexAward].Index;
			Statistics_WeaponDamage[IndexWeapon].AwardDeath_Trigger = AwardTrigger_WeaponDamageDeath[IndexAward].Trigger;
			Statistics_WeaponDamage[IndexWeapon].bAwardDeath = True;

			return;
		}
	}
}

protected function Award_BindByIndex_WeaponDamageKill (int IndexWeapon)
{
	local int IndexAward;

	for (IndexAward = 0; IndexAward < AwardTrigger_WeaponDamageKill.Length; ++IndexAward)
	{
		if (Statistics_WeaponDamage[IndexWeapon].Class_WeaponDamageType != AwardTrigger_WeaponDamageKill[IndexAward].Class_WeaponDamageType)
		{
			continue;
		}
		else
		{
			Statistics_WeaponDamage[IndexWeapon].AwardKill_Index = AwardTrigger_WeaponDamageKill[IndexAward].Index;
			Statistics_WeaponDamage[IndexWeapon].AwardKill_Trigger = AwardTrigger_WeaponDamageKill[IndexAward].Trigger;
			Statistics_WeaponDamage[IndexWeapon].bAwardKill = True;

			return;
		}
	}
}

protected function Award_BindByIndex_WeaponDamageSuicide (int IndexWeapon)
{
	local int IndexAward;

	for (IndexAward = 0; IndexAward < AwardTrigger_WeaponDamageSuicide.Length; ++IndexAward)
	{
		if (Statistics_WeaponDamage[IndexWeapon].Class_WeaponDamageType != AwardTrigger_WeaponDamageSuicide[IndexAward].Class_WeaponDamageType)
		{
			continue;
		}
		else
		{
			Statistics_WeaponDamage[IndexWeapon].AwardSuicide_Index = AwardTrigger_WeaponDamageSuicide[IndexAward].Index;
			Statistics_WeaponDamage[IndexWeapon].AwardSuicide_Trigger = AwardTrigger_WeaponDamageSuicide[IndexAward].Trigger;
			Statistics_WeaponDamage[IndexWeapon].bAwardSuicide = True;

			return;
		}
	}
}

protected function Award_BindByIndex_WeaponDeath (int IndexWeapon)
{
	local int IndexAward;

	for (IndexAward = 0; IndexAward < AwardTrigger_WeaponDeath.Length; ++IndexAward)
	{
		if (Statistics_Weapon[IndexWeapon].Class_Weapon != AwardTrigger_WeaponDeath[IndexAward].Class_Weapon)
		{
			continue;
		}
		else
		{
			Statistics_Weapon[IndexWeapon].AwardDeath_Index = AwardTrigger_WeaponDeath[IndexAward].Index;
			Statistics_Weapon[IndexWeapon].AwardDeath_Trigger = AwardTrigger_WeaponDeath[IndexAward].Trigger;
			Statistics_Weapon[IndexWeapon].bAwardDeath = True;

			break;
		}
	}
}

protected function Award_BindByIndex_WeaponKill (int IndexWeapon)
{
	local int IndexAward;

	for (IndexAward = 0; IndexAward < AwardTrigger_WeaponKill.Length; ++IndexAward)
	{
		if (Statistics_Weapon[IndexWeapon].Class_Weapon != AwardTrigger_WeaponKill[IndexAward].Class_Weapon)
		{
			continue;
		}
		else
		{
			Statistics_Weapon[IndexWeapon].AwardKill_Index = AwardTrigger_WeaponKill[IndexAward].Index;
			Statistics_Weapon[IndexWeapon].AwardKill_Trigger = AwardTrigger_WeaponKill[IndexAward].Trigger;
			Statistics_Weapon[IndexWeapon].bAwardKill = True;

			return;
		}
	}
}

protected function Award_BindByIndex_WeaponSuicide (int IndexWeapon)
{
	local int IndexAward;

	for (IndexAward = 0; IndexAward < AwardTrigger_WeaponSuicide.Length; ++IndexAward)
	{
		if (Statistics_Weapon[IndexWeapon].Class_Weapon != AwardTrigger_WeaponSuicide[IndexAward].Class_Weapon)
		{
			continue;
		}
		else
		{
			Statistics_Weapon[IndexWeapon].AwardSuicide_Index = AwardTrigger_WeaponSuicide[IndexAward].Index;
			Statistics_Weapon[IndexWeapon].AwardSuicide_Trigger = AwardTrigger_WeaponSuicide[IndexAward].Trigger;
			Statistics_Weapon[IndexWeapon].bAwardSuicide = True;

			return;
		}
	}
}

function Statistics_RecordDeath
(
	class<DamageType> Class_DamageType,
	optional RePRI Subject,
	optional bool bIsTeamKill
)
{
	local int Index;

	++Metrics.Deaths;

	if (bIsTeamKill)
	{
		++Metrics.TeamDeaths;
	}

	if (class<WeaponDamageType>(Class_DamageType) != None)
	{
		Index = Statistics_GetIndex_WeaponDamage (class<WeaponDamageType>(Class_DamageType));
		++Statistics_WeaponDamage[Index].Count_Deaths;

		if
		(
			Statistics_WeaponDamage[Index].bAwardDeath
			&&
			(
				(Statistics_WeaponDamage[Index].AwardDeath_Trigger == 0)
				||
				(Statistics_WeaponDamage[Index].AwardDeath_Trigger == Statistics_WeaponDamage[Index].Count_Deaths)
			)
		)
		{
			++Statistics_WeaponDamage[Index].AwardDeath_Multiplier;

			Statistics_RecordAward
			(
				AwardClass_WeaponDeath,
				Statistics_WeaponDamage[Index].AwardDeath_Index,
				Statistics_WeaponDamage[Index].AwardDeath_Multiplier
			);

			AwardAnnounce_WeaponDamageDeath (Index);
		}

		// Weapon Death

		Index = Statistics_GetIndex_Weapon (Statistics_WeaponDamage[Index].Class_Weapon);
		++Statistics_Weapon[Index].Count_Deaths;

		if
		(
			Statistics_Weapon[Index].bAwardDeath
			&&
			(
				(Statistics_Weapon[Index].AwardDeath_Trigger == 0)
				||
				(Statistics_Weapon[Index].AwardDeath_Trigger == Statistics_Weapon[Index].Count_Deaths)
			)
		)
		{
			++Statistics_Weapon[Index].AwardDeath_Multiplier;

			Statistics_RecordAward
			(
				AwardClass_WeaponDeath,
				Statistics_Weapon[Index].AwardDeath_Index,
				Statistics_Weapon[Index].AwardDeath_Multiplier
			);

			AwardAnnounce_WeaponDeath (Index);
		}
	}
	else if (class<VehicleDamageType>(Class_DamageType) != None)
	{
	}
	else
	{
		// Damage Death

		Index = Statistics_GetIndex_Damage (Class_DamageType);
		++Statistics_Damage[Index].Count_Deaths;

		if
		(
			Statistics_Damage[Index].bAwardDeath
			&&
			(
				(Statistics_Damage[Index].AwardDeath_Trigger == 0)
				||
				(Statistics_Damage[Index].AwardDeath_Trigger == Statistics_Damage[Index].Count_Deaths)
			)
		)
		{
			++Statistics_Damage[Index].AwardDeath_Multiplier;

			Statistics_RecordAward
			(
				AwardClass_Death,
				Statistics_Damage[Index].AwardDeath_Index,
				Statistics_Damage[Index].AwardDeath_Multiplier
			);

			AwardAnnounce_DamageDeath (Index);
		}
	}
}

// - Kill branches to vehicle damage, weapon damage, non-weapon damage.

function Statistics_RecordKill
(
	class<DamageType> Class_DamageType,
	optional RePRI Victim, // The player this PRI killed.
	optional bool bIsTeamKill
)
{
	local int Index;

	// Increment

	++Metrics.Kills;

	++ReplicatedKills;


	if (bIsTeamKill)
	{
		++Metrics.Teamkills;
	}

	// Classify

	if (Class<WeaponDamageType>(Class_DamageType) != None)
	{
		if (Statistics_IsHeadShot (Class_DamageType)) // Head Shot -> Headhunter
		{
			Index = Statistics_AddEvent ('HeadShot'); // To record total head shot kill count to trigger Headhunter.

			++Metrics.TotalHeadShotKills;
		}

		// Weapon Damage Kill

		Index = Statistics_GetIndex_WeaponDamage (class<WeaponDamageType>(Class_DamageType));
		++Statistics_WeaponDamage[Index].Count_Kills;

		if
		(
			Statistics_WeaponDamage[Index].bAwardKill
			&&
			(
				(Statistics_WeaponDamage[Index].AwardKill_Trigger == 0)
				||
				(Statistics_WeaponDamage[Index].AwardKill_Trigger == Statistics_WeaponDamage[Index].Count_Kills)
			)
		)
		{
			if (!bIsTeamKill)
			{
				++Statistics_WeaponDamage[Index].AwardKill_Multiplier;

				Statistics_RecordAward
				(
					AwardClass_WeaponKill,
					Statistics_WeaponDamage[Index].AwardKill_Index,
					Statistics_WeaponDamage[Index].AwardKill_Multiplier
				);

				AwardAnnounce_WeaponDamageKill (Index);
			}
		}

		// Weapon

		Index = Statistics_GetIndex_Weapon (Statistics_WeaponDamage[Index].Class_Weapon);
		++Statistics_Weapon[Index].Count_Kills;

		if
		(
			Statistics_Weapon[Index].bAwardKill
			&&
			(
				(Statistics_Weapon[Index].AwardKill_Trigger == 0)
				||
				(Statistics_Weapon[Index].AwardKill_Trigger == Statistics_Weapon[Index].Count_Kills)
			)
		)
		{
			if (!bIsTeamKill)
			{
				++Statistics_Weapon[Index].AwardKill_Multiplier;

				Statistics_RecordAward
				(
					AwardClass_WeaponKill,
					Statistics_Weapon[Index].AwardKill_Index,
					Statistics_Weapon[Index].AwardKill_Multiplier
				);

				AwardAnnounce_WeaponKill (Index);
			}
		}
	}
	else if (Class<VehicleDamageType>(Class_DamageType) != None)
	{
	}
	else
	{
		Index = Statistics_GetIndex_Damage (Class_DamageType);
		++Statistics_Damage[Index].Count_Kills;

		if
		(
			Statistics_Damage[Index].bAwardKill
			&&
			(
				(Statistics_Damage[Index].AwardKill_Trigger == 0)
				||
				(Statistics_Damage[Index].AwardKill_Trigger == Statistics_Damage[Index].Count_Kills)
			)
		)
		{
			if (!bIsTeamKill)
			{
				++Statistics_Damage[Index].AwardKill_Multiplier;

				Statistics_RecordAward
				(
					AwardClass_Kill,
					Statistics_Damage[Index].AwardKill_Index,
					Statistics_Damage[Index].AwardKill_Multiplier
				);

				AwardAnnounce_DamageKill (Index);
			}
		}
	}
}

function Statistics_RecordSuicide (class<DamageType> Class_DamageType)
{
	local int Index;

	++Metrics.Suicides;

	if (class<WeaponDamageType>(Class_DamageType) != None)
	{
		Index = Statistics_GetIndex_WeaponDamage (class<WeaponDamageType>(Class_DamageType));
		++Statistics_WeaponDamage[Index].Count_Suicides;

		if
		(
			Statistics_WeaponDamage[Index].bAwardSuicide
			&&
			(
				(Statistics_WeaponDamage[Index].AwardSuicide_Trigger == 0)
				||
				(Statistics_WeaponDamage[Index].AwardSuicide_Trigger == Statistics_WeaponDamage[Index].Count_Suicides)
			)
		)
		{
			++Statistics_WeaponDamage[Index].AwardSuicide_Multiplier;

			Statistics_RecordAward
			(
				AwardClass_WeaponSuicide,
				Statistics_WeaponDamage[Index].AwardSuicide_Index,
				Statistics_WeaponDamage[Index].AwardSuicide_Multiplier
			);

			AwardAnnounce_WeaponDamageSuicide (Index);
		}

		Index = Statistics_GetIndex_Weapon (Statistics_WeaponDamage[Index].Class_Weapon);
		++Statistics_Weapon[Index].Count_Suicides;

		if
		(
			Statistics_Weapon[Index].bAwardSuicide
			&&
			(
				(Statistics_Weapon[Index].AwardSuicide_Trigger == 0)
				||
				(Statistics_Weapon[Index].AwardSuicide_Trigger == Statistics_Weapon[Index].Count_Suicides)
			)
		)
		{
			++Statistics_Weapon[Index].AwardSuicide_Multiplier;

			Statistics_RecordAward
			(
				AwardClass_WeaponSuicide,
				Statistics_Weapon[Index].AwardSuicide_Index,
				Statistics_Weapon[Index].AwardSuicide_Multiplier
			);

			AwardAnnounce_WeaponSuicide (Index);
		}
	}
	else if (class<VehicleDamageType>(Class_DamageType) != None)
	{
	}
	else
	{
		Index = Statistics_GetIndex_Damage (Class_DamageType);
		++Statistics_Damage[Index].Count_Suicides;

		if
		(
			Statistics_Damage[Index].bAwardSuicide
			&&
			(
				(Statistics_Damage[Index].AwardSuicide_Trigger == 0)
				||
				(Statistics_Damage[Index].AwardSuicide_Trigger == Statistics_Damage[Index].Count_Suicides)
			)
		)
		{
			++Statistics_Damage[Index].AwardSuicide_Multiplier;

			Statistics_RecordAward
			(
				AwardClass_Suicide,
				Statistics_Damage[Index].AwardSuicide_Index,
				Statistics_Damage[Index].AwardSuicide_Multiplier
			);

			AwardAnnounce_DamageSuicide (Index);
		}
	}
}

// ## Events ##

function int Statistics_AddEvent (name Name_Event, optional float Count_Points)
{
	local int IndexEvent;

	IndexEvent = Statistics_RecordEventByName (Name_Event);

	if
	(
		(Statistics_Event[IndexEvent].Award_Trigger == 0)
		||
		(Statistics_Event[IndexEvent].Award_Trigger == Statistics_Event[IndexEvent].Count_Triggered)
	)
	{
		++Statistics_Event[IndexEvent].Award_Multiplier;

		Statistics_RecordAward
		(
			AwardClass_Event,
			Statistics_Event[IndexEvent].Award_Index,
			Statistics_Event[IndexEvent].Award_Multiplier
		);

		AwardAnnounce_Event (IndexEvent, Count_Points);
	}

	return IndexEvent;
}

// - To record an event by name (in contrast to by index).
// - The initial event:
// -- (Count_Triggered == 1)

protected function int Statistics_RecordEventByName (name Name_Event)
{
	local int IndexEvent;

	for (IndexEvent = 0; IndexEvent < Statistics_Event.Length; ++IndexEvent)
	{
		if (Statistics_Event[IndexEvent].Name_Event != Name_Event)
		{
			continue;
		}
		else
		{
			++Statistics_Event[IndexEvent].Count_Triggered;

			return IndexEvent;
		}
	}

	IndexEvent = Statistics_Event.Length;
	Statistics_Event.Length = IndexEvent + 1;

	Statistics_Event[IndexEvent].Name_Event = Name_Event;
	++Statistics_Event[IndexEvent].Count_Triggered;

	Award_BindByIndex_Event (IndexEvent);

	return IndexEvent;
}

// - Called the first time an event is recorded.
// - To cache the AwadTrigger_Event (to reduce lookups).

protected function Award_BindByIndex_Event (int IndexEvent)
{
	local int IndexAward;

	for (IndexAward = 0; IndexAward < AwardTrigger_Event.Length; ++IndexAward)
	{
		if (AwardTrigger_Event[IndexAward].Name_Event != Statistics_Event[IndexEvent].Name_Event)
		{
			continue;
		}
		else
		{
			Statistics_Event[IndexEvent].Award_Index = AwardTrigger_Event[IndexAward].Index;
			Statistics_Event[IndexEvent].Award_Trigger = AwardTrigger_Event[IndexAward].Trigger;
			Statistics_Event[IndexEvent].Award_Multiplier = 0; // Not really needed, since default-initialized to zero.

			return;
		}
	}

	Warn ("AwardTrigger Missing - Statistics_Event["$IndexEvent$"].Name_Event"@Statistics_Event[IndexEvent].Name_Event);
}

// #

// - Clients spectating this player will see the award.

protected function AwardMulticast (Class<Award> AwardClass, int AwardIndex, int Points, int Multiplier)
{
	local Controller Other;

	for (Other = Level.ControllerList; Other != None; Other = Other.NextController)
	{
		if
		(
			Other.bIsPlayer
			&&
			Other.IsA ('RePlayer')
			&&
			(
				(PlayerController(Other).RealViewTarget == Controller(Owner))
				||
				(PlayerController(Other).ViewTarget == Controller(Owner))
				||
				(Other.PlayerReplicationInfo == Self)
			)
		)
		{
			RePlayer(Other).ClientAnnounceAward
			(
				AwardClass,
				AwardIndex,
				Points,
				Multiplier,
				PlayerID
			);
		}
	}
}

// Events
// - To send an alert to a network client.
// - Uses the award system to send notifications such as Last Man Out and
// other things that are not tracked by statistics.

function AlertEvent (name Name_Event)
{
	local int Index;

	for (Index = 0; Index < AwardTrigger_Event.Length; ++Index)
	{
		if (AwardTrigger_Event[Index].Name_Event != Name_Event)
		{
			continue;
		}
		else
		{
			AwardMulticast (AwardClass_Event, AwardTrigger_Event[Index].Index, 0, 0);

			break;
		}
	}
}

protected function AwardAnnounce_Event (int Index, optional float Count_Points)
{
	AwardMulticast (AwardClass_Event, Statistics_Event[Index].Award_Index, Count_Points, Statistics_Event[Index].Award_Multiplier);
}

// Deaths
// 1. AwardAnnounce_DamageDeath (int Index)
// 2. AwardAnnounce_WeaponDamageDeath (int Index)
// 3. AwardAnnounce_WeaponDeath (int Index)

// - PRI got killed by non-weapon damage.

protected function AwardAnnounce_DamageDeath (int Index)
{
	AwardMulticast (AwardClass_Death, Statistics_Damage[Index].AwardDeath_Index, 0, Statistics_Damage[Index].AwardDeath_Multiplier);
}

// - PRI got killed by weapon damage.
// - Recorded together with weapon.

protected function AwardAnnounce_WeaponDamageDeath (int Index)
{
	AwardMulticast (AwardClass_WeaponDeath, Statistics_WeaponDamage[Index].AwardDeath_Index, 0, Statistics_WeaponDamage[Index].AwardDeath_Multiplier);
}

// - PRI got killed by weapon.
// - Recorded together with weapon damage type.

protected function AwardAnnounce_WeaponDeath (int Index)
{
	AwardMulticast (AwardClass_WeaponDeath, Statistics_Weapon[Index].AwardDeath_Index, 0, Statistics_Weapon[Index].AwardDeath_Multiplier);
}

// Kills
// 1. AwardAnnounce_DamageKill (int Index)
// 2. AwardAnnounce_WeaponDamageKill (int Index)
// 3. AwardAnnounce_WeaponKill (int Index)

protected function AwardAnnounce_DamageKill (int Index)
{
	AwardMulticast (AwardClass_Kill, Statistics_Damage[Index].AwardKill_Index, 0, Statistics_Damage[Index].AwardKill_Multiplier);
}

protected function AwardAnnounce_WeaponDamageKill
(
	int Index,
	optional float Points
)
{
	AwardMulticast (AwardClass_WeaponKill, Statistics_WeaponDamage[Index].AwardKill_Index, Points, Statistics_WeaponDamage[Index].AwardKill_Multiplier);
}

protected function AwardAnnounce_WeaponKill (int Index, optional float Points)
{
	AwardMulticast (AwardClass_WeaponKill, Statistics_Weapon[Index].AwardKill_Index, Points, Statistics_Weapon[Index].AwardKill_Multiplier);
}

// Suicides
// 1. AwardAnnounce_DamageSuicide (int Index)
// 2. AwardAnnounce_WeaponDamageSuicide (int Index)
// 3. AwardAnnounce_WeaponSuicide (int Index)

protected function AwardAnnounce_DamageSuicide (int Index)
{
	AwardMulticast (AwardClass_Suicide, Statistics_Damage[Index].AwardSuicide_Index, 0, Statistics_Damage[Index].AwardSuicide_Multiplier);
}

protected function AwardAnnounce_WeaponDamageSuicide (int Index)
{
	AwardMulticast (AwardClass_Suicide, Statistics_WeaponDamage[Index].AwardSuicide_Index, 0, Statistics_WeaponDamage[Index].AwardSuicide_Multiplier);
}

protected function AwardAnnounce_WeaponSuicide (int Index)
{
	AwardMulticast (AwardClass_WeaponSuicide, Statistics_Weapon[Index].AwardSuicide_Index, 0, Statistics_Weapon[Index].AwardSuicide_Multiplier);
}

// - Unrolled loop for the first eight events.

function int Statistics_GetMultiplier_Event (name Name_Event)
{
	local int Index;
	local int Count;

	Count = Statistics_Event.Length;

	if // |
	(
		(0 < Count)
		&&
		(Name_Event == Statistics_Event[0].Name_Event)
	)
	{
		return Statistics_Event[0].Award_Multiplier;
	}
	else if // ||
	(
		(1 < Count)
		&&
		(Name_Event == Statistics_Event[1].Name_Event)
	)
	{
		return Statistics_Event[1].Award_Multiplier;
	}
	else if // |||
	(
		(2 < Count)
		&&
		(Name_Event == Statistics_Event[2].Name_Event)
	)
	{
		return Statistics_Event[2].Award_Multiplier;
	}
	else if // ||||
	(
		(3 < Count)
		&&
		(Name_Event == Statistics_Event[3].Name_Event)
	)
	{
		return Statistics_Event[3].Award_Multiplier;
	}
	else if // |||| |
	(
		(4 < Count)
		&&
		(Name_Event == Statistics_Event[4].Name_Event)
	)
	{
		return Statistics_Event[4].Award_Multiplier;
	}
	else if // |||| ||
	(
		(5 < Count)
		&&
		(Name_Event == Statistics_Event[5].Name_Event)
	)
	{
		return Statistics_Event[5].Award_Multiplier;
	}
	else if // |||| |||
	(
		(6 < Count)
		&&
		(Name_Event == Statistics_Event[6].Name_Event)
	)
	{
		return Statistics_Event[6].Award_Multiplier;
	}
	else if // |||| ||||
	(
		(7 < Count)
		&&
		(Name_Event == Statistics_Event[7].Name_Event)
	)
	{
		return Statistics_Event[7].Award_Multiplier;
	}
	else
	{
		for (Index = 8; Index < Count; ++Index)
		{
			if (Name_Event != Statistics_Event[Index].Name_Event)
			{
				continue;
			}
			else
			{
				return Statistics_Event[Index].Award_Multiplier;
			}
		}
	}

	return 0;
}

// #Resets

protected function ResetAwards ()
{
	AwardRecord.Length = 0;
}

simulated function ResetMetrics ()
{
	Metrics.Accuracy = 0.0f;
	Metrics.TotalWeaponDamageDone_Enemies = 0;
	Metrics.TotalWeaponDamageDone_Self = 0;
	Metrics.TotalWeaponDamageDone_Team = 0;
	Metrics.TotalWeaponDamageTaken_Enemies = 0;
	Metrics.TotalWeaponDamageTaken_Team = 0;
	Metrics.TotalWeaponHitsTaken_Enemies = 0;
	Metrics.TotalWeaponHitsTaken_Self = 0;
	Metrics.TotalWeaponHitsTaken_Team = 0;
	Metrics.TotalHeadShotHits = 0;
	Metrics.TotalHeadShotKills = 0;
	Metrics.LongestHeadShotDistance = 0;
	Metrics.Kills = 0;
	Metrics.Deaths = 0;
	Metrics.Suicides = 0;
	Metrics.Teamkills = 0;
	Metrics.TeamDeaths = 0;
	Metrics.AdrenalineGained = 0;
	Metrics.AdrenalineLost = 0;
}

protected function ResetStatistics ()
{
	Statistics_Damage.Length = 0;
	Statistics_Weapon.Length = 0;
	Statistics_WeaponDamage.Length = 0;
	Statistics_Event.Length = 0;
	Statistics_AdrenalineCombo.Length = 0;
}

// #

simulated function NotifyClientBecomeSpectator (PlayerController PC)
{
	ClientWeaponRecord.Length = 0;

	ResetMetrics ();
}

function NotifyServerBecomeSpectator (PlayerController PC)
{
	ReplicatedKills = 0;


	ResetAwards ();

	ResetStatistics ();

	ResetMetrics ();
}

// #Queries

// Protocol:
// - Return a percentage [0, 100].

function byte Statistics_GetPlayed ()
{
	Metrics.Played = 100;

	return Metrics.Played;
}

// #

DefaultProperties
{
	AwardClass_Death = Class'Award_Death'
	AwardTrigger_DamageDeath(0) = (Class_DamageType=Class'DamTypeCamping',Trigger=0,Index=0) // Camper!

	AwardClass_WeaponDeath = Class'Award_WeaponDeath'
	AwardTrigger_WeaponDamageDeath(0) = (Class_WeaponDamageType=Class'XWeapons.DamTypeFlakChunk',Trigger=5,Index=0)
	AwardTrigger_WeaponDeath(0) = (Class_Weapon=Class'XWeapons.SniperRifle',Trigger=6,Index=1)

	// ## Award Trigger - Kills ##

	AwardClass_Kill = Class'Award_Kill'
	AwardTrigger_DamageKill(0) = (Class_DamageType=Class'Gameplay.Burned',Trigger=3,Index=0)
	AwardTrigger_DamageKill(1) = (Class_DamageType=Class'Engine.FellLava',Trigger=5,Index=1)
	AwardTrigger_DamageKill(2) = (Class_DamageType=Class'Engine.Fell',Trigger=3,Index=2)

	AwardClass_WeaponKill = Class'Award_WeaponKill'
	AwardTrigger_WeaponDamageKill(0) = (Class_WeaponDamageType=Class'XWeapons.DamTypeFlakChunk',Trigger=10,Index=5)
	AwardTrigger_WeaponDamageKill(1) = (Class_WeaponDamageType=Class'XWeapons.DamTypeFlakShell',Trigger=10,Index=7)
	AwardTrigger_WeaponDamageKill(2) = (Class_WeaponDamageType=Class'XWeapons.DamTypeRocketHoming',Trigger=0,Index=12)
	AwardTrigger_WeaponDamageKill(3) = (Class_WeaponDamageType=Class'XWeapons.DamTypeShockBeam',Trigger=10,Index=11)
	AwardTrigger_WeaponDamageKill(4) = (Class_WeaponDamageType=Class'XWeapons.DamTypeShockBall',Trigger=10,Index=17)
	AwardTrigger_WeaponDamageKill(5) = (Class_WeaponDamageType=Class'XWeapons.DamTypeShockCombo',Trigger=15,Index=1) // #5.
	AwardTrigger_WeaponDamageKill(6) = (Class_WeaponDamageType=Class'XWeapons.DamTypeAssaultBullet',Trigger=10,Index=14)
	AwardTrigger_WeaponDamageKill(7) = (Class_WeaponDamageType=Class'XWeapons.DamTypeBioGlob',Trigger=5,Index=6)
	AwardTrigger_WeaponDamageKill(8) = (Class_WeaponDamageType=Class'XWeapons.DamTypeIonBlast',Trigger=3,Index=16)
	AwardTrigger_WeaponDamageKill(9) = (Class_WeaponDamageType=Class'XWeapons.DamTypeLinkPlasma',Trigger=10,Index=9)
	AwardTrigger_WeaponDamageKill(10) = (Class_WeaponDamageType=Class'XWeapons.DamTypeLinkShaft',Trigger=10,Index=10)
	AwardTrigger_WeaponDamageKill(11) = (Class_WeaponDamageType=Class'XWeapons.DamTypeShieldImpact',Trigger=5,Index=15)
	AwardTrigger_WeaponDamageKill(12) = (Class_WeaponDamageType=Class'XWeapons.DamTypeSniperHeadShot',Trigger=0,Index=2)
	AwardTrigger_WeaponDamageKill(13) = (Class_WeaponDamageType=Class'UTClassic.DamTypeClassicHeadshot',Trigger=0,Index=3)
	AwardTrigger_WeaponDamageKill(14) = (Class_WeaponDamageType=Class'DamTypeShockBeamHeadshot',Trigger=0,Index=22)

	AwardTrigger_WeaponKill(0) = (Class_Weapon=Class'XWeapons.FlakCannon',Trigger=15,Index=0) // #0.
	AwardTrigger_WeaponKill(1) = (Class_Weapon=Class'XWeapons.RocketLauncher',Trigger=5,Index=8)
	AwardTrigger_WeaponKill(2) = (Class_Weapon=Class'XWeapons.SniperRifle',Trigger=6,Index=13) // Electric Six
	AwardTrigger_WeaponKill(3) = (Class_Weapon=Class'XWeapons.Minigun',Trigger=8,Index=21) // Sweating Bullets

	// ## AwardTrigger - Suicides ##

	AwardClass_Suicide = Class'Award_Suicide'
	AwardTrigger_WeaponDamageSuicide(0) = (Class_WeaponDamageType=Class'XWeapons.DamTypeShockBall',Trigger=4,Index=1) // Core Dump
	AwardTrigger_DamageSuicide(0) = (Class_DamageType=Class'Gameplay.Burned',Trigger=3,Index=0) // Drawn to the Fire

	AwardClass_WeaponSuicide = Class'Award_WeaponSuicide'
	AwardTrigger_WeaponSuicide(0) = (Class_Weapon=Class'XWeapons.RocketLauncher',Trigger=3,Index=0) // Rocket Scientist

	// #AwardTrigger_Event

	AwardClass_Event = Class'Award_Event'

	AwardTrigger_Event(0) = (Name_Event=KillingSpree,Index=0) // Getting killing spree without knowing why? Check if you have added the event to this list (so it can find a match, or else it would always return the first index).
	AwardTrigger_Event(1) = (Name_Event=Rampage,Index=1)
	AwardTrigger_Event(2) = (Name_Event=Dominating,Index=2)
	AwardTrigger_Event(3) = (Name_Event=Unstoppable,Index=3)
	AwardTrigger_Event(4) = (Name_Event=Godlike,Index=4)
	AwardTrigger_Event(5) = (Name_Event=WickedSick,Index=5)
	AwardTrigger_Event(6) = (Name_Event=WickedSickStack,Index=6)
	AwardTrigger_Event(7) = (Name_Event=DoubleKill,Index=7)
	AwardTrigger_Event(8) = (Name_Event=MultiKill,Index=8)
	AwardTrigger_Event(9) = (Name_Event=MegaKill,Index=9)
	AwardTrigger_Event(10) = (Name_Event=UltraKill,Index=10)
	AwardTrigger_Event(11) = (Name_Event=MonsterKill,Index=11)
	AwardTrigger_Event(12) = (Name_Event=LudicrousKill,Index=12)
	AwardTrigger_Event(13) = (Name_Event=HolyShit,Index=13)
	AwardTrigger_Event(14) = (Name_Event=HolyShitStack,Index=14)

	//	[15, 16] are used internally by the award class.

	AwardTrigger_Event(17) = (Name_Event=ItEndsHere,Index=17)
	AwardTrigger_Event(18) = (Name_Event=HeadShot,Trigger=15,Index=18) // #18
	AwardTrigger_Event(19) = (Name_Event=FirstBloodMatch,Index=19)
	AwardTrigger_Event(20) = (Name_Event=FirstKillRound,Index=20)
	AwardTrigger_Event(21) = (Name_Event=FinalKillMatch,Index=21)
	AwardTrigger_Event(22) = (Name_Event=FinalKillRound,Index=22)
	AwardTrigger_Event(23) = (Name_Event=HatTrick,Index=23)
	AwardTrigger_Event(24) = (Name_Event=Retribution,Index=24)
	AwardTrigger_Event(25) = (Name_Event=FlawlessVictoryTeam,Index=25)
	AwardTrigger_Event(26) = (Name_Event=FlawlessVictoryTeamRound,Index=26)
	AwardTrigger_Event(27) = (Name_Event=HumiliatingDefeatTeam,Index=27)
	AwardTrigger_Event(28) = (Name_Event=HumiliatingDefeatTeamRound,Index=28)
	AwardTrigger_Event(29) = (Name_Event=Overkill,Index=29) // Alert
	AwardTrigger_Event(30) = (Name_Event=Fumbled,Index=30)

	//	[31, 32] are used internally by the award class.

	AwardTrigger_Event(33) = (Name_Event=DarkHorseRound,Index=33)
	AwardTrigger_Event(34) = (Name_Event=DarkHorseMatch,Index=34)

	// [35] is used internally by the award class.

	AwardTrigger_Event(36) = (Name_Event=FirstSuicideMatch,Index=36)
	AwardTrigger_Event(37) = (Name_Event=FirstOutMatch,Index=37)
	AwardTrigger_Event(38) = (Name_Event=FirstOutRound,Index=38)
	AwardTrigger_Event(39) = (Name_Event=LastManOutTeam,Index=39) // Alert
	AwardTrigger_Event(40) = (Name_Event=Confusion,Index=40)
	AwardTrigger_Event(41) = (Name_Event=LastManOutRound,Index=41)
	AwardTrigger_Event(42) = (Name_Event=OwnageRound,Index=42)

	// ReFreon

	AwardTrigger_Event(43) = (Name_Event=Thawesome,Trigger=15,Index=43)
	AwardTrigger_Event(44) = (Name_Event=GodofThaw,Trigger=20,Index=44)

	//

	AwardTrigger_Event(45) = (Name_Event=ComboStealer,Trigger=0,Index=45)

	//

	AwardTrigger_Event(46) = (Name_Event=Stombo,Trigger=0,Index=46)

	//

	AwardTrigger_Event(47) = (Name_Event=LastManStanding,Trigger=0,Index=47)
}

// Awards
// - The award system is simple.  It associates a single damage type with
// a single award context (death, kill, suicide).  Awards may be triggered
// each time they are added (Trigger == 0), or once when the trigger
//	is reached (Trigger > 0).
// -(Trigger < 0) is reserved for awards that trigger every %Trigger% time.
// - Awards are compiled into a single stream, ready to be sent to clients.
// - Some awards are alerts and are not counted.
// - Damage types have different contexts; suicide by, death by, kill by damage type.
// - Awards, when added, are bound to the indices of these damage types, if they exist.
// - Using separate arrays for seperate damage types means the array size is
// smaller, and quicker to search.
// - RePlayer fetches statistics.
