// *F* Developed 2011-2013 rejecht <rejecht at outlook dot com>

class Award_Event extends Award
	abstract
	;

// #Awards

const AwardCount = 48;
var private localized const string CaptionString[AwardCount];

// #Sounds

var(ConstantNames) private const name
	HatTrickSoundName,
	HeadhunterSoundName,
	FirstBloodSoundName,
	RetributionSoundName,
	ConfusionSoundName,
	OwnageSoundName,
	LastManStandingSoundName;

var private const byte
	Ownage_TimeDelay,
	DarkHorseSoundDelay,
	LastManStandingSoundDelay;

var(ConstantSounds) private const Sound
	DarkHorseSound,
	OverkillSound;

// #Queries

static function Color GetCaptionColor (int AwardIndex)
{
	if // Recurring Multi-Step
	(
		(
			(AwardIndex >= 0)
			&&
			(AwardIndex <= 6)
		)
	)
	{
		return Default.CyanCaptionColor;
    }
    else if // Recurring Single-Step
	(
		(AwardIndex == 17)
		||
		(AwardIndex == 20)
		||
		(AwardIndex == 21)
		||
		(AwardIndex == 22)
		||
		(AwardIndex == 24)
		||
		(AwardIndex == 29)
		||
		(AwardIndex == 33)
		||
		(AwardIndex == 34)
		||
		(AwardIndex == 39)
		||
		(AwardIndex == 42)
		||
		(AwardIndex == 45)
		||
		(
			(AwardIndex >= 7)
			&&
			(AwardIndex <= 16)
		)
	)
	{
		return Default.WhiteCaptionColor;
    }
    else if // Orange (Unique)
	(
		(AwardIndex == 19)
		||
		(AwardIndex == 25)
		||
		(AwardIndex == 26)
		||
		(AwardIndex == 27)
		||
		(AwardIndex == 28)
		||
		(AwardIndex == 47)
	)
	{
		return Default.OrangeCaptionColor;
    }
    else if // Purple
	(
		(AwardIndex == 35)
		||
		(AwardIndex == 18)
		||
		(AwardIndex == 23)
		||
		(AwardIndex == 35)
		||
		(AwardIndex == 43)
		||
		(AwardIndex == 44)
	)
	{
		return Default.PurpleCaptionColor;
    }
    else if // Gray
	(
		(AwardIndex == 30)
		||
		(AwardIndex == 36)
		||
		(AwardIndex == 37)
		||
		(AwardIndex == 38)
		||
		(AwardIndex == 40)
		||
		(AwardIndex == 41)
		||
		(AwardIndex == 46)
	)
	{
		return Default.GrayCaptionColor;
    }
    else // White
    {
		return Default.CaptionColor;
    }
}

// - While some events are unique for logging, many
// share the same message.

static function string GetCaptionString (int AwardIndex)
{
	switch (AwardIndex)
	{
		case 19:
			return Default.CaptionString[19];
		case 20:
			return Default.CaptionString[20];
		case 21:
		case 22:
			return Default.CaptionString[21];
		case 25:
		case 26:
			return Default.CaptionString[25];
		case 27:
		case 28:
			return Default.CaptionString[27];
		case 33:
		case 34:
			return Default.CaptionString[33];
		case 36:
		case 37:
		case 38:
			return Default.CaptionString[36];

		case 43:
			return Default.CaptionString[43];
		case 44:
			return Default.CaptionString[44];

		case 45:
			return Default.CaptionString[45];

		case 46:
			return Default.CaptionString[46];

		default:
			if (Default.CaptionString[AwardIndex] == "")
			{
				Warn ("AwardIndex:" @ AwardIndex);
			}
    }

	return Default.CaptionString[AwardIndex];
}

// #Dispatch

static function ClientReceive
(
	PlayerController PC,
	optional int AwardIndex,
	optional int Count_Points,
	optional int Multiplier,
	optional int D
)
{
	if (True)
	{
		switch (AwardIndex)
		{
			case 17:
				break; // Currently no sound for It Ends Here.

			case 18:
				PC.PlayRewardAnnouncement (default.HeadhunterSoundName, 1, True);
				break;

			case 19:
				break;

			case 20:
				break;

			case 23:
				PC.PlayRewardAnnouncement (default.HatTrickSoundName, 1, True);
				break;

			case 24:
				PC.PlayRewardAnnouncement (default.RetributionSoundName, 1, True);
				break;

			case 25:
			case 26:
				// ** PlayEndOfMatchMessage takes care of the announcement.
				break;

			case 27:
			case 28:
				// ** PlayEndOfMatchMessage takes care of the announcement.
				break;

			case 29:
				if (Class'PlayerSettings'.Default.bAnnounceOverkill)
				{
					PC.ClientPlaySound
					(
						Default.OverkillSound,
						True,
						2.0f * FClamp (0.1f + float(Class'PlayerController'.Default.AnnouncerVolume) * 0.225f, 0.2f, 1.0f),
						SLOT_Talk
					);
				}
				break;

			case 33:
			case 34:
				UnrealPlayer(PC).ClientDelayedAnnouncement (default.DarkHorseSound, default.DarkHorseSoundDelay);
				break;

			case 35:
				PC.PlayRewardAnnouncement (default.HeadhunterSoundName, 1, True);
				break;

			case 39:
				if (Class'PlayerSettings'.Default.LastManOutSoundVolume > 0.0f)
				{
					PC.ClientPlaySound
					(
						Class'PlayerSettings'.Default.LastManOut_Sound,
						True,
						FClamp (0.1f + Class'PlayerSettings'.Default.LastManOutSoundVolume * 0.225f, 0.2f, 1.0f),
						SLOT_None
					);
	            }
				break;

			case 40:
			case 41:
				PC.PlayRewardAnnouncement (default.ConfusionSoundName, 1, True);
				break;

			case 42:
				UnrealPlayer(PC).ClientDelayedAnnouncementNamed (default.OwnageSoundName, default.Ownage_TimeDelay);
				break;

			case 47:
				UnrealPlayer(PC).ClientDelayedAnnouncementNamed (default.LastManStandingSoundName, Default.LastManStandingSoundDelay);
				break;

			default:
				break;
		}
	}

	Super.ClientReceive (PC, AwardIndex, Count_Points, Multiplier, D);
}

// #

// - Show a different message.

protected static function HandleLocalMessage
(
	PlayerController PC,
	PlayerReplicationInfo RelatedPRI_1_Subject,
	int AwardIndex
)
{
	switch (AwardIndex)
	{
		case 0:
		case 1:
		case 2:
		case 3:
		case 4:
		case 5:
		case 6:
			PC.ReceiveLocalizedMessage (Class'UnrealGame.KillingSpreeMessage', Clamp (AwardIndex, 0, 5), RelatedPRI_1_Subject);
			return;

		case 7:
		case 8:
		case 9:
		case 10:
		case 11:
		case 12:
		case 13:
		case 14:
		case 15:
			PC.ReceiveLocalizedMessage (Class'XGame.MultiKillMessage', Clamp (AwardIndex - 7 + 1, 0, 6), RelatedPRI_1_Subject);
			return;

		case 19:
			PC.ReceiveLocalizedMessage (Class'UnrealGame.FirstBloodMessage', 0, RelatedPRI_1_Subject);
			return;

		// Pass these to a different style message.

		case 21: // Final Kill
		case 22: // Final Kill
		case 24: // Retribution
		case 25: // Flawless
		case 30: // Fumbled
		case 33: // Dark Horse
		case 34:
		case 36: // First Out
		case 42: // Ownage
		case 47: // Last Man Standing
			PC.ReceiveLocalizedMessage (Class'LocalMessage_AwardLocal', AwardIndex, None, None, Default.Class);
			return;
    }


	Super.HandleLocalMessage (PC, RelatedPRI_1_Subject, AwardIndex);
}

// #

DefaultProperties
{
	CaptionString(0) = "Killing Spree"
	CaptionString(1) = "Rampage"
	CaptionString(2) = "Dominating"
	CaptionString(3) = "Unstoppable"
	CaptionString(4) = "Godlike"
	CaptionString(5) = "Wicked Sick"
	CaptionString(6) = "Wicked Sick!"

	CaptionString(7) = "Double Kill"
	CaptionString(8) = "Multi Kill"
	CaptionString(9) = "Mega Kill"
	CaptionString(10) = "Ultra Kill"
	CaptionString(11) = "Monster Kill"
	CaptionString(12) = "Ludicrous"
	CaptionString(13) = "Holy Shit"
	CaptionString(14) = "Holy Shit!"
	CaptionString(15) = "Ludicrous" // bNoMatureLanguage
	CaptionString(16) = "Ludicrous!" // bNoMatureLanguage

	CaptionString(17) = "It Ends Here"

	CaptionString(18) = "Headhunter"
	HeadhunterSoundName = Headhunter

	CaptionString(19) = "First Blood"
	FirstBloodSoundName = First_Blood

	CaptionString(20) = "Preemptive Strike"

	CaptionString(21) = "Final Word"

	HatTrickSoundName = HatTrick
	CaptionString(23) = "Hat-Trick"

	RetributionSoundName = Retribution
	CaptionString(24) = "Retribution"

	CaptionString(25) = "Flawless Victory"
	CaptionString(27) = "Humiliating Defeat"

	CaptionString(29) = "Overkill"
	OverkillSound = Sound'Re.Sounds.Overkill'

	CaptionString(30) = "Fumbled"

	CaptionString(33) = "Dark Horse"
	DarkHorseSound = Sound'Sounds.DarkHorse'
	DarkHorseSoundDelay = 18

	CaptionString(35) = "Headhunter"

	CaptionString(36) = "First Out"

	CaptionString(39) = "Last Man Out"

	ConfusionSoundName = Confusion
	CaptionString(40) = "Team Kill"
	CaptionString(41) = "Team Killed"

	CaptionString(42) = "Ownage"
	OwnageSoundName = Ownage
	Ownage_TimeDelay = 17

	CaptionString(43) = "Thawesome"
	CaptionString(44) = "God of Thaw"

	CaptionString(45) = "Combo Stealer"

	CaptionString(46) = "Stombo"

	CaptionString(47) = "Last Man Standing"
	LastManStandingSoundName = last_man_standing
	LastManStandingSoundDelay = 23
}
