// Developed 2011-2013 rejecht <rejecht at outlook dot com>

class LocalMessage_Pickup_Local extends LocalMessage_Pickup
	abstract
	;

var protected Material Static_Pickup_Material;
var protected IntBox Static_Pickup_OffsetMaterial;

var private class<Pickup> LastPickupClass;
var private float LastPickupTime;
var private int AccumulatedPickupAmount;

// #Rendering

static event RenderComplexMessage
(
	Canvas Canvas,
	out float Width_Text,
	out float Height_Font,
	optional string MessageString,
	optional int Amount,
	optional PlayerReplicationInfo RelatedPRI_1,
	optional PlayerReplicationInfo RelatedPRI_2,
	optional Object O
)
{
	local float
		X,
		Y,
		Width_Message,
		OffsetX_Text,
		ScaledSize_Icon,
		USize_Material,
		VSize_Material;

	if (Class'PlayerSettings'.Default.PickupMessageType < 2)
	{
		Canvas.SetPos
		(
			Canvas.CurX,
			Canvas.CurY - (Height_Font * 0.5f)
		);

		Canvas.DrawText (MessageString, False);
	}
	else if (Default.Static_Pickup_Material != None)
	{
		if (UsePlus (O))
		{
			MessageString = " +" @ Amount;
		}
		else
		{
			MessageString = " x" @ Amount;
		}

		Canvas.StrLen (MessageString, Width_Message, Height_Font);

		USize_Material = Default.Static_Pickup_OffsetMaterial.X2 - Default.Static_Pickup_OffsetMaterial.X1;
		VSize_Material = Default.Static_Pickup_OffsetMaterial.Y2 - Default.Static_Pickup_OffsetMaterial.Y1;

		ScaledSize_Icon = (Height_Font * 1.5f) / VSize_Material;
		OffsetX_Text = USize_Material * ScaledSize_Icon;
		Width_Message += OffsetX_Text;

		X = (Canvas.ClipX * Default.PosX) - (Width_Message * 0.5f);
		Y = Canvas.CurY;

		Canvas.SetPos
		(
			X,
			Y - ((VSize_Material * ScaledSize_Icon) * 0.5f)
		);

		Canvas.DrawTile
		(
			Default.Static_Pickup_Material,
			USize_Material * ScaledSize_Icon,
			VSize_Material * ScaledSize_Icon,
			Default.Static_Pickup_OffsetMaterial.X1,
			Default.Static_Pickup_OffsetMaterial.Y1,
			USize_Material,
			VSize_Material
		);

		// Draw Amount

		Canvas.SetPos
		(
			X + OffsetX_Text,
			Y - (Height_Font * 0.5f)
		);

		Canvas.DrawText (MessageString, False);
	}
}

protected static final function bool UsePlus (Object O)
{
	return
	(
		(Class<TournamentHealth>(O) != None)
		||
		(Class<ShieldPickup>(O) != None)
	);
}

// Preconditions:
// - (Class<Pickup>(O) != None)

static event string GetString
(
	optional int Amount,
	optional PlayerReplicationInfo RelatedPRI_1, // != None
	optional PlayerReplicationInfo RelatedPRI_2, // == None
	optional Object O
)
{
	local Class<Pickup> PickupClass;
	local int PickupMessageType;

	PickupClass = Class<Pickup>(O);
	PickupMessageType = Clamp (Class'PlayerSettings'.Default.PickupMessageType, Class'PlayerSettings'.Default.MinPickupMessageType, Class'PlayerSettings'.Default.MaxPickupMessageType);

	if
	(
		(PickupClass != None)
		&&
		(RelatedPRI_1 != None)
	)
	{
		if (PickupMessageType == 0)
		{
			return PickupClass.Static.GetLocalString (0, RelatedPRI_1, RelatedPRI_2);
		}
		else if
		(
			(PickupMessageType == 1)
			||
			(PickupMessageType == 2)
		)
		{
			if (UsePlus (O))
			{
				return
				(
					Default.ItemStringPrefix
					$ GetShortPickupNameString (PickupClass)
					$ Default.ItemStringSuffix
					@ "+"
					@ Amount
				);
			}
			else
			{
				return
				(
					Default.ItemStringPrefix
					$ GetShortPickupNameString (PickupClass)
					$ Default.ItemStringSuffix
					@ "x"
					@ Amount
				);
			}
		}
	}
	else
	{
		return "";
	}
}

// #Dispatch

// Preconditions:
// - (Amount > 0)
// - (RelatedPRI_1 != None)
// - (Class<Pickup>(O) != None)

static event ClientReceive
(
	PlayerController PC,
	optional int Amount,
	optional PlayerReplicationInfo RelatedPRI_1,
	optional PlayerReplicationInfo RelatedPRI_2,
	optional Object O
)
{
	local Class<Pickup> PickupClass;
	local int PickupMessageType;

	PickupClass = Class<Pickup>(O);
	RelatedPRI_1 = PC.PlayerReplicationInfo;
	PickupMessageType = Clamp (Class'PlayerSettings'.Default.PickupMessageType, Class'PlayerSettings'.Default.MinPickupMessageType, Class'PlayerSettings'.Default.MaxPickupMessageType);

	if
	(
		(PickupClass != None)
		&&
		(RelatedPRI_1 != None)
	)
	{
		if (Class'PlayerSettings'.Default.bPickupsBumpReticle)
		{
			if (PC.MyHUD != None)
			{
				PickupClass.Static.UpdateHUD (PC.MyHUD); // Bumps Reticle
			}
		}

		if (PickupClass != Default.LastPickupClass)
		{
			Default.AccumulatedPickupAmount = Amount;
			Default.LastPickupClass = PickupClass;
			Default.LastPickupTime = PC.Level.TimeSeconds + Default.Lifetime;
		}
		else
		{
			if ((PC.Level.TimeSeconds - Default.LastPickupTime) > Default.Lifetime)
			{
				Default.AccumulatedPickupAmount = Amount;
			}
			else
			{
				Default.AccumulatedPickupAmount += Amount;
			}

			Default.LastPickupTime = PC.Level.TimeSeconds;
		}


		Amount = Default.AccumulatedPickupAmount;


		if (PC.bEnablePickupForceFeedback)
		{
			PC.ClientPlayForceFeedback (PickupClass.Default.PickupForce);
		}


		if (PickupMessageType == 2)
		{
			GetIcon (PickupClass, Default.Static_Pickup_Material, Default.Static_Pickup_OffsetMaterial);
		}
		else
		{
			Default.Static_Pickup_Material = None;
		}


		if (Class'PlayerSettings'.Default.bShowPickupMessage)
		{
			Super(LocalMessage).ClientReceive (PC, Max (1, Amount), RelatedPRI_1, RelatedPRI_2, O);
		}
	}
	else
	{
		Default.LastPickupClass = None;

		Super(LocalMessage).ClientReceive (PC, Max (1, Amount), RelatedPRI_1, RelatedPRI_2, O);
	}
}

// #

DefaultProperties
{
	// LocalMessage

	bComplexString = True
	bFadeMessage = True
	bIsConsoleMessage = True
	bIsSpecial = True
	bIsUnique = True

	DrawColor = (R=255,G=255,B=255,A=255)
	DrawPivot = DP_MiddleMiddle
	Lifetime = 3
	PosX = 0.5
	PosY = 0.92
	StackMode = SM_None
}
