class ReFreon_SquadAI extends SquadAI
	;

// #

// Thawing pawns are not attackable/enemies.

event bool ValidEnemy (Pawn NewEnemy)
{
	return
	(
		(NewEnemy != None)
		&&
		!NewEnemy.bAmbientCreature
		&&
		(NewEnemy.Health > 0)
		&&
		(NewEnemy.Controller != None)
		&&
		!FriendlyToward (NewEnemy)
		&&
		(
			!NewEnemy.IsA ('ReFreon_Pawn') // Done checking if not a standard game type pawn.
			||
			!ReFreon_Pawn(NewEnemy).IsThawing () // It's a game type pawn, and it's not in a thawing state.
		)
	);
}

// - Thawing players who are out still have a pawn, however,
// they are marked out (bOutofLives) and should be ignored.

event bool TellBotToFollow (Bot B, Controller C)
{
	if
	(
		(C == None)
		||
		C.bDeleteMe
	)
	{
		PickNewLeader ();


		C = SquadLeader;
	}


	if (B == C)
	{
		return False; // Do not follow self.
    }


	if
	(
		(C.PlayerReplicationInfo == None)
		||
		(C.PlayerReplicationInfo.bOutofLives)
	)
	{
		return False;
    }


    return Super.TellBotToFollow (B, C);
}

// (Call-Frequency) High

event bool CheckSquadObjectives (Bot B)
{
	if
	(
		(B != None)
		&&
		(ReGRI(Level.GRI).IsLive ())
	)
	{
		if (ReFreon_TeamAI(Team.AI).Thawing.Length > 0)
		{
			if (TryingToThaw (B))
			{
				return True;
			}
			else if (AssignedThawingTeammateTo (B))
			{
				return True;
			}
	    }

		return Super.CheckSquadObjectives (B);
    }

	return False;
}

// Check for thawing teammates, assign the closest one to B.

protected function bool AssignedThawingTeammateTo (Bot B)
{
	local Pawn ThawingPawn;
	local float Distance;
	local float ShortestDistance;
	local int Index;
	local int IndexClosest;
	local int Count;

	Count = ReFreon_TeamAI(Team.AI).Thawing.Length;

	if
	(
		(Count > 0)
		&&
		(ReFreon_TeamAI(Team.AI).Thawing[0] != None)
	)
	{
		IndexClosest = 0;

		ShortestDistance = VSize (B.Pawn.Location - ReFreon_TeamAI(Team.AI).Thawing[0].Location);


		for (Index = 1; Index < Count; ++Index)
		{
			ThawingPawn = ReFreon_TeamAI(Team.AI).Thawing[Index];

			if (ThawingPawn != None)
			{
				Distance = VSize (B.Pawn.Location - ThawingPawn.Location);

				if (Distance < ShortestDistance)
				{
					ShortestDistance = Distance;

					IndexClosest = Index;
		        }
            }
	    }


		return SetThawTarget (B, ReFreon_TeamAI(Team.AI).Thawing[IndexClosest]);
    }
    else
    {
		return False;
    }
}

// #Thawing

protected function bool TryingToThaw (Bot B)
{
	local ReFreon_PRI OtherPRI;

	if (ReFreon_Pawn(B.MoveTarget) != None)
	{
		OtherPRI = ReFreon_PRI(ReFreon_Pawn(B.MoveTarget).PlayerReplicationInfo);
    }


	if
	(
		(OtherPRI != None)
		&&
		(OtherPRI.ThawingPawn != None)
	)
	{
		return TellBotToThaw (B, OtherPRI.ThawingPawn);
	}
	else
	{
		return False;
	}
}

protected function bool TellBotToThaw (Bot B, Pawn P)
{
	if (B.Pawn.ReachedDestination(P))
	{
		if (B.FindBestPathToward (P, False, True))
		{
			B.RouteGoal = P;
			B.GoalString = "Best path to thaw" @ P;
			B.SetAttractionState ();

			return True;
		}
	}


	if (B.ActorReachable (P))
	{
		B.RouteGoal = P;
		B.MoveTarget = P;
		B.GoalString = "Reachable to thaw" @ P;
		B.SetAttractionState ();

		return True;
	}


	B.MoveTarget = B.FindPathToward (P, B.Pawn.bCanPickupInventory && (Vehicle(B.Pawn) == None));


	if (B.MoveTarget != None)
	{
		B.RouteGoal = P;
		B.GoalString = "Moving to thaw" @ P;
		B.SetAttractionState ();

		return True;
	}


	return False;
}

//

protected function bool SetThawTarget (Bot B, Pawn P)
{
	local ReFreon_PRI ThawingPRI;

	ThawingPRI = ReFreon_PRI(P.PlayerReplicationInfo);

	if
	(
		(ThawingPRI != None)
		&&
		(ThawingPRI.ThawingPawn != None)
	)
	{
		return TellBotToThaw (B, P);
	}
	else
	{
		return False;
	}
}

// #

DefaultProperties
{
}
