// Developed 2011-2013 rejecht <rejecht at outlook dot com>

class RePRI extends AwardPRI
	HideDropDown
	;

// #IndexTeam

const IndexTeam_Max = 2;
const IndexTeam_A = 0;
const IndexTeam_B = 1;
const IndexTeam_Unspecified = 255;


const PlayerName_Characters_Max = 20;
const PlayerName_Characters_MaxColorCoded = 80;


var protected const Vector NullVector;
var protected const Rotator NullRotation;


var(ClientCached) Texture TextureFlag;
var(ClientCached) IntBox OffsetTextureFlag;


var(ServerCachedOnce) string PlayerIDHashString;


var float TeamName_TimeSeconds_LastRelevant;


var(Proxy) bool bHadPacketLoss;


var float TimeSeconds_Joined;


var int KillingSpree_Count;


var int Multikill_Count;
var float Multikill_TimeSeconds_Last;


var int ClientMultikill_Count;
var float ClientMultikill_TimeSeconds_Last;


var(Classes) string RealtimePlayerInfoClassName;
var(Replicated) RealtimePlayerInfo RealtimePlayerInfo;


var int ClientEventBeacons_Code;
var protected int ClientEventBeacons_Type;
var bool ClientEventBeacons_bPlayedDeath;
var Color ClientEventBeacons_ColorMaterial;
var float ClientEventBeacons_TimeSeconds;
var Vector ClientEventBeacons_Location;
var IntBox ClientEventBeacons_OffsetMaterial;


var bool bReceivedCamperWarning;
var int Camping_Count_Penalties;
var int Camping_Count_ConsecutivePenalties;


var float ScaledReverseFriendlyFire;
var int ReverseFFTeammateDamage;


var int Retribution_PlayerID_LastEnemyKiller;


var Combiner ClientSkins_CombinerSkin_Body;

// #Replication

var(CachedTeamData) Vector Position; // Updated through HUD.
var(CachedTeamData) float NormalizedHitPoints; // Updated through HUD. [> 0.0f, 1.0f]
var(CachedTeamData) int HitPoints; // Updated through HUD.
var(CachedTeamData) byte Flags; // Updated through HUD.

var(Replicated) int IndexRound_Joined;


var(Classes) string TeamSummaryClassName;
var(Replicated) TeamSummary TeamSummary;


var(Replicated) string CountryCode; // Set through GameInfo.Login for players.


replication
{
	reliable if (bNetInitial && (Role == ROLE_Authority))
		CountryCode;

	reliable if (bNetDirty && (Role == ROLE_Authority))
		IndexRound_Joined;

	reliable if (bNetOwner && bNetDirty && (Role == ROLE_Authority))
		TeamSummary;

	reliable if (bNetInitial && (Role == ROLE_Authority))
		RealtimePlayerInfo;
}

// #Lifetime

simulated event PostBeginPlay ()
{
	Super.PostBeginPlay ();


	if (Level.NetMode != NM_DedicatedServer)
	{
		ClientLocalResource_PostBeginPlay ();
	}


	if (Role == ROLE_Authority)
	{
		if (Controller(Owner).bIsPlayer)
		{
			RealtimePlayerInfo_PostBeginPlay ();


			if (Owner.IsA ('PlayerController'))
			{
				TeamSummary_PostBeginPlay ();
	        }
        }


		AwardUpdate ();
    }
}

simulated event Destroyed ()
{
	if (Level.NetMode != NM_DedicatedServer)
	{
		TextureFlag = None;

		ClientLocalResource_Destroy ();
	}


	if (Role == ROLE_Authority)
	{
		if (RealtimePlayerInfo != None)
		{
			RealtimePlayerInfo.Destroy ();
			RealtimePlayerInfo = None;
        }
    }


	Super.Destroyed ();
}

// #

protected function RealtimePlayerInfo_PostBeginPlay ()
{
	local Class<RealtimePlayerInfo> LoadedRealtimePlayerInfoClass;

    LoadedRealtimePlayerInfoClass = Class<RealtimePlayerInfo>(DynamicLoadObject (RealtimePlayerInfoClassName, Class'Core.Class'));

    if (LoadedRealtimePlayerInfoClass != None)
    {
		RealtimePlayerInfo = Spawn (LoadedRealtimePlayerInfoClass, Self);

		if (RealtimePlayerInfo == None)
		{
			Warn ("(RealtimePlayerInfo == None) - RealtimePlayerInfoClassName: <" $ RealtimePlayerInfoClassName $ ">");
		}
    }
    else
    {
		Warn ("(LoadedRealtimePlayerInfoClass == None) - RealtimePlayerInfoClassName: <" $ RealtimePlayerInfoClassName $ ">");
    }
}

simulated function bool IsCurrentPlayerStatus (name QueriedPlayerStateName)
{
	return
	(
		(RealtimePlayerInfo != None)
		&&
		RealtimePlayerInfo.IsCurrentPlayerStatus (QueriedPlayerStateName)
	);
}

// #

protected function TeamSummary_PostBeginPlay ()
{
	local Class<TeamSummary> LoadedTeamSummaryClass;

    LoadedTeamSummaryClass = Class<TeamSummary>(DynamicLoadObject (TeamSummaryClassName, Class'Core.Class'));

    if (LoadedTeamSummaryClass != None)
    {
		TeamSummary = Spawn (LoadedTeamSummaryClass, Owner);

		if (TeamSummary == None)
		{
			Warn ("(TeamSummary == None) - TeamSummaryClassName: <" $ TeamSummaryClassName $ ">");
		}
    }
    else
    {
		Warn ("(LoadedTeamSummaryClass == None) - TeamSummaryClassName: <" $ TeamSummaryClassName $ ">");
    }
}

// #Resources

protected simulated function ClientLocalResource_PostBeginPlay ()
{
	ClientSkins_PostBeginPlay ();
}

protected simulated function ClientLocalResource_Destroy ()
{
	ClientSkins_Destroy ();
}

// #CountryCode

function SetCountryCode (string PlayerCountryCode)
{
	CountryCode = PlayerCountryCode;
}

simulated function string GetCountryCodeString ()
{
	return CountryCode;
}

// #Multikill

protected function ResetMultikill ()
{
	MultiKills[0] = 0;
	MultiKills[1] = 0;
	MultiKills[2] = 0;
	MultiKills[3] = 0;
	MultiKills[4] = 0;
	MultiKills[5] = 0;
	MultiKills[6] = 0;

	Multikill_Count = 0.0f;
	Multikill_TimeSeconds_Last = 0.0f;
}

// #KillingSpree

protected function ResetKillingSpree ()
{
	Spree[0] = 0;
	Spree[1] = 0;
	Spree[2] = 0;
	Spree[3] = 0;
	Spree[4] = 0;
	Spree[5] = 0;

	KillingSpree_Count = 0;
}

// #

function NotifyServerBecomeSpectator (PlayerController PC)
{
	IndexRound_Joined = 0;

	StartTime = 0;

	TimeSeconds_Joined = 0.0f;

	ResetKillingSpree ();

	ResetMultikill ();

	Super.NotifyServerBecomeSpectator (PC);
}

// #Award

// - Roundbased games, or rather, TAM, reduced the triggers for various awards.

protected function AwardUpdate ()
{
	if (ReGRI(Level.GRI).bRoundbased)
	{
		AwardTrigger_WeaponKill[0].Trigger = 5; // Flak Monkey
		AwardTrigger_WeaponKill[0].Index = 18;

		AwardTrigger_Event[18].Trigger = 3; // Headhunter
		AwardTrigger_Event[18].Index = 35;

		AwardTrigger_WeaponDamageKill[5].Trigger = 5; // Combo Whore
		AwardTrigger_WeaponDamageKill[5].Index = 20;
    }
}

// #Queries

// Protocol:
// - Return a percentage [0, 100].

function byte Statistics_GetPlayed ()
{
	local float TimeSeconds_Duration;
	local float Count_RoundsPlayed;

	if (ReGRI(Level.GRI).bRoundbased)
	{
		if (IndexRound_Joined != 0)
		{
			Count_RoundsPlayed = (ReGRI(Level.GRI).IndexRound_Current - IndexRound_Joined) + 1.0f;
		}


		Metrics.Played = FClamp ((Count_RoundsPlayed / ReGRI(Level.GRI).IndexRound_Current) * 100.0f, 0.0f, 100.0f);
    }
	else if (Level.GRI.TimeLimit != 0)
	{
		TimeSeconds_Duration = ReGRI(Level.GRI).TimeSeconds_MatchEnded - StartTime;


		Metrics.Played = FClamp ((TimeSeconds_Duration / ReGRI(Level.GRI).TimeSeconds_MatchDuration) * 100.0f, 0.0f, 100.0f);
	}
	else
	{
		Metrics.Played = 100;
    }

	return Metrics.Played;
}

// #ClientEventBeacons

simulated function ClientEventBeacons_Tag
(
	int Code,
	Vector Location_Event,
	float TimeSeconds,
	int Type,
	IntBox OffsetMaterial,
	Color ColorMaterial
)
{
	ClientEventBeacons_Code = Code;
	ClientEventBeacons_Location = Location_Event;
	ClientEventBeacons_TimeSeconds = TimeSeconds;
	ClientEventBeacons_Type = Type;
	ClientEventBeacons_OffsetMaterial = OffsetMaterial;
	ClientEventBeacons_ColorMaterial = ColorMaterial;
}

simulated function ClientEventBeacons_Reset ()
{
	ClientEventBeacons_TimeSeconds = Level.TimeSeconds;
}

simulated function ClientEventBeacons_Died ()
{
	if (ClientEventBeacons_Code == 1)
	{
		ClientEventBeacons_TimeSeconds *= 0.05f;
    }
}

// ## Client Skins ##

// - The combiner is unique for each pawn because each pawn may
// have its own unique skin.
// -- Not True for forced character.
// - The ConstantColor materials are allocated once in the PlayerController.
// Preconditions:
// - (Level.NetMode != NM_DedicatedServer)
// - (GetBodySkinColoring () == ES_Relative)
// - (ClientSkins_CombinerSkin_Body == None)

protected simulated function ClientSkins_PostBeginPlay ()
{
	ClientSkins_CombinerSkin_Body = Combiner_Allocate ();

	if (ClientSkins_CombinerSkin_Body != None)
	{
		ClientSkins_CombinerSkin_Body.CombineOperation = CO_Add;
		ClientSkins_CombinerSkin_Body.AlphaOperation = AO_Use_Mask;

		Level.AddPrecacheMaterial (ClientSkins_CombinerSkin_Body);
	}
	else
	{
		Warn ("(ClientSkins_CombinerSkin_Body == None) - Feature Disabled");
	}
}

// Postconditions:
// - (ClientSkins_CombinerSkin_Body == None)
// Preconditions:
// - (Level.NetMode != NM_DedicatedServer)

protected simulated function ClientSkins_Destroy ()
{
	ClientSkins_CombinerSkin_Body = Combiner_Free (ClientSkins_CombinerSkin_Body);
}

// #Objects

// Tip: Do not use FallbackMaterial here.  It will mess up with
// the skins, due to SetOverlayMaterial, most likely.

simulated final function Combiner Combiner_Allocate ()
{
	return Combiner(Level.ObjectPool.AllocateObject (class'Engine.Combiner'));
}

simulated final function Combiner Combiner_Free (Combiner O)
{
	if (O != None)
	{
		O.FallBackMaterial = None;
		O.Mask = None;
		O.Material2 = None;
		O.Material1 = None;

		Level.ObjectPool.FreeObject (O);

		return None;
	}

	return None;
}

// #

event SetWaitingPlayer (bool b)
{
	Super.SetWaitingPlayer (b);

	NetUpdateTime = Level.TimeSeconds - 1; // *UDKFix*
}

// #Reset

event Reset ()
{
	Super.Reset ();


	NetUpdateTime = Level.TimeSeconds - 1; // *UDKFix*
}

// Contexts:
// a. Controller.InitPlayerReplicationInfo () -> PlayerReplicationInfo.SetPlayerName ()
// - To make sure the player has a name (the default name is Player).
// b. GameInfo.ChangeName (string) -> PlayerReplicationInfo.SetPlayerName (string)

event SetPlayerName (string Text_RequestedPlayerName)
{
	Super.SetPlayerName (Text_RequestedPlayerName);

	NetUpdateTime = Level.TimeSeconds - 1; // *UDKFix*
}

simulated event Material GetPortrait ()
{
	if (Rec.Portrait != None)
	{
		return Rec.Portrait;
	}
	else
	{
		return Material(DynamicLoadObject ("PlayerPictures.cDefault", Class'Engine.Material'));
	}
}

// #Utility

// Preconditions:
// - Code [0, N]

static function bool SkipCharacterCode (int Code)
{
	return
	(
		(Code < 0x20) // Controls C0
		||
		(
			(Code > 0x7E) // Controls C0
			&&
			(Code < 0xA1) // Controls A1 + Non-breaking space (0xA0), not < C1
		)
	);
}

static function string GetFilteredPlayerNameString (coerce string SourcePlayerNameString)
{
	local string FinalString;
	local string SingleCharacterString;
	local int Index;
	local int Count_Characters;

    Index = InStr (SourcePlayerNameString, Chr (0x1B)); // First Color Code Prefix

	while (Index > -1)
	{
		SourcePlayerNameString = Left (SourcePlayerNameString, Index) $ Mid (SourcePlayerNameString, Index + 4); // Skip Color Code
		Index = InStr (SourcePlayerNameString, Chr (0x1B)); // Next Color Code Prefix
	}


	Count_Characters = Len (SourcePlayerNameString);


	SourcePlayerNameString = Left (SourcePlayerNameString, Min (Count_Characters, PlayerName_Characters_Max)); // Use known limits to avoid parsing any extra characters.


	for (Index = 0; Index < Count_Characters; ++Index)
	{
		SingleCharacterString = Mid (SourcePlayerNameString, Index, 1);


		if (!SkipCharacterCode (Asc (SingleCharacterString)))
		{
			if (SingleCharacterString == " ")
			{
				SingleCharacterString = "_";
			}
			else if (SingleCharacterString == "|")
			{
				SingleCharacterString = "I";
			}


			FinalString $= SingleCharacterString;
		}
	}


	return FinalString;
}

// #

simulated event DisplayDebug (Canvas Canvas, out float YL, out float YPos)
{
	local float XL;

	Canvas.Font = Canvas.SmallFont;
	Canvas.SetPos (4, YPos);
	Canvas.StrLen ("*", XL, YL);
	YPos = YPos + YL;

	Super.DisplayDebug (Canvas, YL, YPos);
}

// #LocationName

simulated event string GetLocationName ()
{
	local string VehicleString;
	local Vehicle V;

    if
	(
		(PlayerVolume == None)
		&&
		(PlayerZone == None)
	)
    {
    	if
		(
			(Owner != None)
			&&
			Controller(Owner).IsInState ('Dead')
		)
    	{
        	return StringDead;
        }
        else
    	{
			return StringSpectating;
        }
    }

	// Vehicle

    if (Owner != None)
    {
		if (Vehicle(Controller(Owner).Pawn) != None)
		{
			VehicleString = Vehicle(Controller(Owner).Pawn).GetVehiclePositionString ();
        }
	}
	else if (Level.NetMode == NM_Client)
	{
		if
		(
			(CurrentVehicle != None)
			&&
			(CurrentVehicle.PlayerReplicationInfo == Self)
		)
		{
			VehicleString = CurrentVehicle.GetVehiclePositionString ();
        }
		else if (Level.TimeSeconds - Level.LastVehicleCheck > 1)
		{
			ForEach DynamicActors (Class'Vehicle', V)
			{
				if (V.PlayerReplicationInfo == Self)
				{
					VehicleString = V.GetVehiclePositionString ();
					CurrentVehicle = V;
					break;
				}
            }

			if (V == None)
			{
				Level.LastVehicleCheck = Level.TimeSeconds;
            }
		}
	}

	// Finalize

	if
	(
		(PlayerVolume != None)
		&&
		(PlayerVolume.LocationName != Class'Volume'.Default.LocationName)
	)
	{
		if (Len (PlayerVolume.LocationName @ VehicleString) > 32)
		{
			return PlayerVolume.LocationName;
        }
        else if (VehicleString == "") // *UT2004Fix*
        {
			return PlayerVolume.LocationName;
        }
        else
        {
			return PlayerVolume.LocationName @ VehicleString;
        }
	}
	else if
	(
		(PlayerZone != None)
		&&
		(PlayerZone.LocationName != "")
	)
	{
		if (Len (PlayerZone.LocationName @ VehicleString) > 32)
		{
			return PlayerZone.LocationName;
        }
        else if (VehicleString == "") // *UT2004Fix*
        {
			return PlayerZone.LocationName;
        }
        else
        {
			return PlayerZone.LocationName @ VehicleString;
        }
	}
	else if (VehicleString != "")
	{
		return VehicleString;
    }
    else if (Level.Title != Level.Default.Title)
    {
		return Level.Title;
    }
	else
	{
        return StringUnknown;
    }
}

// #

DefaultProperties
{
	RealtimePlayerInfoClassName = "Re.RealtimePlayerInfo"
	TeamSummaryClassName = "Re.TeamSummary"

	// #Retribution

	Retribution_PlayerID_LastEnemyKiller = -1
}
