// *F* Developed 2011-2014 rejecht <rejecht at outlook dot com>

class RePawn extends xPawn
	HideDropDown
	;

const IndexTeam_Max = 2;
const IndexTeam_A = 0;
const IndexTeam_B = 1;
const IndexTeam_Unspecified = 255;

var(Constants) protected const Vector NullVector;
var(Constants) protected const Vector ZVector;

var(Development) config bool bLogInventory;


const CampingLocationHistoryCount = 10;
var(Camping) Vector CampingLocationHistory[CampingLocationHistoryCount];
var(Camping) int CampingIndexLocation;
var(Camping) bool bCheckCampingBehavior;
var(Camping) int CampingTimeSecondsLastWarning;


var(Skins) protected bool bCustomSkinColorApplied; // Set whenever a pawn got spawned while this feature was active.

var(Skins) protected Material
	ClientSkins_CachedMaterial_LastOverlay,
	ClientSkins_CachedMaterial_Skin[4];


var(SpawnProtection) TemporalFluxEffect TemporalFluxEffect;

// #

replication
{
	// Everyone will see this effect

	reliable if (Role == ROLE_Authority)
		TemporalFluxEffect;
}

// #Lifetime

simulated event Destroyed ()
{
	if (Level.NetMode != NM_DedicatedServer)
	{
		ClientSkins_CachedMaterial_Skin[0] = None;
		ClientSkins_CachedMaterial_Skin[1] = None;
		ClientSkins_CachedMaterial_Skin[2] = None;
		ClientSkins_CachedMaterial_Skin[3] = None;

		ClientSkins_CachedMaterial_LastOverlay = None;
	}


	RemovePowerups (); // *UT2004Fix* Kill off lingering emitters.


	if (PlayerShadow != None)
	{
		PlayerShadow.Destroy (); // ShadowProjector cleans up references.
		PlayerShadow = None;
	}


	if (DeResFX != None)
	{
		DeResFX.Emitters[0].SkeletalMeshActor = None; // Object Reference
		DeResFX.Kill ();
		DeResFX.Destroyed ();
		DeResFX = None;
	}


	if (Shadow != None)
	{
		Shadow.Destroy (); // Engine.Pawn
		Shadow = None;
	}


	Relinquish ();


	DestroyInventory ();


	Super(Actor).Destroyed ();
}

// #Inventory

protected function DestroyInventory ()
{
	if (Level.NetMode != NM_Client)
	{
		Weapon = None;
		SelectedItem = None;


		while (Inventory != None)
		{
			if (bLogInventory)
			{
				Log ("Pawn:" @ Class'ReUtility'.Static.GetShortTimeString (Level) @ Self $ ".DestroyInventory () - Inventory:" @ Inventory);
			}

			Inventory.Destroy ();
		}

		Inventory = None;
	}
}

// #Possession

// - To set view rotation.
// Context:
// - Controller.Possess ()

event PossessedBy (Controller Other)
{
	if (bSetPCRotOnPossess)
	{
		if (Other.IsA ('PlayerController'))
		{
			SetRotation (GetRotationToward ()); // *UT2004Fix*
		}
		else
		{
			SetRotation (Rotation);
		}
	}


	Super.PossessedBy (Other);


	if
	(
		Other.IsA ('PlayerController')
		&&
		(Weapon != None)
	)
	{
		Weapon.ClientWeaponSet (False); // *UDKFix*
	}
}

// - To relinquish this pawn from its controller.

protected function Relinquish ()
{
	if
	(
		(Controller != None)
		&&
		(Controller.Pawn == Self)
	)
	{
		Controller.PawnDied (Self); // Tear off, acknowledge unpossessed, removes reference to this pawn.


		if (Controller != None)
		{
			Controller.Unpossess (); // Calls Pawn.Unpossessed ().

			Controller = None;
		}
	}
}

// #Skins

simulated event SetOverlayMaterial (Material M, float T, bool bOverride)
{
	if
	(
		bCustomSkinColorApplied
		&&
		(T > 0.0f)
	)
	{
		T = 0.4f;
	}

	Super.SetOverlayMaterial (M, T, bOverride);
}

simulated event TickFX (float DeltaTime)
{
	Super.TickFX (DeltaTime);


	if
	(
		(OverlayMaterial != ClientSkins_CachedMaterial_LastOverlay)
		&&
		!bPlayedDeath
	)
	{
		ClientSkins_CachedMaterial_LastOverlay = OverlayMaterial;

		if (OverlayMaterial == Material'XGameShaders.PlayerShaders.PlayerShieldSh')
		{
			ClientSkins_UpdateOverlay (OverlayMaterial, 0);
		}
		else if (OverlayMaterial == Material'XGameShaders.PlayerShaders.LightningHit')
		{
			ClientSkins_UpdateOverlay (OverlayMaterial, 1);
		}
		else if (OverlayMaterial == Material'UT2004Weapons.ShockHitShader')
		{
			ClientSkins_UpdateOverlay (OverlayMaterial, 2);
		}
		else if (OverlayMaterial == Material'XGameShaders.PlayerShaders.LinkHit')
		{
			ClientSkins_UpdateOverlay (OverlayMaterial, 3);
		}
		else
		{
			ClientSkins_UpdateTeam ();
		}
	}
}


protected simulated function ClientSkins_UpdateTeam ()
{
	local RePRI OwnerRePRI;
	local int IndexTeam_Owner;

	OwnerRePRI = RePRI(PlayerReplicationInfo);
	IndexTeam_Owner = GetTeamNum ();

	if
	(
		(OwnerRePRI != None)
		&&
		(OwnerRePRI.ClientSkins_CombinerSkin_Body != None)
		&&
		(ClientSkins_CachedMaterial_Skin[0] != None)
		&&
		(IndexTeam_Owner < IndexTeam_Max)
	)
	{
		OwnerRePRI.ClientSkins_CombinerSkin_Body.Material1 = ClientSkins_CachedMaterial_Skin[0];
		OwnerRePRI.ClientSkins_CombinerSkin_Body.Material2 = ReGRI(Level.GRI).ClientSkins_ConstantColor_Team[IndexTeam_Owner];
	}
}


protected simulated function ClientSkins_UpdateOverlay (Material M, int Index)
{
	local RePRI OwnerRePRI;

	OwnerRePRI = RePRI(PlayerReplicationInfo);

	if (OwnerRePRI != None)
	{
		if
		(
			(OwnerRePRI != None)
			&&
			(OwnerRePRI.ClientSkins_CombinerSkin_Body != None)
			&&
			(ClientSkins_CachedMaterial_Skin[0] != None)
		)
		{
			OwnerRePRI.ClientSkins_CombinerSkin_Body.Material1 = ClientSkins_CachedMaterial_Skin[0];
			OwnerRePRI.ClientSkins_CombinerSkin_Body.Material2 = ReGRI(Level.GRI).ClientSkins_ConstantColor_Overlay[Index];
		}
		else
		{
			OwnerRePRI.ClientSkins_CombinerSkin_Body.Material1 = ClientSkins_CachedMaterial_Skin[0];
			OwnerRePRI.ClientSkins_CombinerSkin_Body.Material2 = ReGRI(Level.GRI).ClientSkins_ConstantColor_Overlay[Index];
		}
	}
}


simulated event Tick (float DeltaTime)
{
	if (Level.NetMode != NM_DedicatedServer)
	{
		if (Controller != None)
		{
			OldController = Controller;
		}


		TickFX (DeltaTime);


		if (bDeRes)
		{
			TickDeRes (DeltaTime);
		}
	}
	else
	{
		Disable ('Tick');  // *UT2004Fix*
	}
}

// ## Pickup ##

// Pawn.HandlePickup -> Controller.HandlePickup
// - To conditionally broadcast pickup messages.

event HandlePickup (Pickup Item)
{
	ReTeamGame(Level.Game).PickupBroadcast (Self, Item);

	Super.HandlePickup (Item); // Default handling of pickups continues.
}

// ## Adrenaline ##

protected function ResetAdrenaline ()
{
	if (Controller != None)
	{
		Controller.Adrenaline = 0.0f;
	}
}

// ## Powerups ##

// - To have the pawn clean up after itself.
// --  after
// combos, for instance.
// (Call-Context 1) Destroyed
// (Call-Context 2) Died

event RemovePowerups ()
{
	local bool bWasTeamCombo;

	if (CurrentCombo != None)
	{
		bWasTeamCombo = CurrentCombo.IsA ('TeamCombo');

		CurrentCombo.Destroy ();
		CurrentCombo = None;

		if (!bWasTeamCombo)
		{
			ResetAdrenaline ();
        }
	}


	if (UDamageTimer != None)
	{
		UDamageTimer.Destroy ();
		UDamageTimer = None;

		DisableUDamage ();
	}


	DestroyTemporalFlux ();
}

// ## Overlays ##

// - Makes sense to check, not sure if it's wise.
// - Done in conjunction with cleaning up in Destroyed to remove
// lingering emitter effects from booster adrenaline etc.

event SetWeaponOverlay (Material M, float Time, bool bOverride)
{
	if (!bDeleteMe)
	{
		Super.SetWeaponOverlay (M, Time, bOverride);
	}
}

// ## Skins ##

protected simulated function ClientSkins_OnDeath ()
{
	local int Index;

	for (Index = 0; Index < Skins.Length; ++Index)
	{
		if (ClientSkins_CachedMaterial_Skin[Index] != None)
		{
			RealSkins[Index] = ClientSkins_CachedMaterial_Skin[Index];
			Skins[Index] = ClientSkins_CachedMaterial_Skin[Index];
		}
	}
}

// ## Character ##

// - Ignores the UT2004.ini setting and only respects the client-side setting.
// -- In other words, the server can not force a player model.

simulated event bool ForceDefaultCharacter ()
{
	local PlayerController PC;
	local int IndexTeam_Local;
	local int IndexTeam_Pawn;

	// - Abort if executed server-side (dedicated).
	// - Abort if the client-side has no specific character.

	if
	(
		(Level.NetMode == NM_DedicatedServer)
		||
		(
			!Class'PlayerSettings'.default.bReplaceTeammateCharacters
			&&
			!Class'PlayerSettings'.default.bReplaceEnemyCharacters
		)
	)
	{
		return False;
	}

	PC = Level.GetLocalPlayerController ();

	if
	(
		(PC != None)
		&&
		(PC.PlayerReplicationInfo != None)
	)
	{
		IndexTeam_Local = PC.GetTeamNum (); // {0, 1, 255}
		IndexTeam_Pawn = GetTeamNum (); // {0, 1, 255}

		if
		(
			PC.PlayerReplicationInfo.bOnlySpectator
			||
			(IndexTeam_Local == IndexTeam_Unspecified)
			||
			(IndexTeam_Pawn == IndexTeam_Unspecified)
		)
		{
			return False; // Do not override character model.  Not enough information.
		}
		else if
		(
			Class'PlayerSettings'.default.bReplaceTeammateCharacters
			&&
			(IndexTeam_Local == IndexTeam_Pawn) // {0, 1} == {0, 1}
		)
		{
			if (IsValidCharacterNameString (Class'PlayerSettings'.default.TeammateReplacementCharacterNameString))
			{
				PlacedFemaleCharacterName = Class'PlayerSettings'.default.TeammateReplacementCharacterNameString;
				PlacedCharacterName = Class'PlayerSettings'.default.TeammateReplacementCharacterNameString;

				return True;
			}
			else
			{
				PlacedFemaleCharacterName = "Tamika";
				PlacedCharacterName = "Jakob";

				return False; // Do not override character model.  Invalid character name.
			}
		}
		else if
		(
			Class'PlayerSettings'.default.bReplaceEnemyCharacters
			&&
			(IndexTeam_Local != IndexTeam_Pawn) // {0, 1} != {0, 1}
		)
		{
			if (IsValidCharacterNameString (Class'PlayerSettings'.default.EnemyReplacementCharacterNameString))
			{
				PlacedFemaleCharacterName = Class'PlayerSettings'.default.EnemyReplacementCharacterNameString;
				PlacedCharacterName = Class'PlayerSettings'.default.EnemyReplacementCharacterNameString;

				return True;
			}
			else
			{
				PlacedFemaleCharacterName = "Tamika";
				PlacedCharacterName = "Jakob";

				return False; // Do not override character model.  Invalid character name.
			}
		}
	}

	return False; // Do not override character model.  Not enough information.
}

// - Allow client to override character models.

simulated event string GetDefaultCharacter ()
{
	// - To guarantee a valid character name (for spawning the right character model).

	if (Level.IsDemoBuild ())
	{
		PlacedFemaleCharacterName = "Tamika";
		PlacedCharacterName = "Jakob";
	}
	else
	{
		if (!IsValidCharacterNameString (PlacedFemaleCharacterName))
		{
			PlacedFemaleCharacterName = "Tamika";
		}

		if (!IsValidCharacterNameString (PlacedCharacterName))
		{
			PlacedCharacterName = "Jakob";
		}
	}


	if
	(
		(PlayerReplicationInfo != None)
		&&
		PlayerReplicationInfo.bIsFemale
	)
	{
		return PlacedFemaleCharacterName;
	}


	return PlacedCharacterName;
}

// - Called whenever pawn enters scope (always in scope in standalone).
// Context:
// - PostNetReceive () -> Setup () -> SpeciesType.static.Setup ()

simulated event Setup (xUtil.PlayerRecord PR, optional bool bLoadNow)
{
	// Client-side

	if
	(
		(Level.NetMode != NM_DedicatedServer)
		&&
		(Level.GRI != None)
	)
	{
		switch (Class'PlayerSettings'.static.GetBodySkinColoring ())
		{
			case ES_Relative:
				Level.GRI.bNoTeamSkins = False;
				Level.GRI.bForceTeamSkins = True;
				break;

			case ES_AbsoluteStandardBright:
				bCustomSkinColorApplied = False;
				Level.GRI.bNoTeamSkins = False;
				Level.GRI.bForceTeamSkins = True;
				class'UnrealGame.DMMutator'.default.bBrightSkins = True;
				class'UnrealGame.DMMutator'.static.StaticSaveConfig ();
				break;

			case ES_AbsoluteStandard:
				bCustomSkinColorApplied = False;
				Level.GRI.bNoTeamSkins = False;
				Level.GRI.bForceTeamSkins = True;
				class'UnrealGame.DMMutator'.default.bBrightSkins = False;
				class'UnrealGame.DMMutator'.static.StaticSaveConfig ();
				break;

			case ES_Individual:
				bCustomSkinColorApplied = False;
				Level.GRI.bNoTeamSkins = True;
				Level.GRI.bForceTeamSkins = False;
				break;
		}
	}

	// Client-side and server-side

	if
	(
		(PR.Species == None)
		||
		ForceDefaultCharacter ()
	)
	{
		PR = Class'XGame.xUtil'.Static.FindPlayerRecord (GetDefaultCharacter ()); // bAlreadySetup, TeamSkin
	}


	RagdollOverride = PR.Ragdoll;
	Species = PR.Species;


	if (Species.Static.Setup (Self, PR))
	{
		ResetPhysicsBasedAnim ();
	}
	else
	{
		PR = Class'XGame.xUtil'.Static.FindPlayerRecord (GetDefaultCharacter ());

		if (Species.Static.Setup (Self, PR))
		{
			ResetPhysicsBasedAnim ();
		}
	}

	// Client-side

	if (Level.NetMode != NM_DedicatedServer)
	{
		ClientSkins_Store ();

		ClientSetup ();
	}
}

// - The default skins have already been set up at this point.

protected simulated function ClientSetup ()
{
	if (ClientSkins_ShouldOverride ())
	{
		ClientSkins_Setup ();
	}
	else
	{
		ClientSkins_Restore ();
	}

	// Lighting


	if (Class'PlayerSettings'.Default.bUnlitCharacters)
	{
		bAcceptsProjectors = False;
		bUnlit = True;
		bUseDynamicLights = False;
		AmbientGlow = GetAmbientGlow ();
		MaxLights = 0;
	}
    else
    {
		bAcceptsProjectors = Default.bAcceptsProjectors;
		bUnlit = Default.bUnlit;
		bUseDynamicLights = Default.bUseDynamicLights;
		AmbientGlow = GetAmbientGlow ();
		MaxLights = Default.MaxLights;
    }
}

protected simulated function byte GetAmbientGlow ()
{
	local PlayerController PC;
	local int IndexTeam_Local, IndexTeam_Pawn;

	PC = Level.GetLocalPlayerController ();

	if
	(
		(PC != None)
		&&
		(PC.PlayerReplicationInfo != None)
		&&
		!PC.PlayerReplicationInfo.bOnlySpectator
	)
	{
		if (IndexTeam_Local == IndexTeam_Pawn)
		{
			return Class'PlayerSettings'.Default.CustomTeammateAmbientGlow;
        }
        else if (IndexTeam_Local != IndexTeam_Pawn)
        {
			return Class'PlayerSettings'.Default.CustomEnemyAmbientGlow;
		}
	}
}

protected simulated function bool ClientSkins_ShouldOverride ()
{
	return
	(
		ClientSkins_CustomColors ()
	);
}

protected simulated function bool ClientSkins_CustomColors ()
{
	return (Class'PlayerSettings'.Static.GetBodySkinColoring () == ES_Relative);
}

static function bool IsValidCharacterNameString (string CharacterNameString)
{
	return
	(
		IsValidFemaleCharacterNameString (CharacterNameString)
		||
		IsValidMaleCharacterNameString (CharacterNameString)
	);
}

protected static final function bool IsValidFemaleCharacterNameString (string CharacterNameString)
{
	return
	(
		(CharacterNameString ~= "Tamika")
		||
		(CharacterNameString ~= "Sapphire")
		||
		(CharacterNameString ~= "Enigma")
		||
		(CharacterNameString ~= "Cathode")
		||
		(CharacterNameString ~= "Rylisa")
		||
		(CharacterNameString ~= "Ophelia")
		||
		(CharacterNameString ~= "Zarina")
	);
}

protected static final function bool IsValidMaleCharacterNameString (string CharacterNameString)
{
	return
	(
		(CharacterNameString ~= "Jakob")
		||
		(CharacterNameString ~= "Gorge")
		||
		(CharacterNameString ~= "Malcolm")
		||
		(CharacterNameString ~= "Xan")
		||
		(CharacterNameString ~= "Brock")
		||
		(CharacterNameString ~= "Gaargod")
		||
		(CharacterNameString ~= "Axon")
	);
}

protected simulated function bool ClientSkins_Applied ()
{
	return (bCustomSkinColorApplied);
}

// - Override normal skins.

protected simulated function ClientSkins_Setup ()
{
	local RePRI OwnerPRI;
	local int IndexTeam_Owner;

	OwnerPRI = RePRI(PlayerReplicationInfo);
	IndexTeam_Owner = GetTeamNum ();

	if
	(
		(OwnerPRI != None)
		&&
		(OwnerPRI.ClientSkins_CombinerSkin_Body != None)
		&&
		(IndexTeam_Owner < IndexTeam_Max)
	)
	{
		OwnerPRI.ClientSkins_CombinerSkin_Body.Material1 = ClientSkins_CachedMaterial_Skin[0];
		OwnerPRI.ClientSkins_CombinerSkin_Body.Material2 = ReGRI(Level.GRI).ClientSkins_ConstantColor_Team[IndexTeam_Owner];

		Skins[0] = OwnerPRI.ClientSkins_CombinerSkin_Body;
		Skins[1] = ClientSkins_CachedMaterial_Skin[1];

		if (Skins.Length > 1)
		{
			Skins[2] = ClientSkins_CachedMaterial_Skin[2];
		}

		if (Skins.Length > 2)
		{
			Skins[3] = ClientSkins_CachedMaterial_Skin[3];
		}

		bCustomSkinColorApplied = True;
	}
}

// - Equivalent to xPawn.RealSkins but used for restoring skins upon player death
// and for re-applying custom colored skins.

protected simulated function ClientSkins_Store ()
{
	local int Index;

	for (Index = 0; Index < Skins.Length; ++Index)
	{
		if (ClientSkins_CachedMaterial_Skin[Index] == None)
		{
			ClientSkins_CachedMaterial_Skin[Index] = Skins[Index];
		}
	}
}

protected simulated function ClientSkins_Restore ()
{
	local int Index;

	for (Index = 0; Index < Skins.Length; ++Index)
	{
		if (ClientSkins_CachedMaterial_Skin[Index] != None)
		{
			Skins[Index] = ClientSkins_CachedMaterial_Skin[Index];
		}
	}

	bCustomSkinColorApplied = False;
}

// #Effect

// - To fix rotation of emitter when spawning at start of match with
// no spawn protection.

event PlayTeleportEffect (bool bOut, bool bSound)
{
	local int IndexTeam; // Default-Initialized: IndexTeam_A (0)

	if
	(
		(PlayerReplicationInfo != None)
		&&
		(PlayerReplicationInfo.Team != None)
	)
	{
		IndexTeam = PlayerReplicationInfo.Team.TeamIndex; // Expected [0, 1]
	}


	if
	(
		!bSpawnIn
		&&
		(
			(Level.TimeSeconds - SpawnTime)
			<
			DeathMatch(Level.Game).SpawnProtectionTime
		)
	)
	{
		bSpawnIn = True;


		SetSpawnProtectionMaterial ();


		SpawnVisualTeleportEffect (IndexTeam);
	}
	else if (bOut)
	{
		if ((Level.TimeSeconds - SpawnTime) > 0.0f)
		{
			DoTranslocateOut (Location);
		}
		else
		{
			SpawnVisualRelativeTeleportEffect (IndexTeam);
		}
	}
	else
	{
		SpawnVisualTeleportEffect (IndexTeam);
	}


	MakeNoise (1.0f);
}

protected function SpawnVisualTeleportEffect (int IndexTeam)
{
	if (TransEffects[0] != None)
	{
		Spawn
		(
			TransEffects[IndexTeam],
			Self,
			,
			Location + CollisionHeight * Vect (0.0f, 0.0f, 0.75f)
		);
	}
}

protected function SpawnVisualRelativeTeleportEffect (int IndexTeam)
{
	if (TransOutEffect[0] != None)
	{
		Spawn
		(
			TransOutEffect[IndexTeam],
			Self,
			,
			Location + CollisionHeight * Vect (0.0f, 0.0f, 0.75f),
			Rotation
		);
	}
}

// #Animation

simulated event PlayDirectionalHit (Vector HitLoc)
{
	// Do not interrupt animations when hit.
}

simulated event PlayDyingAnimation (Class<DamageType> DamageTypeClass, Vector DyingHitLocation)
{
	if
	(
		(Level.NetMode == NM_Client)
		&&
		(DamageTypeClass != None)
		&&
		DamageTypeClass.Default.bAlwaysGibs // *UT2004Fix* @Eliot
	)
	{
		ChunkUp (Rotation, DamageTypeClass.Default.GibPerterbation);
	}
	else
	{
		Super.PlayDyingAnimation (DamageTypeClass, DyingHitLocation);
	}
}

// #

simulated event PlayDying (Class<DamageType> DamageTypeClass, Vector HitLoc)
{
	if (Level.NetMode != NM_DedicatedServer)
	{
		NotifyProxyDied ();
	}

	Super.PlayDying (DamageTypeClass, HitLoc);
}

// #

simulated event ChunkUp (Rotator HitRotation, float ChunkPerterbation)
{
	if (Level.NetMode != NM_DedicatedServer)
	{
		NotifyProxyDied ();
	}

	Super.ChunkUp (HitRotation, ChunkPerterbation);
}

// Preconditions:
// - (Level.NetMode != NM_DedicatedServer)

protected simulated function NotifyProxyDied (optional bool bImmediate)
{
	bAcceptsProjectors = Default.bAcceptsProjectors;
	bUnlit = Default.bUnlit;
	bUseDynamicLights = Default.bUseDynamicLights;
	AmbientGlow = Default.AmbientGlow;
	MaxLights = Default.MaxLights;

	ClientSkins_Restore ();

	Proxy_ClientEventBeacons_Died ();
}


protected simulated function Proxy_ClientEventBeacons_Died ()
{
	if (RePRI(PlayerReplicationInfo) != None)
	{
		RePRI(PlayerReplicationInfo).ClientEventBeacons_Died ();
	}
}


event KilledBy (Pawn EventInstigator)
{
	local Controller Killer;

	Health = 0;

	if (EventInstigator != None)
	{
		Killer = EventInstigator.Controller;

		LastHitBy = None; // *UDKFix*
	}

	Died (Killer, Class'Engine.Suicided', Location);
}

// - To give the pawn a good facing-direction when spawned.

protected function Rotator GetRotationToward ()
{
	local NavigationPoint Other;
	local Rotator NewRotation;
	local Vector OffsetVector, SourceLocation;
	local float Score, HighestScore;

	SourceLocation = Location;

	OffsetVector = ZVector * BaseEyeHeight;

	ForEach RadiusActors (class'Engine.NavigationPoint', Other, 2048.0f, SourceLocation)
	{
		Score = VSize (SourceLocation - Other.Location);

		if (FastTrace (SourceLocation, Other.Location + OffsetVector))
		{
			Score *= 2.0f;
		}

		if (Score > HighestScore)
		{
			HighestScore = Score;

			NewRotation = Rotator (Other.Location - SourceLocation);
		}
	}

	NewRotation.Pitch = 0;
	NewRotation.Roll = 0;
	NewRotation.Yaw = NewRotation.Yaw & 65535;

	return NewRotation;
}

// #

event Landed (Vector HitNormal)
{
	Super(UnrealPawn).Landed (HitNormal);


	MultiJumpRemaining = MaxMultiJump;

	// Forces landing sound effect to play. @Eliot

 	if
	(
		(Health > 0)
		||
		!bHidden
	)
	{
		PlayOwnedSound
		(
			GetSound (EST_Land),
			SLOT_Interact,
			FMin (1.0f, -0.3f * Velocity.Z / JumpZ),
			True
		);
	}
}

// #Dying

simulated state Dying
{
	simulated event Landed (Vector HitNormal)
	{
		if (Level.NetMode != NM_DedicatedServer)
		{
			if (Shadow != None)
			{
				Shadow.Destroy ();
				Shadow = None; // *UT2004Fix*
			}
		}
	}

	exec event SwitchToLastWeapon () {} // *UT2004Fix*

	simulated event BeginState ()
	{
		local int Index;

		RemoveEffects ();

		AmbientSound = None; // xPawn

		SetCollision (True, False, False); // UnrealPawn

		if
		(
			bTearOff
			&&
			(Level.NetMode == NM_DedicatedServer)
		)
		{
			LifeSpan = 1.0f;
		}
		else
		{
			SetTimer (2.0f, False);
		}

		SetPhysics (PHYS_Falling);

		bInvulnerableBody = True;

		if (Controller != None)
		{
			if (Controller.bIsPlayer)
			{
				Relinquish ();
			}
			else
			{
				Controller.Destroy ();
			}
		}

		for (Index = 0; Index < Attached.Length; ++Index)
		{
			if (Attached[Index] != None)
			{
				Attached[Index].PawnBaseDied ();
			}
		}

		DestroyEffects (); // *UT2004Fix*
	}

	simulated event FellOutofWorld (EKillZType KillType)
	{
		if (!bDeRes)
		{
			Super.FellOutofWorld (KillType);
        }
        else
        {
			Destroy ();
        }
	}

	simulated event EndState ()
	{
		Super.EndState ();

		AmbientSound = None; // xPawn
		HighDetailOverlay = None; // Actor
		OverlayMaterial = None; // Actor
    }
}

// *UT2004Fix* Was not declared outside of state.

simulated event ReduceCylinder () {}
simulated event LieStill () {}
simulated event LandThump () {}

// #

// *UT2004Fix*
// - Attempted workaround for the "Skeleton_Alien Attachment not found for
// bone %bone% (rthigh) to Attach %Emitter% (AlienBloodJet0)".

protected simulated function DestroyEffects ()
{
	local Actor OtherActor;
	local int Index;

	for (Index = 0; Index < Attached.Length; ++Index)
	{
		if
		(
			(Attached[Index] != None)
			&&
			Attached[Index].IsA ('xEmitter')
		)
		{
			OtherActor = Attached[Index];
			Attached[Index] = None;

			Attached.Remove (Index--, 1);

			OtherActor.Destroy ();
		}
	}
}

simulated event RemoveFlamingEffects ()
{
	local int Index, Count;

	if (Level.NetMode != NM_DedicatedServer)
	{
		Count = Attached.Length;

		for (Index = 0; Index < Count; ++Index)
		{
			if
			(
				(Attached[Index] != None) // *UT2004Fix*
				&&
				Attached[Index].IsA ('xEmitter')
				&&
				!Attached[Index].IsA ('BloodJet')
			)
			{
				xEmitter(Attached[Index]).mRegen = False;
			}
		}
	}
}

// - Match or round can end while pawn is in water.

event TakeDrowningDamage ()
{
	if (Controller != None)
	{
		Super.TakeDrowningDamage (); // *UT2004Fix*
	}
}

// - The check for Drowned was not correct in UT2004.
// - After fixing it, found there are no drowning sounds, thus skipping playing drowning sound.

event PlayTakeHit (Vector HitLocation, int Damage, Class<DamageType> Class_DamageType)
{
	PlayDirectionalHit (HitLocation);


	if ((Level.TimeSeconds - LastPainSound) >= MinTimeBetweenPainSounds)
	{
		LastPainSound = Level.TimeSeconds;

		if
		(
			(Damage > 0)
			&&
			(Health > 0)
		)
		{
			if (!HeadVolume.bWaterVolume)
			{
				PlaySound
				(
					SoundGroupClass.Static.GetHitSound (),
					SLOT_Pain,
					2.0f * TransientSoundVolume,
					,
					200
				);
			}
			else
			{
				PlaySound (GetSound (EST_HitUnderwater), SLOT_Pain, 1.5f * TransientSoundVolume);
			}
		}
	}
}

// ## DeRes ##

// - To make the deres color neutral.

simulated event StartDeRes ()
{
	if (Level.NetMode != NM_DedicatedServer)
	{
		TexScaler(Combiner(Shader(FinalBlend(DeResMat0).Material).Diffuse).Material1).Material = Texture'Textures.DeRezRGB8';
		TexRotator(TexPanner(Combiner(Shader(FinalBlend(DeResMat0).Material).Diffuse).Material2).Material).Material = Texture'Textures.DeRezA8';

		TexScaler(Combiner(Shader(FinalBlend(DeResMat1).Material).Diffuse).Material1).Material = Texture'Textures.DeRezRGB8';
		TexRotator(TexPanner(Combiner(Shader(FinalBlend(DeResMat1).Material).Diffuse).Material2).Material).Material = Texture'Textures.DeRezA8';

	    if (Physics == PHYS_KarmaRagdoll)
	    {
			KSetBlockKarma (False); // "You MUST turn collision off (KSetBlockKarma) before using bone lifters!"
		}

		Super.StartDeRes ();

		if (DeResFX != None)
		{
			DeResFX.Emitters[0].ColorScale[0].Color.R = 255;
			DeResFX.Emitters[0].ColorScale[0].Color.G = 255;
			DeResFX.Emitters[0].ColorScale[0].Color.B = 255;

			DeResFX.Emitters[0].ColorScale[1].Color.R = 190;
			DeResFX.Emitters[0].ColorScale[1].Color.G = 190;
			DeResFX.Emitters[0].ColorScale[1].Color.B = 190;

			DeResFX.Emitters[0].ColorScale[2].Color.R = 190;
			DeResFX.Emitters[0].ColorScale[2].Color.G = 190;
			DeResFX.Emitters[0].ColorScale[2].Color.B = 190;
		}
	}
}

// #TearOff

simulated event TurnOff ()
{
	if (Role == ROLE_Authority)
	{
		RemoteRole = ROLE_SimulatedProxy; // *UDKFix*
	}

	Super.TurnOff ();
}

// Minor optimizations.

event TakeDamage (int Damage, Pawn PawnInstigator, Vector HitLocation, Vector HitMomentum, Class<Damagetype> DamageTypeClass)
{
	local int ActualDamage;
	local Controller Killer;

	if
	(
		(Role == ROLE_Authority)
		&&
		(Health > 0)
	)
	{
		if
		(
			(
				(PawnInstigator == None)
				||
				(PawnInstigator.Controller == None)
			)
			&&
			DamageTypeClass.Default.bDelayedDamage
			&&
			(DelayedDamageInstigatorController != None)
		)
		{
			PawnInstigator = DelayedDamageInstigatorController.Pawn;
		}


		if
		(
			(Physics == PHYS_None)
			&&
			(DrivenVehicle == None)
		)
		{
			SetMovementPhysics ();
		}


		if
		(
			(Physics == PHYS_Walking)
			&&
			DamageTypeClass.Default.bExtraMomentumZ
		)
		{
			HitMomentum.Z = FMax (HitMomentum.Z, 0.4f * VSize (HitMomentum));
		}


		if (PawnInstigator == Self)
		{
			HitMomentum *= 0.6f;
		}


		HitMomentum /= Mass;


		if (Weapon != None)
		{
			Weapon.AdjustPlayerDamage (Damage, PawnInstigator, HitLocation, HitMomentum, DamageTypeClass); // Momentum
		}


		if (DrivenVehicle != None)
		{
			DrivenVehicle.AdjustDriverDamage (Damage, PawnInstigator, HitLocation, HitMomentum, DamageTypeClass); // Momentum
		}


		if (HitLocation == NullVector)
		{
			HitLocation = Location;
		}


		if
		(
			(PawnInstigator != None)
			&&
			PawnInstigator.HasUDamage ()
		)
		{
			Damage *= 2;
		}


		ActualDamage = Level.Game.ReduceDamage (Damage, self, PawnInstigator, HitLocation, HitMomentum, DamageTypeClass);


		if
		(
			DamageTypeClass.Default.bArmorStops
			&&
			(ActualDamage > 0)
		)
		{
			ActualDamage = ShieldAbsorb (actualDamage);
		}


		Health -= ActualDamage;


		if (HitLocation == NullVector)
		{
			HitLocation = Location;
		}


		PlayHit (ActualDamage, PawnInstigator, HitLocation, DamageTypeClass, HitMomentum);


		if (Health > 0)
		{
			AddVelocity (HitMomentum);


			if (Controller != None)
			{
				Controller.NotifyTakeHit (PawnInstigator, HitLocation, ActualDamage, DamageTypeClass, HitMomentum);
			}


			if
			(
				(PawnInstigator != Self)
				&&
				(PawnInstigator != None)
			)
			{
				LastHitBy = PawnInstigator.Controller;
			}
		}
		else
		{
			if
			(
				DamageTypeClass.Default.bCausedByWorld
				&&
				(
					(PawnInstigator == None)
					||
					(PawnInstigator == Self)
				)
				&&
				(LastHitBy != None)
			)
			{
				Killer = LastHitBy;
			}
			else if (PawnInstigator != None)
			{
				Killer = PawnInstigator.GetKillerController ();
			}


			if
			(
				(Killer == None)
				&&
				DamageTypeClass.Default.bDelayedDamage
			)
			{
				Killer = DelayedDamageInstigatorController;
			}


			if (bPhysicsAnimUpdate)
			{
				TearOffMomentum = HitMomentum;
			}


			Died (Killer, DamageTypeClass, HitLocation);
		}


		MakeNoise (1.0f);
	}
	else
	{
		if (Role < ROLE_Authority)
		{
			Warn ("(Role < ROLE_Authority)" @ DamageTypeClass);
		}

		if
		(
			(DamageTypeClass == None)
			&&
			(PawnInstigator != None)
		)
		{
			Warn ("(DamageTypeClass == None) && (PawnInstigator != None)" @ PawnInstigator);
		}
	}
}


event TakeFallingDamage ()
{
	local float Shake, EffectiveSpeed;

	if (Velocity.Z < (-0.5f * MaxFallSpeed))
	{
		if (Role == ROLE_Authority)
		{
			MakeNoise (1.0f);

			if (Velocity.Z < (-1.0f * MaxFallSpeed))
			{
				EffectiveSpeed = Velocity.Z;

				if (TouchingWaterVolume ())
				{
					EffectiveSpeed = FMin (0.0f, EffectiveSpeed + 100.0f);
				}

				if (EffectiveSpeed < (-1.0f * MaxFallSpeed))
				{
					TakeDamage
					(
						-100 * (EffectiveSpeed + MaxFallSpeed) / MaxFallSpeed,
						None,
						Location,
						NullVector,
						class'Engine.Fell'
					);
				}
			}
		}

		if (Controller != None)
		{
			Shake = FMin (1.0f, (-1.0f * Velocity.Z) / MaxFallSpeed);

			Controller.DamageShake (Shake);
		}
	}
	else if (Velocity.Z < (-1.4f * JumpZ))
	{
		MakeNoise (0.5f);
	}
	else if (Velocity.Z < (-0.8f * JumpZ))
	{
		MakeNoise (0.2f); // *UDKFix* To add less noise for shorter falls.
	}
}

// #Adrenaline

// At this point it is assumed the combo can execute.
// xPlayer.DoCombo -> xPlayer.ServerDoCombo () -> DoCombo ()

event DoCombo (Class<Combo> ComboClass)
{
	if (ComboClass != None)
	{
		if (ReGRI(Level.GRI).bTeamCombos)
		{
			ReGRI(Level.GRI).TeamComboManager[PlayerReplicationInfo.Team.TeamIndex].NotifyDoCombo (Controller, ComboClass); // Sets CurrentCombo
		}
		else if (!InCurrentCombo ())
		{
			CurrentCombo = Spawn (ComboClass, Self);

			if (CurrentCombo != None)
			{
				NotifyDoCombo (ComboClass);
			}
		}
	}
	else
	{
		if (ComboClass == None)
		{
			Warn ("(ComboClass == None)");
		}


		if (InCurrentCombo ())
		{
			Warn ("OK: InCurrentCombo ()" @ ComboClass);
		}
	}
}

//

protected function NotifyDoCombo (Class<Combo> ComboClass)
{
	local PlayerController PC;

	// Clients

	PC = PlayerController(Owner);

	if
	(
		(PC != None)
		&&
		(ComboClass.Default.ExecMessage != "")
	)
	{
		PC.ReceiveLocalizedMessage
		(
			Class'LocalMessage_Combo',
			,
			,
			,
			ComboClass
		);
	}

	// Statistics

	UnrealMPGameInfo(Level.Game).SpecialEvent
	(
		PlayerReplicationInfo,
		string(ComboClass)
	);

	if (RePRI(PlayerReplicationInfo) != None)
	{
		RePRI(PlayerReplicationInfo).Statistics_DoCombo (ComboClass);
	}
}

// #Weapons

simulated event ChangedWeapon () // *UT2004Fix* Previously broke simulated.
{
	Super.ChangedWeapon ();
}

// #Debug

simulated event DisplayDebug (Canvas Canvas, out float YL, out float YPos)
{
	local float XL;

	Canvas.Font = Canvas.SmallFont;
	Canvas.StrLen ("*", XL, YL);
	YPos = YPos + YL;
	Canvas.SetPos (XL, YPos);

	Super.DisplayDebug (Canvas, YL, YPos);
}

// #

event PlayDyingSound ()
{
	if (!bGibbed) // *UT2004Fix* Do not play dying sound when gibbed (or at least try not to).
	{
		Super.PlayDyingSound ();
	}
}

// #

simulated event RemoveEffects ()
{
	local Emitter E;
	local int Index, EmitterIndex, EmitterCount, Count;

	Count = Attached.Length;

	for (Index = 0; Index < Count; ++Index)
	{
		if (Attached[Index] != None)
		{
			if (Attached[Index].IsA ('xEmitter'))
			{
				xEmitter(Attached[Index]).mRegen = False;
			}
			else if (Attached[Index].IsA ('Emitter'))
			{
				E = Emitter(Attached[Index]);

				EmitterCount = E.Emitters.Length;

				for (EmitterIndex = 0; EmitterIndex < EmitterCount; ++EmitterIndex)
				{
					E.Emitters[EmitterIndex].SkeletalMeshActor = None; // Object Reference
				}

				E.Kill ();
			}
		}
	}
}

// #TossWeapon

event TossWeapon (Vector TossVelocity)
{
	if (ReGRI(Level.GRI).bTossWeapon)
	{
		Super.TossWeapon (TossVelocity);
	}
	else if (Weapon != None)
	{
		Weapon.Destroy ();
    }
}

// #

// #SpawnProtection

protected function SetSpawnProtectionMaterial ()
{
	if (TemporalFluxEffect == None)
	{
		TemporalFluxEffect = Spawn (Class'TemporalFluxEffect', Self, , Location, Rotation); // Attaches to the pawn
    }


	SetOverlayMaterial
	(
		ShieldHitMat,
		DeathMatch(Level.Game).SpawnProtectionTime,
		False
	);
}

event DeactivateSpawnProtection ()
{
	DestroyTemporalFlux ();


	Super.DeactivateSpawnProtection ();
}

protected function DestroyTemporalFlux ()
{
	if (TemporalFluxEffect != None)
	{
		TemporalFluxEffect.Destroy ();
		TemporalFluxEffect = None;
    }
}

// #

DefaultProperties
{
	bLogInventory = False

	ZVector = (X=0.0,Y=0.0,Z=1.0)

	// xPawn

	TransEffects(0) = Class'Emitter_SpawnEffectRed'
	TransEffects(1) = Class'Emitter_SpawnEffectBlue'

	TransOutEffect(0) = Class'Emitter_SpawnEffectRedOut'
	TransOutEffect(1) = Class'Emitter_SpawnEffectBlueOut'

	// Pawn

	ControllerClass = Class'ReBot'
}
