// *F* Developed 2011-2014 rejecht <rejecht at outlook dot com>

class ReStatistics extends Scoreboard
	abstract
	config (User)
	;

const IndexTeam_Max = 2;
const IndexTeam_A = 0;
const IndexTeam_B = 1;
const IndexTeam_Unspecified = 255;


var(Tick) protected UnrealPlayer PlayerOwner;
var(Tick) protected RePRI PlayerOwnerPRI;
var(Tick) protected HUD PlayerOwnerHUD;

var protected int IndexTeam_Local;


const Metrics_Count_Columns = 20; // Fixed limit for the number of entries.
const Weapons_Count_Columns = 16; // The number of columns (Hit, EWD, TWD, ...)


var(Colors) GlobalConfig Color
	CountryCodeColor,
	CountryFlagColor,
	CornerBackgroundColor,
	TeamSymbolColor,
	HeaderBackgroundColor,
	NeutralTeamColor, // Used when replication has not coughed up enough information.
	Metrics_ColorTexture_Background,
	ScoreColor,
	Metrics_ColorTexture_Symbol,
	LegendColor,
	ContentColor,
	CaptionColor,
	BodyBackgroundColor,
	PortraitBackgroundColor,
	BodySectionColor,
	Weapons_ColorText_Value,
	Weapons_ColorText_Weapon,
	MatchInfoColor;


var(Tick) protected Color SpectatedPlayerTeamColor;


var(ConstantStrings) protected localized const string
	StandaloneMatchString,
	RankedMatchIDString,
	UnrankedMatchString,
	ScoreLegendString[Metrics_Count_Columns],
	CombosString,
	WeaponString,
	WeaponLegendString[Weapons_Count_Columns],
	AwardsString;


var(ConstantStrings) protected const string
	ScoreLongestString[Metrics_Count_Columns],
	Weapons_Text_Highest[Weapons_Count_Columns];


var(ConstantTextures) protected const Texture
	SpectatorSymbolTexture,
	HeaderBackgroundTexture,
	Metrics_Texture_Bar,
	ScoreBarTexture,
	Metrics_Texture_Up,
	Metrics_Texture_Down,
	Metrics_Texture_Equal,
	BodyBackgroundTexture,
	LeftSectionTexture,
	RightSectionTexture;


var(Constants) protected const IntBox
	OffsetTexture_Spectator,
	Metrics_OffsetTexture_Up,
	Metrics_OffsetTexture_Down,
	Metrics_OffsetTexture_Equal;


var(Constants) protected const byte
	HeaderOpacity,
	ScoreBarAlpha;


var(Constants) protected const int
	Metrics_Count_Used,
	Weapons_Count_Used;


var(ConstantClasses) protected const class<WeaponDamageTypeMap> WeaponDamageTypeMapClass;


var(Tick) protected RePRI SpectatedPRI;


var(Tick) protected float LastRequestAcceptedTime;


var(Tick) protected int
	IndexSizeFont_Large,
	IndexSizeFont_Normal,
	IndexSizeFont_Small;


var(Tick) protected float
	Cached_YMetrics,
	Cached_YBody,
	Cached_UWeapons,
	Cached_UGeneral,
	Cached_XAwards,
	Cached_UAwards,
	Cached_XCombos,
	Cached_YCombos,
	Cached_UCombos,
	Cached_WidthFont_Normal,
	Cached_HeightFont_Normal;


var(Tick) protected string
	MatchTypeString,
	GameString,
	LevelTitleString,
	ServerNameString,
	ServerMessageString,
	ScoreDataString[Metrics_Count_Columns],
	Weapons_Text_Value[Metrics_Count_Columns],
	TextColorCode_Multiplier;

// #Allocated

var protected Combiner LeftPortraitCombiner;
var protected FinalBlend LeftPortraitFinalBlend;

var protected Combiner RightPortraitCombiner;
var protected FinalBlend RightPortraitFinalBlend;

// #Lifetime

event PostBeginPlay ()
{
	Super.PostBeginPlay ();


	TextColorCode_Multiplier = Class'ReUtility'.Static.GetColorCodeString (Class'Award'.Default.AwardMultiplierColor);

	// Left Combiner

	LeftPortraitCombiner = AllocateCombiner ();
	LeftPortraitFinalBlend = AllocateFinalBlend ();

	if
	(
		(LeftPortraitCombiner != None)
		&&
		(LeftPortraitFinalBlend != None)
	)
	{
		LeftPortraitCombiner.CombineOperation = CO_AlphaBlend_With_Mask;
		LeftPortraitCombiner.AlphaOperation = AO_Use_Mask;
		LeftPortraitCombiner.Material1 = None;
		LeftPortraitCombiner.Material2 = Texture'Textures.LeftPortraitAlpha';
		LeftPortraitCombiner.Mask = Texture'Textures.LeftPortraitAlpha';
		LeftPortraitCombiner.InvertMask = False;
		LeftPortraitCombiner.Modulate2X = False;
		LeftPortraitCombiner.Modulate4X = False;
		LeftPortraitCombiner.FallbackMaterial = None;

		LeftPortraitFinalBlend.FrameBufferBlending = FB_Brighten;
		LeftPortraitFinalBlend.ZWrite = False;
		LeftPortraitFinalBlend.ZTest = False;
		LeftPortraitFinalBlend.AlphaTest = False;
		LeftPortraitFinalBlend.TwoSided = False;
		LeftPortraitFinalBlend.AlphaRef = 0;
		LeftPortraitFinalBlend.Material = LeftPortraitCombiner;
		LeftPortraitFinalBlend.FallbackMaterial = None;
	}

	// Right Combiner

	RightPortraitCombiner = AllocateCombiner ();
	RightPortraitFinalBlend = AllocateFinalBlend ();

	if
	(
		(RightPortraitCombiner != None)
		&&
		(RightPortraitFinalBlend != None)
	)
	{
		RightPortraitCombiner.CombineOperation = CO_AlphaBlend_With_Mask;
		RightPortraitCombiner.AlphaOperation = AO_Use_Mask;
		RightPortraitCombiner.Material1 = None;
		RightPortraitCombiner.Material2 = Texture'Textures.RightPortraitAlpha';
		RightPortraitCombiner.Mask = Texture'Textures.RightPortraitAlpha';
		RightPortraitCombiner.InvertMask = False;
		RightPortraitCombiner.Modulate2X = False;
		RightPortraitCombiner.Modulate4X = False;
		RightPortraitCombiner.FallbackMaterial = None;

		RightPortraitFinalBlend.FrameBufferBlending = FB_Brighten;
		RightPortraitFinalBlend.ZWrite = False;
		RightPortraitFinalBlend.ZTest = False;
		RightPortraitFinalBlend.AlphaTest = False;
		RightPortraitFinalBlend.TwoSided = False;
		RightPortraitFinalBlend.AlphaRef = 0;
		RightPortraitFinalBlend.Material = RightPortraitCombiner;
		RightPortraitFinalBlend.FallbackMaterial = None;
	}
}

event Destroyed ()
{
	RightPortraitFinalBlend = FreeFinalBlend (RightPortraitFinalBlend);
	RightPortraitCombiner = FreeCombiner (RightPortraitCombiner);

	LeftPortraitFinalBlend = FreeFinalBlend (LeftPortraitFinalBlend);
	LeftPortraitCombiner = FreeCombiner (LeftPortraitCombiner);


	Super.Destroyed ();
}

// #Resources

protected final function Combiner AllocateCombiner ()
{
	return Combiner(Level.ObjectPool.AllocateObject (class'Engine.Combiner'));
}

protected final function Combiner FreeCombiner (Combiner O)
{
	if (O != None)
	{
		O.FallBackMaterial = None;
		O.Mask = None;
		O.Material2 = None;
		O.Material1 = None;

		Level.ObjectPool.FreeObject (O);

		return None;
	}

	return None;
}

protected final function FinalBlend AllocateFinalBlend ()
{
	return FinalBlend(Level.ObjectPool.AllocateObject (class'Engine.FinalBlend'));
}

protected final function FinalBlend FreeFinalBlend (FinalBlend O)
{
	if (O != None)
	{
		O.Material = None;
		O.FallbackMaterial = None;

		Level.ObjectPool.FreeObject (O);

		return None;
	}

	return None;
}

// #Main

// Preconditions:
// - (PlayerOwnerPRI != None)

event UpdateScoreboard (Canvas Canvas)
{
	Canvas.Style = ERenderStyle.STY_Alpha;


	if (SpectatedPRI == None)
	{
		SpectatedPRI = PlayerOwnerPRI; // Fall back to local player.
	}


	if ((Level.TimeSeconds - LastRequestAcceptedTime) > 2.5f)
	{
		if (RePlayer(PlayerOwner) != None)
		{
			RePlayer(PlayerOwner).Statistics_ServerRequest_Update (SpectatedPRI);
		}

		LastRequestAcceptedTime = Level.TimeSeconds;
	}
	else
	{
	}

	// Cache

	Canvas.Font = PlayerOwnerHUD.GetFontSizeIndex (Canvas, IndexSizeFont_Normal);

	Canvas.StrLen ("*", Cached_WidthFont_Normal, Cached_HeightFont_Normal);

	Cached_XAwards = Cached_WidthFont_Normal;
	Cached_UAwards = Cached_WidthFont_Normal * 28.0f;

	Cached_XCombos = Canvas.ClipX -  Cached_WidthFont_Normal;

	// Draw

	DrawHeader (Canvas);

	Cached_YBody = DrawMetrics (Canvas);

	DrawBody (Canvas, Cached_YBody);
	DrawAwards (Canvas, Cached_YBody);
	DrawWeapons (Canvas, Cached_YBody);
	DrawCombos (Canvas, Cached_YBody);
}

// #Header

protected function DrawHeader (Canvas Canvas)
{
	local string CachedString;
	local float Width_Character, StringHeight, Width_Bar, Height_Bar, Size_Portrait, OffsetX_Side, Y_Text, MidX, MidY;

	// Local Player Name

	Canvas.Font = PlayerOwner.MyHUD.GetFontSizeIndex (Canvas, IndexSizeFont_Large);

	Canvas.StrLen ("1", Width_Character, StringHeight);

	Width_Bar = Canvas.ClipX;
	Height_Bar = StringHeight * 2.0f;
	Cached_YMetrics = Height_Bar;
	Size_Portrait = Height_Bar;

	MidX = Canvas.ClipX * 0.5f;
	MidY = Height_Bar * 0.5f;

	Y_Text = (Height_Bar * 0.5f) - (StringHeight * 0.8f);
	OffsetX_Side = StringHeight * 0.5f;

	// Draw Background

	if (PlayerOwnerPRI != SpectatedPRI)
	{
		if (SpectatedPRI.Team != None)
		{
			SpectatedPlayerTeamColor = Class'PlayerSettings'.Static.GetRelativeTeamColor (IndexTeam_Local, SpectatedPRI.Team.TeamIndex);
		}
		else if (PlayerOwnerPRI.bOnlySpectator)
		{
			SpectatedPlayerTeamColor = Class'PlayerSettings'.Default.CustomPlayerColor;
		}
		else
		{
			SpectatedPlayerTeamColor = NeutralTeamColor;
		}
	}
	else
	{
		SpectatedPlayerTeamColor = Class'PlayerSettings'.Default.CustomPlayerColor;
	}

	Canvas.DrawColor = SpectatedPlayerTeamColor;
	Canvas.DrawColor.A = 191;

	Canvas.SetPos (0.0f, 0.0f);

	Canvas.DrawTilePartialStretched
	(
		HeaderBackgroundTexture,
		Width_Bar,
		Height_Bar
	);

	// <DrawCountryCode>
	if (SpectatedPRI.CountryCode != "")
	{
		if (SpectatedPRI.TextureFlag == None)
		{
			Class'CountryCodes'.Static.GetTextureFlag (SpectatedPRI.CountryCode, SpectatedPRI.TextureFlag, SpectatedPRI.OffsetTextureFlag);
		}

		if (SpectatedPRI.TextureFlag != None)
		{
			Canvas.DrawColor = CountryFlagColor;

			Canvas.SetPos (0.0f + (OffsetX_Side * 0.5f), 0.0f);

			Canvas.DrawTile
			(
				SpectatedPRI.TextureFlag,
				Size_Portrait,
				Size_Portrait,
				SpectatedPRI.OffsetTextureFlag.X1,
				SpectatedPRI.OffsetTextureFlag.Y1,
				SpectatedPRI.OffsetTextureFlag.X2,
				SpectatedPRI.OffsetTextureFlag.Y2
			);
		}
		else
		{
			Canvas.DrawColor = CountryCodeColor;

			CachedString = SpectatedPRI.GetCountryCodeString ();

			Canvas.SetPos
			(
				0.0f + OffsetX_Side,
				Y_Text
			);

			Canvas.DrawText (CachedString, False);
		}
	}
	// </DrawCountryCode>

	// Set Player Name Font

	Canvas.Font = PlayerOwner.MyHUD.GetFontSizeIndex (Canvas, IndexSizeFont_Large);

	// Set Player Name Color

	if (PlayerOwnerPRI != SpectatedPRI)
	{
		if (SpectatedPRI.Team != None)
		{
			Canvas.DrawColor = Class'PlayerSettings'.Static.GetRelativeTeamColor (IndexTeam_Local, SpectatedPRI.Team.TeamIndex);
		}
		else
		{
			Canvas.DrawColor = NeutralTeamColor;
		}
	}
	else
	{
		Canvas.DrawColor = Class'PlayerSettings'.Default.CustomPlayerColor;
	}

	// Draw Player Name

	Canvas.Font = PlayerOwner.MyHUD.GetFontSizeIndex (Canvas, IndexSizeFont_Large + 1);

	CachedString = SpectatedPRI.PlayerName;

	Canvas.StrLen (CachedString, Width_Character, StringHeight);

	Canvas.SetPos
	(
		MidX - (Width_Character * 0.5f),
		MidY - (StringHeight * 0.5f)
	);

	Canvas.DrawText (CachedString, False);
}

// #Metrics

protected function float DrawMetrics (Canvas Canvas)
{
	local float StringWidth, StringHeight, Width_Value, Width_Highest, Width_Legend, Width_Bar, Height_Bar, X, Y, ScaledScore_Local;
	local Color ValueColor;
	local int Index;

	UpdatePlayerScores ();

	Canvas.Font = PlayerOwner.MyHUD.GetFontSizeIndex (Canvas, IndexSizeFont_Normal);

	Canvas.StrLen ("*", StringWidth, StringHeight);

	Width_Bar = Canvas.ClipX;
	Height_Bar = StringHeight * 3.0f;

	X = Canvas.ClipX - StringWidth;
	Y = Cached_YMetrics + (Height_Bar * 0.5f);

	// Score Bar

	if
	(
		!PlayerOwnerPRI.bOnlySpectator
		&&
		(PlayerOwnerPRI != SpectatedPRI)
	)
	{
		ScaledScore_Local = PlayerOwnerPRI.Score / FMax (PlayerOwnerPRI.Score + SpectatedPRI.Score, 1.0f);

		Canvas.DrawColor = Class'PlayerSettings'.Default.CustomPlayerColor;
		Canvas.DrawColor.A = ScoreBarAlpha;

		Canvas.SetPos (0.0f, Cached_YMetrics);

		Canvas.DrawTile
		(
			Metrics_Texture_Bar,
			Width_Bar * ScaledScore_Local,
			Height_Bar,
			0.0f,
			0.0f,
			Metrics_Texture_Bar.USize,
			Metrics_Texture_Bar.VSize
		);

		// Other Player

		if (SpectatedPRI.Team != None)
		{
			ValueColor = Class'PlayerSettings'.static.GetRelativeTeamColor (IndexTeam_Local, SpectatedPRI.Team.TeamIndex);
			ValueColor.A = 255;

			Canvas.DrawColor = ValueColor;
			Canvas.DrawColor.A = ScoreBarAlpha;
		}
		else
		{
			ValueColor = NeutralTeamColor;

			Canvas.DrawColor = ValueColor;
			Canvas.DrawColor.A = ScoreBarAlpha;
		}

		// Score Bar

		if ((1.0f - ScaledScore_Local) > 0)
		{
			Canvas.SetPos
			(
				Width_Bar * ScaledScore_Local,
				Cached_YMetrics
			);

			Canvas.DrawTile
			(
				Metrics_Texture_Bar,
				Canvas.ClipX - (Width_Bar * ScaledScore_Local),
				Height_Bar,
				0.0f,
				0.0f,
				Metrics_Texture_Bar.USize,
				Metrics_Texture_Bar.VSize
			);
		}

		// Score Symbol

		Canvas.DrawColor = Metrics_ColorTexture_Symbol;

		if (PlayerOwnerPRI.Score > SpectatedPRI.Score)
		{
			Canvas.SetPos
			(
				(Canvas.ClipX * 0.5f) - (Height_Bar * 0.5f),
				Cached_YMetrics
			);

			Canvas.DrawTile
			(
				Metrics_Texture_Up,
				Height_Bar,
				Height_Bar,
				Metrics_OffsetTexture_Up.X1,
				Metrics_OffsetTexture_Up.Y1,
				Metrics_OffsetTexture_Up.X2 - Metrics_OffsetTexture_Up.X1,
				Metrics_OffsetTexture_Up.Y2 - Metrics_OffsetTexture_Up.Y1
			);
		}
		else if (PlayerOwnerPRI.Score < SpectatedPRI.Score)
		{
			Canvas.SetPos
			(
				(Canvas.ClipX * 0.5f) - (Height_Bar * 0.5f),
				Cached_YMetrics
			);

			Canvas.DrawTile
			(
				Metrics_Texture_Down,
				Height_Bar,
				Height_Bar,
				Metrics_OffsetTexture_Down.X1,
				Metrics_OffsetTexture_Down.Y1,
				Metrics_OffsetTexture_Down.X2 - Metrics_OffsetTexture_Down.X1,
				Metrics_OffsetTexture_Down.Y2 - Metrics_OffsetTexture_Down.Y1
			);
		}
		else
		{
			Canvas.SetPos
			(
				(Canvas.ClipX * 0.5f) - (Height_Bar * 0.5f),
				Cached_YMetrics
			);

			Canvas.DrawTile
			(
				Metrics_Texture_Equal,
				Height_Bar,
				Height_Bar,
				Metrics_OffsetTexture_Equal.X1,
				Metrics_OffsetTexture_Equal.Y1,
				Metrics_OffsetTexture_Equal.X2 - Metrics_OffsetTexture_Equal.X1,
				Metrics_OffsetTexture_Equal.Y2 - Metrics_OffsetTexture_Equal.Y1
			);
		}
	}
	else if (PlayerOwnerPRI != SpectatedPRI)
	{
		if (SpectatedPRI.Team != None)
		{
			ValueColor = Class'PlayerSettings'.static.GetRelativeTeamColor (IndexTeam_Local, SpectatedPRI.Team.TeamIndex);
			ValueColor.A = 255;

			Canvas.DrawColor = ValueColor;
			Canvas.DrawColor.A = ScoreBarAlpha;
		}
		else
		{
			ValueColor = NeutralTeamColor;

			Canvas.DrawColor = ValueColor;
			Canvas.DrawColor.A = ScoreBarAlpha;
		}

		Canvas.SetPos (0.0f, Cached_YMetrics);

		Canvas.DrawTile
		(
			ScoreBarTexture,
			Width_Bar,
			Height_Bar,
			0.0f,
			0.0f,
			ScoreBarTexture.USize,
			ScoreBarTexture.VSize
		);
	}
	else if (!PlayerOwnerPRI.bOnlySpectator)
	{
		ValueColor = Class'PlayerSettings'.default.CustomPlayerColor;
		ValueColor.A = 255;

		Canvas.DrawColor = ValueColor;
		Canvas.DrawColor.A = ScoreBarAlpha;

		Canvas.SetPos (0.0f, Cached_YMetrics);

		Canvas.DrawTile
		(
			ScoreBarTexture,
			Width_Bar,
			Height_Bar,
			0.0f,
			0.0f,
			ScoreBarTexture.USize,
			ScoreBarTexture.VSize
		);
	}
	else // Local player is a spectator
	{
		ValueColor = ScoreColor;

		Canvas.DrawColor = Metrics_ColorTexture_Background;
		Canvas.DrawColor.A = ScoreBarAlpha;

		Canvas.SetPos (0.0f, Cached_YMetrics);

		Canvas.DrawTile
		(
			ScoreBarTexture,
			Width_Bar,
			Height_Bar,
			0.0f,
			0.0f,
			ScoreBarTexture.USize,
			ScoreBarTexture.VSize
		);
	}

	// Legend / Values

	for (Index = Metrics_Count_Used - 1; Index > -1; --Index)
	{
		// X-Centered Legend

		Canvas.DrawColor = LegendColor;

		Canvas.StrLen (ScoreLongestString[Index], Width_Highest, StringHeight);
		Canvas.StrLen (ScoreLegendString[Index], Width_Legend, StringHeight);

		if (Width_Legend > Width_Highest)
		{
			Width_Highest = Width_Legend;
		}

		Width_Highest += StringHeight; // Margin-Horizontal

		X -= Width_Highest;

		Canvas.SetPos
		(
			X + (Width_Highest * 0.5f) - (Width_Legend * 0.5f),
			Y - StringHeight
		);

		Canvas.DrawText (ScoreLegendString[Index], False);

		// X-Centered Value

		Canvas.DrawColor = ScoreColor;

		Canvas.StrLen (ScoreDataString[Index], Width_Value, StringHeight);

		Canvas.SetPos
		(
			X + (Width_Highest * 0.5f) - (Width_Value * 0.5f),
			Y
		);

		Canvas.DrawText (ScoreDataString[Index], False);
	}


	return Cached_YMetrics + Height_Bar;
}


protected function UpdatePlayerScores ();


protected function DrawBody (Canvas Canvas, optional float YBase)
{
	local string CachedString;
	local float X, Y, MidX, MidY, SizeUV, CornerSize, StringWidth, StringHeight, PortraitWidth, PortraitHeight, Ratio;

	// Initialize

	MidX = Canvas.ClipX * 0.5f;
	MidY = (Canvas.ClipY + YBase)  * 0.5f;
	SizeUV = Canvas.ClipY  * 0.55f;
	CornerSize = Canvas.ClipY * 0.35f;

	// Background

	Canvas.DrawColor = BodyBackgroundColor;

	Canvas.SetPos
	(
		0.0f,
		YBase
	);

	Canvas.DrawTileStretched
	(
		BodyBackgroundTexture,
		Canvas.ClipX,
		Canvas.ClipY - YBase
	);

	// Portraits

	if (Class'PlayerSettings'.Default.bStatisticsShowPortraits)
	{
		if // Left Portrait
		(
			!PlayerOwnerPRI.bOnlySpectator
			&&
			(PlayerOwnerPRI == SpectatedPRI)
			&&
			(LeftPortraitFinalBlend != None)
		)
		{
			LeftPortraitCombiner.Material1 = PlayerOwnerPRI.GetPortrait ();

			PortraitHeight = Canvas.ClipY - YBase; // Assumes (ClipY > YBase)
			Ratio = PortraitHeight / 512.0f;

			Canvas.DrawColor = PortraitBackgroundColor;

			Canvas.SetPos (0.0f, YBase);

			Canvas.DrawTile
			(
				LeftPortraitFinalBlend,
				256.0f * Ratio,
				PortraitHeight,
				0.0f,
				0.0f,
				256.0f - 2.0f,
				512.0f - 2.0f
			);
		}

		if // Right Portrait
		(
			(PlayerOwnerPRI != SpectatedPRI)
			&&
			(RightPortraitFinalBlend != None)
		)
		{
			RightPortraitCombiner.Material1 = SpectatedPRI.GetPortrait ();

			PortraitHeight = Canvas.ClipY - YBase; // Assumes (ClipY > YBase)
			Ratio = PortraitHeight / 512.0f; // PortraitHeight / Source Portraight Height
			PortraitWidth = 256.0f * Ratio; // Source Portrait Width * Ratio

			Canvas.DrawColor = PortraitBackgroundColor;

			Canvas.SetPos (Canvas.ClipX - PortraitWidth, YBase);

			Canvas.DrawTile
			(
				RightPortraitFinalBlend,
				PortraitWidth,
				PortraitHeight,
				2.0f,
				2.0f,
				256.0f - 2.0f,
				512.0f - 2.0f
			);
		}
	}

	// Corner Highlights

	Canvas.DrawColor = SpectatedPlayerTeamColor;
	Canvas.DrawColor.A = CornerBackgroundColor.A;

	Canvas.SetPos (0.0f, YBase);

	Canvas.DrawTile
	(
		Texture'Textures.Splatter',
		CornerSize,
		CornerSize,
		0.0f,
		0.0f,
		Texture'Textures.Splatter'.USize,
		Texture'Textures.Splatter'.VSize
	);


	Canvas.SetPos
	(
		Canvas.ClipX,
		YBase
	);

	Canvas.DrawTile
	(
		Texture'Textures.Splatter',
		-CornerSize,
		CornerSize,
		0.0f,
		0.0f,
		Texture'Textures.Splatter'.USize,
		Texture'Textures.Splatter'.VSize
	);

	// Initialize

	Canvas.Font = PlayerOwnerHUD.GetFontSizeIndex (Canvas, -3);

	Canvas.TextSize ("1", StringWidth, StringHeight);

	X = Canvas.ClipX - StringHeight;
	Y = Canvas.ClipY - (StringHeight * 2.0f);

	Canvas.DrawColor = MatchInfoColor;

	// Server Name

	if (Level.NetMode == NM_Client)
	{
		UpdateServerNameString ();

		Canvas.StrLen (ServerNameString, StringWidth, StringHeight);
		Canvas.SetPos (X - StringWidth, Y);
		Canvas.DrawText (ServerNameString, False);

		Y -= StringHeight;
	}

	// Draw Match Type

	UpdateMatchTypeString ();

	Canvas.StrLen (MatchTypeString, StringWidth, StringHeight);
	Canvas.SetPos (X - StringWidth, Y);
	Canvas.DrawText (MatchTypeString, False);

	Y -= StringHeight;

	// Time

	CachedString = Class'ReUtility'.Static.GetLocalTimeString (Level);

	Canvas.StrLen (CachedString, StringWidth, StringHeight);
	Canvas.SetPos (X - StringWidth, Y);
	Canvas.DrawText (CachedString, False);

	Y -= StringHeight;

	// Level Title

	UpdateLevelTitleString ();

	Canvas.StrLen (LevelTitleString, StringWidth, StringHeight);
	Canvas.SetPos (X - StringWidth, Y);
	Canvas.DrawText (LevelTitleString, False);

	Y -= StringHeight;

	// Game Type

	UpdateGameString ();

	Canvas.StrLen (GameString, StringWidth, StringHeight);
	Canvas.SetPos (X - StringWidth, Y);
	Canvas.DrawText (GameString, False);
}

// #Weapons

protected function float DrawWeapons (Canvas Canvas, optional float YBase)
{
	local float BarX, BarU, StartX, OffsetY, StringWidth, StringHeight, Width_Value, Width_Highest, Width_Legend, T, X[8];
	local int Index, IndexWeapon, IndexWeaponDamageType, Count;

	if (SpectatedPRI.ClientWeaponRecord.Length == 0)
	{
		return YBase;
	}

	Canvas.Font = PlayerOwnerHUD.GetFontSizeIndex (Canvas, IndexSizeFont_Normal);

	// Legend

	Canvas.StrLen ("*", StringWidth, StringHeight);

	YBase += Cached_HeightFont_Normal * 0.5f;

	//

	BarU = Cached_UWeapons + StringHeight + StringHeight;
	BarX = (Canvas.ClipX * 0.5f) - (BarU * 0.5f);

	StartX = (Canvas.ClipX * 0.5f) + (Cached_UWeapons * 0.5f);

	// Goes right-to-left.

	Canvas.DrawColor = LegendColor;

	for (IndexWeapon = Weapons_Count_Used - 1; IndexWeapon > -1; --IndexWeapon)
	{
		Canvas.StrLen (Weapons_Text_Highest[IndexWeapon], Width_Highest, StringHeight);
		Canvas.StrLen (WeaponLegendString[IndexWeapon], Width_Legend, StringHeight);

		if (Width_Legend > Width_Highest)
		{
			Width_Highest = Width_Legend;
		}

		Width_Highest += StringWidth; // Adds a single character for padding on each side of field when centered.

		T += Width_Highest;

		StartX -= Width_Highest; // StartX moves toward the left by Width_Highest each iteration.

		X[IndexWeapon] = StartX + (Width_Highest * 0.5f); // Store the centered x-position.

		// Legend

		Canvas.SetPos
		(
			X[IndexWeapon] - (Width_Legend * 0.5f),
			YBase
		);

		Canvas.DrawText (WeaponLegendString[IndexWeapon], False);
	}

	Cached_UWeapons = T;

	// Values

	OffsetY = YBase + StringHeight;

	Count = SpectatedPRI.ClientWeaponRecord.Length;

	//

	for (IndexWeapon = 0; IndexWeapon < Count; ++IndexWeapon)
	{
		if (!SpectatedPRI.ClientWeaponRecord[IndexWeapon].bHasData)
		{
			continue;
		}
		else
		{
			Canvas.DrawColor = Weapons_ColorText_Weapon;
			Canvas.DrawColor.A = 191;

			Canvas.SetPos
			(
				Canvas.ClipX * 0.26f,
				OffsetY
			);

			Canvas.DrawTilePartialStretched
			(
				Texture'Textures.RowHighlight',
				Canvas.ClipX * 0.48f,
				StringHeight
			);


			Weapons_UpdateSpectatedWeaponValues (SpectatedPRI.ClientWeaponRecord[IndexWeapon]);

			StartX = Cached_UWeapons;

			Canvas.DrawColor = Weapons_ColorText_Weapon;

			for (Index = Weapons_Count_Used - 1; Index > -1; --Index)
			{
				Canvas.StrLen (Weapons_Text_Value[Index], Width_Value, StringHeight);

				Canvas.SetPos
				(
					X[Index] - (Width_Value * 0.5f),
					OffsetY
				);

				Canvas.DrawText (Weapons_Text_Value[Index], False);
			}

			OffsetY += StringHeight; // Newline

			// Weapon Damage Type (Subtotals)

			Canvas.DrawColor = Weapons_ColorText_Value;

			for
			(
				IndexWeaponDamageType = 0;
				IndexWeaponDamageType < SpectatedPRI.ClientWeaponRecord[IndexWeapon].WeaponDamageRecord.Length;
				++IndexWeaponDamageType
			)
			{
				if (!SpectatedPRI.ClientWeaponRecord[IndexWeapon].WeaponDamageRecord[IndexWeaponDamageType].bHasData)
				{
					continue;
				}
				else
				{
// 					Canvas.DrawColor = SpectatedPRI.ClientWeaponRecord[IndexWeapon].Class_Weapon.Default.HUDColor * 0.9f;

					Weapons_UpdateSpectatedWeaponDamageTypeValues (SpectatedPRI.ClientWeaponRecord[IndexWeapon].WeaponDamageRecord[IndexWeaponDamageType]);

					for (Index = Weapons_Count_Used - 1; Index > -1; --Index)
					{
						Canvas.StrLen (Weapons_Text_Value[Index], Width_Value, StringHeight);

						Canvas.SetPos
						(
							X[Index] - (Width_Value * 0.5f),
							OffsetY
						);

						Canvas.DrawText (Weapons_Text_Value[Index], False);
					}

					OffsetY += StringHeight; // Newline
				}
			}
		}
	}

	return OffsetY;
}

// - Using out to avoid copying.
// Postconditions:
// - Out_ReadOnly_WeaponRecord contents remain unchanged.

protected function Weapons_UpdateSpectatedWeaponValues (out AwardPRI.SWeaponRecord Out_ReadOnly_WeaponRecord)
{
}

// - Using out to avoid copying.
// Postconditions:
// - Out_ReadOnly_WeaponDamageRecord contents remain unchanged.

protected function Weapons_UpdateSpectatedWeaponDamageTypeValues (out AwardPRI.SWeaponDamageRecord Out_ReadOnly_WeaponDamageRecord)
{
}


protected function float DrawCombos (Canvas Canvas, optional float YBase)
{
	local string CachedString;
	local float X, StartX, Y, StringWidth, StringHeight;
	local int Index, Count;

	Count = SpectatedPRI.Statistics_AdrenalineCombo.Length;

	if (Count == 0)
	{
		return YBase;
	}

	Canvas.Font = PlayerOwnerHUD.GetFontSizeIndex (Canvas, IndexSizeFont_Normal);

	Canvas.StrLen ("*", StringWidth, StringHeight);

	X = Canvas.ClipX - Cached_WidthFont_Normal;
	Y = YBase;

	// Draw Background

	Canvas.DrawColor = BodySectionColor;

	Canvas.SetPos
	(
		X - RightSectionTexture.USize,
		Y + (StringHeight * 0.5f)
	);

	Canvas.DrawTileStretched
	(
		RightSectionTexture,
		RightSectionTexture.USize,
		(Count + 1) * StringHeight
	);

	X -= Cached_WidthFont_Normal;
	Y += Cached_HeightFont_Normal * 0.5f; // Newline

	// Legend

	Canvas.DrawColor = LegendColor;

	CachedString = CombosString @ "(" $ SpectatedPRI.Statistics_AdrenalineCombo.Length $ ")";

	Canvas.StrLen (CachedString, StringWidth, StringHeight);

	Canvas.SetPos
	(
		X - StringWidth,
		Y
	);

	Canvas.DrawText (CachedString, False);

	//

	X -= Cached_WidthFont_Normal;
	Y += StringHeight; // Newline

	StartX = X;

	// Content

	Canvas.DrawColor = ContentColor;

	if (Count > 0)
	{
		for (Index = 0; Index < SpectatedPRI.Statistics_AdrenalineCombo.Length; ++Index)
		{
			X = StartX;

			Canvas.DrawColor = Class'Award'.Default.AwardMultiplierColor;

			CachedString = "x" $ SpectatedPRI.Statistics_AdrenalineCombo[Index].Count_Activations;

			Canvas.StrLen (CachedString, StringWidth, StringHeight);

			X -= StringWidth;

			Canvas.SetPos (X, Y);

			Canvas.DrawText (CachedString, False);

			//

			Canvas.DrawColor = ContentColor;

			CachedString
			=
			Left (SpectatedPRI.Statistics_AdrenalineCombo[Index].Class_Combo.Default.ExecMessage, 20)
			$ " "
			;

			Canvas.StrLen (CachedString, StringWidth, StringHeight);

			X -= StringWidth;

			Canvas.SetPos (X, Y);

			Canvas.DrawText (CachedString, False);

			Y += StringHeight; // Newline
		}
	}

	return Y;
}

// #Awards

protected function float DrawAwards (Canvas Canvas, optional float YBase)
{
	local float X, Y, StringWidth, StringHeight;
	local int Index, Count;

	Count = SpectatedPRI.AwardRecord.Length;

	if (Count == 0)
	{
		return YBase;
	}

	Canvas.Font = PlayerOwnerHUD.GetFontSizeIndex (Canvas, IndexSizeFont_Normal);

	Canvas.StrLen ("*", StringWidth, StringHeight);

	X = Cached_XAwards;
	Y = YBase;

	// Draw Background

	Canvas.DrawColor = BodySectionColor;

	Canvas.SetPos (X, Y + (StringHeight * 0.5f));

	Canvas.DrawTileStretched
	(
		LeftSectionTexture,
		LeftSectionTexture.USize,
		(Count + 1) * StringHeight
	);

	X += Cached_WidthFont_Normal;

	Y += Cached_HeightFont_Normal * 0.5f;

	// Legend

	Canvas.DrawColor = LegendColor;

	Canvas.SetPos
	(
		X,
		Y
	);

	Canvas.DrawText (AwardsString @ "(" $ SpectatedPRI.AwardRecord.Length $ ")", False);

	X += Cached_WidthFont_Normal;
	Y += StringHeight;

	for (Index = 0; Index < Count; ++Index)
	{
		Canvas.DrawColor = SpectatedPRI.AwardRecord[Index].CaptionColor;

		Canvas.SetPos (X, Y);

		if (SpectatedPRI.AwardRecord[Index].Award_Multiplier > 1)
		{
			Canvas.DrawText
			(
				SpectatedPRI.AwardRecord[Index].Class_Award.Static.GetCaptionString
				(
					SpectatedPRI.AwardRecord[Index].Award_Index
				)
				@ TextColorCode_Multiplier
				$ "x"
				$ SpectatedPRI.AwardRecord[Index].Award_Multiplier,
				False
			);
		}
		else
		{
			Canvas.DrawText
			(
				SpectatedPRI.AwardRecord[Index].Class_Award.static.GetCaptionString
				(
					SpectatedPRI.AwardRecord[Index].Award_Index
				),
				False
			);
		}

		Y += StringHeight;
	}

	return Y;
}

// #

function RequestStats (RePRI RequestedPRI)
{
	if (RequestedPRI != None)
	{
		SpectatedPRI = RequestedPRI;
	}
}

// #

event NextStats ()
{
	local int IndexOuter;
	local int IndexInner;

	if
	(
		(PlayerOwner == None)
		||
		(PlayerOwner.GameReplicationInfo == None)
	)
	{
		return;
	}

	for
	(
		IndexOuter = 0;
		IndexOuter < (PlayerOwner.GameReplicationInfo.PRIArray.Length - 1);
		++IndexOuter
	)
	{
		if (SpectatedPRI == PlayerOwner.GameReplicationInfo.PRIArray[IndexOuter])
		{
			for
			(
				IndexInner = IndexOuter + 1;
				IndexInner < PlayerOwner.GameReplicationInfo.PRIArray.Length;
				++IndexInner
			)
			{
				SpectatedPRI = RePRI(PlayerOwner.GameReplicationInfo.PRIArray[IndexInner]);

				if
				(
					(SpectatedPRI != None)
					&&
					!SpectatedPRI.bOnlySpectator
				)
				{
					LastRequestAcceptedTime = 0;

					return;
				}
			}
		}
	}

	SpectatedPRI = RePRI(PlayerOwner.GameReplicationInfo.PRIArray[0]);

	LastRequestAcceptedTime = 0;
}

// #

event DrawScoreboard (Canvas Canvas)
{
	if
	(
		UpdateGRI ()
		&&
		(PlayerOwnerPRI != None)
		&&
		UpdateMetrics ()
	)
	{
		if
		(
			(PlayerOwnerPRI.Team == None)
			||
			PlayerOwnerPRI.bOnlySpectator
		)
		{
			IndexTeam_Local = IndexTeam_Unspecified;
		}
		else
		{
			IndexTeam_Local = PlayerOwnerPRI.Team.TeamIndex;
		}


		UpdateScoreboard (Canvas);
	}
	else
	{
		if (PlayerOwner == None)
		{
			PlayerOwner = UnrealPlayer(Level.GetLocalPlayerController ());
		}


		if (PlayerOwner != None)
		{
			PlayerOwnerPRI = RePRI(PlayerOwner.PlayerReplicationInfo);
			PlayerOwnerHUD = PlayerOwner.MyHUD;
		}
		else
		{
			PlayerOwnerPRI = None;
		}
	}
}

event bool UpdateGRI ()
{
	local int IndexSource;
	local int IndexDestination;
	local PlayerReplicationInfo PRI;

	if (GRI != None)
	{
		for (IndexSource = 0; IndexSource < (GRI.PRIArray.Length - 1); ++IndexSource)
		{
			for (IndexDestination = IndexSource + 1; IndexDestination < GRI.PRIArray.Length; ++IndexDestination)
			{
				if (!InOrder (GRI.PRIArray[IndexSource], GRI.PRIArray[IndexDestination]))
				{
					PRI = GRI.PRIArray[IndexSource];
					GRI.PRIArray[IndexSource] = GRI.PRIArray[IndexDestination];
					GRI.PRIArray[IndexDestination] = PRI;
				}
			}
		}

		return true;
	}
	else
	{
		InitGRI ();
	}


	return False;
}

protected function UpdateMatchTypeString ()
{
	if (Level.NetMode == NM_Standalone)
	{
		if (Level.Game.CurrentGameProfile != None)
		{
			MatchTypeString = StandaloneMatchString @ "(" $ Class'XInterface.ScoreboardDeathMatch'.Default.SkillLevel[Clamp (Level.Game.CurrentGameProfile.BaseDifficulty, 0, 7)] $ ")";
		}
		else
		{
			MatchTypeString = StandaloneMatchString @ "(" $ Class'XInterface.ScoreboardDeathMatch'.Default.SkillLevel[Clamp (Level.Game.GameDifficulty, 0, 7)] $ ")";
		}
	}
	else if
	(
		(GRI != None)
		&&
		(GRI.bMatchHasBegun)
	)
	{
		if (GRI.MatchID != 0)
		{
			MatchTypeString = RankedMatchIDString $ GRI.MatchID;
		}
		else
		{
			MatchTypeString = UnrankedMatchString;
		}


		if (GRI.BotDifficulty >= 0)
		{
			MatchTypeString @= "(" $ Class'XInterface.ScoreboardDeathMatch'.Default.SkillLevel[Clamp (GRI.BotDifficulty, 0, 7)] $ ")";
		}
	}
	else
	{
		MatchTypeString = "";
	}
}

protected function UpdateGameString ()
{
	if
	(
		(GameString == "")
		&&
		(GRI != None)
	)
	{
		GameString = Caps (GRI.GameName);
	}
}

protected function UpdateLevelTitleString ()
{
	if
	(
		(LevelTitleString == "")
		&&
		(GRI != None)
	)
	{
		LevelTitleString = Class'ReUtility'.static.GetPlainString (Level.Title);
	}
}

protected function bool UpdateMetrics ()
{
	return True;
}

// Preconditions: (Level.NetMode == NM_Client)
// - (Reminder) GRI resets ServerName if this is a client machine.

protected function UpdateServerNameString ()
{
	if
	(
		(ServerNameString == "")
		&&
		(GRI != None)
	)
	{
		ServerNameString = Class'ReUtility'.Static.GetPlainString (GRI.ServerName);
	}
}

// #

protected function UpdateServerMessageString ()
{
	if
	(
		(ServerMessageString == "")
		&&
		(GRI != None)
	)
	{
		ServerMessageString = Class'ReUtility'.Static.GetPlainString (GRI.MessageoftheDay);

		ServerMessageString = Repl (GRI.MessageoftheDay, Chr (160), ' ');
	}
}

// #

DefaultProperties
{
	IndexSizeFont_Large = 0
	IndexSizeFont_Normal = -3
	IndexSizeFont_Small = -4

	TeamSymbolColor = (R=255,G=255,B=255,A=15)

	SpectatorSymbolTexture = Texture'Textures.Symbols'
	OffsetTexture_Spectator = (X1=768,Y1=512,X2=1024,Y2=768)

	// Header

	HeaderOpacity = 255

	HeaderBackgroundColor = (R=255,G=255,B=255,A=223)
	HeaderBackgroundTexture = Texture'Textures.RowHighlight'

	NeutralTeamColor = (R=255,G=255,B=0,A=255)

	Metrics_ColorTexture_Background = (R=127,G=127,B=127,A=223)
	ScoreBarTexture = Texture'Textures.ScoreBar'
	ScoreBarAlpha = 191
	Metrics_Texture_Bar = Texture'Textures.ScoreBar'

	Metrics_ColorTexture_Symbol = (R=255,G=255,B=255,A=127)
	Metrics_Texture_Up = Texture'Textures.Symbols'
	Metrics_OffsetTexture_Up = (X1=960,Y1=0,X2=1024,Y2=64)
	Metrics_Texture_Down = Texture'Textures.Symbols'
	Metrics_OffsetTexture_Down = (X1=0,Y1=64,X2=64,Y2=128)
	Metrics_Texture_Equal = Texture'Textures.Symbols'
	Metrics_OffsetTexture_Equal = (X1=768,Y1=0,X2=832,Y2=64)

	BodyBackgroundColor = (R=16,G=26,B=16,A=159)
	BodyBackgroundTexture = Texture'Engine.WhiteTexture'

	CornerBackgroundColor = (R=64,G=136,B=64,A=255)

	BodySectionColor = (R=255,G=255,B=255,A=127)

	LeftSectionTexture = Texture'Textures.SectionYellowLeft'
	RightSectionTexture = Texture'Textures.SectionYellowRight'

	// Colors

	CountryCodeColor = (R=178,G=178,B=0,A=255)
	CountryFlagColor = (R=255,G=255,B=255,A=150)
	MatchInfoColor = (R=64,G=192,B=64,A=143)
	CaptionColor = (R=255,G=255,B=255,A=255)
	ContentColor = (R=223,G=223,B=223,A=255)
	LegendColor = (R=255,G=255,B=255,A=255)
	ScoreColor = (R=255,G=255,B=191,A=255)

	PortraitBackgroundColor = (R=255,G=255,B=255,A=255)

	Weapons_ColorText_Weapon = (R=135,G=255,B=135,A=255)
	Weapons_ColorText_Value = (R=255,G=255,B=191,A=255)


	AwardsString = "Awards"
	CombosString = "Combos"
	WeaponString = "Weapon"

	WeaponDamageTypeMapClass = Class'WeaponDamageTypeMap'


	StandaloneMatchString = "Standalone Match"
	RankedMatchIDString = "Ranked Match ID #"
	UnrankedMatchString = "Unranked Match"
}
