// *F* Developed 2011-2014 rejecht <rejecht at outlook dot com>

// - To allow changes to ammunition.
// - To inject subclasses used to track weapons fire for statistics.

class ReTAM_Mutator extends DMMutator
	CacheExempt
	config
	HideDropDown
	;

// #CheckRelevance #CheckReplacement

const bRemove = False;
const bKeep = True;

// #

var(Replicated) bool
	bKeepDoubleDamage,
	bKeepRedeemer,
	bKeepSuperShield,
	bKeepSuperHealth;

// #

replication
{
	reliable if (bNetInitial && (Role == ROLE_Authority))
		bKeepDoubleDamage,
		bKeepRedeemer,
		bKeepSuperShield,
		bKeepSuperHealth;
}

// #

var class<PlayerReplicationInfo> PlayerReplicationInfoClass;

// #Lifetime

event PreBeginPlay ()
{
	Super.PreBeginPlay ();


	bKeepDoubleDamage = ReTAM(Level.Game).bKeepDoubleDamage;
	bKeepRedeemer = ReTAM(Level.Game).bKeepRedeemer;
	bKeepSuperShield = ReTAM(Level.Game).bKeepSuperShield;
	bKeepSuperHealth = ReTAM(Level.Game).bKeepSuperHealth;
}


simulated event BeginPlay ()
{
	DestroyAllWeaponLockers ();


	Super.BeginPlay();
}

// - Wait for bNetInitial replicated state before acting.

simulated event PostNetBeginPlay ()
{
	if (Role < ROLE_Authority)
	{
		HideAllPickupBases ();
    }

	Super.PostNetBeginPlay ();
}

// #Protocol

event bool MutatorIsAllowed ()
{
	return
	(
		Super.MutatorIsAllowed ()
		&&
		(ReTAM(Level.Game) != None)
	);
}

// #

protected simulated function bool KeepPickup (Actor Other)
{
	return
	(
		(
			(
				Other.IsA ('UDamagePack')
				||
				Other.IsA ('UDamageReward')
			)
			&&
			bKeepDoubleDamage
		)
		||
		(
			Other.IsA ('RedeemerPickup')
			&&
			bKeepRedeemer
		)
		||
		(
			Other.IsA ('SuperHealthPack')
			&&
			bKeepSuperHealth
		)
		||
		(
			Other.IsA ('SuperShieldPack')
			&&
			bKeepSuperShield
		)
	);
}

protected simulated function bool KeepPickupBase (Actor Other)
{
	return
	(
		(
			Other.IsA ('UDamageCharger')
			&&
			bKeepDoubleDamage
		)
		||
		(
			Other.IsA ('xWeaponBase')
			&&
			(Class<Redeemer>(xWeaponBase(Other).WeaponType) != None)
			&&
			bKeepRedeemer
		)
		||
		(
			Other.IsA ('SuperHealthCharger')
			&&
			bKeepSuperHealth
		)
		||
		(
			Other.IsA ('SuperShieldCharger')
			&&
			bKeepSuperShield
		)
	);
}

// - Client reflects server-side removal of pickup bases.

protected simulated function HideAllPickupBases ()
{
	local xPickupBase Other;

	ForEach AllActors (Class'Engine.xPickupBase', Other)
	{
		if
		(
			!Other.bGameRelevant
			&&
			!KeepPickupBase (Other)
		)
		{
			if (Other.myEmitter != None)
			{
				Other.myEmitter.Destroy ();
				Other.myEmitter = None;
			}


			if (!Other.bHidden)
			{
				Other.bHidden = True;
				Other.ResetStaticFilterState ();
            }


			if (Other.myPickup != None)
			{
				Other.myPickup.GotoState ('Disabled');
				Other.Powerup = None;
			}
        }
	}
}

protected simulated function DestroyAllWeaponLockers ()
{
	local WeaponLocker WL;

	ForEach AllActors (Class'XWeapons.WeaponLocker', WL)
	{
		WL.GotoState ('Disabled'); // Destroys any dynamically spawned WeaponLocker and its Effects, sets bHidden, disables collision, calls ResetStaticFilterState.
	}
}

// Actor.PreBeginPlay, unless (Actor.bGameRelevant == True), calls CheckRelevance which
// eventually calls this function, and if True, the Actor destroys itself.
// Protocol:
// - True keeps the actor.
// - False destroys the actor.

event bool CheckReplacement (Actor Other, out byte Out_bSuperRelevant)
{
	Out_bSuperRelevant = 0;

	if
	(
		Other.IsA ('Controller')
		&&
		Controller(Other).bIsPlayer // Applies to usable controllers only.
	)
	{
		Controller(Other).PlayerReplicationInfoClass = PlayerReplicationInfoClass;
	}
	else if (!Level.bStartup)
	{
		if (Other.IsA ('Pickup')) // Remove Pickup
		{
			if (KeepPickup (Other))
			{
				return bKeep;
            }
			else if
			(
				Other.bStatic
				||
				Other.bNoDelete
			)
			{
				Other.GotoState ('Disabled');
			}

			return bRemove;
		}
	}
	else
	{
		if (Other.IsA ('xPickupBase')) // // Remove Pickup Base
		{
			if (KeepPickupBase (Other))
			{
				return bKeep;
			}
			else
			{
				if
				(
					!Other.bHidden
					&&
					Other.bStatic
				)
				{
					Other.bHidden = True;
					Other.ResetStaticFilterState ();
				}
            }

			return bRemove;
		}

		// Remove Pickup

		if (Other.IsA ('Pickup'))
		{
			if (KeepPickup (Other))
			{
				return bKeep;
            }
			else if
			(
				Other.bStatic
				||
				Other.bNoDelete
			)
			{
				Other.GotoState ('Disabled');
			}

			return bRemove;
		}
    }


	return bKeep; // Keep
}

// #

DefaultProperties
{
	PlayerReplicationInfoClass = Class'RePRI'

	bKeepDoubleDamage = False
	bKeepSuperShield = False
	bKeepSuperHealth = False

	//

	bAlwaysRelevant = True
	bNetTemporary = True // Close Replication Channel

	RemoteRole = ROLE_SimulatedProxy // Replicate
}
