// *F* Developed 2011-2014 rejecht <rejecht at outlook dot com>

class ReTDM_HUD extends ReTeamGameHUD
	config (User)
	HideDropDown
	;

// #

const Symbol_Left = 0;
const Symbol_Right = 1;

var protected const Texture Texture_Symbols;

var(Opacity) GlobalConfig byte TeamInfoOpacity, SpectatedTeamInfoOpacity;

var(Colors) GlobalConfig Color TeammateRadar_BackgroundTextureColor;
var protected const Texture TeamRadar_Texture_Level;
var protected const IntBox TeamRadar_OffsetTexture_Level;

var protected const float
	TeamScoreBackgroundAlternativePosX,
	TeamScoreBackgroundDiscAlternativePosX,
	TeamSymbolsAlternativePosX,
	TeamSymbolsAlternativeOffsetX[IndexTeam_Max],
	ScoreTeamAlternativePosX,
	ScoreTeamAlternativeOffsetX[IndexTeam_Max];

// #TeamName

var protected const float TeamName_ScaledPosition_Y_Base;

var(Colors) GlobalConfig Color
	TeamName_Color_Playing,
	TeamName_Color_Dead,
	TeamName_Color_Spawning,
	TeamName_Color_Out,
	Teamname_ColorTexture_Background,
	TeamName_Color_LocationName;

var protected const Texture Teamname_Texture_Background;

// #TeamName #Teammates

var(TickGlocals) protected string DrawTeamInfo_L_CachedString;

var(TickGlocals) protected float
	DrawTeamInfo_L_Height_Row,
	DrawTeamInfo_L_Width_Character,
	DrawTeamInfo_L_Width_Text,
	DrawTeamInfo_L_Height_Text,
	DrawTeamInfo_L_Width_Slot,
	DrawTeamInfo_L_PlayerNameCharacters,
	DrawTeamInfo_L_LocationNameCharacters,
	DrawTeamInfo_L_Height_Slot,
	DrawTeamInfo_L_X_Status,
	DrawTeamInfo_L_Y_Slot,
	DrawTeamInfo_L_X_PlayerName,
	DrawTeamInfo_L_X_CountryFlag,
	DrawTeamInfo_L_X_CountryText,
	DrawTeamInfo_L_Y_Text,
	DrawTeamInfo_L_X_LocationName,
	DrawTeamInfo_L_Y_LocationName,
	DrawTeamInfo_L_OffsetY_Text,
	DrawTeamInfo_L_X_Health;

var(TickGlocals) protected int
	DrawTeamInfo_L_Count,
	DrawTeamInfo_L_Drawn,
	DrawTeamInfo_L_Index,
	DrawTeamInfo_L_Rows;

var(TickGlocals) protected byte
	DrawTeamInfo_L_Opacity;

// #TeamName #Enemies

var(TickGlocals) protected string DrawEnemyInfo_L_CachedString;

var(TickGlocals) protected float
	DrawEnemyInfo_L_Width_Text,
	DrawEnemyInfo_L_Width_Character,
	DrawEnemyInfo_L_Height_Text,
	DrawEnemyInfo_L_Width_Slot,
	DrawEnemyInfo_L_Height_Slot,
	DrawEnemyInfo_L_Height_Row,
	DrawEnemyInfo_L_PlayerNameCharacters,
	DrawEnemyInfo_L_X_Slot,
	DrawEnemyInfo_L_Y_Slot,
	DrawEnemyInfo_L_X_Background,
	DrawEnemyInfo_L_Y_Text,
	DrawEnemyInfo_L_X_Status,
	DrawEnemyInfo_L_X_CountryFlag,
	DrawEnemyInfo_L_X_CountryText,
	DrawEnemyInfo_L_X_PlayerName,
	DrawEnemyInfo_L_OffsetY_Text;

var(Glocals) protected int
	DrawEnemyInfo_L_Index,
	DrawEnemyInfo_L_Rows;

// #TeamResources

var() protected SpriteWidget TeamResourceBackgroundDisc[IndexTeam_Max];
var() protected SpriteWidget TeamPlayerCountOverlay[IndexTeam_Max];
var() protected SpriteWidget TeamResourceSymbol[IndexTeam_Max];
var() protected NumericWidget TeamResources[IndexTeam_Max];
var() protected SpriteWidget TeamResourceSymbolOverlay[IndexTeam_Max];
var() protected SpriteWidget TeamResourceFinal[IndexTeam_Max]; // Flashes a ! when the respawn pool is empty.

// #

var protected array<RePRI>
	Teammates,
	Enemies;

var protected int
	TeamPlayerCount[IndexTeam_Max],
	TeamResourceCount[IndexTeam_Max],
	MaxTeamResourceCount[IndexTeam_Max];

// #

protected function UpdateReset ()
{
	Super.UpdateReset ();

	Teammates.Length = 0;
	Enemies.Length = 0;

	TeamPlayerCount[IndexTeam_A] = 0;
	TeamPlayerCount[IndexTeam_B] = 0;

	ResetTeamResources ();
}

// Preconditions:
// - (OtherRePRI != None)
// - (!OtherRePRI.bOnlySpectator)
// - (OtherRePRI.Team != None))
// - (OtherRePRI.Team.TeamIndex < IndexTeam_Max)

protected function UpdatePlayer (RePRI OtherRePRI)
{
	Super.UpdatePlayer (OtherRePRI);


	if (OtherRePRI.Team.TeamIndex == TeammateTeamIndex)
	{
		Teammates[Teammates.Length] = OtherRePRI;


		if (OtherRePRI.RealtimePlayerInfo != None)
		{
			if (ShouldDrawTeamInfo (OtherRePRI))
			{
				OtherRePRI.TeamName_TimeSeconds_LastRelevant = Level.TimeSeconds;
			}
		}
	}
	else
	{
		Enemies[Enemies.Length] = OtherRePRI;


		if (OtherRePRI.RealtimePlayerInfo != None)
		{
			if (TeamName_ShouldDrawEnemy (OtherRePRI))
			{
				OtherRePRI.TeamName_TimeSeconds_LastRelevant = Level.TimeSeconds;
			}
		}
	}


	CalculatePlayerResources (OtherRePRI);
}

// #Resources

protected function ResetTeamResources () 
{
	TeamResourceCount[IndexTeam_A] = 0;
	TeamResourceCount[IndexTeam_B] = 0;

	MaxTeamResourceCount[IndexTeam_A] = 0;
	MaxTeamResourceCount[IndexTeam_B] = 0;
}

protected function CalculatePlayerResources (RePRI OtherRePRI)
{
	local int OtherTeamIndex;

	OtherTeamIndex = OtherRePRI.Team.TeamIndex;

	// Count how many lives remain

	if (!OtherRePRI.bOutofLives)
	{
		if (ReGRI.MaxLives > 1)
		{
			TeamResourceCount[OtherTeamIndex] += ReGRI.MaxLives - OtherRePRI.Deaths; // *X* Bug when bCountTeamKills == False, NumLives and Deaths get imbalanced, because NumLives is not replicated.
		}
		else
		{
			++TeamResourceCount[OtherTeamIndex];
		}
	}

	++TeamPlayerCount[OtherTeamIndex];
}

// Called after running through all players.

protected function CalculateTeamResources ()
{
	local int MaxLives;

	MaxLives = ReGRI.MaxLives;

	if (MaxLives > 0)
	{
		MaxTeamResourceCount[TeammateTeamIndex] = TeamPlayerCount[TeammateTeamIndex] * MaxLives;
		MaxTeamResourceCount[EnemyTeamIndex] = TeamPlayerCount[EnemyTeamIndex] * MaxLives;
	}
}

// - To update anything common to players, spectators, and spectating players.
// - To do the relative team coloring.
// Callers:
// - DrawHUD () before drawing.
// - LinkActors () before returning.

event UpdateTeamHUD ()
{
	local GameReplicationInfo GRI;
	local Color TeamColor;


	if (PlayerOwner != None)
	{
		GRI = PlayerOwner.GameReplicationInfo;
	}


	if (GRI != None)
	{
		// A

		if (GRI.Teams[TeammateTeamIndex] != None)
		{
			ScoreTeam[Symbol_Left].Value = Min (GRI.Teams[TeammateTeamIndex].Score, 999); // Left
		}

		TeamResources[Symbol_Left].Value = Min (TeamResourceCount[TeammateTeamIndex], 999);
		TeamResourceSymbolOverlay[Symbol_Left].Scale = float(TeamResourceCount[TeammateTeamIndex]) / float(MaxTeamResourceCount[TeammateTeamIndex]);
		TeamSymbol_Update (GRI, Symbol_Left, TeammateTeamIndex);

		// B

		if (GRI.Teams[EnemyTeamIndex] != None)
		{
			ScoreTeam[Symbol_Right].Value = Min (GRI.Teams[EnemyTeamIndex].Score, 999); // Right
		}

		TeamResources[Symbol_Right].Value = Min (TeamResourceCount[EnemyTeamIndex], 999);
		TeamResourceSymbolOverlay[Symbol_Right].Scale = float(TeamResourceCount[EnemyTeamIndex]) / float(MaxTeamResourceCount[EnemyTeamIndex]);
		TeamSymbol_Update (GRI, Symbol_Right, EnemyTeamIndex);
	}


	if (TeammateTeamIndex == IndexTeam_A)
	{
		if (!bUsingCustomHUDColor)
		{
			HUDColorRed = TeammateColor;
			HUDColorBlue = EnemyColor;

			HUDColorTeam[IndexTeam_A] = TeammateColor;
			HUDColorTeam[IndexTeam_B] = EnemyColor;
		}
		else
		{
			HUDColorRed = CustomHUDColor;
			HUDColorBlue = CustomHUDColor;

			HUDColorTeam[IndexTeam_A] = TeammateColor;
			HUDColorTeam[IndexTeam_B] = EnemyColor;
		}
	}
	else if (TeammateTeamIndex == IndexTeam_B)
	{
		if (!bUsingCustomHUDColor)
		{
			HUDColorRed = EnemyColor;
			HUDColorBlue = TeammateColor;

			HUDColorTeam[IndexTeam_A] = EnemyColor;
			HUDColorTeam[IndexTeam_B] = TeammateColor;
		}
		else
		{
			HUDColorRed = CustomHUDColor;
			HUDColorBlue = CustomHUDColor;

			HUDColorTeam[IndexTeam_A] = TeammateColor;
			HUDColorTeam[IndexTeam_B] = EnemyColor;
		}
	}

	// Team Symbols

	TeamColor = TeammateColor;

	TeamColor.A = TeamSymbols[0].Tints[IndexTeam_A].A;
	TeamSymbols[0].Tints[IndexTeam_A] = TeamColor;
	TeamSymbols[0].Tints[IndexTeam_B] = TeamColor;

	TeamColor.A = TeamResourceSymbol[0].Tints[IndexTeam_A].A;
	TeamResourceSymbol[0].Tints[IndexTeam_A] = TeamColor;
	TeamResourceSymbol[0].Tints[IndexTeam_B] = TeamColor;

	TeamColor.A = TeamResourceSymbolOverlay[Symbol_Left].Tints[IndexTeam_A].A;
	TeamResourceSymbolOverlay[Symbol_Left].Tints[IndexTeam_A] = TeamColor;
	TeamResourceSymbolOverlay[Symbol_Left].Tints[IndexTeam_B] = TeamColor;
	TeamPlayerCountOverlay[0].Tints[IndexTeam_A] = TeamColor;
	TeamPlayerCountOverlay[0].Tints[IndexTeam_B] = TeamColor;

	// Enemy Colors

	TeamColor = EnemyColor;

	TeamColor.A = TeamSymbols[1].Tints[IndexTeam_A].A;
	TeamSymbols[1].Tints[IndexTeam_A] = TeamColor;
	TeamSymbols[1].Tints[IndexTeam_B] = TeamColor;

	TeamColor.A = TeamResourceSymbol[1].Tints[IndexTeam_A].A;
	TeamResourceSymbol[1].Tints[IndexTeam_A] = TeamColor;
	TeamResourceSymbol[1].Tints[IndexTeam_B] = TeamColor;

	TeamColor.A = TeamResourceSymbolOverlay[Symbol_Right].Tints[IndexTeam_A].A;
	TeamResourceSymbolOverlay[Symbol_Right].Tints[IndexTeam_A] = TeamColor;
	TeamResourceSymbolOverlay[Symbol_Right].Tints[IndexTeam_B] = TeamColor;
	TeamPlayerCountOverlay[1].Tints[IndexTeam_A] = TeamColor;
	TeamPlayerCountOverlay[1].Tints[IndexTeam_B] = TeamColor;
}

//

event DrawSpectatingHUD (Canvas Canvas)
{
	Super.DrawSpectatingHUD (Canvas);


	if
	(
		!bIsCinematic
		&&
		(ReGRI != None)
	)
	{
		UpdateTeamHUD ();


		if (ReGRI.bMatchhasBegun)
		{
			if (!bMatchOver)
			{
				ShowTeamScorePassA (Canvas);

				ShowTeamScorePassC (Canvas);
			}


			if
			(
				(PlayerOwnerPRI != None)
				&&
				!bMatchOver
			)
			{
				if
				(
					Class'PlayerSettings'.Default.bShowTeamInfo
					&&
					PlayerOwnerPRI.IsA ('RePRI')
					&&
					(RePRI(PlayerOwnerPRI).TeamSummary != None)
				)
				{
					DrawTeamInfo (Canvas);
				}

				if (Class'PlayerSettings'.Default.bShowEnemyInfo)
				{
					DrawEnemyInfo (Canvas);
				}
			}
		}
	}
}

protected function DrawSpectatingPlayerHUD (Canvas Canvas)
{
	Super.DrawSpectatingPlayerHUD (Canvas);


	if
	(
		(ReGRI != None)
		&&
		!bMatchOver
		&&
		Class'PlayerSettings'.Default.bShowTeammateRadar
		&&
		PlayerOwnerPRI.IsA ('RePRI')
		&&
		(RePRI(PlayerOwnerPRI).TeamSummary != None)
	)
	{
		DrawTeamRadar (Canvas);
	}
}

event ShowTeamScorePassA (Canvas Canvas)
{
	local GameReplicationInfo GRI;
	local bool bShowResources;

	GRI = PlayerOwner.GameReplicationInfo;

	bShowResources = Class'PlayerSettings'.Default.bShowTeamResources;

	if (bShowResources)
	{
		// Discs

		DrawWidgetAsTile (Canvas, TeamResourceBackgroundDisc[IndexTeam_A]);
		DrawWidgetAsTile (Canvas, TeamResourceBackgroundDisc[IndexTeam_B]);

		// Overlay

		if
		(
			(GRI != None)
			&&
			(GRI.MaxLives > 0)
		)
		{
			// Background

			DrawWidgetAsTile (Canvas, TeamResourceSymbol[IndexTeam_A]);
			DrawWidgetAsTile (Canvas, TeamResourceSymbol[IndexTeam_B]);

			// Numbers / Flash ! if one player remains on a team of more than one player

			if (TeamResourceCount[TeammateTeamIndex] > 1)
			{
				DrawWidgetAsTile (Canvas, TeamResourceSymbolOverlay[Symbol_Left]);
				DrawNumericWidgetAsTiles (Canvas, TeamResources[Symbol_Left], DigitsBig);
	        }
	        else if
	        (
				(TeamResourceCount[TeammateTeamIndex] == 1)
			)
	        {
				if (bMatchLive && (TeamPlayerCount[TeammateTeamIndex] > 1))
				{
					DrawWidgetAsTile (Canvas, TeamResourceFinal[Symbol_Left]);
                }
                else if (TeamPlayerCount[TeammateTeamIndex] == 1)
                {
					DrawWidgetAsTile (Canvas, TeamResourceSymbolOverlay[Symbol_Left]);
					DrawNumericWidgetAsTiles (Canvas, TeamResources[Symbol_Left], DigitsBig);
                }
	        }


			if (TeamResourceCount[EnemyTeamIndex] > 1)
			{
				DrawWidgetAsTile (Canvas, TeamResourceSymbolOverlay[Symbol_Right]);
				DrawNumericWidgetAsTiles (Canvas, TeamResources[Symbol_Right], DigitsBig);
	        }
	        else if
	        (
				(TeamResourceCount[EnemyTeamIndex] == 1)
			)
	        {
				if (bMatchLive && (TeamPlayerCount[EnemyTeamIndex] > 1))
				{
					DrawWidgetAsTile (Canvas, TeamResourceFinal[Symbol_Right]);
                }
                else if (TeamPlayerCount[EnemyTeamIndex] == 1)
                {
					DrawWidgetAsTile (Canvas, TeamResourceSymbolOverlay[Symbol_Right]);
					DrawNumericWidgetAsTiles (Canvas, TeamResources[Symbol_Right], DigitsBig);
                }
	        }
		}
		else
		{
			// Background

			DrawWidgetAsTile (Canvas, TeamPlayerCountOverlay[IndexTeam_A]);
			DrawWidgetAsTile (Canvas, TeamPlayerCountOverlay[IndexTeam_B]);

			DrawNumericWidgetAsTiles (Canvas, TeamResources[IndexTeam_A], DigitsBig);
			DrawNumericWidgetAsTiles (Canvas, TeamResources[IndexTeam_B], DigitsBig);
        }

		// Score Background

		TeamScoreBackground[IndexTeam_A].PosX = Default.TeamScoreBackground[IndexTeam_A].PosX;
		TeamScoreBackground[IndexTeam_B].PosX = Default.TeamScoreBackground[IndexTeam_B].PosX;

		TeamScoreBackgroundDisc[IndexTeam_A].PosX = Default.TeamScoreBackgroundDisc[IndexTeam_A].PosX;
		TeamScoreBackgroundDisc[IndexTeam_B].PosX = Default.TeamScoreBackgroundDisc[IndexTeam_B].PosX;

		TeamSymbols[IndexTeam_A].PosX = Default.TeamSymbols[IndexTeam_A].PosX;
		TeamSymbols[IndexTeam_B].PosX = Default.TeamSymbols[IndexTeam_B].PosX;
		TeamSymbols[IndexTeam_A].OffsetX = Default.TeamSymbols[IndexTeam_A].OffsetX;
		TeamSymbols[IndexTeam_B].OffsetX = Default.TeamSymbols[IndexTeam_B].OffsetX;

		ScoreTeam[IndexTeam_A].PosX = Default.ScoreTeam[IndexTeam_A].PosX;
		ScoreTeam[IndexTeam_B].PosX = Default.ScoreTeam[IndexTeam_B].PosX;
		ScoreTeam[IndexTeam_A].OffsetX = Default.ScoreTeam[IndexTeam_A].OffsetX;
		ScoreTeam[IndexTeam_B].OffsetX = Default.ScoreTeam[IndexTeam_B].OffsetX;
	}
	else
	{
		TeamScoreBackground[IndexTeam_A].PosX = TeamScoreBackgroundAlternativePosX;
		TeamScoreBackground[IndexTeam_B].PosX = TeamScoreBackgroundAlternativePosX;

		TeamScoreBackgroundDisc[IndexTeam_A].PosX = TeamScoreBackgroundDiscAlternativePosX;
		TeamScoreBackgroundDisc[IndexTeam_B].PosX = TeamScoreBackgroundDiscAlternativePosX;

		TeamSymbols[IndexTeam_A].PosX = TeamSymbolsAlternativePosX;
		TeamSymbols[IndexTeam_B].PosX = TeamSymbolsAlternativePosX;
		TeamSymbols[IndexTeam_A].OffsetX = TeamSymbolsAlternativeOffsetX[IndexTeam_A];
		TeamSymbols[IndexTeam_B].OffsetX = TeamSymbolsAlternativeOffsetX[IndexTeam_B];

		ScoreTeam[IndexTeam_A].PosX = ScoreTeamAlternativePosX;
		ScoreTeam[IndexTeam_B].PosX = ScoreTeamAlternativePosX;
		ScoreTeam[IndexTeam_A].OffsetX = ScoreTeamAlternativeOffsetX[IndexTeam_A];
		ScoreTeam[IndexTeam_B].OffsetX = ScoreTeamAlternativeOffsetX[IndexTeam_B];
	}

	//

	if (bShowPoints)
	{
		DrawWidgetAsTile (Canvas, TeamScoreBackground[IndexTeam_A]);
		DrawWidgetAsTile (Canvas, TeamScoreBackground[IndexTeam_B]);

		DrawWidgetAsTile (Canvas, TeamScoreBackgroundDisc[IndexTeam_A]);
		DrawWidgetAsTile (Canvas, TeamScoreBackgroundDisc[IndexTeam_B]);


		if (TeamSymbols[IndexTeam_A].WidgetTexture != None)
		{
			DrawWidgetAsTile (Canvas, TeamSymbols[IndexTeam_A]);
		}

		if (TeamSymbols[IndexTeam_B].WidgetTexture != None)
		{
			DrawWidgetAsTile (Canvas, TeamSymbols[IndexTeam_B]);
		}


	 	DrawNumericWidgetAsTiles (Canvas, ScoreTeam[IndexTeam_A], DigitsBig);
		DrawNumericWidgetAsTiles (Canvas, ScoreTeam[IndexTeam_B], DigitsBig);
	}

	//

	if (bShowPoints || bShowResources)
	{
		ShowVersusIcon (Canvas);
	}
}

// Preconditions:
// - (PlayerOwner != None)

event DrawHUDPassC (Canvas Canvas)
{
	Super.DrawHUDPassC (Canvas);


	if
	(
		(PlayerOwnerPRI != None)
		&&
		(ReGRI != None)
		&&
		!bMatchOver
	)
	{
		if
		(
			PlayerOwnerPRI.IsA ('RePRI')
			&&
			(RePRI(PlayerOwnerPRI).TeamSummary != None)
		)
		{
			if (Class'PlayerSettings'.Default.bShowTeamInfo)
			{
				DrawTeamInfo (Canvas);
			}

			if
			(
				(Class'PlayerSettings'.Default.bShowTeammateRadar)
				&&
				!PlayerOwnerPRI.bOnlySpectator
			)
			{
				DrawTeamRadar (Canvas);
			}
		}


		if (Class'PlayerSettings'.Default.bShowEnemyInfo)
		{
			DrawEnemyInfo (Canvas);
		}
	}
}

// #TeamRadar

protected function DrawTeamRadar (Canvas Canvas)
{
	local PlayerReplicationInfo SpectatingPRI;
	local RePRI TeammatePRI;
	local Vector SourcePosition;
	local Rotator SourceRotation;
	local float SourceYaw, X, Y, SizeRadar, RadiusRadar, SizeBlip, RadiusBlip, OffsetY, OffsetScale, TeammateAngle, ScaledSizeRadar;
	local int Index;

	Canvas.DrawColor = TeammateRadar_BackgroundTextureColor;

	SizeRadar = Canvas.ClipX * Class'PlayerSettings'.Default.TeammateRadarSize * HUDScale * 0.5f;

	X = (Canvas.ClipX * Class'PlayerSettings'.Default.TeammateRadarPosX) - (SizeRadar * 0.5f);
	Y = (Canvas.ClipY * Class'PlayerSettings'.Default.TeammateRadarPosY) - (SizeRadar * 0.5f);

	Canvas.SetPos (X, Y);

	Canvas.DrawTile
	(
		TeamRadar_Texture_Level,
		SizeRadar,
		SizeRadar,
		128.0f,
		64.0f,
		64.0f,
		64.0f
	);


	if (PawnOwner == None)
	{
		SourcePosition = PlayerOwner.Location;
		SourceYaw = PlayerOwner.Rotation.Yaw;
		SpectatingPRI = PlayerOwnerPRI;
	}
	else if (!PlayerOwner.bBehindView)
	{
		SourcePosition = PawnOwner.Location;
		SourceYaw = PawnOwner.Rotation.Yaw;
		SpectatingPRI = PawnOwnerPRI;
	}
	else
	{
		Canvas.GetCameraLocation (SourcePosition, SourceRotation);

		SourceYaw = SourceRotation.Yaw;
		SpectatingPRI = PlayerOwnerPRI;
	}


	RadiusRadar = SizeRadar * 0.5f;

	ScaledSizeRadar = Class'PlayerSettings'.Default.TeammateRadarSize * HUDScale * 0.000167f * Canvas.ClipX;

	SizeBlip = 8.0f * Canvas.ClipX * HUDScale / 1600.0f;
	RadiusBlip = SizeBlip * 0.5f;

	X += RadiusRadar;
	OffsetY = (Y / Canvas.ClipY) + (RadiusRadar / Canvas.ClipY);

	for (Index = 0; Index < 16; ++Index) // TeamSummaryTicker.TeamSizeLimit (Value: 16)
	{
		TeammatePRI = RePRI(PlayerOwnerPRI).TeamSummary.RePRI[Index];

		if
		(
			(TeammatePRI != None)
			&&
			(TeamRadar_IsPlayerIn (TeammatePRI))
			&&
			(TeammatePRI != SpectatingPRI)
		)
		{
			Canvas.DrawColor = TeamRadar_Color (TeammatePRI);

			OffsetScale = ScaledSizeRadar * FMin (VSize (TeammatePRI.Position - SourcePosition), 1500.0f);
			TeammateAngle = ((Rotator(TeammatePRI.Position - SourcePosition).Yaw - SourceYaw) & 65535.0f) * 6.2832f / 65536.0f;

			Canvas.SetPos
			(
				X + OffsetScale * Sin (TeammateAngle) - RadiusBlip,
				OffsetY * Canvas.ClipY - OffsetScale * Cos (TeammateAngle) - RadiusBlip
			);

			Canvas.DrawTile
			(
				TeamRadar_Texture_Level,
				SizeBlip,
				SizeBlip,
				192.0f,
				64.0f,
				64.0f,
				64.0f
			);
		}
	}
}

protected function bool TeamRadar_IsPlayerIn (PlayerReplicationInfo OtherPRI)
{
	return
	(
		!OtherPRI.bOutofLives
	);
}

protected function Color TeamRadar_Color (RePRI OtherPRI)
{
	return GetColorHealth (OtherPRI.NormalizedHitPoints);
}

// #TeamName

protected function TeamName_DrawPosition
(
	Canvas Canvas,
	Actor SourceActor,
	RePRI OtherPRI,
	float X,
	float Y,
	float SizeUV
)
{
	local float Yaw, Distance, Angle, SizeUV_Symbol, SizeU_Line, SizeV_Line, Radius_Background, Radius_Height, Radius_Symbol;

	// Background Disc

	Canvas.DrawColor = BlackColor;

	Canvas.SetPos (X, Y);

	Canvas.DrawTile
	(
		Texture'HudContent.Generic.HUD',
		SizeUV,
		SizeUV,
		117.0f,
		256.0f,
		58.0f,
		58.0f
	);

	// Player Status

	Canvas.DrawColor = TeamRadar_Color (OtherPRI);
	Canvas.DrawColor.A = 223;

	Radius_Background = SizeUV * 0.5f;

	SizeUV_Symbol = SizeUV;
	Radius_Symbol = SizeUV_Symbol * 0.5f;

	Canvas.SetPos
	(
		X + Radius_Background - Radius_Symbol,
		Y + Radius_Background - Radius_Symbol
	);

	Canvas.DrawTile
	(
		Texture'HudContent.Generic.HUD',
		SizeUV_Symbol,
		SizeUV_Symbol,
		336.0f,
		430.0f,
		80.0f,
		80.0f
	);

	// Height Dot

	if (Class'PlayerSettings'.Default.bDrawTeammateRadarHeightDot)
	{
		Canvas.DrawColor = CyanColor;

		Radius_Height = Radius_Background * 0.9f;

		Distance = FClamp (OtherPRI.Position.Z - SourceActor.Location.Z, -1024.0f, 1024.0f) / 1024.0f;

		SizeU_Line = Radius_Background;
		SizeV_Line = FMax (6.0f, SizeUV * 0.1f);

		Canvas.SetPos
		(
			(X + Radius_Background) - (SizeU_Line * 0.5f),
			(Y + Radius_Background) - (Radius_Height * Distance) - (SizeV_Line * 0.5f)
		);

		Canvas.DrawTile
		(
			Texture'HudContent.Generic.HUD',
			SizeU_Line,
			SizeV_Line,
			340.0f,
			432.0f,
			78.0f,
			78.0f
		);
	}

	// Position Dot

	Canvas.DrawColor = WhiteColor;

	Yaw = Rotator(OtherPRI.Position - SourceActor.Location).Yaw;
	Angle = ((Yaw - SourceActor.Rotation.Yaw) & 65535.0f) * 6.2832f / 65536.0f;

	SizeUV_Symbol = FMax (5.0f, SizeUV * 0.4f);
	Radius_Symbol = SizeUV_Symbol * 0.5f;

	Radius_Height = Radius_Background * 0.75f;

	Canvas.SetPos
	(
		(X + Radius_Background) + (Radius_Height * Sin (Angle)) - Radius_Symbol,
		(Y + Radius_Background) - (Radius_Height * Cos (Angle)) - Radius_Symbol
	);

	Canvas.DrawTile
	(
		Texture'HudContent.Generic.HUD',
		SizeUV_Symbol,
		SizeUV_Symbol,
		336.0f,
		430.0f,
		80.0f,
		80.0f
	);
}

protected function TeamName_SetBackgroundColor (Canvas Canvas, RePRI OtherRePRI)
{
	Canvas.DrawColor = Teamname_ColorTexture_Background;
}

protected function TeamName_SetNameColor (Canvas Canvas, RePRI OtherRePRI)
{
	if (OtherRePRI.IsCurrentPlayerStatus ('Out'))
	{
		Canvas.DrawColor = TeamName_Color_Out;
	}
	else if (OtherRePRI.IsCurrentPlayerStatus ('Dead'))
	{
		Canvas.DrawColor = TeamName_Color_Spawning;
	}
	else
	{
		Canvas.DrawColor = TeamName_Color_Playing;
	}
}

protected function TeamName_SetHealthColor (Canvas Canvas, RePRI OtherRePRI)
{
	if (!OtherRePRI.IsCurrentPlayerStatus ('Out'))
	{
		Canvas.DrawColor = GetColorHealth (OtherRePRI.NormalizedHitPoints);
	}
	else
	{
		Canvas.DrawColor = TeamName_Color_Dead;
	}
}

protected function TeamName_SetLocationColor (Canvas Canvas, RePRI OtherRePRI)
{
	if (!OtherRePRI.IsCurrentPlayerStatus ('Out'))
	{
		Canvas.DrawColor = TeamName_Color_LocationName;
	}
	else
	{
		Canvas.DrawColor = TeamName_Color_Dead;
	}
}

// #TeamName #Allies

protected function bool ShouldDrawTeamInfo (RePRI OtherRePRI)
{
	return
	(
		(OtherRePRI != PlayerOwnerPRI)
		&&
		!OtherRePRI.IsCurrentPlayerStatus ('Out')
	);
}

protected function DrawTeamInfo (Canvas Canvas)
{
	local Actor Source;
	local RePRI OtherRePRI;
	local Color DrawnColor;
	local TeamSummary OwnTeamSummary;
	local bool bShowCountryCode, bShowStatus, bShowLocationName;

	// Reset

	DrawTeamInfo_L_Drawn = 0;

	// Settings

	bShowCountryCode = Class'PlayerSettings'.Default.bShowTeamInfoCountry;
	bShowStatus = Class'PlayerSettings'.Default.bShowTeamInfoStatus;
	bShowLocationName = Class'PlayerSettings'.Default.bShowTeamInfoLocation;

	// Source

	if (PawnOwner == None)
	{
		Source = PlayerOwner;
	}
	else if (!PlayerOwner.bBehindView)
	{
		Source = PawnOwner; // Spectated player is in control over rotation.
	}
	else
	{
		Source = PlayerOwner; // Spectator is in control over rotation.
	}

	// Font Relative Size / Offset

	Canvas.Font = GetFontSizeIndex (Canvas, Clamp (Class'PlayerSettings'.Default.TeamInfoFontSize, -8, 8));
	Canvas.TextSize ("*", DrawTeamInfo_L_Width_Text, DrawTeamInfo_L_Height_Text);

	DrawTeamInfo_L_Width_Character = DrawTeamInfo_L_Width_Text; // Width of a single character
	DrawTeamInfo_L_Y_Slot = Canvas.ClipY * TeamName_ScaledPosition_Y_Base;
	DrawTeamInfo_L_Height_Row = DrawTeamInfo_L_Height_Text * 1.25f;

	if (bShowLocationName)
	{
		DrawTeamInfo_L_Height_Slot = (DrawTeamInfo_L_Height_Text + DrawTeamInfo_L_Height_Text) * 1.125f;
	}
	else
	{
		DrawTeamInfo_L_Height_Slot = DrawTeamInfo_L_Height_Text * 1.25f;
	}

	DrawTeamInfo_L_OffsetY_Text = (DrawTeamInfo_L_Height_Row * 0.5f) - (DrawTeamInfo_L_Height_Text * 0.5f) + 1.0f;
	DrawTeamInfo_L_Rows = Min (Clamp (Class'PlayerSettings'.Default.TeamInfoRows, Class'PlayerSettings'.Default.MinTeamInfoRows, Class'PlayerSettings'.Default.MaxTeamInfoRows), Teammates.Length);
	DrawTeamInfo_L_Width_Slot = DrawTeamInfo_L_Width_Character * 17;
	DrawTeamInfo_L_X_Health = DrawTeamInfo_L_Width_Slot;

	// Offsets / Characters

	if (bShowStatus)
	{
		if (bShowCountryCode) // [Status][Country][Player]
		{
			DrawTeamInfo_L_PlayerNameCharacters = 7;
			DrawTeamInfo_L_X_Status = ((DrawTeamInfo_L_Width_Character * 3) * 0.5f) - (DrawTeamInfo_L_Height_Row * 0.5f);
			DrawTeamInfo_L_X_CountryFlag = (DrawTeamInfo_L_Width_Character * 3) + (((DrawTeamInfo_L_Width_Character * 3) * 0.5f) - (DrawTeamInfo_L_Height_Row * 0.5f));
			DrawTeamInfo_L_X_CountryText = ((DrawTeamInfo_L_Width_Character * 3) + ((DrawTeamInfo_L_Width_Character * 3) * 0.5f)) - DrawTeamInfo_L_Width_Character;
			DrawTeamInfo_L_X_PlayerName = DrawTeamInfo_L_Width_Character * (3 + 3);
		}
		else // [Status][Player]
		{
			DrawTeamInfo_L_PlayerNameCharacters = 10;
			DrawTeamInfo_L_X_Status = ((DrawTeamInfo_L_Width_Character * 3) * 0.5f) - (DrawTeamInfo_L_Height_Row * 0.5f);
			DrawTeamInfo_L_X_PlayerName = DrawTeamInfo_L_Width_Character * 3;
		}

		DrawTeamInfo_L_LocationNameCharacters = 13;
		DrawTeamInfo_L_X_LocationName = DrawTeamInfo_L_Width_Character * 3;
	}
	else if (bShowCountryCode)
	{
		if (bShowStatus) // [Status][Country][Player]
		{
			DrawTeamInfo_L_PlayerNameCharacters = 7;
			DrawTeamInfo_L_X_Status = ((DrawTeamInfo_L_Width_Character * 3) * 0.5f) - (DrawTeamInfo_L_Height_Row * 0.5f);
			DrawTeamInfo_L_X_CountryFlag = (DrawTeamInfo_L_Width_Character * 3) + (((DrawTeamInfo_L_Width_Character * 3) * 0.5f) - (DrawTeamInfo_L_Height_Row * 0.5f));
			DrawTeamInfo_L_X_CountryText = ((DrawTeamInfo_L_Width_Character * 3) + ((DrawTeamInfo_L_Width_Character * 3) * 0.5f)) - DrawTeamInfo_L_Width_Character;
			DrawTeamInfo_L_X_PlayerName = DrawTeamInfo_L_Width_Character * (3 + 3);
		}
		else // [Country][Player]
		{
			DrawTeamInfo_L_PlayerNameCharacters = 10;
			DrawTeamInfo_L_X_CountryFlag = ((DrawTeamInfo_L_Width_Character * 3) * 0.5f) - (DrawTeamInfo_L_Height_Row * 0.5f);
			DrawTeamInfo_L_X_CountryText = ((DrawTeamInfo_L_Width_Character * 3) * 0.5f) - DrawTeamInfo_L_Width_Character;
			DrawTeamInfo_L_X_PlayerName = DrawTeamInfo_L_Width_Character * 3;
		}

		DrawTeamInfo_L_LocationNameCharacters = 13;
		DrawTeamInfo_L_X_LocationName = DrawTeamInfo_L_Width_Character * 3;
	}
	else // [Player]
	{
		DrawTeamInfo_L_PlayerNameCharacters = 12;
		DrawTeamInfo_L_X_PlayerName = DrawTeamInfo_L_Width_Character * 1;

		DrawTeamInfo_L_LocationNameCharacters = 13;
		DrawTeamInfo_L_X_LocationName = DrawTeamInfo_L_Width_Character * 3;
	}

	//

 	if
 	(
		(!PlayerOwnerPRI.bOnlySpectator)
		&&
	 	(RePRI(PlayerOwnerPRI).TeamSummary != None)
	)
	{
		DrawTeamInfo_L_Y_Text = DrawTeamInfo_L_Y_Slot + DrawTeamInfo_L_OffsetY_Text; // </NextLine>
		DrawTeamInfo_L_Y_LocationName = DrawTeamInfo_L_Y_Slot + DrawTeamInfo_L_OffsetY_Text + DrawTeamInfo_L_Height_Text; // </NextLine>

		DrawnColor = BlackColor;
		DrawnColor.A = TeamInfoOpacity;

		// Draw

		OwnTeamSummary = RePRI(PlayerOwnerPRI).TeamSummary;

		for (DrawTeamInfo_L_Index = 0; DrawTeamInfo_L_Index < 16; ++DrawTeamInfo_L_Index) // TeamSummaryTicker.TeamSizeLimit (Value: 16)
		{
			OtherRePRI = OwnTeamSummary.RePRI[DrawTeamInfo_L_Index];

			if
			(
				(OtherRePRI != None)
				&&
				((Level.TimeSeconds - OtherRePRI.TeamName_TimeSeconds_LastRelevant) < 1.5f)
			)
			{
				++DrawTeamInfo_L_Drawn;

				// <DrawBackground>
				Canvas.DrawColor = DrawnColor;

				Canvas.SetPos (0.0f, DrawTeamInfo_L_Y_Slot);

				Canvas.DrawTile
				(
					Teamname_Texture_Background,
					DrawTeamInfo_L_Width_Slot,
					DrawTeamInfo_L_Height_Slot,
					0.0f,
					0.0f,
					Teamname_Texture_Background.USize,
					Teamname_Texture_Background.VSize
				);
				// </DrawBackground>


				// <DrawPosition>
				if (bShowStatus)
				{
					TeamName_DrawPosition (Canvas, Source, OtherRePRI, DrawTeamInfo_L_X_Status, DrawTeamInfo_L_Y_Slot, DrawTeamInfo_L_Height_Row);
				}
				// </DrawPosition>

				TeamName_SetNameColor (Canvas, OtherRePRI);

				// <DrawCountry>
				if (bShowCountryCode)
				{
					if (OtherRePRI.TextureFlag == None)
					{
						Class'CountryCodes'.Static.GetTextureFlag (OtherRePRI.CountryCode, OtherRePRI.TextureFlag, OtherRePRI.OffsetTextureFlag);
					}

					if (OtherRePRI.TextureFlag != None)
					{
						Canvas.SetPos (DrawTeamInfo_L_X_CountryFlag, DrawTeamInfo_L_Y_Slot);

						Canvas.DrawTile
						(
							OtherRePRI.TextureFlag,
							DrawTeamInfo_L_Height_Row,
							DrawTeamInfo_L_Height_Row,
							OtherRePRI.OffsetTextureFlag.X1,
							OtherRePRI.OffsetTextureFlag.Y1,
							OtherRePRI.OffsetTextureFlag.X2,
							OtherRePRI.OffsetTextureFlag.Y2
						);
					}
					else // Country code but no flag.
					{
						Canvas.SetPos
						(
							DrawTeamInfo_L_X_CountryText,
							DrawTeamInfo_L_Y_Text
						);

						Canvas.DrawText (OtherRePRI.CountryCode, False);
					}
				}
				// </DrawCountry>


				// <DrawPlayerName>
				DrawTeamInfo_L_CachedString = Left (OtherRePRI.PlayerName, DrawTeamInfo_L_PlayerNameCharacters);
				Canvas.SetPos (DrawTeamInfo_L_X_PlayerName, DrawTeamInfo_L_Y_Text);
				Canvas.DrawText (DrawTeamInfo_L_CachedString, False);
				// </DrawPlayerName>


				// <DrawHealth>
				TeamName_SetHealthColor (Canvas, OtherRePRI);

				if (OtherRePRI.PlayerZone != None)
				{
					DrawTeamInfo_L_CachedString = string(OtherRePRI.HitPoints);
				}
				else
				{
					DrawTeamInfo_L_CachedString = "0";
				}

				Canvas.TextSize (DrawTeamInfo_L_CachedString, DrawTeamInfo_L_Width_Text, DrawTeamInfo_L_Height_Text);

				Canvas.SetPos
				(
					DrawTeamInfo_L_X_Health - DrawTeamInfo_L_Width_Text,
					DrawTeamInfo_L_Y_Text
				);

				Canvas.DrawText (DrawTeamInfo_L_CachedString, False);
				// </DrawHealth>


				if (bShowLocationName)
				{
					TeamName_SetLocationColor (Canvas, OtherRePRI);
					DrawTeamInfo_L_Opacity = Canvas.DrawColor.A;

					Canvas.SetPos (DrawTeamInfo_L_X_LocationName, DrawTeamInfo_L_Y_LocationName);
					Canvas.DrawText (Left (OtherRePRI.GetLocationName (), DrawTeamInfo_L_LocationNameCharacters), False);


					if ((OtherRePRI.Flags & 1) == 1)
					{
						Canvas.DrawColor = WhiteColor;
						Canvas.DrawColor.A = DrawTeamInfo_L_Opacity;

						Canvas.SetPos
						(
							DrawTeamInfo_L_X_Health - DrawTeamInfo_L_Height_Text,
							DrawTeamInfo_L_Y_LocationName
						);

						Canvas.DrawTile
						(
							Texture_Symbols,
							DrawTeamInfo_L_Height_Text,
							DrawTeamInfo_L_Height_Text,
							0.0f,
							0.0f,
							64.0f,
							64.0f
						);
                    }


					DrawTeamInfo_L_Y_Slot += DrawTeamInfo_L_Height_Slot + 1.0f; // </NextSlot>
					DrawTeamInfo_L_Y_Text = DrawTeamInfo_L_Y_Slot + DrawTeamInfo_L_OffsetY_Text;

					DrawTeamInfo_L_Y_LocationName = DrawTeamInfo_L_Y_Slot + DrawTeamInfo_L_OffsetY_Text + DrawTeamInfo_L_Height_Text; // </NextLine>
				}
				else
				{
					DrawTeamInfo_L_Y_Slot += DrawTeamInfo_L_Height_Slot + 1.0f; // </NextSlot>
					DrawTeamInfo_L_Y_Text = DrawTeamInfo_L_Y_Slot + DrawTeamInfo_L_OffsetY_Text;
				}
			}


			if (DrawTeamInfo_L_Drawn >= DrawTeamInfo_L_Rows)
			{
				break;
			}
		}
	}
	else // Red Team Spectator Perspective
	{
		DrawTeamInfo_L_PlayerNameCharacters += 4; // Health
		DrawTeamInfo_L_Height_Slot = DrawTeamInfo_L_Height_Row;
		DrawTeamInfo_L_Y_Text = DrawTeamInfo_L_Y_Slot + DrawTeamInfo_L_OffsetY_Text; // </NextLine>

		DrawnColor = TeammateColor;
		DrawnColor.A = SpectatedTeamInfoOpacity;

		DrawTeamInfo_L_Count = Teammates.Length;

		for (DrawTeamInfo_L_Index = 0; DrawTeamInfo_L_Index < DrawTeamInfo_L_Count; ++DrawTeamInfo_L_Index)
		{
			OtherRePRI = Teammates[DrawTeamInfo_L_Index];

			if ((Level.TimeSeconds - OtherRePRI.TeamName_TimeSeconds_LastRelevant) < 1.5f)
			{
				++DrawTeamInfo_L_Drawn;

				// <DrawBackground>
				Canvas.DrawColor = DrawnColor;

				Canvas.SetPos (0.0f, DrawTeamInfo_L_Y_Slot);

				Canvas.DrawTile
				(
					Teamname_Texture_Background,
					DrawTeamInfo_L_Width_Slot,
					DrawTeamInfo_L_Height_Slot,
					0,
					0,
					Teamname_Texture_Background.USize,
					Teamname_Texture_Background.VSize
				);
				// </DrawBackground>


				// <DrawPosition>
				if (bShowStatus)
				{
					TeamName_DrawSpectatorSymbol (Canvas, OtherRePRI, DrawTeamInfo_L_X_Status, DrawTeamInfo_L_Y_Slot, DrawTeamInfo_L_Height_Row);
				}
				// </DrawPosition>


				TeamName_SetNameColor (Canvas, OtherRePRI);


				// <DrawCountry>
				if (bShowCountryCode)
				{
					if (OtherRePRI.TextureFlag == None)
					{
						Class'CountryCodes'.Static.GetTextureFlag (OtherRePRI.CountryCode, OtherRePRI.TextureFlag, OtherRePRI.OffsetTextureFlag);
					}

					if (OtherRePRI.TextureFlag != None)
					{
						Canvas.SetPos (DrawTeamInfo_L_X_CountryFlag, DrawTeamInfo_L_Y_Slot);

						Canvas.DrawTile
						(
							OtherRePRI.TextureFlag,
							DrawTeamInfo_L_Height_Row,
							DrawTeamInfo_L_Height_Row,
							OtherRePRI.OffsetTextureFlag.X1,
							OtherRePRI.OffsetTextureFlag.Y1,
							OtherRePRI.OffsetTextureFlag.X2,
							OtherRePRI.OffsetTextureFlag.Y2
						);
					}
					else // Country code but no flag.
					{
						Canvas.TextSize (OtherRePRI.CountryCode, DrawTeamInfo_L_Width_Text, DrawTeamInfo_L_Height_Text);

						Canvas.SetPos
						(
							DrawTeamInfo_L_X_CountryText - (DrawTeamInfo_L_Width_Text * 0.5f),
							DrawTeamInfo_L_Y_Text
						);

						Canvas.DrawText (OtherRePRI.CountryCode, False);
					}
				}
				// </DrawCountry>


				// <DrawPlayerName>
				DrawTeamInfo_L_CachedString = Left (OtherRePRI.PlayerName, DrawTeamInfo_L_PlayerNameCharacters);
				Canvas.SetPos (DrawTeamInfo_L_X_PlayerName, DrawTeamInfo_L_Y_Text);
				Canvas.DrawText (DrawTeamInfo_L_CachedString, False);
				// </DrawPlayerName>


				DrawTeamInfo_L_Y_Slot += DrawTeamInfo_L_Height_Slot + 1.0f; // </NextSlot>
				DrawTeamInfo_L_Y_Text = DrawTeamInfo_L_Y_Slot + DrawTeamInfo_L_OffsetY_Text; // </NextLine>
			}


			if (DrawTeamInfo_L_Drawn >= DrawTeamInfo_L_Rows)
			{
				break;
			}
		}
	}
}

// #TeamName #Enemies

protected function bool TeamName_ShouldDrawEnemy (RePRI EnemyRePRI)
{
	return
	(
		!EnemyRePRI.IsCurrentPlayerStatus ('Out')
	);
}

protected function TeamName_DrawSpectatorSymbol (Canvas Canvas, RePRI OtherRePRI, float X, float Y, float SizeUV)
{
	if (OtherRePRI.IsCurrentPlayerStatus ('Out'))
	{
		Canvas.DrawColor = TeamName_Color_Out;

		Canvas.SetPos (X, Y);

		Canvas.DrawTile
		(
			Texture_Symbols,
			SizeUV,
			SizeUV,
			OffsetTexture_Out.X1,
			OffsetTexture_Out.Y1,
			64.0f,
			64.0f
		);
	}
	else if (OtherRePRI.IsCurrentPlayerStatus ('Dead'))
	{
		Canvas.DrawColor = TeamName_Color_Spawning;

		Canvas.SetPos (X, Y);

		Canvas.DrawTile
		(
			Texture_Symbols,
			SizeUV,
			SizeUV,
			OffsetTexture_Spawning.X1,
			OffsetTexture_Spawning.Y1,
			64.0f,
			64.0f
		);
	}
	else
	{
		Canvas.DrawColor = TeamName_Color_Playing;

		Canvas.SetPos (X, Y);

		Canvas.DrawTile
		(
			Texture_Symbols,
			SizeUV,
			SizeUV,
			OffsetTexture_Playing.X1,
			OffsetTexture_Playing.Y1,
			64.0f,
			64.0f
		);
	}
}

protected function DrawEnemyInfo (Canvas Canvas)
{
	local RePRI OtherRePRI;
	local Color DrawnColor;
	local int Count, Drawn;
	local bool bShowCountryCode, bShowStatus;

	// Reset

	Drawn = 0;

	// Settings

	bShowCountryCode = Class'PlayerSettings'.Default.bShowEnemyInfoCountry;
	bShowStatus = Class'PlayerSettings'.Default.bShowEnemyInfoStatus;

	// Initialize

	Canvas.Font = GetFontSizeIndex (Canvas, Clamp (Class'PlayerSettings'.default.EnemyInfoFontSize, -8, 8));
	Canvas.TextSize ("*", DrawEnemyInfo_L_Width_Text, DrawEnemyInfo_L_Height_Text);

	DrawEnemyInfo_L_Width_Character = DrawEnemyInfo_L_Width_Text;
	DrawEnemyInfo_L_Width_Slot = DrawEnemyInfo_L_Width_Character * 17;


	DrawEnemyInfo_L_X_Slot = Canvas.ClipX;
	DrawEnemyInfo_L_Y_Slot = Canvas.ClipY * TeamName_ScaledPosition_Y_Base;

	DrawEnemyInfo_L_X_Background = DrawEnemyInfo_L_X_Slot - DrawEnemyInfo_L_Width_Slot;

	DrawEnemyInfo_L_Height_Row = DrawEnemyInfo_L_Height_Text * 1.25f;
	DrawEnemyInfo_L_Height_Slot = DrawEnemyInfo_L_Height_Row;

	DrawEnemyInfo_L_OffsetY_Text = (DrawEnemyInfo_L_Height_Slot * 0.5f) - (DrawEnemyInfo_L_Height_Text * 0.5f) + 1.0f;

	DrawEnemyInfo_L_Rows = Min (Clamp (Class'PlayerSettings'.Default.EnemyInfoRows, Class'PlayerSettings'.Default.MinEnemyInfoRows, Class'PlayerSettings'.Default.MaxEnemyInfoRows), Enemies.Length);

	DrawEnemyInfo_L_Y_Text = DrawEnemyInfo_L_Y_Slot + DrawEnemyInfo_L_OffsetY_Text; // </NextLine>

	// <Status>
	// Offsets / Characters

	if (bShowStatus)
	{
		if (bShowCountryCode) // [Status][Country][Player]
		{
			DrawEnemyInfo_L_PlayerNameCharacters = 8 + 4;
			DrawEnemyInfo_L_X_Status = DrawEnemyInfo_L_X_Slot - ((DrawEnemyInfo_L_Width_Character * 3) * 0.5f) - (DrawEnemyInfo_L_Height_Row * 0.5f);
			DrawEnemyInfo_L_X_CountryFlag = DrawEnemyInfo_L_X_Slot - ((DrawEnemyInfo_L_Width_Character * 4) * 1.0f) - (DrawEnemyInfo_L_Height_Row * 0.5f);
			DrawEnemyInfo_L_X_CountryText = DrawEnemyInfo_L_X_Slot - ((DrawEnemyInfo_L_Width_Character * 4) * 1.0f) - (DrawEnemyInfo_L_Height_Row * 0.5f);
			DrawEnemyInfo_L_X_PlayerName = DrawEnemyInfo_L_X_Slot - (DrawEnemyInfo_L_Width_Character * (3 + 3));
		}
		else // [Status][Player]
		{
			DrawEnemyInfo_L_PlayerNameCharacters = 10 + 4;
			DrawEnemyInfo_L_X_Status = DrawEnemyInfo_L_X_Slot - ((DrawEnemyInfo_L_Width_Character * 3) * 0.5f) - (DrawEnemyInfo_L_Height_Row * 0.5f);
			DrawEnemyInfo_L_X_PlayerName = DrawEnemyInfo_L_X_Slot - (DrawEnemyInfo_L_Width_Character * 3);
		}
	}
	else if (bShowCountryCode)
	{
		if (bShowStatus) // [Status][Country][Player]
		{
			DrawEnemyInfo_L_PlayerNameCharacters = 8 + 4;
			DrawEnemyInfo_L_X_Status = DrawEnemyInfo_L_X_Slot - ((DrawEnemyInfo_L_Width_Character * 3) * 0.5f) - (DrawEnemyInfo_L_Height_Row * 0.5f);
			DrawEnemyInfo_L_X_CountryFlag = DrawEnemyInfo_L_X_Slot - ((DrawEnemyInfo_L_Width_Character * 4) * 1.0f) - (DrawEnemyInfo_L_Height_Row * 0.5f);
			DrawEnemyInfo_L_X_CountryText = DrawEnemyInfo_L_X_Slot - ((DrawEnemyInfo_L_Width_Character * 4) * 1.0f) - (DrawEnemyInfo_L_Height_Row * 0.5f);
			DrawEnemyInfo_L_X_PlayerName = DrawEnemyInfo_L_X_Slot - (DrawEnemyInfo_L_Width_Character * (3 + 3)) - DrawEnemyInfo_L_Width_Character;
		}
		else // [Country][Player]
		{
			DrawEnemyInfo_L_PlayerNameCharacters = 10 + 4;
			DrawEnemyInfo_L_X_CountryFlag = DrawEnemyInfo_L_X_Slot - ((DrawEnemyInfo_L_Width_Character * 3) * 0.5f) - (DrawEnemyInfo_L_Height_Row * 0.5f);
			DrawEnemyInfo_L_X_CountryText = DrawEnemyInfo_L_X_Slot - ((DrawEnemyInfo_L_Width_Character * 3) * 0.5f) - (DrawEnemyInfo_L_Height_Row * 0.5f);
			DrawEnemyInfo_L_X_PlayerName = DrawEnemyInfo_L_X_Slot - (DrawEnemyInfo_L_Width_Character * 3);
		}
	}
	else // [Player]
	{
		DrawEnemyInfo_L_PlayerNameCharacters = 14;
		DrawEnemyInfo_L_X_PlayerName = DrawEnemyInfo_L_X_Slot - (DrawEnemyInfo_L_Width_Character * 1);
	}
	// </Country>

	Count = Enemies.Length;

	if (!PlayerOwnerPRI.bOnlySpectator)
	{
		DrawnColor = BlackColor;
	    DrawnColor.A = TeamInfoOpacity;
    }
    else
    {
		DrawnColor = EnemyColor;
	    DrawnColor.A = SpectatedTeamInfoOpacity;
    }

	for (DrawEnemyInfo_L_Index = 0; DrawEnemyInfo_L_Index < Count; ++DrawEnemyInfo_L_Index)
	{
		OtherRePRI = Enemies[DrawEnemyInfo_L_Index];

		if ((Level.TimeSeconds - OtherRePRI.TeamName_TimeSeconds_LastRelevant) < 1.5f)
		{
			++Drawn;

			// <DrawEnemyBackground>
			Canvas.DrawColor = DrawnColor;

			Canvas.SetPos (DrawEnemyInfo_L_X_Background, DrawEnemyInfo_L_Y_Slot);

			Canvas.DrawTile
			(
				Teamname_Texture_Background,
				DrawEnemyInfo_L_Width_Slot,
				DrawEnemyInfo_L_Height_Slot,
				0.0f,
				0.0f,
				Teamname_Texture_Background.USize - 2.0f,
				Teamname_Texture_Background.VSize - 2.0f
			);
			// <DrawEnemyBackground>


			// <DrawEnemyStatus>
			if (bShowStatus)
			{
				TeamName_DrawSpectatorSymbol (Canvas, OtherRePRI, DrawEnemyInfo_L_X_Status, DrawEnemyInfo_L_Y_Slot, DrawEnemyInfo_L_Height_Row);
			}
			// </DrawEnemyStatus>


			TeamName_SetNameColor (Canvas, OtherRePRI);


			// <DrawCountry>
			if (bShowCountryCode)
			{
				if (OtherRePRI.TextureFlag == None)
				{
					Class'CountryCodes'.Static.GetTextureFlag (OtherRePRI.CountryCode, OtherRePRI.TextureFlag, OtherRePRI.OffsetTextureFlag);
				}

				if (OtherRePRI.TextureFlag != None)
				{
					Canvas.SetPos (DrawEnemyInfo_L_X_CountryFlag, DrawEnemyInfo_L_Y_Slot);

					Canvas.DrawTile
					(
						OtherRePRI.TextureFlag,
						DrawEnemyInfo_L_Height_Row,
						DrawEnemyInfo_L_Height_Row,
						OtherRePRI.OffsetTextureFlag.X1,
						OtherRePRI.OffsetTextureFlag.Y1,
						OtherRePRI.OffsetTextureFlag.X2,
						OtherRePRI.OffsetTextureFlag.Y2
					);
				}
				else // Country code but no flag.
				{
					Canvas.SetPos (DrawEnemyInfo_L_X_CountryText, DrawEnemyInfo_L_Y_Text);
					Canvas.DrawText (OtherRePRI.CountryCode, False);
				}
			}
			// </DrawCountry>


			// <DrawEnemyPlayerName>
			DrawEnemyInfo_L_CachedString = Left (OtherRePRI.PlayerName, DrawEnemyInfo_L_PlayerNameCharacters);
			Canvas.TextSize (DrawEnemyInfo_L_CachedString, DrawEnemyInfo_L_Width_Text, DrawEnemyInfo_L_Height_Text);
			Canvas.SetPos (DrawEnemyInfo_L_X_PlayerName - DrawEnemyInfo_L_Width_Text, DrawEnemyInfo_L_Y_Text);
			Canvas.DrawText (DrawEnemyInfo_L_CachedString, False);
			// </DrawEnemyPlayerName>


			DrawEnemyInfo_L_Y_Slot += DrawEnemyInfo_L_Height_Slot + 1.0f; // </NextSlot>
			DrawEnemyInfo_L_Y_Text = DrawEnemyInfo_L_Y_Slot + DrawEnemyInfo_L_OffsetY_Text;
		}


		if (Drawn >= DrawEnemyInfo_L_Rows)
		{
			break;
		}
	}
}

// #ChatPortrait

protected static function Color ChatPortrait_GetColorHealth (RePRI OtherRePRI)
{
	return GetColorHealth (OtherRePRI.NormalizedHitPoints);
}

// Preconditions:
// - (Portrait != None)

protected function ChatPortrait_Draw (Canvas Canvas)
{
	local TeamInfo TeamInfo_Local;
	local float X_Portrait, Y_Portrait, Y_Bar, Width_Text, Height_Text, Height_Bar, ScaledValue;

	// #

	if (PlayerOwner.PlayerReplicationInfo != None)
	{
		TeamInfo_Local = PlayerOwner.PlayerReplicationInfo.Team;
	}

	// #

	ChatPortrait_Draw_L_SizeUV_Portrait = 0.125f * Canvas.ClipY; // 128

	X_Portrait = (-ChatPortrait_Draw_L_SizeUV_Portrait) * PortraitX;
	Y_Portrait = (Canvas.ClipY * 0.5f) - (ChatPortrait_Draw_L_SizeUV_Portrait * 0.5f);

	// #

	Height_Bar = ChatPortrait_Draw_L_SizeUV_Portrait * 0.15f;
	Y_Bar = Y_Portrait + ChatPortrait_Draw_L_SizeUV_Portrait - Height_Bar;

	// #Portrait

	Canvas.DrawColor = ChatPortrait_ColorTexture_Portrait;

	Canvas.SetPos (X_Portrait, Y_Portrait);

	Canvas.DrawTile
	(
		Portrait,
		ChatPortrait_Draw_L_SizeUV_Portrait,
		ChatPortrait_Draw_L_SizeUV_Portrait,
		0.0f,
		0.0f,
		ChatPortrait_Draw_L_SizeUVSource_Portrait,
		ChatPortrait_Draw_L_SizeUVSource_Portrait
	);

	// #Health

	if (PortraitPRI != None)
	{
		ScaledValue = RePRI(PortraitPRI).NormalizedHitPoints;

		if (PlayerOwnerPRI.bOnlySpectator)
		{
			Canvas.DrawColor = ChatPortrait_GetColorHealth (RePRI(PortraitPRI));
			Canvas.DrawColor.A = ChatPortrait_ColorAlphaTexture_Health;

			Canvas.SetPos (X_Portrait, Y_Bar);

			Canvas.DrawTile
			(
				Texture'Engine.WhiteTexture',
				ChatPortrait_Draw_L_SizeUV_Portrait * ScaledValue,
				Height_Bar,
				0.0f,
				0.0f,
				Texture'Engine.WhiteTexture'.USize,
				Texture'Engine.WhiteTexture'.VSize
			);
		}
		else if (TeamInfo_Local == PortraitPRI.Team)
		{
			Canvas.DrawColor = ChatPortrait_GetColorHealth (RePRI(PortraitPRI));
			Canvas.DrawColor.A = ChatPortrait_ColorAlphaTexture_Health;

			Canvas.SetPos (X_Portrait, Y_Bar);

			Canvas.DrawTile
			(
				Texture'Engine.WhiteTexture',
				ChatPortrait_Draw_L_SizeUV_Portrait * ScaledValue,
				Height_Bar,
				0.0f,
				0.0f,
				Texture'Engine.WhiteTexture'.USize,
				Texture'Engine.WhiteTexture'.VSize
			);
		}
	}

	// #Effect

	Canvas.DrawColor = ChatPortrait_ColorTexture_Effect;

	Canvas.SetPos (X_Portrait, Y_Portrait);

	Canvas.DrawTile
	(
		ChatPortrait_Material_Noise,
		ChatPortrait_Draw_L_SizeUV_Portrait,
		ChatPortrait_Draw_L_SizeUV_Portrait,
		0.0f,
		0.0f,
		ChatPortrait_USize_Material_Noise,
		ChatPortrait_VSize_Material_Noise
	);

	// #Border

	Canvas.DrawColor = ChatPortrait_ColorTexture_Frame;

	Canvas.SetPos
	(
		X_Portrait,
		Y_Portrait
	);

	Canvas.DrawTilePartialStretched
	(
		Texture'Textures.PortraitFrame',
		ChatPortrait_Draw_L_SizeUV_Portrait,
		ChatPortrait_Draw_L_SizeUV_Portrait
	);

	// #Name

	if (RePRI(PortraitPRI) != None)
	{
		Canvas.DrawColor = ChatPortrait_ColorText_PlayerName;

		Canvas.Font = GetFontSizeIndex (Canvas, -2);
		Canvas.StrLen (ChatPortrait_Draw_L_Text_PlayerName, Width_Text, Height_Text);

		if (Width_Text > ChatPortrait_Draw_L_SizeUV_Portrait)
		{
			Canvas.Font = GetFontSizeIndex (Canvas, -3);
			Canvas.StrLen (ChatPortrait_Draw_L_Text_PlayerName, Width_Text, Height_Text);

			if (Width_Text > ChatPortrait_Draw_L_SizeUV_Portrait)
			{
				Canvas.Font = GetFontSizeIndex (Canvas, -4);
				Canvas.StrLen (ChatPortrait_Draw_L_Text_PlayerName, Width_Text, Height_Text);
			}
		}

		Canvas.SetPos
		(
			X_Portrait + (ChatPortrait_Draw_L_SizeUV_Portrait * 0.5f) - (Width_Text * 0.5f),
			Y_Portrait + ChatPortrait_Draw_L_SizeUV_Portrait
		);

		Canvas.DrawText (ChatPortrait_Draw_L_Text_PlayerName, False);
	}
}

// #TeamSymbols

protected function TeamSymbol_Update (GameReplicationInfo GRI, int Side, int IndexTeam)
{
	if (GRI.TeamSymbols[IndexTeam] != None) // Team Symbol
	{
		TeamSymbols[Side].WidgetTexture = GRI.TeamSymbols[IndexTeam];
		TeamSymbols[Side].TextureCoords.X1 = 0;
		TeamSymbols[Side].TextureCoords.Y1 = 0;
		TeamSymbols[Side].TextureCoords.X2 = GRI.TeamSymbols[IndexTeam].MaterialUSize ();
		TeamSymbols[Side].TextureCoords.Y2 = GRI.TeamSymbols[IndexTeam].MaterialVSize ();
		TeamSymbols[Side].TextureScale = 0.09f;
	}
}

// #Queries

event bool CustomHUDColorAllowed ()
{
	return True;
}

// #Commands

exec function ShowPoints ()
{
	bShowPoints = !bShowPoints;

	SaveConfig ();
}

// #Commands

exec function ShowRadar ()
{
	Class'PlayerSettings'.Static.ShowRadar ();
}

exec function MoveRadar (int X, int Y)
{
	Class'PlayerSettings'.Static.MoveRadar (X, Y);
}

exec function SizeRadar (int UV)
{
	Class'PlayerSettings'.Static.SizeRadar (UV);
}

exec function ResetRadar ()
{
	Class'PlayerSettings'.Static.ResetRadar ();
}

exec function ShowResources ()
{
	Class'PlayerSettings'.static.ShowResources ();
}

exec function ShowTeam ()
{
	Class'PlayerSettings'.static.ShowTeam ();
}

exec function ShowEnemy ()
{
	Class'PlayerSettings'.static.ShowEnemy ();
}

// #

DefaultProperties
{
	TeamInfoOpacity = 200
	SpectatedTeamInfoOpacity = 88

	Texture_Symbols = Texture'Textures.Symbols'

	TeamName_ScaledPosition_Y_Base = 0.10370370

	TeamName_Color_Playing = (R=200,G=200,B=200,A=255)
	TeamName_Color_Dead = (R=150,G=150,B=150,A=255)
	TeamName_Color_Spawning = (R=223,G=200,B=200,A=255)
	TeamName_Color_Out = (R=100,G=100,B=100,A=255)

	Teamname_ColorTexture_Background = (R=102,G=111,B=126,A=100)
	Teamname_Texture_Background = Texture'Textures.TeamInfo'

	TeamName_Color_LocationName = (R=172,G=124,B=0,A=255)

	TeamRadar_Texture_Level = Texture'Textures.Symbols'
	TeamRadar_OffsetTexture_Level = (X1=768,Y1=0,X2=831,Y2=63)

	TeammateRadar_BackgroundTextureColor = (R=255,G=255,B=255,A=159)

	// TeamResources

	TeamResourceBackgroundDisc(0) = (WidgetTexture=Texture'HudContent.Generic.HUD',PosX=0.5,PosY=0.0,OffsetX=-32,OffsetY=7,DrawPivot=DP_UpperRight,RenderStyle=STY_Alpha,TextureCoords=(X1=119,Y1=258,X2=173,Y2=313),TextureScale=0.6,ScaleMode=SM_Right,Scale=1.000000,Tints[0]=(G=255,R=255,B=255,A=255),Tints[1]=(G=255,R=255,B=255,A=255))
	TeamResourceBackgroundDisc(1) = (WidgetTexture=Texture'HudContent.Generic.HUD',PosX=0.5,PosY=0.0,OffsetX=32,OffsetY=7,DrawPivot=DP_UpperLeft,RenderStyle=STY_Alpha,TextureCoords=(X1=119,Y1=258,X2=173,Y2=313),TextureScale=0.6,ScaleMode=SM_Right,Scale=1.000000,Tints[0]=(G=255,R=255,B=255,A=255),Tints[1]=(G=255,R=255,B=255,A=255))

	TeamResourceSymbol(0) = (WidgetTexture=Texture'Textures.Symbols',PosX=0.5,PosY=0.0,OffsetX=-43,OffsetY=13,DrawPivot=DP_UpperRight,RenderStyle=STY_Alpha,TextureCoords=(X1=192,Y1=64,X2=255,Y2=127),TextureScale=0.47,ScaleMode=SM_Right,Scale=1.000000,Tints[0]=(B=127,G=127,R=127,A=50),Tints[1]=(B=127,G=127,R=127,A=50))
	TeamResourceSymbol(1) = (WidgetTexture=Texture'Textures.Symbols',PosX=0.5,PosY=0.0,OffsetX=44,OffsetY=13,DrawPivot=DP_UpperLeft,RenderStyle=STY_Alpha,TextureCoords=(X1=192,Y1=64,X2=255,Y2=127),TextureScale=0.47,ScaleMode=SM_Right,Scale=1.000000,Tints[0]=(B=127,G=127,A=50),Tints[1]=(B=127,G=127,R=127,A=50))

	TeamResourceSymbolOverlay(0) = (WidgetTexture=Texture'Textures.Symbols',PosX=0.5,PosY=0.0,OffsetX=-43,OffsetY=13,DrawPivot=DP_UpperRight,RenderStyle=STY_Alpha,TextureCoords=(X1=192,Y1=64,X2=255,Y2=127),TextureScale=0.47,ScaleMode=SM_Up,Scale=1.000000,Tints[0]=(B=127,G=127,R=127,A=150),Tints[1]=(B=127,G=127,R=127,A=150))
	TeamResourceSymbolOverlay(1) = (WidgetTexture=Texture'Textures.Symbols',PosX=0.5,PosY=0.0,OffsetX=44,OffsetY=13,DrawPivot=DP_UpperLeft,RenderStyle=STY_Alpha,TextureCoords=(X1=192,Y1=64,X2=255,Y2=127),TextureScale=0.47,ScaleMode=SM_Up,Scale=1.000000,Tints[0]=(B=127,G=127,A=150),Tints[1]=(B=127,G=127,R=127,A=150))

	TeamPlayerCountOverlay(0) = (WidgetTexture=Texture'Textures.Symbols',PosX=0.5,PosY=0.0,OffsetX=-68,OffsetY=28,DrawPivot=DP_UpperRight,RenderStyle=STY_Alpha,TextureCoords=(X1=256,Y1=0,X2=320,Y2=64),TextureScale=0.35,ScaleMode=SM_Up,Scale=1.000000,Tints[0]=(B=127,G=127,R=127,A=150),Tints[1]=(B=127,G=127,R=127,A=150))
	TeamPlayerCountOverlay(1) = (WidgetTexture=Texture'Textures.Symbols',PosX=0.5,PosY=0.0,OffsetX=68,OffsetY=28,DrawPivot=DP_UpperLeft,RenderStyle=STY_Alpha,TextureCoords=(X1=256,Y1=0,X2=320,Y2=64),TextureScale=0.35,ScaleMode=SM_Up,Scale=1.000000,Tints[0]=(B=127,G=127,A=150),Tints[1]=(B=127,G=127,R=127,A=150))

	TeamResourceFinal(0) = (WidgetTexture=Material'HudContent.Generic.HUDPulse',PosX=0.5,PosY=0.0,OffsetX=-105,OffsetY=40,DrawPivot=DP_UpperRight,RenderStyle=STY_Alpha,TextureCoords=(X1=341,Y1=211,X2=366,Y2=271),TextureScale=0.3,ScaleMode=SM_Right,Scale=1.000000,Tints[0]=(G=255,R=255,B=255,A=255),Tints[1]=(G=255,R=255,B=255,A=255))
	TeamResourceFinal(1) = (WidgetTexture=Material'HudContent.Generic.HUDPulse',PosX=0.5,PosY=0.0,OffsetX=105,OffsetY=40,DrawPivot=DP_UpperLeft,RenderStyle=STY_Alpha,TextureCoords=(X1=341,Y1=211,X2=366,Y2=271),TextureScale=0.3,ScaleMode=SM_Right,Scale=1.000000,Tints[0]=(G=255,R=255,B=255,A=255),Tints[1]=(G=255,R=255,B=255,A=255))

	TeamResources(0) = (RenderStyle=STY_Alpha,MinDigitCount=1,TextureScale=0.24,DrawPivot=DP_MiddleMiddle,PosX=0.5,PosY=0.0,OffsetX=-149,OffsetY=88,Tints[0]=(B=255,G=255,R=255,A=255),Tints[1]=(B=255,G=255,R=255,A=255))
	TeamResources(1) = (RenderStyle=STY_Alpha,MinDigitCount=1,TextureScale=0.24,DrawPivot=DP_MiddleMiddle,PosX=0.5,PosY=0.0,OffsetX=149,OffsetY=88,Tints[0]=(B=255,G=255,R=255,A=255),Tints[1]=(B=255,G=255,R=255,A=255))

	TeamSymbolsAlternativePosX = 0.5
	TeamScoreBackgroundDiscAlternativePosX = 0.5
	TeamScoreBackgroundAlternativePosX = 0.5
	TeamSymbolsAlternativeOffsetX(0) = -240
	TeamSymbolsAlternativeOffsetX(1) = 240

	ScoreTeamAlternativePosX = 0.5
	ScoreTeamAlternativeOffsetX(0) = -90
	ScoreTeamAlternativeOffsetX(1) = 90

	// HUDCTeamDeathMatch

	TeamScoreBackground(0) = (WidgetTexture=Texture'HudContent.Generic.HUD',PosX=0.442,PosY=0.0,OffsetX=-50,OffsetY=10,DrawPivot=DP_UpperRight,RenderStyle=STY_Alpha,TextureCoords=(X1=168,Y1=211,X2=334,Y2=255),TextureScale=0.53,ScaleMode=SM_Right,Scale=1.000000,Tints[0]=(G=0,R=0,B=0,A=150),Tints[1]=(G=0,R=0,B=0,A=150))
	TeamScoreBackground(1) = (WidgetTexture=Texture'HudContent.Generic.HUD',PosX=0.558,PosY=0.0,OffsetX=50,OffsetY=10,DrawPivot=DP_UpperLeft,RenderStyle=STY_Alpha,TextureCoords=(X1=168,Y1=211,X2=334,Y2=255),TextureScale=0.53,ScaleMode=SM_Right,Scale=1.000000,Tints[0]=(G=0,R=0,B=0,A=150),Tints[1]=(G=0,R=0,B=0,A=150))

	TeamScoreBackgroundDisc(0) = (WidgetTexture=Texture'HudContent.Generic.HUD',PosX=0.442,PosY=0.0,OffsetX=-35,OffsetY=5,DrawPivot=DP_UpperRight,RenderStyle=STY_Alpha,TextureCoords=(X1=119,Y1=258,X2=173,Y2=313),TextureScale=0.53,ScaleMode=SM_Right,Scale=1.000000,Tints[0]=(G=255,R=255,B=255,A=255),Tints[1]=(G=255,R=255,B=255,A=255))
	TeamScoreBackgroundDisc(1) = (WidgetTexture=Texture'HudContent.Generic.HUD',PosX=0.558,PosY=0.0,OffsetX=35,OffsetY=5,DrawPivot=DP_UpperLeft,RenderStyle=STY_Alpha,TextureCoords=(X1=119,Y1=258,X2=173,Y2=313),TextureScale=0.53,ScaleMode=SM_Right,Scale=1.000000,Tints[0]=(G=255,R=255,B=255,A=255),Tints[1]=(G=255,R=255,B=255,A=255))

	TeamSymbols(0) = (RenderStyle=STY_Alpha,TextureCoords=(X2=256,Y2=256),TextureScale=0.100000,DrawPivot=DP_UpperRight,PosX=0.445,PosY=0.0,OffsetX=-260,OffsetY=60,Tints[0]=(B=100,G=100,R=255,A=150),Tints[1]=(B=32,G=32,R=255,A=150))
	TeamSymbols(1) = (RenderStyle=STY_Alpha,TextureCoords=(X2=256,Y2=256),TextureScale=0.100000,PosX=0.555,PosY=0.0,OffsetX=260,OffsetY=60,Tints[0]=(B=255,G=128,A=150),Tints[1]=(B=255,G=210,R=32,A=150))

	ScoreTeam(0) = (RenderStyle=STY_Alpha,MinDigitCount=1,TextureScale=0.54,DrawPivot=DP_MiddleRight,PosX=0.442,PosY=0.0,OffsetX=-90,OffsetY=32,Tints[0]=(B=255,G=255,R=255,A=255),Tints[1]=(B=255,G=255,R=255,A=255))
	ScoreTeam(1) = (RenderStyle=STY_Alpha,MinDigitCount=1,TextureScale=0.54,DrawPivot=DP_MiddleLeft,PosX=0.558,PosY=0.0,OffsetX=90,OffsetY=32,Tints[0]=(B=255,G=255,R=255,A=255),Tints[1]=(B=255,G=255,R=255,A=255))
}
