 // *F* Developed 2011-2014 rejecht <rejecht at outlook dot com>

class ReTeamGame extends xTeamGame
	abstract
	config
	CacheExempt
	HideDropDown
	;

const IndexTeam_Max = 2;
const IndexTeam_A = 0;
const IndexTeam_B = 1;
const IndexTeam_Unspecified = 255;

// Verbosity

const bAdvanced = True;
const bMultiplayerOnly = True;

// Semantics

var const Vector NullVector;
var const Rotator NullRotator;

// #TeamGameEvents

var protected const GlobalConfig string TeamGameEventsClassNameString;
var protected Class<TeamGameEvents> TeamGameEventsClass;
var protected TeamGameEvents TeamGameEvents;

// #ServerBrowser

var protected GlobalConfig bool bIncludeBotNames;

var protected Color DefaultValueColor, CustomValueColor, PlayerTeamColor[IndexTeam_Max];

var protected string
	DefaultValueColorCodeString,
	CustomValueColorCodeString,
	PlayerTeamColorColorCodeString[IndexTeam_Max];

// #

const MaxCustomTeamNameLength = 20;

var const bool bGarbageCollection;

// #Spawning

var(Spawning) protected bool bVersus; // Set whenever two players are left; one on each team.

struct SPlayerSpawnInfo
{
	var Vector PawnLocation;
	var float PawnCollisionSize;
	var int PawnTeamIndex;
	var byte ZoneNumber;
};

var(Spawning) array<SPlayerSpawnInfo> PlayerSpawnInfo; // Cached info for RatePlayerStart optimization.

// #bIsServer

var(Cached) protected bool bIsServer;

// #Adrenaline

var(Adrenaline) protected float ADR_MajorKill_Retribution;

//

var(Rules) protected const bool
	bCheerfulBots, // Bots are forced to celebrate
	bRoundbased, // To help make different decisions
	bFreezeAdrenaline, // Players keep their adrenaline (this feature is no longer fully supported)
	bResetSpree, // Players get their spree reset when a round resets (even if they were alive)
	bFreezeEndRound; // Gameplay stops when round ends

// #PlayInfo

var(PlayInfo) protected const string ShortGameName;

var(PlayInfo) protected const string
	PlayInfoGroupGame,
	PlayInfoGroupTeams,
	PlayInfoGroupPickups,
	PlayInfoGroupStats,
	PlayInfoGroupPlayers,
	PlayInfoGroupCamping,
	PlayInfoGroupAdrenaline;

// #Masquerading

// This settings applies to the whole server instance.

var(GlobalRules) protected const GlobalConfig bool bMasquerade; // Used to advertise server under a fixed game type.
var(GlobalRules) protected const GlobalConfig string MasqueradingGameType;
var(GlobalRules) protected const GlobalConfig bool bShowGameType; // The active game type is shown in level title "MapName (GameTypeAcronym)"

// #PlayerRegion

var(Region) protected const GlobalConfig string PlayerRegionClassNameString;
var(Region) protected Class<PlayerRegion> PlayerRegionClass;
var(Region) protected PlayerRegion PlayerRegion; // Default: Handles the CountryCode.

// #Messages

var(Messages) protected const Class<LocalMessage>
	AwardBroadcastLocalMessageClass,
	DrawRoundLocalMessageClass;

// #bShowBalanceTeams

var protected const config bool bShowBalanceTeams;

// #TeamNames

var config protected string RedTeamNameString;
var(PlayInfo) private const string RedTeamNamePropertyText;
var(PlayInfo) private localized const string
	RedTeamNameDisplayText,
	RedTeamNameDescriptionText;

var config protected string BlueTeamNameString;
var(PlayInfo) private const string BlueTeamNamePropertyText;
var(PlayInfo) private localized const string
	BlueTeamNameDisplayText,
	BlueTeamNameDescriptionText;

// #MatchStatistics

// :Constants

const ContentString = "\"";
const SeparatorString = ",";

const RowStringLengthLimit = 1024;

// :bLogMatchStatistics

var(Rules) protected GlobalConfig bool bLogMatchStatistics;
var(PlayInfo) private const string LogMatchStatisticsPropertyText;
var(PlayInfo) private localized const string
	LogMatchStatisticsDisplayText,
	LogMatchStatisticsDescriptionText;

// :bIncludePlayerGUID

var(Rules) protected GlobalConfig bool bIncludePlayerGUID;
var(PlayInfo) private const string LogPlayerGUIDPropertyText;
var(PlayInfo) private localized const string
	LogPlayerGUIDDisplayText,
	LogPlayerGUIDDescriptionText;

// :bPublicMatchStatistics

var(Rules) protected GlobalConfig bool bPublicMatchStatistics;
var(PlayInfo) private const string PublicMatchStatisticsPropertyText;
var(PlayInfo) private localized const string
	PublicMatchStatisticsDisplayText,
	PublicMatchStatisticsDescriptionText;

// :Implementation

var protected bool bLogMatchStatisticsOpenFailed;
var protected bool bMatchStatisticsCompiled;

var protected MatchLog MatchStatistics_MatchLog;

var array<string> MatchStatistics_Rows;

// #bTossWeapon

var(Rules) protected config bool bTossWeapon;
var(PlayInfo) private const string TossWeaponPropertyText;
var(PlayInfo) private localized const string
	TossWeaponDisplayText,
	TossWeaponDescriptionText;

// #DamageBonus

var(Adrenaline) protected float ADR_DamageBonus;

var(Rules) protected config bool bDamageBonus;
var(PlayInfo) private const string DamageBonusPropertyText;
var(PlayInfo) private localized const string
	DamageBonusDisplayText,
	DamageBonusDescriptionText;

// #bAllowTrackingDamageIndicators

var(Rules) protected config bool bAllowTrackingDamageIndicators;
var(PlayInfo) private const string AllowTrackingDamageIndicatorsPropertyText;
var(PlayInfo) private localized const string
	AllowTrackingDamageIndicatorsDisplayText,
	AllowTrackingDamageIndicatorsDescriptionText;

const CampingLocationHistoryCount = 10;
var(Messages) protected const class<LocalMessage> CampingLocalMessageClass;
var protected const class<DamageType> CampingDamageTypeClass;

// #bCampingCheck

var(Rules) protected config bool bCampingCheck;
var(PlayInfo) private const string CampingCheckPropertyText;
var(PlayInfo) private localized const string
	CampingCheckDisplayText,
	CampingCheckDescriptionText;

// #

var(Rules) protected config bool bCampingCheckSendLocation;
var(PlayInfo) private const string CampingCheckSendLocationPropertyText;
var(PlayInfo) private localized const string
	CampingCheckSendLocationDisplayText,
	CampingCheckSendLocationDescriptionText;

// #CampingCheckVolume

var(Limits) protected const float
	CampingCheckVolume_Min,
	CampingCheckVolume_Max;

var(Rules) protected config float CampingCheckVolume;
var(PlayInfo) private const string CampingCheckVolumePropertyText;
var(PlayInfo) private localized const string
	CampingCheckVolumeDisplayText,
	CampingCheckVolumeDescriptionText;

// #Camping

// :bCampingCheckKickRepeatedOffenders

var(Rules) protected config bool bCampingCheckKickRepeatedOffenders;
var(PlayInfo) private const string CampingCheckKickRepeatedOffendersPropertyText;
var(PlayInfo) private localized const string
	CampingCheckKickRepeatedOffendersDisplayText,
	CampingCheckKickRepeatedOffendersDescriptionText;

// :CampingCheckInterval

var(Rules) protected int CampingCheckInterval;

// #bDenySuicide

var(Rules) protected config bool bDenySuicide;
var(PlayInfo) private const string DenySuicidePropertyText;
var(PlayInfo) private localized const string
	DenySuicideDisplayText,
	DenySuicideDescriptionText;

// #

enum EEndGameCause
{
	Unspecified,
	MatchOver_NoParticipants,
	MatchOver_Forfeit,
	MatchOver_ScoreLimit,
	MatchOver_TimeLimit,
	MatchOver_LastPlayerStanding,
	MatchOver_LastTeamStanding
};

var protected EEndGameCause EndGameCause;

// #AdrenalineCombo

const AdrenalineComboControl_Count = 6;

var(PlayInfo) private const string AdrenalineComboControlPropertyText[AdrenalineComboControl_Count];
var(PlayInfo) private localized const string
	AdrenalineComboControlDisplayText[AdrenalineComboControl_Count],
	AdrenalineComboControlDescriptionText[AdrenalineComboControl_Count];

var protected config const array<const string> DenyCustomAdrenalineCombo;

var protected array<string> DenyAdrenalineCombo;

var(Rules) protected config bool
	bDisableSpeedCombo,
	bDisableBoosterCombo,
	bDisableInvisibleCombo,
	bDisableBerserkCombo,
	bDisablePintsizedCombo,
	bDisableCamouflageCombo;

// #TeamCombos

// :bTeamCombos

var(Rules) protected config bool bTeamCombos;
var(PlayInfo) private const string TeamCombosPropertyText;
var(PlayInfo) private localized const string
	TeamCombosDescriptionText,
	TeamCombosDisplayText;

// :Implementation

var(Spawned) protected TeamComboManager TeamComboManager[IndexTeam_Max];

// #PickupNotifications

var(Rules) protected config bool bPickupNotifications;
var(PlayInfo) private const string PickupNotificationsPropertyText;
var(PlayInfo) private localized const string
	PickupNotificationsDisplayText,
	PickupNotificationsDescriptionText;

// #ScoreRecovery #Internal

var protected const bool bScoreRecoveryEnabled;
var protected const Class<ScoreRecoveryObject_Re> Class_ScoreRecoveryObject;
var protected transient array <ScoreRecoveryObject_Re> ScoreRecoveryGroup;

enum EKillEventContext
{
	KillEventContext_None,
	KillEventContext_EnemyKill,
	KillEventContext_TeamKill,
	KillEventContext_Suicide,
	KillEventContext_Death
};

struct SKillEventData
{
	var Controller
		Killer,
		Victim,
		DarkHorse;

	var Class<DamageType> Class_DamageType;

	var float TimeSeconds_Updated;

	var int IndexTeam_Killer, IndexTeam_Victim;

	var EKillEventContext Context;

	var bool
		bFirstKill,
		bFirstDeath,
		bVictimOutofLives,
		bVictimRespawnsDepleted;
};

var SKillEventData KillEventData; // Updated each time Killed gets called.

struct STeamEventData
{
	var Controller Last[IndexTeam_Max];

	var float TimeSeconds_Updated;

	var int
		Count_Out[IndexTeam_Max],
		Count_In[IndexTeam_Max],
		Count_Lives[IndexTeam_Max],
		Count_Deaths[IndexTeam_Max];
};

var STeamEventData TeamEventData;

// #RageQuit #Internal

var protected bool bRageQuitAnnouncement;
const RageQuit_TimeSeconds_Interval = 3.5f;
var int RageQuit_Count_Ragequitters;

// #SpawnManager

var(Spawning) protected array<PlayerStart> SpawnManager_PlayerStartGroup;
var(Spawning) protected array<Teleporter> SpawnManager_TeleportGroup;
var(Spawning) protected array<NavigationPoint> SpawnManager_SpawnPointGroup;
var(Spawning) protected int SpawnManager_ZoneCount;

var(Spawning) protected float SpawnManager_LongestDistance;

// #Limits

var(Limits) const int
	GoalScore_Min,
	GoalScore_Max,
	MaxLives_Min,
	MaxLives_Max,
	MaxPlayers_Min,
	MaxPlayers_Max,
	MaxSpectators_Min,
	MaxSpectators_Max,
	ResetTimeDelay_Min,
	ResetTimeDelay_Max,
	TimeLimit_Min,
	TimeLimit_Max,
	MinNetPlayers_Min,
	MinNetPlayers_Max,
	NetWait_Min,
	NetWait_Max,
	RestartWait_Min,
	RestartWait_Max;

var(Limits) const float
	SpawnProtectionTime_Min,
	SpawnProtectionTime_Max;

var protected const class<UnrealPawn> SpawnBot_Class_Pawn;

// #Lifetime

event PreBeginPlay ()
{
	bIsServer = (Level.NetMode == NM_DedicatedServer) || (Level.NetMode == NM_ListenServer);


	CustomValueColorCodeString = Class'ReUtility'.Static.GetColorCodeString (CustomValueColor);
	DefaultValueColorCodeString = Class'ReUtility'.Static.GetColorCodeString (DefaultValueColor);

	PlayerTeamColorColorCodeString[IndexTeam_A] = Class'ReUtility'.Static.GetColorCodeString (PlayerTeamColor[IndexTeam_A]);
	PlayerTeamColorColorCodeString[IndexTeam_B] = Class'ReUtility'.Static.GetColorCodeString (PlayerTeamColor[IndexTeam_B]);


	Super.PreBeginPlay ();
}


event Destroyed ()
{
	TeamCombos_Destroyed ();

	MatchStatistics_Destroy ();

	PlayerRegion_Destroy ();

	TeamGameEvents_Destroy ();

	ScoreRecoveryObjectGroup_Destroy ();

	Super.Destroyed ();
}

// Override team names.

event UnrealTeamInfo GetRedTeam (int TeamBots)
{
	local UnrealTeamInfo RedTeamInfo;
	local string TeamNameString;

	TeamNameString = Left (Class'ReUtility'.Static.GetPlainString (RedTeamNameString), MaxCustomTeamNameLength);

	RedTeamInfo = Super.GetRedTeam (TeamBots);

	if
	(
		(TeamNameString != "")
		&&
		(RedTeamInfo != None)
	)
	{
		RedTeamInfo.TeamName = TeamNameString;
    }

	return RedTeamInfo;
}

event UnrealTeamInfo GetBlueTeam (int TeamBots)
{
	local UnrealTeamInfo BlueTeamInfo;
	local string TeamNameString;

	TeamNameString = Left (Class'ReUtility'.Static.GetPlainString (BlueTeamNameString), MaxCustomTeamNameLength);

	BlueTeamInfo = Super.GetBlueTeam (TeamBots);

	if
	(
		(TeamNameString != "")
		&&
		(BlueTeamInfo != None)
	)
	{
		BlueTeamInfo.TeamName = TeamNameString;
    }

	return BlueTeamInfo;
}

// #

event InitGame (string OptionsString, out string Out_ErrorMessageString)
{
	OptionsString = Class'ReUtility'.Static.GetPlainString (OptionsString);

	Super.InitGame (OptionsString, Out_ErrorMessageString);

	// Toss Weapon

	bTossWeapon = GetBooleanOption (OptionsString, TossWeaponPropertyText, bTossWeapon);

	// Team Combos

	bTeamCombos = GetBooleanOption (OptionsString, TeamCombosPropertyText, bTeamCombos);

	// Standard Adrenaline Combos

	bDisableSpeedCombo = GetBooleanOption (OptionsString, "bDisableSpeedCombo", bDisableSpeedCombo);
	bDisableBoosterCombo = GetBooleanOption (OptionsString, "bDisableBoosterCombo", bDisableBoosterCombo);
	bDisableInvisibleCombo = GetBooleanOption (OptionsString, "bDisableInvisibleCombo", bDisableInvisibleCombo);
	bDisableBerserkCombo = GetBooleanOption (OptionsString, "bDisableBerserkCombo", bDisableBerserkCombo);
	bDisablePintsizedCombo = GetBooleanOption (OptionsString, "bDisablePintsizedCombo", bDisablePintsizedCombo);
	bDisableCamouflageCombo = GetBooleanOption (OptionsString, "bDisableCamouflageCombo", bDisableCamouflageCombo);

	// Damage Bonus

	bDamageBonus = GetBooleanOption (OptionsString, DamageBonusPropertyText, bDamageBonus);

	// Match Statistics

	bLogMatchStatistics = GetBooleanOption (OptionsString, LogMatchStatisticsPropertyText, bLogMatchStatistics);
	bIncludePlayerGUID = GetBooleanOption (OptionsString, LogPlayerGUIDPropertyText, bIncludePlayerGUID);
	bPublicMatchStatistics = GetBooleanOption (OptionsString, PublicMatchStatisticsPropertyText, bPublicMatchStatistics);

	// Allow Tracking Damage Indicators

	bAllowTrackingDamageIndicators = GetBooleanOption (OptionsString, AllowTrackingDamageIndicatorsPropertyText, bAllowTrackingDamageIndicators);

	// Pickup Notifications

	bPickupNotifications = GetBooleanOption (OptionsString, PickupNotificationsPropertyText, bPickupNotifications);

	// Camping Check

	bCampingCheck = GetBooleanOption (OptionsString, CampingCheckPropertyText, bCampingCheck);
	bCampingCheckSendLocation = GetBooleanOption (OptionsString, CampingCheckSendLocationPropertyText, bCampingCheckSendLocation);
	CampingCheckVolume = FClamp (GetFloatOption (OptionsString, CampingCheckVolumePropertyText, CampingCheckVolume), CampingCheckVolume_Min, CampingCheckVolume_Max);
	bCampingCheckKickRepeatedOffenders = GetBooleanOption (OptionsString, CampingCheckKickRepeatedOffendersPropertyText, bCampingCheckKickRepeatedOffenders);

	// Suicide

	bDenySuicide = GetBooleanOption (OptionsString, DenySuicidePropertyText, bDenySuicide);

	// GameInfo (Limits)

	GoalScore = Clamp (GetIntOption (OptionsString, "GoalScore", GoalScore), GoalScore_Min, GoalScore_Max);
	MaxLives = Clamp (GetIntOption (OptionsString, "MaxLives", MaxLives), MaxLives_Min, MaxLives_Max);
	MaxPlayers = Clamp (GetIntOption (OptionsString, "MaxPlayers", MaxPlayers), MaxPlayers_Min, MaxPlayers_Max);
	MaxSpectators = Clamp (GetIntOption (OptionsString, "MaxSpectators", MaxSpectators), MaxSpectators_Min, MaxSpectators_Max);
	ResetTimeDelay = Clamp (GetIntOption (OptionsString, "ResetTimeDelay", ResetTimeDelay), ResetTimeDelay_Min, ResetTimeDelay_Max);
	TimeLimit = Clamp (GetIntOption (OptionsString, "TimeLimit", TimeLimit), TimeLimit_Min, TimeLimit_Max);

	// DeathMatch (Limits)

	MinNetPlayers = Clamp (GetIntOption (OptionsString, "MinNetPlayers", MinNetPlayers), MinNetPlayers_Min, MinNetPlayers_Max);
	NetWait = Clamp (GetIntOption (OptionsString, "NetWait", NetWait), NetWait_Min, NetWait_Max);
	RestartWait = Clamp (GetIntOption (OptionsString, "RestartWait", RestartWait), RestartWait_Min, RestartWait_Max);
	SpawnProtectionTime = FClamp (GetIntOption (OptionsString, "SpawnProtectionTime", SpawnProtectionTime), SpawnProtectionTime_Min, SpawnProtectionTime_Max);

	// Spawn Manager

	SpawnManager_InitGame ();

	// Adrenaline Combo Control

	AdrenalineComboControl_InitGame ();

	// Team Game Events

	TeamGameEvents_InitGame ();

	// Player Region

	PlayerRegion_InitGame ();

	// Match Statistics

	MatchStatistics_InitGame ();

	// About

	Class'About'.Static.About (Level);

	// Save

	SaveConfig ();
}

event PostBeginPlay ()
{
	Super.PostBeginPlay ();

	// Team Combos

	TeamCombos_PostBeginPlay ();
}

// #TeamCombos

protected function TeamCombos_PostBeginPlay ()
{
	if (ReGRI(Level.GRI).bTeamCombos)
	{
		TeamCombos_SpawnTeamComboManager (IndexTeam_A);
		ReGRI(Level.GRI).TeamComboManager[IndexTeam_A] = TeamComboManager[IndexTeam_A];
		ReGRI(Level.GRI).TeamComboReplicationInfo[IndexTeam_A] = TeamComboManager[IndexTeam_A].TCRI;

		TeamCombos_SpawnTeamComboManager (IndexTeam_B);
		ReGRI(Level.GRI).TeamComboManager[IndexTeam_B] = TeamComboManager[IndexTeam_B];
		ReGRI(Level.GRI).TeamComboReplicationInfo[IndexTeam_B] = TeamComboManager[IndexTeam_B].TCRI;
	}
}

protected function TeamCombos_Destroyed ()
{
	if (ReGRI(Level.GRI).bTeamCombos)
	{
		TeamCombos_DestroyTeamComboManager (IndexTeam_A);
		TeamCombos_DestroyTeamComboManager (IndexTeam_B);
	}
}

protected function TeamCombos_SpawnTeamComboManager (int TeamIndex)
{
	TeamComboManager[TeamIndex] = Spawn (Class'TeamComboManager');

	if (TeamComboManager[TeamIndex] != None)
	{
		TeamComboManager[TeamIndex].InitializeTeamIndex (TeamIndex);
	}
	else
	{
		Warn ("TeamComboManager[" $ TeamIndex $ "]");
	}
}

protected function TeamCombos_DestroyTeamComboManager (int TeamIndex)
{
	if (TeamComboManager[TeamIndex] != None)
	{
		TeamComboManager[TeamIndex].Destroy ();
		TeamComboManager[TeamIndex] = None;
	}
}

//

event Reset ()
{
	bOvertime = False;
	bOvertimeBroadcast = False;

	KillEvent_Reset ();
}

// #

protected function MatchStatistics_InitGame ()
{
}

protected function MatchStatistics_Destroy ()
{
	if (MatchStatistics_MatchLog != None)
	{
		MatchStatistics_MatchLog.Destroyed ();
		MatchStatistics_MatchLog = None;
	}
}

protected function bool MatchStatistics_Opened ()
{
	if (MatchStatistics_MatchLog != None)
	{
		return True;
	}
	else if (!bLogMatchStatisticsOpenFailed)
	{
		MatchStatistics_MatchLog = Spawn (Class'MatchLog');

		if (MatchStatistics_MatchLog != None)
		{
			return True;
		}
		else
		{
			Warn ("(MatchStatisticsLog == None)");

			bLogMatchStatisticsOpenFailed = True;
		}
	}

	return False;
}

// Log match statistics.

protected function MatchStatistics_MatchEnded ()
{
	if
	(
		bPublicMatchStatistics
		||
		bLogMatchStatistics
	)
	{
		MatchStatistics_CompileRows ();

		bMatchStatisticsCompiled = True;

		if (bLogMatchStatistics)
		{
			MatchStatistics_Logged ();
		}
	}
}

protected function bool MatchStatistics_Logged ()
{
	local int Index;

	if (MatchStatistics_Opened ())
	{
		for (Index = 0; Index < MatchStatistics_Rows.Length; ++Index)
		{
			if (!MatchStatistics_LogRow (MatchStatistics_Rows[Index]))
			{
				return False;
			}
		}

		return True;
	}
	else
	{
		return False;
	}
}

function bool MatchStatistics_Compiled ()
{
	return bMatchStatisticsCompiled;
}

protected function MatchStatistics_CompileRows ()
{
	local PlayerReplicationInfo OtherPRI;
	local Controller Other;
	local int Index, Count;
	local string CachedString;

	// First Row / Header Row

	CachedString = MatchStatistics_CompileHeaderRow ();

	if (Len (CachedString) <= RowStringLengthLimit)
	{
		MatchStatistics_Rows[MatchStatistics_Rows.Length] = CachedString;
	}
	else
	{
		MatchStatistics_Rows[MatchStatistics_Rows.Length] = Left (CachedString, RowStringLengthLimit);

		Warn ("Truncated: (Len (CachedString) > " $ RowStringLengthLimit $ ") - Length:" @ Len (CachedString));
	}

	// Consecutive Rows / Data Rows

	SortPRIArray ();

	Count = GameReplicationInfo.PRIArray.Length;

	for (Index = 0; Index < Count; ++Index)
	{
		OtherPRI = GameReplicationInfo.PRIArray[Index];

		Other = Controller(OtherPRI.Owner);

		if
		(
			(Other != None)
			&&
			Other.bIsPlayer
			&&
			(Other.PlayerReplicationInfo != None)
			&&
			!Other.PlayerReplicationInfo.bOnlySpectator
		)
		{
			CachedString = MatchStatistics_CompileDataRow (Other);

			if (Len (CachedString) <= RowStringLengthLimit)
			{
				MatchStatistics_Rows[MatchStatistics_Rows.Length] = CachedString;
			}
			else
			{
				MatchStatistics_Rows[MatchStatistics_Rows.Length] = Left (CachedString, RowStringLengthLimit);

				Warn ("Truncated: (Len (CachedString) > " $ RowStringLengthLimit $ ") - Length:" @ Len (CachedString));
			}
		}
	}
}

protected final function SortPRIArray ()
{
    local int i,j;
    local PlayerReplicationInfo tmp;

    for (i=0; i<GameReplicationInfo.PRIArray.Length-1; i++)
    {
        for (j=i+1; j<GameReplicationInfo.PRIArray.Length; j++)
        {
            if( !InOrder( GameReplicationInfo.PRIArray[i], GameReplicationInfo.PRIArray[j] ) )
            {
                tmp = GameReplicationInfo.PRIArray[i];
                GameReplicationInfo.PRIArray[i] = GameReplicationInfo.PRIArray[j];
                GameReplicationInfo.PRIArray[j] = tmp;
            }
        }
    }
}

protected final function bool InOrder (PlayerReplicationInfo P1, PlayerReplicationInfo P2)
{
    if( P1.bOnlySpectator )
    {
        if( P2.bOnlySpectator )
            return true;
        else
            return false;
    }
    else if ( P2.bOnlySpectator )
        return true;

    if( P1.Score < P2.Score )
        return false;
    if( P1.Score == P2.Score )
    {
		if ( P1.Deaths > P2.Deaths )
			return false;
		if ( (P1.Deaths == P2.Deaths) && (PlayerController(P2.Owner) != None) && (Viewport(PlayerController(P2.Owner).Player) != None) )
			return false;
	}

    return true;
}

protected function string MatchStatistics_CompileHeaderRow ()
{
	return
	(
		ContentString $ "GUID" $ ContentString
		$ SeparatorString $ ContentString $ "Name" $ ContentString
		$ SeparatorString $ ContentString $ "Score" $ ContentString
	);
}

protected function string MatchStatistics_CompileDataRow (Controller Other)
{
	local PlayerReplicationInfo OtherPRI;
	local string GUIDString;

	OtherPRI = Other.PlayerReplicationInfo;

	if
	(
		(Other.IsA ('PlayerController'))
		&&
		bIncludePlayerGUID
	)
	{
		GUIDString = PlayerController(Other).GetPlayerIDHash ();
	}
	else // Bot || !bIncludePlayerGUID
	{
		GUIDString = "";
	}

	return
	(
		ContentString $ GUIDString $ ContentString // GUID
		$ SeparatorString $ ContentString $ OtherPRI.PlayerName $ ContentString // Name
		$ SeparatorString $ ContentString $ OtherPRI.Score $ ContentString // Score
	);
}

protected function bool MatchStatistics_LogRow (string RowString)
{
	if (MatchStatistics_Opened ())
	{
		MatchStatistics_MatchLog.LogString (RowString);

		return True;
	}
	else
	{
		return False;
	}
}

// Preconditions:
// - (Index > -1)
// Returns ("") if Index is greater than the number of array elements.

function string MatchStatistics_GetRowString (int Index)
{
	if (Index < MatchStatistics_Rows.Length)
	{
		return MatchStatistics_Rows[Index];
	}
	else
	{
		return "";
	}
}

// #TeamGameEvents

protected function TeamGameEvents_InitGame ()
{
	local Class<TeamGameEvents> LoadedTeamGameEventsClass;

	LoadedTeamGameEventsClass = Class<TeamGameEvents>(DynamicLoadObject (TeamGameEventsClassNameString, Class'Core.Class'));

	if (LoadedTeamGameEventsClass == None)
	{
		LoadedTeamGameEventsClass = Class<TeamGameEvents>(DynamicLoadObject (Default.TeamGameEventsClassNameString, Class'Core.Class'));
    }

	TeamGameEvents = Spawn (LoadedTeamGameEventsClass);

	if (TeamGameEvents != None)
	{
		TeamGameEventsClass = LoadedTeamGameEventsClass;

		Log ("TeamGameEventsClass: <" $ TeamGameEventsClass $ ">");
	}
	else
	{
		Warn ("(TeamGameEvents == None) - LoadedTeamGameEventsClass: <" $ LoadedTeamGameEventsClass $ ">");
	}
}

protected function TeamGameEvents_Destroy ()
{
	if (TeamGameEvents != None)
	{
		TeamGameEvents.Destroy ();
		TeamGameEvents = None;
	}
}

// #PlayerRegion

// Postconditions:
// - PlayerRegionClass and PlayerRegion are valid if PlayerRegion was spawned.

protected function PlayerRegion_InitGame ()
{
	local Class<PlayerRegion> LoadedPlayerRegionClass;

	LoadedPlayerRegionClass = Class<PlayerRegion>(DynamicLoadObject (PlayerRegionClassNameString, Class'Core.Class'));

	if (LoadedPlayerRegionClass == None)
	{
		LoadedPlayerRegionClass = Class<PlayerRegion>(DynamicLoadObject (Default.PlayerRegionClassNameString, Class'Core.Class'));
    }

	PlayerRegion = Spawn (LoadedPlayerRegionClass);

	if (PlayerRegion != None)
	{
		PlayerRegionClass = LoadedPlayerRegionClass;

		Log ("PlayerRegionClass: <" $ PlayerRegionClass $ ">");
	}
	else
	{
		Warn ("(PlayerRegion == None) - LoadedPlayerRegionClass: <" $ LoadedPlayerRegionClass $ ">");
	}
}

protected function PlayerRegion_Destroy ()
{
	if (PlayerRegion != None)
	{
		PlayerRegion.Destroy ();
		PlayerRegion = None;
	}
}

// #Adrenaline

// - Easy access to the stock adrenaline combos.

protected function AdrenalineComboControl_InitGame ()
{
	local int Index;

	if (bDisableSpeedCombo)
	{
		AdrenalineComboControl_Deny ("FFFF");
	}
	else
	{
		AdrenalineComboControl_Allow ("FFFF");
	}

	if (bDisableBoosterCombo)
	{
		AdrenalineComboControl_Deny ("BBBB");
	}
	else
	{
		AdrenalineComboControl_Allow ("BBBB");
	}

	if (bDisableInvisibleCombo)
	{
		AdrenalineComboControl_Deny ("RRLL");
	}
	else
	{
		AdrenalineComboControl_Allow ("RRLL");
	}

	if (bDisableBerserkCombo)
	{
		AdrenalineComboControl_Deny ("FFBB");
	}
	else
	{
		AdrenalineComboControl_Allow ("FFBB");
	}

	if (bDisablePintsizedCombo)
	{
		AdrenalineComboControl_Deny ("LLLL");
	}
	else
	{
		AdrenalineComboControl_Allow ("LLLL");
	}

	if (bDisableCamouflageCombo)
	{
		AdrenalineComboControl_Deny ("RRRR");
	}
	else
	{
		AdrenalineComboControl_Allow ("RRRR");
	}

	// - Ignore stock adrenaline combos.
	// - Append custom adrenaline combos.

	for (Index = 0; Index < DenyCustomAdrenalineCombo.Length; ++Index)
	{
		if (Class'PlayerSettings'.Static.IsKeyComboStringValid (DenyCustomAdrenalineCombo[Index]))
		{
			switch (DenyCustomAdrenalineCombo[Index])
			{
				case "BBBB": // Booster!
				case "FFFF": // Speed!
				case "FFBB": // Berserk!
				case "RRLL": // Invisible!
				case "LLLL": // Camouflage!
				case "RRRR": // Pint-sized!
					break;

				default:
					DenyAdrenalineCombo[DenyAdrenalineCombo.Length] = DenyCustomAdrenalineCombo[Index];
					break;
			}
		}
	}


	for (Index = 0; Index < DenyAdrenalineCombo.Length; ++Index)
	{
		Log ("AdrenalineComboDenied:" @ DenyAdrenalineCombo[Index]);
	}
}

// - By default, all adrenaline combos are allowed.
// - Only combos denied are remembered.
// - Consider that the same combo may be enabled several times.

protected function AdrenalineComboControl_Allow (string KeyComboString)
{
	local int Index;

	for (Index = 0; Index < DenyAdrenalineCombo.Length; ++Index)
	{
		if (KeyComboString ~= DenyAdrenalineCombo[Index])
		{
			DenyAdrenalineCombo.Remove (Index--, 1);
		}
	}
}

protected function AdrenalineComboControl_Deny (string KeyComboString)
{
	local int Index;

	for (Index = 0; Index < DenyAdrenalineCombo.Length; ++Index)
	{
		if (KeyComboString ~= DenyAdrenalineCombo[Index])
		{
			return;
		}
	}

	DenyAdrenalineCombo[DenyAdrenalineCombo.Length] = KeyComboString;
}

// - Called after InitGame.

event InitGameReplicationInfo ()
{
	Super.InitGameReplicationInfo ();


	ReGRI(GameReplicationInfo).bTeamCombos = bTeamCombos;
	ReGRI(GameReplicationInfo).bTossWeapon = bTossWeapon;
	ReGRI(GameReplicationInfo).bFreezeEndRound = bFreezeEndRound;
	ReGRI(GameReplicationInfo).bCampingCheck = bCampingCheck;
	ReGRI(GameReplicationInfo).bCampingCheckSendLocation = bCampingCheckSendLocation;
	ReGRI(GameReplicationInfo).CampingCheckVolume = CampingCheckVolume;
	ReGRI(GameReplicationInfo).bCampingCheckKickRepeatedOffenders = bCampingCheckKickRepeatedOffenders;
	ReGRI(GameReplicationInfo).DenyAdrenalineCombo = DenyAdrenalineCombo;
	ReGRI(GameReplicationInfo).bLogMatchStatistics = bLogMatchStatistics;
	ReGRI(GameReplicationInfo).bIncludePlayerGUID = bIncludePlayerGUID;
	ReGRI(GameReplicationInfo).bPublicMatchStatistics = bPublicMatchStatistics;
	ReGRI(GameReplicationInfo).bDamageBonus = bDamageBonus;
	ReGRI(GameReplicationInfo).bPickupNotifications = bPickupNotifications;
	ReGRI(GameReplicationInfo).bRoundbased = bRoundbased;
	ReGRI(GameReplicationInfo).bDenySuicide = bDenySuicide;
	ReGRI(GameReplicationInfo).bAllowTrackingDamageIndicators = bAllowTrackingDamageIndicators;
	ReGRI(GameReplicationInfo).MaxPlayers = MaxPlayers;
	ReGRI(GameReplicationInfo).MaxSpectators = MaxSpectators;
	ReGRI(GameReplicationInfo).SpawnProtectionTime = SpawnProtectionTime;
}

// #

protected function float GetFloatOption (string OptionsString, string KeyString, float Value)
{
	local string CachedString;

	CachedString = ParseOption (OptionsString, KeyString);

	if (CachedString != "")
	{
		return float(CachedString);
	}
	else
	{
		return Value;
	}
}

protected function bool GetBooleanOption (string OptionsString, string KeyString, bool bValue)
{
	local string CachedString;

	CachedString = ParseOption (OptionsString, KeyString);

	if (CachedString != "")
	{
		return bool(CachedString);
	}
	else
	{
		return bValue;
	}
}

// #Pickup

// (Context) Pawn acquires a pickup.
// - Announce who picked up what to human participants and spectators.
// - The local player gets notified once by itself and then
// made aware that others also received the pickup announcement.

protected function BroadcastPickupEvent
(
	class<LocalMessage> LocalMessageClass,
	PlayerReplicationInfo OtherPRI,
	optional int Contextual,
	optional Object O
)
{
	local Controller Other;

	for (Other = Level.ControllerList; Other != None; Other = Other.NextController)
	{
		if
		(
			Other.IsA ('PlayerController')
			&&
			Other.bIsPlayer
		)
		{
			PlayerController(Other).ReceiveLocalizedMessage
			(
				LocalMessageClass,
				Contextual,
				,
				OtherPRI,
				O
			);
		}
	}
}

protected function bool ShouldBroadcastPickup (Pickup OtherItem)
{
	return
	(
		OtherItem.IsA ('WeaponPickup')
		||
		OtherItem.IsA ('ShieldPickup')
		||
		OtherItem.IsA ('SuperHealthPack')
		||
		OtherItem.IsA ('UDamagePack')
	);
}

// - To notify everyone when a pawn picks up certain items.
// Context:
// - RePawn.HandlePickup ()

function PickupBroadcast (Pawn OtherPawn, Pickup OtherItem)
{
	if
	(
		ReGRI(Level.GRI).bPickupNotifications
		&&
		ShouldBroadcastPickup (OtherItem)
	)
	{
		if (!OtherItem.IsA ('WeaponPickup'))
		{
			BroadcastPickupEvent
			(
				Class'LocalMessage_Pickup_Remote',
				OtherPawn.PlayerReplicationInfo,
				,
				OtherItem.Class
			);
		}
		else if
		(
			!OtherItem.bDropped
			&&
			!WeaponPickup(OtherItem).bThrown
		)
		{
			if (!Level.Game.bWeaponStay)
			{
				BroadcastPickupEvent
				(
					Class'LocalMessage_Pickup_Remote',
					OtherPawn.PlayerReplicationInfo,
					,
					OtherItem.class
				);
			}
			else if
			(
				(OtherItem.InventoryType == Class'XWeapons.Redeemer') // Announced regardless of weapon stay.
				&&
				(
					!OtherItem.bDropped // Do not announce picking up dropped weapons.
					||
					!WeaponPickup(OtherItem).bThrown // Do not announce picking up thrown weapons.
				)
			)
			{
				BroadcastPickupEvent
				(
					Class'LocalMessage_Pickup_Remote',
					OtherPawn.PlayerReplicationInfo,
					,
					OtherItem.Class
				);
			}
		}
	}


	TeamGameEvents.PlayerPickedUp (OtherPawn, OtherItem);
}

// (Call-Sequence)
// 1. GameInfo.RestartPlayer (Controller)
// 2. GameInfo.GetDefaultPlayerClass ()
// - Controller.PawnClass may override the pawn class.
// Also:
// 1. GameInfo.Login () PawnClass
// Protocol:
// - To return the pawn to spawn for the network client.

event Class<Pawn> GetDefaultPlayerClass (Controller Other)
{
	if (Other == None)
	{
		Warn ("(Other == None)");
	}

	return (Class<Pawn>(DynamicLoadObject (DefaultPlayerClassName, Class'Core.Class')));
}

// Contexts:
// - PlayerController.ServerViewNextPlayer ()
// Preconditions:
// - (OtherViewTarget.IsA ('Controller'))

event bool CanSpectate
(
	PlayerController Viewer,
	bool bViewerIsSpectator,
	Actor OtherViewTarget
)
{
	if
	(
		(OtherViewTarget == None)
		||
		Controller(OtherViewTarget).PlayerReplicationInfo.bOnlySpectator
	)
	{
		return False;
	}
	else
	{
		return True;
	}
}

// #Spawning

protected function bool ShouldForceBotOut ()
{
	return !ReGRI(GameReplicationInfo).IsLive ();
}

protected function IncrementBotCount ()
{
	++NumBots; // No Clamping
}

protected function int NextPlayerID ()
{
	local int PlayerID;

	PlayerID = CurrentID++;

	return PlayerID;
}

protected function AssignPlayerID (PlayerReplicationInfo OtherPRI)
{
	OtherPRI.PlayerID = NextPlayerID ();

// 	Log ("AssignPlayerID (" $ OtherPRI $ ") OtherPRI.PlayerID:" @ OtherPRI.PlayerID);
}

event bool AddBot (optional string BotNameString)
{
	local Bot AddedBot;

	AddedBot = SpawnBot (BotNameString);

	if (AddedBot != None)
	{
		AssignPlayerID (AddedBot.PlayerReplicationInfo);


		IncrementBotCount ();


		if (ShouldForceBotOut ())
		{
			SetPlayerOut (AddedBot);
		}


		if (Level.NetMode == NM_Standalone)
		{
			RestartPlayer (AddedBot);
		}
		else
		{
			AddedBot.GotoState ('Dead', 'MPStart');
		}


		BroadcastLocalizedMessage (GameMessageClass, 1, AddedBot.PlayerReplicationInfo);


		return True;
	}
	else
	{
		Warn ("(AddedBot == None) BotNameString: <" $ BotNameString $ ">");

		return False;
	}
}

// AddBots -> AddBot -> *SpawnBot* -> RestartPlayer
// - To force any bot to use the controller pawn specified
// for this game type.

event Bot SpawnBot (optional string CharacterNameString)
{
	local Bot B;
	local RosterEntry BotRosterEntry;
	local UnrealTeamInfo BotTeam;

	BotTeam = GetBotTeam ();

	BotRosterEntry = BotTeam.ChooseBotClass (CharacterNameString);

	BotRosterEntry.PawnClass = SpawnBot_Class_Pawn;
	BotRosterEntry.PawnClassName = string(SpawnBot_Class_Pawn);

	if (BotRosterEntry.PawnClass == None)
	{
		BotRosterEntry.Init ();
	}

	B = Bot(Spawn (BotRosterEntry.PawnClass.Default.ControllerClass));

	if (B != None)
	{
		InitializeBot (B, BotTeam, BotRosterEntry);
	}
	else
	{
		Warn ("(B == None) - SpawnBot_Class_Pawn:" @ SpawnBot_Class_Pawn);
	}

	return B;
}

// #CombatLog

protected function KillEvent_Reset ()
{
	KillEventData.Killer = None;
	KillEventData.Victim = None;

	KillEventData.Class_DamageType = None;

	KillEventData.Context = KillEventContext_None;

	KillEventData.bFirstDeath = False;
	KillEventData.bFirstKill = False;

	KillEventData.bVictimOutofLives = False;
	KillEventData.bVictimRespawnsDepleted = False;
}

// - To allow updates to the state before any decisions are made.

protected function KillEvent_OnVictim ()
{
}

// ## Award Retribution ##

// Protocol:
// - If a player kills an enemy player, the victim remembers the ID of the killer.
// - If the victim kills someone other than the last killer, the revenge is off, and thus no sweet adrenaline reward.

protected function AwardCheck_Retribution ()
{
	if (KillEventData.Context == KillEventContext_EnemyKill)
	{
		if
		(
			(RePRI(KillEventData.Killer.PlayerReplicationInfo).Retribution_PlayerID_LastEnemyKiller == KillEventData.Victim.PlayerReplicationInfo.PlayerID)
			&&
			!KillEventData.Killer.PlayerReplicationInfo.bOutofLives
		)
		{
			AwardGive_Retribution ();
		}

		RePRI(KillEventData.Killer.PlayerReplicationInfo).Retribution_PlayerID_LastEnemyKiller = -1;

		RePRI(KillEventData.Victim.PlayerReplicationInfo).Retribution_PlayerID_LastEnemyKiller = KillEventData.Killer.PlayerReplicationInfo.PlayerID;
	}
}


protected function AwardGive_Retribution ()
{
	StatisticsProxy_RecordEvent (KillEventData.Killer.PlayerReplicationInfo, 'Retribution', ADR_MajorKill_Retribution);


	AwardMulticast (AwardBroadcastLocalMessageClass, 1, KillEventData.Killer.PlayerReplicationInfo, KillEventData.Victim.PlayerReplicationInfo);
}

// ## Award Killing Spree ##

// Protocol:
// - Called after incrementing spree.
// Preconditions:
// - (Count_Spree > 0)
// - (KillEventData.Killer.Pawn != None)

protected function AwardCheck_KillingSpree (int Count_Spree)
{
	if
	(
		(Count_Spree > 4)
		&&
		((Count_Spree % 5) == 0)
	)
	{
		AwardGive_KillingSpree (Count_Spree);
	}
}

// - Called for every five kills.
// Preconditions:
// - ((Count_Spree % 5) == 0)

protected function AwardGive_KillingSpree (int Count_Spree)
{
	Count_Spree = (Count_Spree / 5) - 1;

	SpecialEvent (KillEventData.Killer.PlayerReplicationInfo, "spree_" $ (Clamp (Count_Spree, 0, 5) + 1));

	KillEventData.Killer.AwardAdrenaline (ADR_MajorKill);

	switch (Count_Spree)
	{
		case 0:
			StatisticsProxy_RecordEvent (KillEventData.Killer.PlayerReplicationInfo, 'KillingSpree', ADR_MajorKill);
			break;

		case 1:
			StatisticsProxy_RecordEvent (KillEventData.Killer.PlayerReplicationInfo, 'Rampage', ADR_MajorKill);
			break;

		case 2:
			StatisticsProxy_RecordEvent (KillEventData.Killer.PlayerReplicationInfo, 'Dominating', ADR_MajorKill);
			break;

		case 3:
			StatisticsProxy_RecordEvent (KillEventData.Killer.PlayerReplicationInfo, 'Unstoppable', ADR_MajorKill);
			break;

		case 4:
			StatisticsProxy_RecordEvent (KillEventData.Killer.PlayerReplicationInfo, 'Godlike', ADR_MajorKill);
			break;

		case 5:
			StatisticsProxy_RecordEvent (KillEventData.Killer.PlayerReplicationInfo, 'WickedSick', ADR_MajorKill);
			break;

		default:
			StatisticsProxy_RecordEvent (KillEventData.Killer.PlayerReplicationInfo, 'WickedSickStack', ADR_MajorKill);
			break;
	}

	AwardAnnounce_KillingSpree (Count_Spree);
}

protected function AwardAnnounce_KillingSpree (int SpreeCount)
{
	SpreeCount = Clamp (SpreeCount, 0, 5);

	AwardMulticast (Class'UnrealGame.KillingSpreeMessage', SpreeCount, KillEventData.Killer.PlayerReplicationInfo, , , KillEventData.Killer.PlayerReplicationInfo);
}

// Checked at end of match.

protected function AwardCheck_LastManStanding ()
{
}

protected function AwardGive_LastManStanding ()
{
	StatisticsProxy_RecordEvent (KillEventData.Killer.PlayerReplicationInfo, 'LastManStanding');
}

// - All clients except ExceptedPRI receive this message.

protected function AwardMulticast
(
	Class<LocalMessage> LocalMessageClass,
	optional int Context,
	optional PlayerReplicationInfo RelatedPRI_1,
	optional PlayerReplicationInfo RelatedPRI_2,
	optional Object O,
	optional PlayerReplicationInfo ExceptedPRI
)
{
	local Controller Other;

	for (Other = Level.ControllerList; Other != None; Other = Other.NextController)
	{
		if
		(
			Other.bIsPlayer
			&&
			Other.IsA ('PlayerController')
			&&
			(Other.PlayerReplicationInfo != None)
			&&
			(Other.PlayerReplicationInfo != ExceptedPRI)
		)
		{
			PlayerController(Other).ReceiveLocalizedMessage
			(
				LocalMessageClass,
				Context,
				RelatedPRI_1,
				RelatedPRI_2,
				O
			);
		}
	}
}

// - Award killer if ended spree of an enemy player.

protected function AwardCheck_ItEndsHere (Pawn Victim_Pawn)
{
	if
	(
		(Victim_Pawn != None)
		&&
		(Victim_Pawn.GetSpree () > 4)
	)
	{
		AwardGive_ItEndsHere ();
	}
}

protected function AwardGive_ItEndsHere ()
{
	if (KillEventData.Context == KillEventContext_EnemyKill)
	{
		KillEventData.Killer.AwardAdrenaline (ADR_MajorKill);

		StatisticsProxy_RecordEvent (KillEventData.Killer.PlayerReplicationInfo, 'ItEndsHere', ADR_MajorKill);

		AwardMulticast (Class'UnrealGame.KillingSpreeMessage', 0, KillEventData.Victim.PlayerReplicationInfo, KillEventData.Killer.PlayerReplicationInfo);
	}
	else if (KillEventData.Context == KillEventContext_Death)
	{
		AwardMulticast (Class'UnrealGame.KillingSpreeMessage', 1, None, KillEventData.Victim.PlayerReplicationInfo);
	}
}

// #FirstKill

protected function AwardCheck_FirstKill ()
{
	if (KillEventData.Context == KillEventContext_EnemyKill)
	{
		KillEventData.bFirstKill = True;

		if (!bFirstBlood)
		{
			bFirstBlood = True;

			AwardGive_FirstBlood ();
		}
	}
}

protected function AwardGive_FirstBlood ()
{
	SpecialEvent (KillEventData.Killer.PlayerReplicationInfo, "first_blood");

	KillEventData.Killer.AwardAdrenaline (ADR_MajorKill);

	StatisticsProxy_RecordEvent (KillEventData.Killer.PlayerReplicationInfo, 'FirstBloodMatch', ADR_MajorKill);

	AwardMulticast (Class'UnrealGame.FirstBloodMessage', 0, KillEventData.Killer.PlayerReplicationInfo);
}

// #

protected function StatisticsProxy_RecordEvent
(
	PlayerReplicationInfo OtherPRI,
	name Name_Event,
	optional float Count_Points
)
{
	if (RePRI(OtherPRI) != None)
	{
		RePRI(OtherPRI).Statistics_AddEvent (Name_Event, Count_Points);
	}
}

// - This is a proxy for RePRI.AlertEvent ().

protected function AlertEvent
(
	PlayerReplicationInfo OtherPRI,
	name Name_Event
)
{
	if (RePRI(OtherPRI) != None)
	{
		RePRI(OtherPRI).AlertEvent (Name_Event);
	}
}

// #Announcements

// #FirstDeath

// Preconditions:
// - (KillEventData.Victim != None)

protected function AwardCheck_FirstDeath ()
{
	if
	(
		(KillEventData.Context == KillEventContext_EnemyKill)
		||
		(KillEventData.Context == KillEventContext_Death)
	)
	{
		KillEventData.bFirstDeath = True;

		AwardGive_FirstDeath ();
	}
}

// Preconditions:
// - (KillEventData.Victim != None)

protected function AwardGive_FirstDeath ()
{
	if (MaxLives > 0)
	{
		AlertEvent (KillEventData.Victim.PlayerReplicationInfo, 'FirstOutMatch');
	}
	else
	{
		AlertEvent (KillEventData.Victim.PlayerReplicationInfo, 'FirstDeathMatch');
	}
}

// ## UT2004 Fixes ##

// *UT2004Fix*
// - This frequency is no longer operational.

static event Texture GetRandomTeamSymbol (int Index)
{
	Warn ("<None> Texture GetRandomTeamSymbol ("$Index$")");

	return None;
}

// (Replaces) DeathMatch.InitTeamSymbols
// - Sometimes both teams would get the same team symbol.

event InitTeamSymbols ()
{
	local array<string> Group_Text_TeamSymbol;
	local int Count_TeamSymbols;
	local int Index_TeamSymbol[IndexTeam_Max];
	local Texture Texture_TeamSymbol;

	if
	(
		(GameReplicationInfo.TeamSymbols[0] != None)
		&&
		(GameReplicationInfo.TeamSymbols[1] != None)
	)
	{
		return; // Already loaded.
	}

	class'CacheManager'.Static.GetTeamSymbolList (Group_Text_TeamSymbol, True);

	Count_TeamSymbols = Group_Text_TeamSymbol.Length;

	if (Count_TeamSymbols < 1)
	{
		Warn ("(Count_TeamSymbols < 1)");

		return;
	}

	Index_TeamSymbol[0] = Rand (Count_TeamSymbols);
	Index_TeamSymbol[1] = Rand (Count_TeamSymbols);

	if (Index_TeamSymbol[0] == Index_TeamSymbol[1]) // *UT2004Fix*
	{
		++Index_TeamSymbol[1];

		if (Index_TeamSymbol[1] >= Count_TeamSymbols)
		{
			Index_TeamSymbol[1] = 0;
		}
	}

	Texture_TeamSymbol = Texture(DynamicLoadObject (Group_Text_TeamSymbol[Index_TeamSymbol[0]], class'Texture'));

	if (Texture_TeamSymbol != None)
	{
		GameReplicationInfo.TeamSymbols[0] = Texture_TeamSymbol;
	}
	else
	{
		Warn ("(Texture_TeamSymbol[0] == None)"@Group_Text_TeamSymbol[Index_TeamSymbol[0]]);
	}

	Texture_TeamSymbol = Texture(DynamicLoadObject (Group_Text_TeamSymbol[Index_TeamSymbol[1]], class'Texture'));

	if (Texture_TeamSymbol != None)
	{
		GameReplicationInfo.TeamSymbols[1] = Texture_TeamSymbol;
	}
	else
	{
		Warn ("(Texture_TeamSymbol[1] == None)"@Group_Text_TeamSymbol[Index_TeamSymbol[1]]);
	}

	GameReplicationInfo.TeamSymbolNotify ();
}

// #PendingMatch

// - To check and update the player name before passing it to the super function.

event PlayerController Login
(
	string PortalTagNameString,
	string ClientURIQueryString,
	out string Out_ErrorString
)
{
	local PlayerController OtherPC;
	local string ClientNameString;
	local string ClientNameReplacementString;
	local string CountryCodeString;

	// Isolate Name

	ClientNameString = "Name=" $ ParseOption (ClientURIQueryString, "Name");

	// Get Compliant Name

	ClientNameReplacementString = Class'RePRI'.Static.GetFilteredPlayerNameString (ClientNameString);

	// Replace Name

	ClientURIQueryString = Repl (ClientURIQueryString, ClientNameString, ClientNameReplacementString, True);

	// Continue Default Handling

	OtherPC = Super.Login (PortalTagNameString, ClientURIQueryString, Out_ErrorString); // StripColor, ...

	// Verify & Set Country Code

	if
	(
		(PlayerRegion != None)
		&&
		(OtherPC != None)
		&&
		OtherPC.IsA ('RePlayer')
		&&
		(OtherPC.PlayerReplicationInfo != None)
		&&
		OtherPC.PlayerReplicationInfo.IsA ('RePRI')
	)
	{
		CountryCodeString = GetValidCountryCode (ParseOption (ClientURIQueryString, "CountryCode"));

		PlayerRegion.UpdateCountryCode (OtherPC, RePRI(OtherPC.PlayerReplicationInfo), CountryCodeString);
	}


	return OtherPC;
}

// Restricted to two letters using the characters [A, Z].

static function string GetValidCountryCode (string NewCountryCodeString)
{
	local string ValidCountryCodeString;

	ValidCountryCodeString = Caps (NewCountryCodeString);

	if
	(
		(Len (ValidCountryCodeString) == 2)
		&&
		(Asc (Mid (ValidCountryCodeString, 0, 1)) > 0x40)
		&&
		(Asc (Mid (ValidCountryCodeString, 0, 1)) < 0x5B) // [A, Z]
		&&
		(Asc (Mid (ValidCountryCodeString, 1, 1)) > 0x40)
		&&
		(Asc (Mid (ValidCountryCodeString, 1, 1)) < 0x5B) // [A, Z]
	)
	{
		return ValidCountryCodeString;
	}
	else
	{
		return "";
	}
}

// #ScoreRecoveryObject

// Contexts:
// - ReTeamGame.PostLogin
// - RePlayer.BecomeActivePlayer
// Preconditions:
// - (RePC != None)

protected function ScoreRecoveryObject_Re ScoreRecoveryObject_Get
(
	RePlayer RePC,
	string PlayerIdentifierString
)
{
	local ScoreRecoveryObject_Re PDO;
	local int Index;

	for (Index = 0; Index < ScoreRecoveryGroup.Length; ++Index)
	{
		if (PlayerIdentifierString ~= ScoreRecoveryGroup[Index].PlayerIdentifierString)
		{
			return ScoreRecoveryGroup[Index];
		}
	}

	PDO = Spawn (Class_ScoreRecoveryObject);

	if (PDO != None)
	{
		ScoreRecoveryObject_Created (RePC, PDO, PlayerIdentifierString);

		ScoreRecoveryGroup[ScoreRecoveryGroup.Length] = PDO;
	}

	return PDO;
}

// Preconditions:
// - (RePC != None)

protected function ScoreRecoveryOnLogin (PlayerController OtherPC)
{
	local RePlayer RePC;
	local ScoreRecoveryObject_Re PDO;

	if
	(
		bScoreRecoveryEnabled
		&&
		(Level.NetMode != NM_Standalone)
		&&
		(RePRI(OtherPC.PlayerReplicationInfo) != None)
		&&
		!Level.bLevelChange
		&&
		!OtherPC.PlayerReplicationInfo.bOnlySpectator
	)
	{
		RePC = RePlayer(OtherPC);

		if (RePC != None)
		{
			PDO = ScoreRecoveryObject_Get (RePC, RePRI(OtherPC.PlayerReplicationInfo).PlayerIDHashString);

			if (PDO.DisconnectCount > PDO.RestoreCount)
			{
				PDO.Restore (RePC);
			}
		}
	}
}

// Preconditions:
// - (RePC != None)

protected function ScoreRecoveryOnLogout (PlayerController OtherPC)
{
	local RePlayer RePC;
	local ScoreRecoveryObject_Re PDO;

	if
	(
		bScoreRecoveryEnabled
		&&
		(Level.NetMode != NM_Standalone)
		&&
		!Level.bLevelChange
		&&
		!bDeleteMe
		&&
		(RePRI(OtherPC.PlayerReplicationInfo) != None)
		&&
		!OtherPC.PlayerReplicationInfo.bOnlySpectator
	)
	{
		RePC = RePlayer(OtherPC);

		if (RePC != None)
		{
			PDO = ScoreRecoveryObject_Get (RePC, RePRI(OtherPC.PlayerReplicationInfo).PlayerIDHashString);

			if (PDO != None)
			{
				PDO.Update (RePC);
			}
			else
			{
				Warn ("(PDO == None)");
			}
		}
	}
}

// Preconditions:
// - (RePC != None)
// - (PDO != None)

protected function ScoreRecoveryObject_Created (RePlayer RePC, ScoreRecoveryObject PDO, string PlayerIdentifierString)
{
	PDO.CreationTime = Level.TimeSeconds;
	PDO.PlayerIdentifierString = PlayerIdentifierString;
	PDO.RestoreCount = 0;
	PDO.DisconnectCount = 0;
}

// Preconditions:
// - (PDO != None)

protected function ScoreRecoveryObject_Re ScoreRecoveryObject_Destroy (ScoreRecoveryObject PDO)
{
	if (PDO != None)
	{
		PDO.Destroy ();
		PDO = None;
	}
	else
	{
		Warn ("(PDO == None)");
	}

	return None;
}

protected function ScoreRecoveryObjectGroup_Destroy ()
{
	local int Index;

	for (Index = 0; Index < ScoreRecoveryGroup.Length; ++Index)
	{
		ScoreRecoveryGroup[Index] = ScoreRecoveryObject_Destroy (ScoreRecoveryGroup[Index]);
	}

	ScoreRecoveryGroup.Remove (0, ScoreRecoveryGroup.Length);
}

// ## Login ##

// Preconditions:
// - (Other != None)

event PostLogin (PlayerController OtherPC)
{
	local Rotator R;

	Super.PostLogin (OtherPC);

	// Score Recovery Identifier

	RePRI(OtherPC.PlayerReplicationInfo).PlayerIDHashString = OtherPC.GetPlayerIDHash ();

	//

	if (OtherPC.PlayerReplicationInfo.bOnlySpectator)
	{
		OtherPC.ClientGotoState ('Spectating', ''); // *UDKFix*
	}

	//

	if (OtherPC.StartSpot != None)
	{
		R = OtherPC.StartSpot.Rotation;

		R.Pitch = 0;
		R.Roll = 0;

		OtherPC.SetRotation (R);
		OtherPC.ClientSetRotation (R);
	}
}

// Protocol:
// - Return (True) if a player should be forced out of the match or round.

protected function bool ShouldForceLoggedInPlayerOut (PlayerController Other)
{
	return False;
}

protected function LoggedInPlayerForcedOut (PlayerController Other)
{
}

// #

state MatchInProgress
{
	event BeginState ()
	{
		Super.BeginState ();

		if (!bRoundbased)
		{
			ReGRI(GameReplicationInfo).ChangeMatchStatus ('Live', MatchStatusLive);
		}

		ReGRI(GameReplicationInfo).TimeSeconds_MatchStarted = Level.TimeSeconds;
	}

	// Preconditions:
	// - (OtherPC != None)

	event PostLogin (PlayerController OtherPC)
	{
		Global.PostLogin (OtherPC); // Rotation / Spectating State


		ScoreRecoveryOnLogin (OtherPC);


		if (ShouldForceLoggedInPlayerOut (OtherPC))
		{
			LoggedInPlayerForcedOut (OtherPC);
		}
	}

	// Context:
	// - PostLogin
	// Protocol:
	// - In a round-based game, force the player out if the round is in progress.

	protected function bool ShouldForceLoggedInPlayerOut (PlayerController OtherPC)
	{
		return
		(
			!OtherPC.PlayerReplicationInfo.bOnlySpectator
			&&
			Level.Game.bDelayedStart
			&&
			(MaxLives > 0)
			&&
			OtherPC.PlayerReplicationInfo.bOutofLives
		);
	}

	protected function LoggedInPlayerForcedOut (PlayerController OtherPC)
	{
		SetPlayerOut (OtherPC);

		OtherPC.bBehindView = True;

		OtherPC.GotoState ('Spectating');
		OtherPC.ClientGotoState ('Spectating', ''); // *UDKFix*

		OtherPC.ServerViewNextPlayer (); // Reminder: Requires (Spectating State)
	}

	// Preconditions:
	// - (Other != None)

	event Logout (Controller Other)
	{
		if (Other.IsA ('PlayerController'))
		{
			ScoreRecoveryOnLogout (PlayerController(Other));
		}

		Super.Logout (Other);
	}

	event Timer ()
	{
		if (ReGRI(GameReplicationInfo).IsLive ())
		{
			if (bCampingCheck)
			{
				CheckCampers ();
			}
		}
		else if
		(
			ReGRI(GameReplicationInfo).IsRestarting ()
			||
			ReGRI(GameReplicationInfo).IsSpawning ()
		)
		{
			if (bShowBalanceTeams)
			{
				MulticastBalanceTeams ();
            }
        }


		Super.Timer ();
	}
}

protected function MulticastBalanceTeams ()
{
	local Controller Other;
	local array<PlayerController> HumanPlayer, HumanSpectator;
	local TeamInfo BalanceTeamInfo, JoinTeamInfo;
	local int TeamPlayerCount[IndexTeam_Max], OtherTeamIndex, BalanceTeamIndex, Index, Count, Difference, SpectatorDifference;

	for (Other = Level.ControllerList; Other != None; Other = Other.NextController)
	{
		if
		(
			Other.bIsPlayer
			&&
			(Other.PlayerReplicationInfo != None)
		)
		{
			if (Other.PlayerReplicationInfo.Team != None)
			{
				OtherTeamIndex = Other.PlayerReplicationInfo.Team.TeamIndex;

				if (Other.IsA ('PlayerController'))
				{
					HumanPlayer[HumanPlayer.Length] = PlayerController(Other);
	            }

				++TeamPlayerCount[OtherTeamIndex];
            }
            else if (Other.PlayerReplicationInfo.bOnlySpectator)
            {
				if (Other.IsA ('PlayerController'))
				{
					HumanSpectator[HumanSpectator.Length] = PlayerController(Other);
	            }
            }
		}
	}

	// Which team should get the request?

	SpectatorDifference = Abs (TeamPlayerCount[IndexTeam_A] - TeamPlayerCount[IndexTeam_B]);

	if ((TeamPlayerCount[IndexTeam_A] - TeamPlayerCount[IndexTeam_B]) == 0)
	{
		BalanceTeamIndex = IndexTeam_Unspecified;
    }
	else if (TeamPlayerCount[IndexTeam_A] < (TeamPlayerCount[IndexTeam_B] - 1))
	{
		BalanceTeamIndex = IndexTeam_B;
		Difference = TeamPlayerCount[IndexTeam_B] - TeamPlayerCount[IndexTeam_A] - 1;
    }
	else if (TeamPlayerCount[IndexTeam_B] < (TeamPlayerCount[IndexTeam_A] - 1))
	{
		BalanceTeamIndex = IndexTeam_A;
		Difference = TeamPlayerCount[IndexTeam_A] - TeamPlayerCount[IndexTeam_B] - 1;
    }
    else
    {
		BalanceTeamIndex = IndexTeam_Unspecified; // Either equal count or not even number of players.
    }

	// Players

	if
	(
		(BalanceTeamIndex != IndexTeam_Unspecified)
		&&
		(Difference > 0)
	)
	{
		BalanceTeamInfo = GameReplicationInfo.Teams[BalanceTeamIndex];
		JoinTeamInfo = GameReplicationInfo.Teams[(BalanceTeamIndex + 1) % IndexTeam_Max];

		Count = HumanPlayer.Length;

		for (Index = 0; Index < Count; ++Index)
		{
			if (HumanPlayer[Index].PlayerReplicationInfo.Team == BalanceTeamInfo)
			{
				HumanPlayer[Index].ReceiveLocalizedMessage
				(
					Class'LocalMessage_BalanceTeams',
					Difference,
					,
					,
					JoinTeamInfo
				);
            }
        }
    }

	// Spectators

	if
	(
		(BalanceTeamIndex != IndexTeam_Unspecified)
		&&
		(SpectatorDifference >= NumSpectators)
	)
	{
		BalanceTeamInfo = GameReplicationInfo.Teams[BalanceTeamIndex];
		JoinTeamInfo = GameReplicationInfo.Teams[(BalanceTeamIndex + 1) % IndexTeam_Max];

		Count = HumanSpectator.Length;

		for (Index = 0; Index < Count; ++Index)
		{
			HumanSpectator[Index].ReceiveLocalizedMessage
			(
				Class'LocalMessage_BalanceTeams',
				SpectatorDifference,
				,
				,
				JoinTeamInfo
			);
        }
    }
}

// Preconditions:
// - (P != None)
// - (P.PlayerReplicationInfo != None)

protected function bool EnemiesCanSeePawn (Pawn P)
{
	local Pawn OtherPawn;

	ForEach VisibleCollidingActors (Class'Engine.Pawn', OtherPawn, 4096.0f, P.Location, True)
	{
		if
		(
			(OtherPawn.PlayerReplicationInfo != None)
			&&
			(OtherPawn.PlayerReplicationInfo.Team != None)
			&&
			(OtherPawn.PlayerReplicationInfo.Team != P.PlayerReplicationInfo.Team)
		)
		{
			return True;
		}
	}

	return False;
}

// #Camping

// Part 1:
// - Track the location of the pawn.
// - When the location history has filled,
// check if should give a penalty.

protected function bool CamperShouldReceivePenalty (RePawn OtherRePawn)
{
	local RePRI OtherRePRI;
	local Box BoxLocation;
	local float SizeDimension;

	//

	OtherRePRI = RePRI(OtherRePawn.PlayerReplicationInfo);

	// Track Location

	OtherRePawn.CampingLocationHistory[OtherRePawn.CampingIndexLocation] = OtherRePawn.Location;

	++OtherRePawn.CampingIndexLocation;

	// Check Warmup

	if (OtherRePawn.CampingIndexLocation == CampingLocationHistoryCount)
	{
		OtherRePawn.CampingIndexLocation = 0; // Wrap

		OtherRePawn.bCheckCampingBehavior = True;
	}

	// Check Camping Behavior

	if (OtherRePawn.bCheckCampingBehavior)
	{
		BoxLocation.Min.X = OtherRePawn.CampingLocationHistory[0].X;
		BoxLocation.Min.Y = OtherRePawn.CampingLocationHistory[0].Y;
		BoxLocation.Min.Z = OtherRePawn.CampingLocationHistory[0].Z;
		BoxLocation.Max.X = OtherRePawn.CampingLocationHistory[0].X;
		BoxLocation.Max.Y = OtherRePawn.CampingLocationHistory[0].Y;
		BoxLocation.Max.Z = OtherRePawn.CampingLocationHistory[0].Z;

		// ||

		BoxLocation.Min.X = FMin (BoxLocation.Min.X, OtherRePawn.CampingLocationHistory[1].X);
		BoxLocation.Min.Y = FMin (BoxLocation.Min.Y, OtherRePawn.CampingLocationHistory[1].Y);
		BoxLocation.Min.Z = FMin (BoxLocation.Min.Z, OtherRePawn.CampingLocationHistory[1].Z);
		BoxLocation.Max.X = FMax (BoxLocation.Max.X, OtherRePawn.CampingLocationHistory[1].X);
		BoxLocation.Max.Y = FMax (BoxLocation.Max.Y, OtherRePawn.CampingLocationHistory[1].Y);
		BoxLocation.Max.Z = FMax (BoxLocation.Max.Z, OtherRePawn.CampingLocationHistory[1].Z);

		// |||

		BoxLocation.Min.X = FMin (BoxLocation.Min.X, OtherRePawn.CampingLocationHistory[2].X);
		BoxLocation.Min.Y = FMin (BoxLocation.Min.Y, OtherRePawn.CampingLocationHistory[2].Y);
		BoxLocation.Min.Z = FMin (BoxLocation.Min.Z, OtherRePawn.CampingLocationHistory[2].Z);
		BoxLocation.Max.X = FMax (BoxLocation.Max.X, OtherRePawn.CampingLocationHistory[2].X);
		BoxLocation.Max.Y = FMax (BoxLocation.Max.Y, OtherRePawn.CampingLocationHistory[2].Y);
		BoxLocation.Max.Z = FMax (BoxLocation.Max.Z, OtherRePawn.CampingLocationHistory[2].Z);

		// ||||

		BoxLocation.Min.X = FMin (BoxLocation.Min.X, OtherRePawn.CampingLocationHistory[3].X);
		BoxLocation.Min.Y = FMin (BoxLocation.Min.Y, OtherRePawn.CampingLocationHistory[3].Y);
		BoxLocation.Min.Z = FMin (BoxLocation.Min.Z, OtherRePawn.CampingLocationHistory[3].Z);
		BoxLocation.Max.X = FMax (BoxLocation.Max.X, OtherRePawn.CampingLocationHistory[3].X);
		BoxLocation.Max.Y = FMax (BoxLocation.Max.Y, OtherRePawn.CampingLocationHistory[3].Y);
		BoxLocation.Max.Z = FMax (BoxLocation.Max.Z, OtherRePawn.CampingLocationHistory[3].Z);

		// |||| |

		BoxLocation.Min.X = FMin (BoxLocation.Min.X, OtherRePawn.CampingLocationHistory[4].X);
		BoxLocation.Min.Y = FMin (BoxLocation.Min.Y, OtherRePawn.CampingLocationHistory[4].Y);
		BoxLocation.Min.Z = FMin (BoxLocation.Min.Z, OtherRePawn.CampingLocationHistory[4].Z);
		BoxLocation.Max.X = FMax (BoxLocation.Max.X, OtherRePawn.CampingLocationHistory[4].X);
		BoxLocation.Max.Y = FMax (BoxLocation.Max.Y, OtherRePawn.CampingLocationHistory[4].Y);
		BoxLocation.Max.Z = FMax (BoxLocation.Max.Z, OtherRePawn.CampingLocationHistory[4].Z);

		// |||| ||

		BoxLocation.Min.X = FMin (BoxLocation.Min.X, OtherRePawn.CampingLocationHistory[5].X);
		BoxLocation.Min.Y = FMin (BoxLocation.Min.Y, OtherRePawn.CampingLocationHistory[5].Y);
		BoxLocation.Min.Z = FMin (BoxLocation.Min.Z, OtherRePawn.CampingLocationHistory[5].Z);
		BoxLocation.Max.X = FMax (BoxLocation.Max.X, OtherRePawn.CampingLocationHistory[5].X);
		BoxLocation.Max.Y = FMax (BoxLocation.Max.Y, OtherRePawn.CampingLocationHistory[5].Y);
		BoxLocation.Max.Z = FMax (BoxLocation.Max.Z, OtherRePawn.CampingLocationHistory[5].Z);

		// |||| |||

		BoxLocation.Min.X = FMin (BoxLocation.Min.X, OtherRePawn.CampingLocationHistory[6].X);
		BoxLocation.Min.Y = FMin (BoxLocation.Min.Y, OtherRePawn.CampingLocationHistory[6].Y);
		BoxLocation.Min.Z = FMin (BoxLocation.Min.Z, OtherRePawn.CampingLocationHistory[6].Z);
		BoxLocation.Max.X = FMax (BoxLocation.Max.X, OtherRePawn.CampingLocationHistory[6].X);
		BoxLocation.Max.Y = FMax (BoxLocation.Max.Y, OtherRePawn.CampingLocationHistory[6].Y);
		BoxLocation.Max.Z = FMax (BoxLocation.Max.Z, OtherRePawn.CampingLocationHistory[6].Z);

		// |||| ||||

		BoxLocation.Min.X = FMin (BoxLocation.Min.X, OtherRePawn.CampingLocationHistory[7].X);
		BoxLocation.Min.Y = FMin (BoxLocation.Min.Y, OtherRePawn.CampingLocationHistory[7].Y);
		BoxLocation.Min.Z = FMin (BoxLocation.Min.Z, OtherRePawn.CampingLocationHistory[7].Z);
		BoxLocation.Max.X = FMax (BoxLocation.Max.X, OtherRePawn.CampingLocationHistory[7].X);
		BoxLocation.Max.Y = FMax (BoxLocation.Max.Y, OtherRePawn.CampingLocationHistory[7].Y);
		BoxLocation.Max.Z = FMax (BoxLocation.Max.Z, OtherRePawn.CampingLocationHistory[7].Z);

		// |||| |||| |

		BoxLocation.Min.X = FMin (BoxLocation.Min.X, OtherRePawn.CampingLocationHistory[8].X);
		BoxLocation.Min.Y = FMin (BoxLocation.Min.Y, OtherRePawn.CampingLocationHistory[8].Y);
		BoxLocation.Min.Z = FMin (BoxLocation.Min.Z, OtherRePawn.CampingLocationHistory[8].Z);
		BoxLocation.Max.X = FMax (BoxLocation.Max.X, OtherRePawn.CampingLocationHistory[8].X);
		BoxLocation.Max.Y = FMax (BoxLocation.Max.Y, OtherRePawn.CampingLocationHistory[8].Y);
		BoxLocation.Max.Z = FMax (BoxLocation.Max.Z, OtherRePawn.CampingLocationHistory[8].Z);

		// |||| |||| ||

		BoxLocation.Min.X = FMin (BoxLocation.Min.X, OtherRePawn.CampingLocationHistory[9].X);
		BoxLocation.Min.Y = FMin (BoxLocation.Min.Y, OtherRePawn.CampingLocationHistory[9].Y);
		BoxLocation.Min.Z = FMin (BoxLocation.Min.Z, OtherRePawn.CampingLocationHistory[9].Z);
		BoxLocation.Max.X = FMax (BoxLocation.Max.X, OtherRePawn.CampingLocationHistory[9].X);
		BoxLocation.Max.Y = FMax (BoxLocation.Max.Y, OtherRePawn.CampingLocationHistory[9].Y);
		BoxLocation.Max.Z = FMax (BoxLocation.Max.Z, OtherRePawn.CampingLocationHistory[9].Z);

		//

		SizeDimension
		=
		FMax
		(
			FMax (BoxLocation.Max.X - BoxLocation.Min.X, BoxLocation.Max.Y - BoxLocation.Min.Y),
			BoxLocation.Max.Z - BoxLocation.Min.Z
		);

		if // Within camping volume and not yet received camping penalty?
		(
			(SizeDimension < CampingCheckVolume)
			&&
			(OtherRePawn.CampingTimeSecondsLastWarning == 0)
		)
		{
			OtherRePawn.CampingTimeSecondsLastWarning = CampingCheckInterval;

			return True; // Had not yet received a penalty.
		}
		else if (SizeDimension > CampingCheckVolume) // Outside of camping volume?
		{
			OtherRePRI.Camping_Count_ConsecutivePenalties = 0;

			OtherRePRI.bReceivedCamperWarning = False;
		}
		else if (OtherRePawn.CampingTimeSecondsLastWarning > 0) // Keep counting down to next penalty.
		{
			--OtherRePawn.CampingTimeSecondsLastWarning;
		}
	}

	return False; // Still filling location history.
}

// - Calculates the damage done to the player and stores it at the
// index in the int array associated with the pawn.
// Context:
// - Called when it was decided players should receive camping penalty.
// Protocol:
// - Return (False) if pawn survives.  New health and shield strength out.
// - Return (True) if pawn dies.

protected function bool CamperDied
(
	RePawn OtherRePawn,
	out int Out_NewHealth,
	out int Out_NewShieldStrength
)
{
	local RePRI OtherRePRI;
	local int Health;
	local int ShieldStrength;
	local int CampingPenaltyDamage;
	local int HealthTotal;

	OtherRePRI = RePRI(OtherRePawn.PlayerReplicationInfo);


	Health = OtherRePawn.Health;
	ShieldStrength = int(OtherRePawn.ShieldStrength);


	CampingPenaltyDamage = 10 * (OtherRePRI.Camping_Count_Penalties + 1); // {10, 20, 30, ...}

	HealthTotal = Health + ShieldStrength;

	if (HealthTotal <= CampingPenaltyDamage)
	{
		return True; // Died
	}
	else if (ShieldStrength > 0)
	{
		ShieldStrength -= CampingPenaltyDamage;

		if (ShieldStrength < 0)
		{
			Health -= Abs (ShieldStrength); // Damage to shield carried over to health.
			ShieldStrength = 0;
		}

		Out_NewShieldStrength = ShieldStrength;
		Out_NewHealth = Health;

		return False;
	}
	else
	{
		Out_NewShieldStrength = 0;
		Out_NewHealth = Health - CampingPenaltyDamage;

		return False;
	}
}

// Returns True if Other has a living pawn.

protected function bool LivingPlayer (Controller Other)
{
	return
	(
		Other.bIsPlayer
		&&
		(RePRI(Other.PlayerReplicationInfo) != None)
		&&
		!Other.PlayerReplicationInfo.bOnlySpectator
		&&
		!Other.PlayerReplicationInfo.bOutofLives
		&&
		(RePawn(Other.Pawn) != None)
		&&
		(Other.Pawn.Health > 0)
	);
}

// Preconditions:
// - PlayerIsAlive (Other)

protected function bool CampingRelevantPlayer (Controller Other)
{
	return !EnemiesCanSeePawn (Other.Pawn); // Excludes players in close proximity that are in plain sight.
}

//

protected function CheckCampers ()
{
	local Controller Other;
	local RePawn OtherRePawn;
	local array<RePawn> CampingPawnGroup;

	local array<int>
		ShieldStrength,
		LivesRemaining,
		Health;

	local int
		ShieldStrengthNew,
		HealthNew,
		LivesRemainingNew,
		Index,
		PawnOutCount, // (MaxLives > 0)
		LivingPawnCount;

	// Collect Potential Campers
	// - Count Living Players
	// - Collect pawns that are eligible for camping behavior checks.

	for (Other = Level.ControllerList; Other != None; Other = Other.NextController)
	{
		if (LivingPlayer (Other))
		{
			++LivingPawnCount;


			if (CampingRelevantPlayer (Other))
			{
				CampingPawnGroup[CampingPawnGroup.Length] = RePawn(Other.Pawn);
			}
		}
	}

	// Process Potential Campers
	// - Players that receive camping penalty get their new health and shield
	// calculated, and checked whether would die from such penalty damage.
	// -- If MaxLives > 0, check if player is out, and if all are out, act accordingly.

	if (CampingPawnGroup.Length > 0)
	{
		for (Index = 0; Index < CampingPawnGroup.Length; ++Index)
		{
			OtherRePawn = CampingPawnGroup[Index];

			if (CamperShouldReceivePenalty (OtherRePawn))
			{
				if (CamperDied (OtherRePawn, HealthNew, ShieldStrengthNew)) // Health depleted during pass
				{
					HealthNew = 0;
					ShieldStrengthNew = 0;
					LivesRemainingNew = OtherRePawn.PlayerReplicationInfo.NumLives + 1;

					if
					(
						(GameReplicationInfo.MaxLives > 0) // Finite Lives
						&&
						(LivesRemainingNew >= GameReplicationInfo.MaxLives) // Last respawn?
					)
					{
						++PawnOutCount;
                    }
				}
			}
			else // Pawn does not receive penalty at this interval, but has been marked.
			{
				HealthNew = -1;
				ShieldStrengthNew = -1;
				LivesRemainingNew = -1;
			}


			Health[Health.Length] = HealthNew;
			ShieldStrength[ShieldStrength.Length] = ShieldStrengthNew;
			LivesRemaining[LivesRemaining.Length] = LivesRemainingNew;
		}


		if (PawnOutCount == LivingPawnCount) // All living players out as a result of this pass?
		{
			if (bRoundbased) // Can reset round
			{
				EndRoundDraw (1); // Round is reset
            }
			else // Treat as normal pass, and drop to at least 1 health.
            {
				for (Index = 0; Index < CampingPawnGroup.Length; ++Index)
				{
					OtherRePawn = CampingPawnGroup[Index];

					WarnCamper (OtherRePawn);

					AnnounceCamper (OtherRePawn);

					if (CamperUpdated (OtherRePawn))
					{
						OtherRePawn.LastHitBy = None;

						OtherRePawn.ShieldStrength = 0;
						OtherRePawn.Health = 1;

						OtherRePawn.TakeDamage
						(
							0.0f,
							OtherRePawn,
							NullVector,
							NullVector,
							CampingDamageTypeClass
						);

						OtherRePawn.LastHitBy = None;
					}
				}
            }
		}
		else
		{
			for (Index = 0; Index < CampingPawnGroup.Length; ++Index)
			{
				OtherRePawn = CampingPawnGroup[Index];

				if (Health[Index] == 0) // Determined that player died
				{
					WarnCamper (OtherRePawn);

					if (CamperUpdated (OtherRePawn))
					{
						OtherRePawn.LastHitBy = None;

						OtherRePawn.Died (None, CampingDamageTypeClass, OtherRePawn.Location);
					}
				}
				else if (Health[Index] > 0)
				{
					WarnCamper (OtherRePawn);

					AnnounceCamper (OtherRePawn);

					if (CamperUpdated (OtherRePawn))
					{
						OtherRePawn.LastHitBy = None;

						OtherRePawn.ShieldStrength = ShieldStrength[Index];
						OtherRePawn.Health = Health[Index];

						OtherRePawn.TakeDamage
						(
							0.0f,
							OtherRePawn,
							NullVector,
							NullVector,
							CampingDamageTypeClass
						);

						OtherRePawn.LastHitBy = None;
					}
				}
			}
		}
	}
}

protected function WarnCamper (RePawn CamperRePawn)
{
	local Controller CamperController;
	local RePRI CamperRePRI;

	// Initialize

	CamperController = CamperRePawn.Controller;
	CamperRePRI = RePRI(CamperController.PlayerReplicationInfo);

	// Notify Camper

	if (CamperController.IsA ('RePlayer'))
	{
		RePlayer(CamperController).ClientAnnounceCamper
		(
			Min (CamperRePRI.Camping_Count_ConsecutivePenalties, 7)
		);
	}
}

// - Announce camper to all non-camping players and spectators.
// - Bots hear the location of the camper.

protected function AnnounceCamper (RePawn CamperRePawn)
{
	local Controller Other;
	local Controller CamperController;
	local TeamInfo CamperTeam;
	local RePRI CamperRePRI;

	// Initialize

	CamperController = CamperRePawn.Controller;
	CamperRePRI = RePRI(CamperController.PlayerReplicationInfo);
	CamperTeam = CamperRePRI.Team;

	// - Players are notified of campers on the other teams.
	// - Spectators are notified of all campers.

	for (Other = Level.ControllerList; Other != None; Other = Other.NextController)
	{
		if
		(
			Other.IsA ('RePlayer') // Notify Network Clients
			&&
			Other.bIsPlayer
			&&
			(
				(
					(Other.PlayerReplicationInfo != None)
					&&
					Other.PlayerReplicationInfo.bOnlySpectator
				)
				||
				(
					(Other != CamperController)
					&&
					(CamperTeam != Other.PlayerReplicationInfo.Team)
				)
			)
		)
		{
			if (bCampingCheckSendLocation)
			{
				RePlayer(Other).ClientAnnounceCamper
				(
					192,
					CamperRePRI,
					CamperRePawn.Location
				);
			}
			else
			{
				RePlayer(Other).ClientAnnounceCamper
				(
					128,
					CamperController.PlayerReplicationInfo
				);
			}
		}
		else if
		(
			Other.IsA ('Bot')
			&&
			Other.bIsPlayer
			&&
			(Other != CamperController)
		)
		{
			Bot(Other).HearNoise (FClamp (FRand (), 0.75f, 1.0f), CamperRePawn);
		}
	}
}

// Context:
// - The pawn is camping.
// Preconditions:
// - (CamperRePawn != None)
// Protocol:
// - Return (False) if camper was a network client who got kicked for excessive camping
// (assumed idling).

protected function bool CamperUpdated (RePawn CamperRePawn)
{
	local PlayerController CamperPC;
	local RePRI CamperRePRI;

	// Initialize

	CamperRePRI = RePRI(CamperRePawn.PlayerReplicationInfo);

	// Update State || Kick Check

	if (!CamperRePRI.bReceivedCamperWarning)
	{
		CamperRePRI.bReceivedCamperWarning = True;
	}
	else
	{
		++CamperRePRI.Camping_Count_Penalties;
		++CamperRePRI.Camping_Count_ConsecutivePenalties;

		CamperPC = PlayerController(CamperRePawn.Controller);

		if
		(
			(Level.NetMode == NM_DedicatedServer)
			&&
			((CamperRePRI.Ping * 4) < 999)
			&&
			(CamperPC != None)
			&&
			bCampingCheckKickRepeatedOffenders
			&&
			(CamperRePRI.Camping_Count_ConsecutivePenalties > 3)
		)
		{
			KickIdler (CamperPC);

			return False;
		}
	}

	return True;
}

// ## Camping ##

// Context:
// - A: GameInfo.ReduceDamage ()
// Protocol:
// - Return True if player is spawn protected.

protected function bool IsSpawnProtected
(
	Pawn Instigator_Pawn,
	Pawn Injured_Pawn,
	class<DamageType> Instigator_Class_DamageType
)
{
	return
	(
		(Instigator_Pawn != None)
		&&
		(Instigator_Pawn != Injured_Pawn)
		&&
		((Level.TimeSeconds - Injured_Pawn.SpawnTime) < SpawnProtectionTime)
		&&
		(
			(class<WeaponDamageType>(Instigator_Class_DamageType) != None)
			||
			(class<VehicleDamageType>(Instigator_Class_DamageType) != None)
		)
	);
}

// Context:
// - A: GameInfo.ReduceDamage ()
// Preconditions:
// - (Injured_Pawn != None)
// - (Class_DamageType != None)

protected function Controller ReduceDamage_GetController_Instigator
(
	Pawn Instigator_Pawn,
	Pawn Injured_Pawn,
	Class<DamageType> Class_DamageType
)
{
	local Controller Instigator_Controller;

	if (Instigator_Pawn != None)
	{
		Instigator_Controller = Instigator_Pawn.Controller;
	}

	if
	(
		(Instigator_Controller == None)
		&&
		Class_DamageType.Default.bDelayedDamage
	)
	{
		Instigator_Controller = Injured_Pawn.DelayedDamageInstigatorController;
	}

	return Instigator_Controller;
}

// #

state MatchOver
{
	ignores CheckScore, EndGame, SetPlayerDefaults;

	event bool CheckEndGame (PlayerReplicationInfo Decider, string Text)
	{
		return False;
	}

	event bool CheckMaxLives (PlayerReplicationInfo Decider)
	{
		return False;
	}

	event int ReduceDamage
	(
		int Damage,
		Pawn injured,
		Pawn instigatedBy,
		Vector HitLocation,
		out Vector Momentum,
		Class<DamageType> Class_DamageType
	)
	{
		Momentum = NullVector; // *UDKFix*

		return 0;
	}

	// *UT2004Fix*
	// - Latecomers were not set to the GameEnded state.

	event PostLogin (PlayerController PC)
	{
		Global.PostLogin (PC);

		RePlayer(PC).ExtendedGameHasEnded (EndGameFocus);
	}

	// *UDKFix*
	// - To mark newly joined players as out.

	event bool ChangeTeam (Controller Other, int Index_Team, bool bNewTeam)
	{
		SetPlayerOut (Other);

		return True;
	}

	event Timer ()
	{
		Super.Timer ();

		if (EndMessageCounter == 1)
		{
			if (bRoundbased)
			{
				AnnounceRoundBest ();
			}
			else
			{
				AnnounceMatchBest ();
			}
		}

		if (EndMessageCounter == 2)
		{
			AnnounceMatchWinner ();
		}
	}

	event BeginState ()
	{
		Super.BeginState ();


		ReGRI(GameReplicationInfo).ChangeMatchStatus ('Over', MatchStatusOver);


		SnapshotMatchWinner ();


		NotifyMatchEnded ();
	}
}

// #

protected function SnapshotMatchWinner ()
{
}

// #

protected function NotifyMatchEnded ()
{
}

// ## Round ##

protected function EndRoundDraw (optional int Context)
{
}

protected function AnnounceRoundDraw (int Context)
{
	BroadcastLocalizedMessage (DrawRoundLocalMessageClass, Context);
}

// ## Countdown ##

function int GetTimeSecondsProxy_RoundDuration ()
{
	return ReGRI(GameReplicationInfo).GetTimeSeconds_RoundDuration ();
}

function int GetRoundNumberProxy_CurrentRound ()
{
	return ReGRI(GameReplicationInfo).IndexRound_Current;
}

protected function AnnounceMatchBest ()
{
}

protected function AnnounceRoundBest ()
{
	local Controller Other;
	local ReGRI ReGRI;

	ReGRI = ReGRI(GameReplicationInfo);

	for (Other = Level.ControllerList; Other != None; Other = Other.NextController)
	{
		if
		(
			Other.bIsPlayer
			&&
			Other.IsA ('RePlayer') // Never played back in demo recording (based on UnrealPlayer).
			&&
			(Other.PlayerReplicationInfo != None)
		)
		{
			RePlayer(Other).ClientAnnounceRoundBest
			(
				ReGRI.RoundBestMetrics.Flags,
				ReGRI.RoundBestMetrics.IndexRound_Current,
				ReGRI.RoundBestMetrics.TimeSeconds_RoundDuration,
				ReGRI.RoundBestMetrics.Accuracy[0],
				ReGRI.RoundBestMetrics.Accuracy[1],
				ReGRI.RoundBestMetrics.Accuracy_Value,
				ReGRI.RoundBestMetrics.Damage[0],
				ReGRI.RoundBestMetrics.Damage[1],
				ReGRI.RoundBestMetrics.Damage_Value,
				ReGRI.RoundBestMetrics.Headshots[0],
				ReGRI.RoundBestMetrics.Headshots[1],
				ReGRI.RoundBestMetrics.Headshot_Value,
				ReGRI.RoundBestMetrics.Longest[0],
				ReGRI.RoundBestMetrics.Longest[1],
				ReGRI.RoundBestMetrics.Longest_Value
			);
		}
	}
}

protected function AnnounceMatchWinner ()
{
}

// ## Scoring ##

protected function bool ValidEndGameReason (string Text_Reason)
{
	return
	(
		(Text_Reason ~= "RoundLimit")
		||
		(Text_Reason ~= "Triggered")
		||
		(Text_Reason ~= "LastMan")
		||
		(Text_Reason ~= "TimeLimit")
		||
		(Text_Reason ~= "FragLimit")
		||
		(Text_Reason ~= "TeamScoreLimit")
		||
		(Text_Reason ~= "NoParticipants")
	);
}

function bool NoActivePlayers ()
{
	return ((Numbots + NumPlayers) == 0);
}

// Modified DeathMatch Implementation

event bool CheckEndGame (PlayerReplicationInfo Winner, string Text_Reason)
{
	local Controller Other;
	local bool bLastMan;

	if (bOverTime)
	{
		if (NoActivePlayers ())
		{
			return True;
		}
		else
		{
			bLastMan = True;

			for (Other = Level.ControllerList; Other != None; Other = Other.nextController)
			{
				if
				(
					Other.bIsPlayer
					&&
					(Other.PlayerReplicationInfo != None)
					&&
					!Other.PlayerReplicationInfo.bOutofLives
				)
				{
					bLastMan = False;

					break;
				}
			}
		}

		if (bLastMan)
		{
			return True;
		}
	}

	bLastMan = (Text_Reason ~= "LastMan");

	// Game Rules Modifier

	if
	(
		!bLastMan
		&&
		(GameRulesModifiers != None)
		&&
		!GameRulesModifiers.CheckEndGame (Winner, Text_Reason)
	)
	{
		return False;
	}

	// Find Winner

	if (Winner == None)
	{
		for (Other = Level.ControllerList; Other != None; Other = Other.NextController)
		{
			if
			(
				Other.bIsPlayer
				&&
				!Other.PlayerReplicationInfo.bOutOfLives
				&&
				(
					(Winner == None)
					||
					(Other.PlayerReplicationInfo.Score >= Winner.Score)
				)
			)
			{
				Winner = Other.PlayerReplicationInfo;
			}
		}
	}

	// Tie

	if (!bLastMan)
	{
		for (Other = Level.ControllerList; Other != None; Other = Other.NextController)
		{
			if
			(
				Other.bIsPlayer
				&&
				(Winner != Other.PlayerReplicationInfo)
				&&
				(Other.PlayerReplicationInfo.Score == Winner.Score)
				&&
				!Other.PlayerReplicationInfo.bOutOfLives
			)
			{
				if (!bOverTimeBroadcast)
				{
					StartupStage = 7;

					PlayStartupMessage ();

					bOverTimeBroadcast = True;
				}

				return False;
			}
		}
	}

	// Match Ended

	EndTime = Level.TimeSeconds + EndTimeDelay;

	//

	SetEndGameFocus (Winner);

	//

	return True;
}

// #

protected function NotifyGameEnded ()
{
}

// - DeathMatch Implementation (Modified)

event SetEndGameFocus (PlayerReplicationInfo Winner)
{
	local Controller Other;
	local Controller Next;
	local RePlayer RePC;
	local bool bWon;

	//

	GameReplicationInfo.Winner = Winner;

	//

	if (CurrentGameProfile != None)
	{
		CurrentGameProfile.bWonMatch = (PlayerController(Winner.Owner) != None);
	}

	//

	EndGameFocus = Controller(Winner.Owner).Pawn;

	if (EndGameFocus != None)
	{
		EndGameFocus.bAlwaysRelevant = True;
	}

	//

	MatchStatistics_MatchEnded (); // Compile statistics before players can request them.

	//

	Other = Level.ControllerList;

	while (Other != None)
	{
		Next = Other.NextController;


		RePC = RePlayer(Other);

		if (RePC != None)
		{
			bWon = (RePC.PlayerReplicationInfo == Winner);


			if (EndGameFocus != None)
			{
				RePC.ExtendedGameHasEnded (EndGameFocus, bWon);
			}
			else
			{
				RePC.GotoState ('GameEnded');
				RePC.ClientGotoState ('GameEnded', '');
			}
		}
		else
		{
			Other.GameHasEnded ();
		}


		Other = Next;
	}
}

// - Required when games end with no players.

protected function SetEndGame ()
{
	local Controller Other;
	local Controller Next;
	local PlayerController PC;

	GameReplicationInfo.Winner = None;

	if (CurrentGameProfile != None)
	{
		CurrentGameProfile.bWonMatch = False;
	}

	EndGameFocus = None;

	Other = Level.ControllerList;

	while (Other != None)
	{
		Next = Other.NextController;

		PC = PlayerController(Other);

		if (PC != None)
		{
			PC.ClientGameEnded ();
		}

		Other.GameHasEnded ();

		Other = Next;
	}
}

// - To announce the cause.
// - To recognize RoundLimit.
// Contexts:
// - (MatchInProgress)Timer ()
// - CheckMaxLives ()
// - CheckScore ()

event EndGame (PlayerReplicationInfo Decider, string ReasonString)
{
	if (ValidEndGameReason (ReasonString))
	{
		Super(GameInfo).EndGame (Decider, ReasonString);


		if (bGameEnded)
		{
			SetEndGameReason (ReasonString);

			GotoState ('MatchOver');
		}
	}
}

protected function SetEndGameReason (string ReasonString)
{
	switch (ReasonString)
	{
		case "FragLimit":
		case "RoundLimit":
		case "TeamScoreLimit":

			EndGameCause = MatchOver_ScoreLimit;
			break;

		case "LastMan":

			if (Level.GRI.bTeamGame)
			{
				EndGameCause = MatchOver_LastTeamStanding;
			}
			else
			{
				EndGameCause = MatchOver_LastPlayerStanding;
			}

			break;

		case "TimeLimit":

			EndGameCause = MatchOver_TimeLimit;
			break;

		case "NoParticipants":

			EndGameCause = MatchOver_NoParticipants;
			break;

		default:

			Warn ("Unknown ReasonString:"@ReasonString);
			break;
	}
}

// ## Spawning ##

// - To prevent RestartPlayer from restarting this player.
// Preconditions:
// - (!PlayerReplicationInfo.bOnlySpectator)
// - (Other != None)

function SetPlayerOut (Controller Other)
{
	Other.PlayerReplicationInfo.NumLives = MaxLives;
	Other.PlayerReplicationInfo.bOutofLives = True;

	RePRI(Other.PlayerReplicationInfo).RealtimePlayerInfo.ChangePlayerStatus ('Out');
}

// - To allow RestartPlayer to restart this player.
// Preconditions:
// - (Other != None)
// - (!Other.PlayerReplicationInfo.bOnlySpectator)

function ResetPlayerLives (Controller Other)
{
	Other.PlayerReplicationInfo.NumLives = 0;
	Other.PlayerReplicationInfo.bOutofLives = False;

	RePRI(Other.PlayerReplicationInfo).RealtimePlayerInfo.DefaultPlayerStatus ();
}

// - Typically used when entering a teleport that transfers the
// player to another level.  This is passed in the URL.
// Preconditions:
// - (Other != None)

protected function NavigationPoint FindPlayerStartPortal
(
	Controller Other,
	optional byte TeamIndex,
	optional string PortalTagNameString
)
{
	local int Index;

	if (PortalTagNameString != "")
	{
		for (Index = 0; Index < SpawnManager_TeleportGroup.Length; ++Index)
		{
			if (string(SpawnManager_TeleportGroup[Index].Tag) ~= PortalTagNameString)
			{
				return SpawnManager_TeleportGroup[Index];
			}
		}
	}

	return None;
}

// - PostLogin calls FindPlayerStart with (Player == None) to get an initial start
// for the camera view.
// - Randomize which player start to return.

function NavigationPoint FindFirstPlayerStart ()
{
	local int Count;

	Count = SpawnManager_PlayerStartGroup.Length;

	if (Count > 1) // More to choose from.
	{
		return SpawnManager_PlayerStartGroup[Rand (Count)]; // Reminder: Rand returns [0, Count - 1].
	}
	else if (Count == 1) // Only one.
	{
		return SpawnManager_PlayerStartGroup[0];
	}
	else
	{
		Warn ("No PlayerStart Actor");
	}

	return None;
}

protected function bool ValidSpawnPoint (NavigationPoint NP)
{
	return
	(
		!NP.IsA ('Teleporter')
		&&
		!NP.IsA ('JumpPad')
	);
}

protected function SpawnManager_InitGame ()
{
	local Actor CollisionPawn;
	local PlayerStart CollisionPlayerStart;
	local NavigationPoint PotentialSpawnPoint;
	local int ZoneIndex;

	//

	SpawnManager_ZoneCount = 0;
	SpawnManager_PlayerStartGroup.Length = 0;
	SpawnManager_SpawnPointGroup.Length = 0;
	SpawnManager_TeleportGroup.Length = 0;

	// Spawn Collision
	// - Get initial spawn position.
	// - PlayerStart actors should generally support all
	// players, as their collision is checked by the editor.

	ForEach AllActors (Class'Engine.PlayerStart', CollisionPlayerStart)
	{
		CollisionPawn = Spawn (Class'CollisionPawn', None, '', CollisionPlayerStart.Location);

		if (CollisionPawn != None)
		{
			break;
		}
	}

	// - Using AllActors because the level may have a broken chain
	// of NavigationPoint actors (maps where paths must be rebuilt).
	// - Filtering out navigation points where pawns would not fit, assuming
	// default collision size in xPawn.

	if (CollisionPawn != None)
	{
		if (!Level.bPathsRebuilt)
		{
			Log ("Using: AllActors (!Level.bPathsRebuilt)");

			ForEach AllActors (Class'Engine.NavigationPoint', PotentialSpawnPoint)
			{
				if
				(
					(CollisionPawn.SetLocation (PotentialSpawnPoint.Location)) // Fit
					&&
					!FastTrace (PotentialSpawnPoint.Location - Vect (0.0f, 0.0f, 64.0f), PotentialSpawnPoint.Location) // Non-Floating
				)
				{
					if (ValidSpawnPoint (PotentialSpawnPoint))
					{
						SpawnManager_SpawnPointGroup[SpawnManager_SpawnPointGroup.Length] = PotentialSpawnPoint;
					}

					// Always Valid / Further Grouping

					if (PotentialSpawnPoint.IsA ('PlayerStart'))
					{
						SpawnManager_PlayerStartGroup[SpawnManager_PlayerStartGroup.Length] = PlayerStart(PotentialSpawnPoint);
					}
					else if (PotentialSpawnPoint.IsA ('Teleporter')) // To avoid having to do an AllActors call in FindPlayerStart.
					{
						SpawnManager_TeleportGroup[SpawnManager_TeleportGroup.Length] = Teleporter(PotentialSpawnPoint);
					}

					// Zones

					ZoneIndex = int(PotentialSpawnPoint.Region.ZoneNumber); // Reminder: No sign-extension because a byte is unsigned.

					if (ZoneIndex > SpawnManager_ZoneCount)
					{
						SpawnManager_ZoneCount = ZoneIndex;
					}
				}
			}
		}
		else
		{
			Log ("Using: NavigationPointList");

			for (PotentialSpawnPoint = Level.NavigationPointList; PotentialSpawnPoint != None; PotentialSpawnPoint = PotentialSpawnPoint.NextNavigationPoint)
			{
				if
				(
					(CollisionPawn.SetLocation (PotentialSpawnPoint.Location)) // Fit
					&&
					!FastTrace (PotentialSpawnPoint.Location - Vect (0.0f, 0.0f, 64.0f), PotentialSpawnPoint.Location) // Non-Floating
				)
				{
					if (ValidSpawnPoint (PotentialSpawnPoint))
					{
						SpawnManager_SpawnPointGroup[SpawnManager_SpawnPointGroup.Length] = PotentialSpawnPoint;
					}

					// Always Valid / Further Grouping

					if (PotentialSpawnPoint.IsA ('PlayerStart'))
					{
						SpawnManager_PlayerStartGroup[SpawnManager_PlayerStartGroup.Length] = PlayerStart(PotentialSpawnPoint);
					}
					else if (PotentialSpawnPoint.IsA ('Teleporter')) // To avoid having to do an AllActors call in FindPlayerStart.
					{
						SpawnManager_TeleportGroup[SpawnManager_TeleportGroup.Length] = Teleporter(PotentialSpawnPoint);
					}

					// Zones

					ZoneIndex = int(PotentialSpawnPoint.Region.ZoneNumber); // Reminder: No sign-extension because a byte is unsigned.

					if (ZoneIndex > SpawnManager_ZoneCount)
					{
						SpawnManager_ZoneCount = ZoneIndex;
					}
				}
            }
		}
	}

	// Log

	Log ("SpawnPoints:" @ SpawnManager_SpawnPointGroup.Length, 'Re');

	//

	if (CollisionPawn != None)
	{
		CollisionPawn.Destroy ();
		CollisionPawn = None;
	}
}

//

protected function bool IsValidFindPlayerStart (NavigationPoint N)
{
	return
	(
		N.IsA ('PathNode')
		||
		N.IsA ('PlayerStart')
		||
		N.IsA ('JumpSpot')
	);
}

//

function NavigationPoint FindPlayerStart (Controller Other, optional byte RequestedTeamIndex, optional string TeleportTagNameString)
{
	local NavigationPoint N, BestStart;
	local float BestRating, NewRating;
	local int Index;
	local byte OtherTeamIndex;

	// Initial Start

	if
	(
		(Other != None)
		&&
		(Other.StartSpot != None)
		&&
		(Level.NetMode == NM_Standalone)
		&&
		(
			bWaitingToStartMatch
			||
			(
				(Other.PlayerReplicationInfo != None)
				&&
				Other.PlayerReplicationInfo.bWaitingPlayer
			)
		)
	)
	{
		return Other.StartSpot;
	}

	// Mutator Override

	if (GameRulesModifiers != None)
	{
		N = GameRulesModifiers.FindPlayerStart (Other, RequestedTeamIndex, TeleportTagNameString);

		if (N != None)
		{
			return N;
		}
	}

	// Inter-Level Teleport

	N = FindPlayerStartPortal (Other, RequestedTeamIndex, TeleportTagNameString);

	if (N != None)
	{
		return N;
	}

	// Use RequestedTeamIndex if Other doesn't have a OtherTeamIndex yet.

	if
	(
		(Other != None)
		&&
		(Other.PlayerReplicationInfo != None)
	)
	{
		if (Other.PlayerReplicationInfo.Team != None)
		{
			OtherTeamIndex = Other.PlayerReplicationInfo.Team.TeamIndex;
		}
		else
		{
			OtherTeamIndex = RequestedTeamIndex;
		}
	}
	else
	{
		OtherTeamIndex = RequestedTeamIndex;
	}

	// Find Player Start

	for (Index = 0; Index < SpawnManager_SpawnPointGroup.Length; ++Index)
	{
		N = SpawnManager_SpawnPointGroup[Index];


		if (!IsValidFindPlayerStart (N))
		{
			NewRating = 1;
        }
        else
        {
			NewRating = RatePlayerStart (N, OtherTeamIndex, Other);
        }


		if (NewRating > BestRating)
		{
			BestRating = NewRating;
			BestStart = N;
		}
	}

	// Fallback

	if
	(
		(BestStart == None)
		||
		(
			(PlayerStart(BestStart) == None)
			&&
			(Other != None)
			&&
			Other.bIsPlayer
		)
	)
	{
		if (SpawnManager_PlayerStartGroup.Length > 0)
		{
			Log ("FindPlayerStart fallback for" @ Other @ "SpawnPoints:" @ SpawnManager_SpawnPointGroup.Length);
		}
		else
		{
			Warn ("NO PLAYERSTART");
		}

		BestRating = -100000000.0f;

		for (Index = 0; Index < SpawnManager_SpawnPointGroup.Length; ++Index)
		{
			N = SpawnManager_SpawnPointGroup[Index];

			NewRating = RatePlayerStart (N, 0, Other);

			if (N.IsA ('InventorySpot'))
			{
				NewRating -= 50;
			}

			NewRating += 20.0f * FRand ();

			if (NewRating > BestRating)
			{
				BestRating = NewRating;

				BestStart = N;
			}
		}
	}

	return BestStart;
}

// - The server should inform and start switch procedure if in netplay.

protected function bool IsLevelPlayable ()
{
	if (SpawnManager_PlayerStartGroup.Length > 0)
	{
		return True;
	}
	else
	{
		return False;
	}
}

// - (Note) The variables are always initialized to some
// value.

protected function KillEvent_MaxLives ()
{
	// ## Victim

	// ** Player Out (Victim)

	KillEventData.bVictimOutofLives
	=
	(
		(KillEventData.Victim != None)
		&&
		KillEventData.Victim.PlayerReplicationInfo.bOutofLives
	);

	// ** Respawns Depleted (Victim)
	// - To notify player that it is on its last respawn in a match
	// where players have two or more lives.

	KillEventData.bVictimRespawnsDepleted
	=
	(
		(MaxLives > 1)
		&&
		(KillEventData.Victim != None)
		&&
		((MaxLives - KillEventData.Victim.PlayerReplicationInfo.NumLives) == 1)
	);
}

// ## Killed ##

protected function AwardReset_Multikill (Controller Killer)
{
	local RePRI OtherPRI;

	OtherPRI = RePRI(Killer.PlayerReplicationInfo);

	if (OtherPRI != None)
	{
		OtherPRI.Multikill_Count = 0;
	}
}

// **
// - To support the statistics system.
// - To support stacked holy shit announcements.

protected function AwardCheck_Multikill (Controller Killer, float Reward)
{
	local RePRI OtherPRI;
	local int BoundedLevel;

	OtherPRI = RePRI(Killer.PlayerReplicationInfo);

	if (OtherPRI != None)
	{
		if ((Level.TimeSeconds - OtherPRI.Multikill_TimeSeconds_Last) < 4)
		{
			// ** Reflect State

			if (OtherPRI != None)
			{
				BoundedLevel = Min (OtherPRI.Multikill_Count, 7);

				switch (BoundedLevel)
				{
					case 0:
						StatisticsProxy_RecordEvent (OtherPRI, 'DoubleKill', Reward);
						break;

					case 1:
						StatisticsProxy_RecordEvent (OtherPRI, 'MultiKill', Reward);
						break;

					case 2:
						StatisticsProxy_RecordEvent (OtherPRI, 'MegaKill', Reward);
						break;

					case 3:
						StatisticsProxy_RecordEvent (OtherPRI, 'UltraKill', Reward);
						break;

					case 4:
						StatisticsProxy_RecordEvent (OtherPRI, 'MonsterKill', Reward);
						break;

					case 5:
						StatisticsProxy_RecordEvent (OtherPRI, 'LudicrousKill', Reward);
						break;

					case 6:
						StatisticsProxy_RecordEvent (OtherPRI, 'HolyShit', Reward);
						break;

					case 7:
						StatisticsProxy_RecordEvent (OtherPRI, 'HolyShitStack', Reward);
						break;
				}

				BoundedLevel = Min (OtherPRI.Multikill_Count, 6);

				OtherPRI.MultiKills[BoundedLevel] += 1;

				if
				(
					(OtherPRI.Multikill_Count > 0)
					&&
					(OtherPRI.MultiKills[BoundedLevel - 1] > 0)
				)
				{
					OtherPRI.MultiKills[BoundedLevel - 1] -= 1;
				}
			}

			++OtherPRI.Multikill_Count;

			SpecialEvent (OtherPRI, "multikill_" $ OtherPRI.Multikill_Count);

			Killer.AwardAdrenaline (Reward);
		}
		else
		{
			OtherPRI.Multikill_Count = 0;
		}

		OtherPRI.Multikill_TimeSeconds_Last = Level.TimeSeconds;
	}
}

// ## Killed ##

protected function bool IsSuicide (Class<DamageType> DamageTypeClass)
{
	return
	(
		(DamageTypeClass == Class'Engine.Suicided')
		||
		(DamageTypeClass == Class'Engine.DamageType')
	);
}

// To keep the KillEventData up-to-date.

protected function UpdateKillEventData
(
	Controller Killer, // (Killer || None)
	Controller Victim,
	Pawn Victim_Pawn,
	class<DamageType> Class_DamageType,
	out PlayerReplicationInfo Out_Killer_PlayerReplicationInfo,
	out TeamInfo Out_Killer_TeamInfo,
	out PlayerReplicationInfo Out_Victim_PlayerReplicationInfo,
	out TeamInfo Out_Victim_TeamInfo,
	out int Out_Flags
)
{
	KillEventData.TimeSeconds_Updated = Level.TimeSeconds;

	// Update Damage Type

	KillEventData.Class_DamageType = Class_DamageType;

	if
	(
		(Killer != None)
		&&
		Killer.bIsPlayer
		&&
		(Killer.PlayerReplicationInfo != None)
		&&
		(Killer.PlayerReplicationInfo.Team != None)
	)
	{
		KillEventData.Killer = Killer;
		Out_Flags = Out_Flags | 1;

		Out_Killer_PlayerReplicationInfo = Killer.PlayerReplicationInfo;
		Out_Killer_TeamInfo = Killer.PlayerReplicationInfo.Team;

		KillEventData.IndexTeam_Killer = Killer.PlayerReplicationInfo.Team.TeamIndex;
	}
	else
	{
		KillEventData.Killer = None;
		KillEventData.IndexTeam_Killer = IndexTeam_Unspecified;
	}

	// Update Victim

	if
	(
		(Victim != None)
		&&
		Victim.bIsPlayer
		&&
		(Victim.PlayerReplicationInfo != None)
		&&
		(Victim.PlayerReplicationInfo.Team != None)
	)
	{
		KillEventData.Victim = Victim;
		Out_Flags = Out_Flags | 2;

		Out_Victim_PlayerReplicationInfo = Victim.PlayerReplicationInfo;
		Out_Victim_TeamInfo = Victim.PlayerReplicationInfo.Team;

		KillEventData.IndexTeam_Victim = Victim.PlayerReplicationInfo.Team.TeamIndex;
	}
	else
	{
		KillEventData.Victim = None;
		KillEventData.IndexTeam_Victim = IndexTeam_Unspecified;
	}

	// Update Context

	if ((Out_Flags & 3) == 3) // Killer && Victim
	{
		if (Out_Killer_PlayerReplicationInfo != Out_Victim_PlayerReplicationInfo)
		{
			if (bTeamGame)
			{
				if (Out_Killer_TeamInfo != Out_Victim_TeamInfo)
				{
					KillEventData.Context = KillEventContext_EnemyKill;
				}
				else
				{
					KillEventData.Context = KillEventContext_TeamKill;
				}
			}
			else
			{
				KillEventData.Context = KillEventContext_EnemyKill;
			}
		}
		else if (IsSuicide (Class_DamageType))
		{
			KillEventData.Context = KillEventContext_Suicide;
		}
		else
		{
			KillEventData.Context = KillEventContext_Death;
		}
	}
	else
	{
		KillEventData.Context = KillEventContext_None;
	}
}

// ####

protected function UpdateTeamEventData ()
{
	local Controller Other;
	local PlayerReplicationInfo OtherPRI;


	TeamEventData.TimeSeconds_Updated = Level.TimeSeconds;

	TeamEventData.Last[IndexTeam_A] = None;
	TeamEventData.Last[IndexTeam_B] = None;

	TeamEventData.Count_Out[IndexTeam_A] = 0;
	TeamEventData.Count_Out[IndexTeam_B] = 0;

	TeamEventData.Count_In[IndexTeam_A] = 0;
	TeamEventData.Count_In[IndexTeam_B] = 0;

	TeamEventData.Count_Lives[IndexTeam_A] = 0;
	TeamEventData.Count_Lives[IndexTeam_B] = 0;


	for (Other = Level.ControllerList; Other != None; Other = Other.NextController)
	{
		OtherPRI = Other.PlayerReplicationInfo;

		if
		(
			Other.bIsPlayer
			&&
			(OtherPRI != None)
			&&
			!OtherPRI.bOnlySpectator
			&&
			(OtherPRI.Team != None)
		)
		{
			if (OtherPRI.bOutofLives)
			{
				++TeamEventData.Count_Out[OtherPRI.Team.TeamIndex];
			}
			else
			{
				++TeamEventData.Count_In[OtherPRI.Team.TeamIndex];

				TeamEventData.Last[OtherPRI.Team.TeamIndex] = Other;
			}

			TeamEventData.Count_Deaths[OtherPRI.Team.TeamIndex] += OtherPRI.NumLives;
			TeamEventData.Count_Lives[OtherPRI.Team.TeamIndex] += MaxLives - OtherPRI.NumLives;
		}
	}
}

protected function AwardGive_EnemyKill (Controller Killer, Controller Victim, Pawn VictimPawn, Class<DamageType> ClassDamageType)
{
	Killer.AwardAdrenaline (ADR_Kill);
}

// - To intercept the DamageType.Static.ScoreKill function.
// -- Replaces the DeathMatch implementation.
// - bOutofLives is modified in ScoreKill.
// -- Note: TeamGame does not increment NumLives for team kills, but Deaths is always incremented.
// - ScoreKill is never called for players leaving.
// (Call-Sequence 1)
// - Pawn.Died () -> Level.Game.Killed ()
// Contexts:
// - A: Player killed by other player.
// - B: Player killed self.
// - C: Player killed by environmental or game type damage.

event Killed
(
	Controller Killer, // (Killer || None)
	Controller Victim,
	Pawn Victim_Pawn,
	Class<DamageType> DamageTypeClass
)
{
	local Vehicle V;
	local PlayerReplicationInfo Killer_PlayerReplicationInfo, Victim_PlayerReplicationInfo;
	local TeamInfo Killer_TeamInfo, Victim_TeamInfo;
	local string KillInfo;
	local int Score, Flags;

	// Update Killer

	if (Killer == None)
	{
		Killer = Victim;
	}

	// Categorize

	UpdateKillEventData (Killer, Victim, Victim_Pawn, DamageTypeClass, Killer_PlayerReplicationInfo, Killer_TeamInfo, Victim_PlayerReplicationInfo, Victim_TeamInfo, Flags);

	// Update Statistics

	if (KillEventData.Context == KillEventContext_EnemyKill)
	{
		KillEvent
		(
			"K",
			Killer_PlayerReplicationInfo,
			Victim_PlayerReplicationInfo,
			DamageTypeClass
		);


		TeamGameEvents.PlayerKilledEnemy (Killer, Victim, DamageTypeClass);
	}
	else if
	(
		(KillEventData.Context == KillEventContext_Death)
		||
		(KillEventData.Context == KillEventContext_Suicide)
	)
	{
		KillEvent
		(
			"K",
			None,
			Victim_PlayerReplicationInfo,
			DamageTypeClass
		);


		if (KillEventData.Context == KillEventContext_Suicide)
		{
			TeamGameEvents.PlayerSuicided (Victim, DamageTypeClass);
        }
        else if (KillEventData.Context == KillEventContext_Death)
        {
			TeamGameEvents.PlayerDied (Victim, DamageTypeClass);
        }
	}
	else if (KillEventData.Context == KillEventContext_TeamKill)
	{
		KillEvent
		(
			"TK",
			Killer_PlayerReplicationInfo,
			Victim_PlayerReplicationInfo,
			DamageTypeClass
		);

		TeamGameEvents.PlayerKilledTeammate (Killer, Victim, DamageTypeClass);
	}

	// Check Awards

	KillEvent_OnVictim ();


	AwardCheck_Retribution ();


	if (!KillEventData.bFirstDeath)
	{
		AwardCheck_FirstDeath ();
	}


	AwardCheck_ItEndsHere (Victim_Pawn);


	if
	(
		(KillEventData.Context == KillEventContext_EnemyKill)
		&&
		(class<VehicleDamageType> (DamageTypeClass) != None)
	)
	{
		DamageTypeClass.static.ScoreKill (Killer, Victim);
	}


	if (!KillEventData.bFirstKill)
	{
		AwardCheck_FirstKill ();
	}


	switch (KillEventData.Context)
	{
		case KillEventContext_EnemyKill:

			AwardGive_EnemyKill (Killer, Victim, Victim_Pawn, DamageTypeClass);

			AwardCheck_Multikill (Killer, ADR_MajorKill);

			if (Killer.Pawn != None)
			{
				Killer.Pawn.IncrementSpree ();

				AwardCheck_KillingSpree (Killer.Pawn.GetSpree ());
			}

			break;

		case KillEventContext_Suicide:

			if (DamageTypeClass != Class'Engine.DamageType') // Do not punish for changing teams.
			{
				Killer.AwardAdrenaline (ADR_MinorError);
			}

			break;

		case KillEventContext_Death:

			Killer.AwardAdrenaline (ADR_MinorError);

			break;

		case KillEventContext_TeamKill:

			Killer.AwardAdrenaline (ADR_KillTeamMate);

			AwardReset_Multikill (Killer);

			break;

		default:

			break;
	}

	// Handle Vehicles

	if
	(
		((Flags & 1) == 1) // Killer
		&&
		(Vehicle(Victim_Pawn) != None)
		&&
		(
			(Victim != None)
			||
			Vehicle(Victim_Pawn).bEjectDriver
		)
		&&
		Vehicle(Victim_Pawn).IndependentVehicle ()
	)
	{
		Score = VehicleScoreKill (Killer, Victim, Vehicle(Victim_Pawn), KillInfo);

		if (Score > 0)
		{
			if (KillEventData.Context == KillEventContext_TeamKill)
			{
				Score = -Score;

				KillInfo = "TeamKill_" $ KillInfo;
			}

			if (Score != 0)
			{
				Killer_PlayerReplicationInfo.Score += Score;
				Killer_PlayerReplicationInfo.NetUpdateTime = Level.TimeSeconds - 1.0f;

				ScoreEvent (Killer_PlayerReplicationInfo, Score, KillInfo);
			}
		}
	}

	// *May* modify bOutofLives, NumLives, Deaths

	if ((Flags & 3) == 3)
	{
		IncrementDeaths (Victim_PlayerReplicationInfo);

		ScoreKill (Killer, Victim);
	}

	// At this point bOutofLives has been evaluated.

	ChangePlayerStatus (Victim, DamageTypeClass);

	//

	if (Level.GRI.bTeamGame)
	{
		UpdateTeamEventData ();
	}


	if (MaxLives > 0)
	{
		KillEvent_MaxLives ();
	}


	DiscardInventory (Victim_Pawn);


	if
	(
		bAllowVehicles
		&&
		(Level.NetMode == NM_Standalone)
		&&
		(PlayerController(Victim) != None)
	)
	{
		for (V = VehicleList; V != None; V = V.NextVehicle)
		{
			if (Victim.GetTeamNum () == V.Team)
			{
				V.PlayerStartTime = 0;
			}
		}
	}


	NotifyKilled (Killer, Victim, Victim_Pawn);


	if ((Flags & 3) == 3)
	{
		KilledBroadcastMessage (Killer, Victim, DamageTypeClass);
	}
}

//

protected function IncrementDeaths (PlayerReplicationInfo Victim_PlayerReplicationInfo)
{
	if (Victim_PlayerReplicationInfo.Deaths < 99999.0f)
	{
		Victim_PlayerReplicationInfo.Deaths += 1.0f;
	}

	Victim_PlayerReplicationInfo.NetUpdateTime = FMin (Victim_PlayerReplicationInfo.NetUpdateTime, Level.TimeSeconds + 0.3f * FRand ());
}

// To set a consistent status on players (Dead, Out).

protected function ChangePlayerStatus (Controller Victim, Class<DamageType> DamageTypeClass)
{
	local RePRI OtherRePRI;

	if (Victim != None)
	{
		OtherRePRI = RePRI(Victim.PlayerReplicationInfo);
	}

	if
	(
		(OtherRePRI != None)
		&&
		(OtherRePRI.RealtimePlayerInfo != None)
	)
	{
		if (OtherRePRI.bOutofLives)
		{
			OtherRePRI.RealtimePlayerInfo.ChangePlayerStatus ('Out');
		}
		else
		{
			OtherRePRI.RealtimePlayerInfo.ChangePlayerStatus ('Dead');
		}
	}
	else
	{
		Warn ("ChangePlayerStatus () Victim:" @ Victim);
	}
}

protected function KilledBroadcastMessage
(
	Controller Killer, // (Killer || None)
	Controller Victim,
	Class<DamageType> DamageTypeClass
)
{
	BroadcastDeathMessage (Killer, Victim, DamageTypeClass);
}

// - To direct events to the new statistics functions.

event KillEvent
(
	string Text_Type,
	PlayerReplicationInfo Killer,
	PlayerReplicationInfo Victim,
	class<DamageType> DamageTypeClass
)
{
	local RePRI PRI_Killer;
	local RePRI PRI_Victim;
	local bool bIsTeamKill;

	// Context

	if (Text_Type ~= "TK")
	{
		bIsTeamKill = True;
	}

	PRI_Killer = RePRI(Killer);
	PRI_Victim = RePRI(Victim);

	// Death

	if (PRI_Victim != None)
	{
		if
		(
			(Killer == None)
			||
			(Killer == Victim)
		)
		{
			PRI_Victim.Statistics_RecordSuicide (DamageTypeClass);
		}

		PRI_Victim.Statistics_RecordDeath (DamageTypeClass, PRI_Killer, bIsTeamKill);
	}

	// Kill

	if
	(
		(PRI_Killer != None)
		&&
		(PRI_Killer != Victim)
	)
	{
		PRI_Killer.Statistics_RecordKill (DamageTypeClass, PRI_Victim, bIsTeamKill);
	}

	if (GameStats != None)
	{
		GameStats.KillEvent (Text_Type, Killer, Victim, DamageTypeClass);
	}
}

// ** Pawn.KilledBy -> Pawn.Died -> GameInfo.Killed -> GameInfo.BroadcastDeathMessage -> GameInfo.BroadcastLocalized -> BroadcastHandler.AllowBroadcastLocalized -> BroadcastHandler.BroadcastLocalized -> BroadcastHandler.AcceptBroadcastLocalized -> PlayerController.ReceiveLocalizedMessage
//
// - To add more information to the death messages, which beat variable replication to the client.
// -- The Switch value is a bitmap.
// --- The original value was 0 or 1 (occupied one binary digit).
//
// Preconditions:
// - (Killed != None)
// - Killed.bIsPlayer

event BroadcastDeathMessage
(
	Controller Killer,
	Controller Victim,
	class<DamageType> DamageTypeClass
)
{
	local int Context;

	// Additional flags sent.

	if (MaxLives > 0)
	{
		Context = GetBroadcastDeathMessageFlags ();
	}

	// - The standard death message broadcast logic.

	if
	(
		(Killer == Victim) // Suicide
		||
		(Killer == None) // World damage killed player.
	)
	{
		BroadcastLocalized (Self, DeathMessageClass, Context | 0x1, None, Victim.PlayerReplicationInfo, DamageTypeClass);
	}
	else // Player killed player.
	{
		BroadcastLocalized (Self, DeathMessageClass, Context, Killer.PlayerReplicationInfo, Victim.PlayerReplicationInfo, DamageTypeClass);
	}
}

// Context:
// Bit: Value:	Interpretation:
// 0	0		Standard: Player <DamageType> Player
// 0	1		Standard: <DamageType> Player
// 1	1		Extended: Victim is out (PRI.bOutofLives && (GRI.MaxLives > 0))
// 2	1		Extended: Respawns Depleted (((PRI.NumLives + 1) == Level.Game.MaxLives) && (GRI.MaxLives > 1))
// 3	1		Extended: Final Death (GRI.MaxLives > 0) - To slow down karma for dramatic effect.

protected function int GetBroadcastDeathMessageFlags ()
{
	local int Context;

	if (KillEventData.bVictimOutofLives)
	{
		Context = Context | 0x2; // ** Player Out (Victim)
	}
	else if (KillEventData.bVictimRespawnsDepleted)
	{
		Context = Context | 0x4; // ** Respawns Depleted (Victim)
	}

	// - The match may have ended at this point.
	// - If round-based, and more rounds, the level is reset shortly after
	// receiveing this message.
	// - Is set whether there is at least one team or one player left.

	if
	(
		(KillEventData.IndexTeam_Victim != IndexTeam_Unspecified)
		&&
		(TeamEventData.Count_In[KillEventData.IndexTeam_Victim] == 0)
	)
	{
		Context = Context | 0x8; // ** Last Man Standing (!bTeamGame) / Last Team Standing (bTeamGame)
	}

	//

	return Context;
}


event NotifyKilled (Controller Killer, Controller Killed, Pawn KilledPawn)
{
	if (bRageQuitAnnouncement)
	{
		if
		(
			(Killer != None)
			&&
			(RePlayer(Killed) != None)
		)
		{
			RageQuitTrack (RePlayer(Killed));
		}
	}

	Super.NotifyKilled (Killer, Killed, KilledPawn);
}

event NotifyLogout (Controller Exiting)
{
	if (bRageQuitAnnouncement)
	{
		if (RePlayer(Exiting) != None)
		{
			RageQuitCheck (RePlayer(Exiting));
		}
	}

	Super.NotifyLogout (Exiting);
}

// ## Rage Quit ##

protected function RageQuitCheck (RePlayer Exiting)
{
	if
	(
		(Exiting != None)
		&&
		!ReGRI(GameReplicationInfo).IsPending ()
		&&
		!ReGRI(GameReplicationInfo).IsOver ()
		&&
		(
			(Level.TimeSeconds - Exiting.RageQuit_TimeSeconds_Valid)
			<
			RageQuit_TimeSeconds_Interval
		)
	)
	{
		BroadcastLocalizedMessage
		(
			Class'LocalMessage_RageQuit',
			,
			Exiting.PlayerReplicationInfo
		);

		TeamGameEvents.PlayerRageQuit (Exiting);
	}
}

protected function RageQuitTrack (RePlayer Killed)
{
	if (Killed != None)
	{
		Killed.RageQuit_TimeSeconds_Valid = Level.TimeSeconds;
	}
}

// ## Level Reset ##

// Protocol:
// - Return True to call A.Reset ().

protected function bool Level_Reset_Query (Actor A)
{
	if
	(
		A.IsA ('LevelInfo') // To skip garbage collection between resets.
		&&
		!bGarbageCollection
	)
	{
		Level.DefaultGravity = Level.Default.DefaultGravity;

		return False;
	}

	return True;
}

protected function Level_Destroy_Actors ()
{
	local Projectile StrayProjectile;

	ForEach DynamicActors (Class'Engine.Projectile', StrayProjectile)
	{
		StrayProjectile.Destroyed ();
	}
}

protected function Level_Reset ()
{
	Level_Reset_Controllers ();

	Level_Destroy_Actors ();

	Level_Reset_Actors ();

	Reset ();

	SetGameSpeed (GameSpeed);
}

// - If the client is a match participant, its state is reset to PlayerWaiting.
// - If the client is a match spectator, ClientResetSpectator () gets called.
// - The PlayerReplicationInfo is not reset yet, and thus NumLives
// and bOutofLives remain unchanged (handled separately).

protected function Level_Reset_Controllers ()
{
	local Controller Other;
	local Controller Next;
	local NavigationPoint StartSpot;
	local float Adrenaline;

	Other = Level.ControllerList;

	while (Other != None)
	{
		Next = Other.NextController;

		if
		(
			Other.bIsPlayer
			&&
			(Other.PlayerReplicationInfo != None)
		)
		{
			if (!Other.PlayerReplicationInfo.bOnlySpectator)
			{
				StartSpot = Other.StartSpot;


				if
				(
					!bFreezeAdrenaline
					&&
					(Other.Pawn != None)
					&&
					(Other.Pawn.Health > 0)
					&&
					Other.Pawn.InCurrentCombo ()
				)
				{
					Other.Pawn.RemovePowerups ();
				}


				Adrenaline = Other.Adrenaline;


				if (Other.IsA ('PlayerController'))
				{
					PlayerController(Other).ClientReset ();
				}


				Other.Reset ();


				if (Other != None)
				{
					Other.Adrenaline = Adrenaline;
					Other.StartSpot = StartSpot;

					if (Other.IsA ('PlayerController'))
					{
						PlayerController(Other).WaitDelay = 0; // PlayerController.ServerRestartPlayer ()
					}
				}
			}
			else
			{
				if (Other.IsA ('RePlayer'))
				{
					RePlayer(Other).ClientResetSpectator ();
				}
			}
		}

		Other = Next;
	}
}

// - LevelInfo.Reset triggers the garbage collection process.
// -- In unstable environments, this is likely to cause a crash, which
// is great for developers to weed out errors early, but unfortunate for
// servers that crash.
// - The Debugger triggers a crash whenever any Mover is reset.
// -- Seems to be related to the initial state of the mover.
// - The game calls its own reset function when this function returns.

protected function Level_Reset_Actors ()
{
	local Actor A;

	ForEach AllActors (Class'Engine.Actor', A)
	{
		if
		(
			!Level_Reset_Query (A) // Game type specific actors to exclude.  Can override the rest of the checks, but not recommended.
			||
			A.IsA ('Controller') // Already reset.
			||
			A.IsA ('PlayerReplicationInfo') // Handled in Level_RestartPlayers
			||
			A.IsA ('Mover') // Debugger forces crash when these actors are attempted reset (seems to be related to InitialState).
			||
			(A == Self) // Reset last.
		)
		{
			continue;
		}

		A.Reset ();
	}
}

// - To reset network clients and bots between rounds.
// - Bots are restarted by MatchInProgress.Timer.
// -- This is a slow process in NM_Standalone, however.
// - Match spectators (bOnlySpectator) are not reset.

protected function Level_RestartPlayers ()
{
	local Controller Other;
	local Controller Next;

	Other = Level.ControllerList;

	while (Other != None)
	{
		Next = Other.NextController;

		if
		(
			Other.bIsPlayer
			&&
			(Other.PlayerReplicationInfo != None)
			&&
			!Other.PlayerReplicationInfo.bOnlySpectator
		)
		{
			ResetPlayerLives (Other);

			if
			(
				Other.IsA ('PlayerController')
				||
				(Level.NetMode == NM_Standalone)
			)
			{
				RestartPlayer (Other);
			}
			else
			{
				Other.GotoState ('Dead', 'MPStart'); // Adds waiting time before respawning a bot. -> Level.Game.RestartPlayer (Self)
			}
		}

		Other = Next;
	}
}

// #Spawning

event RestartPlayer (Controller Other)
{
	Super.RestartPlayer (Other);

	if
	(
		(Other != None)
		&&
		(Other.Pawn != None)
	)
	{
		PostRestartPlayer (Other, Other.Pawn);

		TeamGameEvents.PlayerSpawned (Other, Other.Pawn);
	}
}

// - In an effort to find spawn points and increase success-rate of Spawn.

protected function CachePlayerPositions ()
{
	local Controller Other;
	local Pawn OtherPawn;
	local int Index;

	// Reset

	PlayerSpawnInfo.Length = 0;

	// Check for a 1-on-1 scenario.

	if
	(
		(TeamEventData.Count_In[IndexTeam_A] == 1)
		&&
		(TeamEventData.Count_In[IndexTeam_B] == 1)
	)
	{
		bVersus = True;
    }
    else
    {
		bVersus = False;
    }

	// Cache position, collision size, and team index of players for quicker lookup.

	for (Other = Level.ControllerList; Other != None; Other = Other.NextController)
	{
		OtherPawn = Other.Pawn;

		if
		(
			(OtherPawn != None)
			&&
			Other.bIsPlayer
		)
		{
			Index = PlayerSpawnInfo.Length;
			PlayerSpawnInfo.Length = PlayerSpawnInfo.Length + 1;

			PlayerSpawnInfo[Index].PawnLocation = OtherPawn.Location;
			PlayerSpawnInfo[Index].PawnCollisionSize = (OtherPawn.CollisionRadius * 2.0f) + OtherPawn.CollisionHeight;

			if
			(
				(OtherPawn.PlayerReplicationInfo != None)
				&&
				(OtherPawn.PlayerReplicationInfo.Team != None)
				&&
				(OtherPawn.PlayerReplicationInfo.Team.TeamIndex < IndexTeam_Max)
			)
			{
				PlayerSpawnInfo[Index].PawnTeamIndex = OtherPawn.PlayerReplicationInfo.Team.TeamIndex;
            }
            else
            {
				PlayerSpawnInfo[Index].PawnTeamIndex = IndexTeam_Unspecified;

				Warn ("IndexTeam_Unspecified");
            }

			PlayerSpawnInfo[Index].ZoneNumber = OtherPawn.Region.ZoneNumber;
		}
	}
}

//

protected function bool IsLatecomer (Controller Other)
{
	return
	(
		bMustJoinBeforeStart
		&&
		(UnrealPlayer(Other) != None)
		&&
		UnrealPlayer(Other).bLatecomer
	);
}

// Context:
// - RestartPlayer ()
// Preconditions:
// - (Other != None)
// - (OtherPawn != None)

protected function PostRestartPlayer (Controller Other, Pawn OtherPawn)
{
	if (RePRI(Other.PlayerReplicationInfo) != None)
	{
		RePRI(Other.PlayerReplicationInfo).RealtimePlayerInfo.DefaultPlayerStatus ();
	}


	NotifyPlayerRestarted (Other, OtherPawn);
}

// #TeamCombos

protected function NotifyPlayerRestarted (Controller Other, Pawn OtherPawn)
{
	if (Level.GRI.bTeamGame)
	{
		UpdateTeamEventData ();
	}

	if // a)
	(
		(KillEventData.DarkHorse != None)
		&&
		(KillEventData.DarkHorse.PlayerReplicationInfo != None)
		&&
		(KillEventData.DarkHorse.PlayerReplicationInfo.Team != None)
		&&
		(KillEventData.DarkHorse.PlayerReplicationInfo.Team.TeamIndex < IndexTeam_Max)
		&&
		(TeamEventData.Count_In[KillEventData.DarkHorse.PlayerReplicationInfo.Team.TeamIndex] > 1)
		&&
		!ReGRI(GameReplicationInfo).IsRestarting ()
	)
	{
		AwardReset_DarkHorse ();
	}

	if (ReGRI(Level.GRI).bTeamCombos)
	{
		ReGRI(Level.GRI).TeamCombos_PlayerRestarted (Other);
	}
}

protected function AwardReset_DarkHorse ()
{
	KillEventData.DarkHorse = None;
}

// #

function float GetScoreLimit ()
{
	return GoalScore;
}

// #Queries

// Preconditions:
// - (Class_DamageType != None)

static function bool GameRulesDamageType (class<DamageType> DamageTypeClass)
{
	return
	(
		(DamageTypeClass == Class'DamTypeCamping')
		||
		(DamageTypeClass == Class'DamTypeOvertime')
		||
		(DamageTypeClass == class'Engine.Suicided')
		||
		(DamageTypeClass == class'Engine.DamageType')
	);
}

// ## Player Names ##

// Contexts:
// - [S] GameInfo.Login () [S->S] ChangeName ()
// - C-> PlayerController.SetName () [C->S] PlayerController.ChangeName () -> GameInfo.ChangeName ()
// - S-> DeathMatch.InitializeBot -> ChangeName ()
// - S-> DeathMatch.InitPlacedBot -> ChangeName ()
// Preconditions:
// - (Other != None)
// - (Other.PlayerReplicationInfo != None)

event ChangeName
(
	Controller Other,
	string RequestedPlayerNameString,
	bool bChange
)
{
	local string PlayerNameString;

	RequestedPlayerNameString = Class'RePRI'.Static.GetFilteredPlayerNameString (RequestedPlayerNameString);

	if (RequestedPlayerNameString == "")
	{
		Warn ("(RequestedPlayerNameString == \"\")");

		return;
	}

	// Check requested name against current name.

	PlayerNameString = Class'RePRI'.Static.GetFilteredPlayerNameString (Other.PlayerReplicationInfo.PlayerName);

	if (PlayerNameString ~= RequestedPlayerNameString)
	{
		Warn ("(PlayerNameString ~= RequestedPlayerNameString)");

		return;
	}

	// Check if name is reserved.

	if
	(
		(RequestedPlayerNameString ~= "WebAdmin")
		||
		(RequestedPlayerNameString ~= "DemoRecSpectator")
	)
	{
		Warn ("(Reserved Names)");
	}

	// (Some redundancy like StripColor..)

	Super.ChangeName (Other, RequestedPlayerNameString, bChange);
}

// #Adrenaline

// a. [S] RePlayer.DoCombo ()
// b. [S] ReBot.TryCombo () [S->S] ReTeamGame.RecommendCombo ()
// - Reminder: The ~= operator is not case-sensitive.

function bool AdrenalineComboDenied (string ComboClassNameString)
{
	local class<Combo> LoadedComboClass;
	local string KeyComboString;
	local int IndexCombo;

	if (Len (ComboClassNameString) > 2) // "A.B"
	{
		LoadedComboClass = Class<Combo>(DynamicLoadObject (ComboClassNameString, Class'Core.Class'));

		if (LoadedComboClass != None)
		{
			KeyComboString = Class'ReUtility'.Static.GetKeyComboString (LoadedComboClass);

			for (IndexCombo = 0; IndexCombo < DenyAdrenalineCombo.Length; ++IndexCombo)
			{
				if (DenyAdrenalineCombo[IndexCombo] ~= KeyComboString)
				{
					return True;
				}
			}
		}
		else
		{
			Warn ("(LoadedComboClass == None)" @ ComboClassNameString);
		}
    }

	return False;
}

// - Bot controllers call this function.
// (Call-Sequence) Bot.TryCombo () -> GameInfo.NewRecommendCombo () -> GameInfo.RecommendCombo ()
// - Bot.TryCombo () has been adapted to expect an empty string in return.
// - A game type can override a mutator by returning a different text string.
// Protocol:
// - Return an empty string to signal that the combo is not enabled.
// - Return a different string to override a mutator.
// Preconditions:
// - Text_Combo is a fully qualified class name (PackageName.ClassName).

event string RecommendCombo (string ComboClassNameString)
{
	if (AdrenalineComboDenied (ComboClassNameString))
	{
		return "";
	}

	return ComboClassNameString;
}

event string NewRecommendCombo (string ComboClassNameString, AIController Other)
{
	local string NewComboClassNameString;

	NewComboClassNameString = RecommendCombo (ComboClassNameString);

	if (NewComboClassNameString ~= ComboClassNameString)
	{
		NewComboClassNameString = BaseMutator.NewRecommendCombo (ComboClassNameString, Other);
    }

	if (!AdrenalineComboDenied (NewComboClassNameString))
	{
		return NewComboClassNameString;
	}
	else
	{
		return "";
	}
}

// #

function PlayerUsedCombo (Controller Player, Class<Combo> ComboClass)
{
	TeamGameEvents.PlayerUsedCombo (Player, ComboClass);
}

// #

function PlayerAwarded (Controller Player, PlayerReplicationInfo PlayerPRI, Class<Award> AwardClass, int AwardIndex, int AwardMultiplier)
{
	TeamGameEvents.PlayerAwarded (Player, PlayerPRI, AwardClass, AwardIndex, AwardMultiplier);
}

// ##

// Contexts:
// a. PlayerController.BecomeActivePlayer () -> AllowBecomeActivePlayer (PlayerController)
// b. UT2K4Tab_PlayerLoginControls.BecomeActivePlayer () -> AllowBecomeActivePlayer (PlayerController)
// - To add a check for bRoundbased where MaxLives do not prevent players from joining.

event bool AllowBecomeActivePlayer (PlayerController PC)
{
	if (PC.PlayerReplicationInfo == None)
	{
		PC.ReceiveLocalizedMessage (GameMessageClass, 13);

		return False;
	}
	else if
	(
		bMustJoinBeforeStart // Configurable Server Requirement
		&&
		GameReplicationInfo.bMatchHasBegun // Server Match State
	)
	{
		PC.ReceiveLocalizedMessage (GameMessageClass, 13);

		return False;
	}
	else if (NumPlayers >= MaxPlayers) // Server Limit
	{
		PC.ReceiveLocalizedMessage (GameMessageClass, 13);

		return False;
	}
	else if
	(
		!bRoundbased
		&&
		(MaxLives > 0)
		&&
		PC.PlayerReplicationInfo.bOutofLives
	)
	{
		PC.ReceiveLocalizedMessage (GameMessageClass, 13);

		return False;
	}
	else if (PC.IsInState ('GameEnded'))
	{
		PC.ReceiveLocalizedMessage (GameMessageClass, 13);

		return False;
	}
	else if (PC.IsInState ('RoundEnded'))
	{
		PC.ReceiveLocalizedMessage (GameMessageClass , 13);

		return False;
	}
	else if (Level.NetMode == NM_Standalone)
	{
		if (NumBots > InitialBots)
		{
			--RemainingBots;

			bPlayerBecameActive = True;
		}

		TeamGameEvents.PlayerBecameParticipant (PC);

		return True;
	}


	TeamGameEvents.PlayerBecameParticipant (PC);

	return True;
}

// ## Player Control ##

event bool BecomeSpectator (PlayerController OtherPC)
{
	if (Super.BecomeSpectator (OtherPC))
	{
		NotifyPlayerBecomeSpectatorAccepted (OtherPC);

		return True;
	}
	else
	{
		NotifyPlayerBecomeSpectatorDenied (OtherPC);

		return False;
	}
}

// ## Player Notifications ##

protected function NotifyPlayerBecomeSpectatorAccepted (PlayerController OtherPC)
{
	DestroyPlayerProperty (OtherPC);


	TeamGameEvents.PlayerBecameSpectator (OtherPC);
}

protected function NotifyPlayerBecomeSpectatorDenied (PlayerController OtherPC)
{
}

// - To solve a problem with a lingering pawn when leaving to
// become a spectator while controlling a Redeemer, and derez bodies.
// - To solve a problem with leftover projectiles.
// *X* To spawn spark emitters to replace the projectiles.

function DestroyPlayerProperty (Controller Other)
{
	local Actor OtherActor;

	ForEach DynamicActors (Class'Engine.Actor', OtherActor)
	{
		if (OtherActor.Owner == Other)
		{
			if (OtherActor.IsA ('Pawn'))
			{
				Pawn(OtherActor).Died (None, Class'Engine.DamageType', NullVector);
			}
			else if (OtherActor.IsA ('Projectile'))
			{
				OtherActor.Destroy ();
			}
		}
	}
}

// #ServerInfo

event GetServerInfo (out ServerResponseLine Out_SRL)
{
	local string MapNameString, MapNameRealGameTypeString;


	Super.GetServerInfo (Out_SRL);

	// #Override

	if (bMasquerade)
	{
		MapNameString = Left (string(Level), InStr(string(Level), "."));

		if (bShowGameType)
		{
			MapNameRealGameTypeString = MapNameString @ "(" $ Acronym $ ")";
        }
        else
        {
			MapNameRealGameTypeString = MapNameString;
        }

		Out_SRL.MapName = MapNameRealGameTypeString; // Example: DM-Rankin (ReFreon)

		Out_SRL.GameType = MasqueradingGameType; // Masquerade as a similar game type.
	}
}

// #PlayInfo

event GetServerDetails (out ServerResponseLine Out_SRL)
{
	Super.GetServerDetails (Out_SRL);


	AddServerDetail (Out_SRL, "Re - Version", Class'About'.Static.GetVersionString ());


	if (bTeamCombos != Default.bTeamCombos)
	{
		AddServerDetail (Out_SRL, CustomValueColorCodeString $ TeamCombosPropertyText, bTeamCombos);
    }
    else
    {
		AddServerDetail (Out_SRL, DefaultValueColorCodeString $ TeamCombosPropertyText, bTeamCombos);
    }


	if (bPublicMatchStatistics != Default.bPublicMatchStatistics)
	{
		AddServerDetail (Out_SRL, CustomValueColorCodeString $ PublicMatchStatisticsPropertyText, bPublicMatchStatistics);
    }
    else
    {
		AddServerDetail (Out_SRL, DefaultValueColorCodeString $ PublicMatchStatisticsPropertyText, bPublicMatchStatistics);
    }


	if (bLogMatchStatistics != Default.bLogMatchStatistics)
	{
		AddServerDetail (Out_SRL, CustomValueColorCodeString $ LogMatchStatisticsPropertyText, bLogMatchStatistics);
    }
    else
    {
		AddServerDetail (Out_SRL, DefaultValueColorCodeString $ LogMatchStatisticsPropertyText, bLogMatchStatistics);
    }


	if (bAllowTrackingDamageIndicators != Default.bAllowTrackingDamageIndicators)
	{
		AddServerDetail (Out_SRL, CustomValueColorCodeString $ AllowTrackingDamageIndicatorsPropertyText, bAllowTrackingDamageIndicators);
    }
    else
    {
		AddServerDetail (Out_SRL, DefaultValueColorCodeString $ AllowTrackingDamageIndicatorsPropertyText, bAllowTrackingDamageIndicators);
    }


	if (bDamageBonus != Default.bDamageBonus)
	{
		AddServerDetail (Out_SRL, CustomValueColorCodeString $ DamageBonusPropertyText, bDamageBonus);
    }
    else
    {
		AddServerDetail (Out_SRL, DefaultValueColorCodeString $ DamageBonusPropertyText, bDamageBonus);
    }


	if (bCampingCheck != Default.bCampingCheck)
	{
		AddServerDetail (Out_SRL, CustomValueColorCodeString $ CampingCheckPropertyText, bCampingCheck);
    }
    else
    {
		AddServerDetail (Out_SRL, DefaultValueColorCodeString $ CampingCheckPropertyText, bCampingCheck);
    }


	if (bCampingCheckSendLocation != Default.bCampingCheckSendLocation)
	{
		AddServerDetail (Out_SRL, CustomValueColorCodeString $ CampingCheckSendLocationPropertyText, bCampingCheckSendLocation);
    }
    else
    {
		AddServerDetail (Out_SRL, DefaultValueColorCodeString $ CampingCheckSendLocationPropertyText, bCampingCheckSendLocation);
    }


	if (bDenySuicide != Default.bDenySuicide)
	{
		AddServerDetail (Out_SRL, CustomValueColorCodeString $ DenySuicidePropertyText, bDenySuicide);
    }
    else
    {
		AddServerDetail (Out_SRL, DefaultValueColorCodeString $ DenySuicidePropertyText, bDenySuicide);
    }


	if (bPickupNotifications != Default.bPickupNotifications)
	{
		AddServerDetail (Out_SRL, CustomValueColorCodeString $ PickupNotificationsPropertyText, bPickupNotifications);
    }
    else
    {
		AddServerDetail (Out_SRL, DefaultValueColorCodeString $ PickupNotificationsPropertyText, bPickupNotifications);
    }


	if (bDisableBerserkCombo != Default.bDisableBerserkCombo)
	{
		AddServerDetail (Out_SRL, CustomValueColorCodeString $ AdrenalineComboControlPropertyText[0], bDisableBerserkCombo);
    }
    else
    {
		AddServerDetail (Out_SRL, DefaultValueColorCodeString $ AdrenalineComboControlPropertyText[0], bDisableBerserkCombo);
    }


	if (bDisableBoosterCombo != Default.bDisableBoosterCombo)
	{
		AddServerDetail (Out_SRL, CustomValueColorCodeString $ AdrenalineComboControlPropertyText[1], bDisableBoosterCombo);
    }
    else
    {
		AddServerDetail (Out_SRL, DefaultValueColorCodeString $ AdrenalineComboControlPropertyText[1], bDisableBoosterCombo);
    }


	if (bDisableInvisibleCombo != Default.bDisableInvisibleCombo)
	{
		AddServerDetail (Out_SRL, CustomValueColorCodeString $ AdrenalineComboControlPropertyText[2], bDisableInvisibleCombo);
    }
    else
    {
		AddServerDetail (Out_SRL, DefaultValueColorCodeString $ AdrenalineComboControlPropertyText[2], bDisableInvisibleCombo);
    }


	if (bDisableSpeedCombo != Default.bDisableSpeedCombo)
	{
		AddServerDetail (Out_SRL, CustomValueColorCodeString $ AdrenalineComboControlPropertyText[3], bDisableSpeedCombo);
    }
    else
    {
		AddServerDetail (Out_SRL, DefaultValueColorCodeString $ AdrenalineComboControlPropertyText[3], bDisableSpeedCombo);
    }


	if (bDisablePintsizedCombo != Default.bDisablePintsizedCombo)
	{
		AddServerDetail (Out_SRL, CustomValueColorCodeString $ AdrenalineComboControlPropertyText[4], bDisablePintsizedCombo);
    }
    else
    {
		AddServerDetail (Out_SRL, DefaultValueColorCodeString $ AdrenalineComboControlPropertyText[4], bDisablePintsizedCombo);
    }


	if (bDisableCamouflageCombo != Default.bDisableCamouflageCombo)
	{
		AddServerDetail (Out_SRL, CustomValueColorCodeString $ AdrenalineComboControlPropertyText[5], bDisableCamouflageCombo);
    }
    else
    {
		AddServerDetail (Out_SRL, DefaultValueColorCodeString $ AdrenalineComboControlPropertyText[5], bDisableCamouflageCombo);
    }
}

// To allow bots to be included

event GetServerPlayers (out ServerResponseLine Out_ServerState)
{
    local Mutator M;
	local Controller Other;
	local PlayerReplicationInfo OtherPRI;
	local int Index, TeamFlag[2];

	Index = Out_ServerState.PlayerInfo.Length;

	TeamFlag[0] = 1 << 29;
	TeamFlag[1] = TeamFlag[0] << 1;

	for (Other = Level.ControllerList; Other != None; Other = Other.NextController)
	{
		OtherPRI = Other.PlayerReplicationInfo;

		if
		(
			(MessagingSpectator(Other) == None)
			&&
			(OtherPRI != None)
		)
		{
			if (!OtherPRI.bBot)
			{
				Out_ServerState.PlayerInfo.Length = Index + 1;
				Out_ServerState.PlayerInfo[Index].PlayerNum = Other.PlayerNum;
				Out_ServerState.PlayerInfo[Index].PlayerName = OtherPRI.PlayerName;
				Out_ServerState.PlayerInfo[Index].Score = OtherPRI.Score;
				Out_ServerState.PlayerInfo[Index].Ping = 4 * OtherPRI.Ping;

				if (OtherPRI.Team != None)
				{
					Out_ServerState.PlayerInfo[Index].StatsID
					=
					Out_ServerState.PlayerInfo[Index].StatsID
					| TeamFlag[OtherPRI.Team.TeamIndex];
	            }

				++Index;
			}
			else if (bIncludeBotNames)
			{
				Out_ServerState.PlayerInfo.Length = Index + 1;
				Out_ServerState.PlayerInfo[Index].PlayerNum = Other.PlayerNum;
				Out_ServerState.PlayerInfo[Index].PlayerName = OtherPRI.PlayerName $ "*";
				Out_ServerState.PlayerInfo[Index].Score = OtherPRI.Score;
				Out_ServerState.PlayerInfo[Index].Ping = 4 * OtherPRI.Ping;

				if (OtherPRI.Team != None)
				{
					Out_ServerState.PlayerInfo[Index].StatsID
					=
					Out_ServerState.PlayerInfo[Index].StatsID
					| TeamFlag[OtherPRI.Team.TeamIndex];
	            }

				++Index;
			}
		}
    }

	// Mutators

	for (M = BaseMutator.NextMutator; M != None; M = M.NextMutator)
	{
		M.GetServerPlayers (Out_ServerState);
    }
}

// #

static event FillPlayInfo (PlayInfo PI)
{
	local string GroupString;
	local int Group;

	Super.FillPlayInfo (PI);

	// * - Teams

	Group = 100;
	GroupString = Default.ShortGameName @ "-" @ Default.PlayInfoGroupTeams;

	PI.AddSetting (GroupString, Default.RedTeamNamePropertyText, Default.RedTeamNameDisplayText, 0, Group++, "Text", "20", "", !bMultiplayerOnly, bAdvanced);
	PI.AddSetting (GroupString, Default.BlueTeamNamePropertyText, Default.BlueTeamNameDisplayText, 0, Group++, "Text", "20", "", !bMultiplayerOnly, bAdvanced);

	// * - Pickups

	Group = 200;
	GroupString = Default.ShortGameName @ "-" @ Default.PlayInfoGroupPickups;

	PI.AddSetting (GroupString, Default.PickupNotificationsPropertyText, Default.PickupNotificationsDisplayText, 0, Group++, "Check", "", "", !bMultiplayerOnly, bAdvanced);

	// * - Stats

	Group = 300;
	GroupString = Default.ShortGameName @ "-" @ Default.PlayInfoGroupStats;

	PI.AddSetting (GroupString, Default.LogMatchStatisticsPropertyText, Default.LogMatchStatisticsDisplayText, 0, Group++, "Check", "", "", !bMultiplayerOnly, bAdvanced);
	PI.AddSetting (GroupString, Default.LogPlayerGUIDPropertyText, Default.LogPlayerGUIDDisplayText, 0, Group++, "Check", "", "", !bMultiplayerOnly, bAdvanced);
	PI.AddSetting (GroupString, Default.PublicMatchStatisticsPropertyText, Default.PublicMatchStatisticsDisplayText, 0, Group++, "Check", "", "", !bMultiplayerOnly, bAdvanced);

	// * - Players

	Group = 400;
	GroupString = Default.ShortGameName @ "-" @ Default.PlayInfoGroupPlayers;

	PI.AddSetting (GroupString, Default.AllowTrackingDamageIndicatorsPropertyText, Default.AllowTrackingDamageIndicatorsDisplayText, 0, Group++, "Check", "", "", !bMultiplayerOnly, bAdvanced);
	PI.AddSetting (GroupString, Default.DamageBonusPropertyText, Default.DamageBonusDisplayText, 0, Group++, "Check", "", "", !bMultiplayerOnly, bAdvanced);
	PI.AddSetting (GroupString, Default.DenySuicidePropertyText, Default.DenySuicideDisplayText, 0, Group++, "Check", "", "", bMultiplayerOnly, bAdvanced);
	PI.AddSetting (GroupString, Default.TossWeaponPropertyText, Default.TossWeaponDisplayText, 0, Group++, "Check", "", "", !bMultiplayerOnly, bAdvanced);

	// * - Camping

	Group = 500;
	GroupString = Default.ShortGameName @ "-" @ Default.PlayInfoGroupCamping;

	PI.AddSetting (GroupString, Default.CampingCheckPropertyText, Default.CampingCheckDisplayText, 0, Group++, "Check", "", "", !bMultiplayerOnly, bAdvanced);
	PI.AddSetting (GroupString, Default.CampingCheckSendLocationPropertyText, Default.CampingCheckSendLocationDisplayText, 0, Group++, "Check", "", "", !bMultiplayerOnly, bAdvanced);
	PI.AddSetting (GroupString, Default.CampingCheckVolumePropertyText, Default.CampingCheckVolumeDisplayText, 0, Group++, "Text", "8;" $ Default.CampingCheckVolume_Min $ ":" $ Default.CampingCheckVolume_Max, "", !bMultiplayerOnly, bAdvanced);
	PI.AddSetting (GroupString, Default.CampingCheckKickRepeatedOffendersPropertyText, Default.CampingCheckKickRepeatedOffendersDisplayText, 0, Group++, "Check", "", "", bMultiplayerOnly, bAdvanced);

	// * - Adrenaline

	Group = 600;
	GroupString = Default.ShortGameName @ "-" @ Default.PlayInfoGroupAdrenaline;

	PI.AddSetting (GroupString, Default.TeamCombosPropertyText, Default.TeamCombosDisplayText, 0, Group++, "Check", "", "", !bMultiplayerOnly, !bAdvanced);
	PI.AddSetting (GroupString, Default.AdrenalineComboControlPropertyText[0], Default.AdrenalineComboControlDisplayText[0], 0, Group++, "Check", "", "", !bMultiplayerOnly, bAdvanced);
	PI.AddSetting (GroupString, Default.AdrenalineComboControlPropertyText[1], Default.AdrenalineComboControlDisplayText[1], 0, Group++, "Check", "", "", !bMultiplayerOnly, bAdvanced);
	PI.AddSetting (GroupString, Default.AdrenalineComboControlPropertyText[2], Default.AdrenalineComboControlDisplayText[2], 0, Group++, "Check", "", "", !bMultiplayerOnly, bAdvanced);
	PI.AddSetting (GroupString, Default.AdrenalineComboControlPropertyText[3], Default.AdrenalineComboControlDisplayText[3], 0, Group++, "Check", "", "", !bMultiplayerOnly, bAdvanced);
	PI.AddSetting (GroupString, Default.AdrenalineComboControlPropertyText[4], Default.AdrenalineComboControlDisplayText[4], 0, Group++, "Check", "", "", !bMultiplayerOnly, bAdvanced);
	PI.AddSetting (GroupString, Default.AdrenalineComboControlPropertyText[5], Default.AdrenalineComboControlDisplayText[5], 0, Group++, "Check", "", "", !bMultiplayerOnly, bAdvanced);
}

// Reminder: Gets called in different context.

static event string GetDisplayText (string PropertyText)
{
	switch (PropertyText)
	{
		case Default.AllowTrackingDamageIndicatorsPropertyText:
			return Default.AllowTrackingDamageIndicatorsDisplayText;


		case Default.TossWeaponPropertyText:
			return Default.TossWeaponDisplayText;


		case Default.PickupNotificationsPropertyText:
			return Default.PickupNotificationsDisplayText;


		case Default.CampingCheckPropertyText:
			return Default.CampingCheckDisplayText;

		case Default.CampingCheckSendLocationPropertyText:
			return Default.CampingCheckSendLocationDisplayText;

		case Default.CampingCheckVolumePropertyText:
			return Default.CampingCheckVolumeDisplayText;

		case Default.CampingCheckKickRepeatedOffendersPropertyText:
			return Default.CampingCheckKickRepeatedOffendersDisplayText;


		case Default.LogMatchStatisticsPropertyText:
			return Default.LogMatchStatisticsDisplayText;

		case Default.LogPlayerGUIDPropertyText:
			return Default.LogPlayerGUIDDisplayText;

		case Default.PublicMatchStatisticsPropertyText:
			return Default.PublicMatchStatisticsDisplayText;


		case Default.DamageBonusPropertyText:
			return Default.DamageBonusDisplayText;


		case Default.AdrenalineComboControlPropertyText[0]:
			return Default.AdrenalineComboControlDisplayText[0];

		case Default.AdrenalineComboControlPropertyText[1]:
			return Default.AdrenalineComboControlDisplayText[1];

		case Default.AdrenalineComboControlPropertyText[2]:
			return Default.AdrenalineComboControlDisplayText[2];

		case Default.AdrenalineComboControlPropertyText[3]:
			return Default.AdrenalineComboControlDisplayText[3];

		case Default.AdrenalineComboControlPropertyText[4]:
			return Default.AdrenalineComboControlDisplayText[4];

		case Default.AdrenalineComboControlPropertyText[5]:
			return Default.AdrenalineComboControlDisplayText[5];


		case Default.DenySuicidePropertyText:
			return Default.DenySuicideDisplayText;

		case Default.TeamCombosPropertyText:
			return Default.TeamCombosDisplayText;


		case Default.RedTeamNamePropertyText:
			return Default.RedTeamNameDisplayText;

		case Default.BlueTeamNamePropertyText:
			return Default.BlueTeamNameDisplayText;
	}

	return Super.GetDisplayText (PropertyText);
}

static event string GetDescriptionText (string PropertyText)
{
	switch (PropertyText)
	{
		case Default.AllowTrackingDamageIndicatorsPropertyText:
			return Default.AllowTrackingDamageIndicatorsDescriptionText;


		case Default.TossWeaponPropertyText:
			return Default.TossWeaponDescriptionText;


		case Default.PickupNotificationsPropertyText:
			return Default.PickupNotificationsDescriptionText;


		case Default.LogMatchStatisticsPropertyText:
			return Default.LogMatchStatisticsDescriptionText;

		case Default.LogPlayerGUIDPropertyText:
			return Default.LogPlayerGUIDDescriptionText;

		case Default.PublicMatchStatisticsPropertyText:
			return Default.PublicMatchStatisticsDescriptionText;


		case Default.DamageBonusPropertyText:
			return Default.DamageBonusDescriptionText;


		case Default.CampingCheckPropertyText:
			return Default.CampingCheckDescriptionText;

		case Default.CampingCheckSendLocationPropertyText:
			return Default.CampingCheckSendLocationDescriptionText;

		case Default.CampingCheckVolumePropertyText:
			return Default.CampingCheckVolumeDescriptionText;

		case Default.CampingCheckKickRepeatedOffendersPropertyText:
			return Default.CampingCheckKickRepeatedOffendersDescriptionText;


		case Default.AdrenalineComboControlPropertyText[0]:
			return Default.AdrenalineComboControlDisplayText[0];

		case Default.AdrenalineComboControlPropertyText[1]:
			return Default.AdrenalineComboControlDisplayText[1];

		case Default.AdrenalineComboControlPropertyText[2]:
			return Default.AdrenalineComboControlDisplayText[2];

		case Default.AdrenalineComboControlPropertyText[3]:
			return Default.AdrenalineComboControlDisplayText[3];

		case Default.AdrenalineComboControlPropertyText[4]:
			return Default.AdrenalineComboControlDisplayText[4];

		case Default.AdrenalineComboControlPropertyText[5]:
			return Default.AdrenalineComboControlDisplayText[5];


		case Default.DenySuicidePropertyText:
			return Default.DenySuicideDescriptionText;


		case Default.TeamCombosPropertyText:
			return Default.TeamCombosDescriptionText;


		case Default.RedTeamNamePropertyText:
			return Default.RedTeamNameDescriptionText;

		case Default.BlueTeamNamePropertyText:
			return Default.BlueTeamNameDescriptionText;
	}

	return Super.GetDescriptionText (PropertyText);
}

// Contexts:
// - GameInfo.Login () - OptionString
// - GameInfo.GetServerInfo () - GameReplicationInfo.ServerName

event string StripColor (string SourceString)
{
	return Class'ReUtility'.Static.GetPlainString (SourceString);
}


static event string MakeColorCode (Color Color)
{
	return
	(
		Chr (0x1B)
		$ Chr (Max (Color.R, 1))
		$ Chr (Max (Color.G, 1))
		$ Chr (Max (Color.B, 1))
	);
}

// #Obsolete

simulated event UpdateAnnouncements ()
{
}

// #Removed

event bool CanShowPathTo (PlayerController Other, int IndexTeam)
{
	return False;
}

event PlayWinMessage (PlayerController Player, bool bWinner) {}

// #Replaced

event EndSpree (Controller Killer, Controller Victim) {}
event NotifySpree (Controller Other, int Count_Spree) {}

// #

protected function bool ShouldFreezeRound ()
{
	return bFreezeEndRound;
}

// #Teams

function PlayerChangedTeams (PlayerController Player, TeamInfo JoinedTeam, TeamInfo AbandonedTeam)
{
	TeamGameEvents.PlayerChangedTeams (Player, JoinedTeam, AbandonedTeam);
}

// #

DefaultProperties
{
	ShortGameName = "ReTeamGame"

	bShowGameType = True

	// Game Specific

	bCheerfulBots = True
	bRoundbased = False
	bFreezeEndRound = False
	bFreezeAdrenaline = False
	bResetSpree = False

	ADR_MajorKill_Retribution = 10

	//

	DefaultValueColor = (R=16,G=255,B=16,A=255)
	CustomValueColor = (R=255,G=255,B=16,A=255)

	PlayerTeamColor(0) = (R=255,G=0,B=0,A=255)
	PlayerTeamColor(1) = (R=128,G=192,B=255,A=255)

	//

	RedTeamNameString = ""
	RedTeamNamePropertyText = "RedTeamNameString"
	RedTeamNameDisplayText = "Red Team Name"
	RedTeamNameDescriptionText = "Use a custom team name. Leave blank for default team name."

	BlueTeamNameString = ""
	BlueTeamNamePropertyText = "BlueTeamNameString"
	BlueTeamNameDisplayText = "Blue Team Name"
	BlueTeamNameDescriptionText = "Use a custom team name. Leave blank for default team name."

	//

	bMasquerade = False
	MasqueradingGameType = "xTeamGame"

	//

	TeamGameEventsClassNameString = "Re.TeamGameEvents"

	//

	PlayerRegionClassNameString = "Re.PlayerRegion"

	// Configurable

	AwardBroadcastLocalMessageClass = Class'LocalMessage_AwardBroadcast'


	DrawRoundLocalMessageClass = Class'LocalMessage_DrawRound'

	// #PlayInfo

	PlayInfoGroupAdrenaline = "Adrenaline"
	PlayInfoGroupCamping = "Camping"
	PlayInfoGroupGame = "Game"
	PlayInfoGroupPickups = "Pickups"
	PlayInfoGroupPlayers = "Players"
	PlayInfoGroupStats = "Stats"
	PlayInfoGroupTeams = "Teams"

	//

	bShowBalanceTeams = True


	bScoreRecoveryEnabled = True
	Class_ScoreRecoveryObject = Class'ScoreRecoveryObject_Re'


	bTossWeapon = True
	TossWeaponPropertyText = "bTossWeapon"
	TossWeaponDisplayText = "Toss Weapon"
	TossWeaponDescriptionText = "Players toss currently held weapon upon death for anyone to pick up."


	bAllowTrackingDamageIndicators = True
	AllowTrackingDamageIndicatorsPropertyText = "bAllowTrackingDamageIndicators"
	AllowTrackingDamageIndicatorsDisplayText = "Allow Tracking Damage Indicators"
	AllowTrackingDamageIndicatorsDescriptionText = "Allow players to use the tracking damage indicators."


	bPickupNotifications = False
	PickupNotificationsPropertyText = "bPickupNotifications"
	PickupNotificationsDisplayText = "Pickup Notifications"
	PickupNotificationsDescriptionText = "Players are notified when weapons or powerful items get picked up."


	CampingDamageTypeClass = Class'DamTypeCamping'
	CampingLocalMessageClass = Class'LocalMessage_Camper'


	CampingCheckInterval = 5 // Constant

	bCampingCheck = False
	CampingCheckPropertyText = "bCampingCheck"
	CampingCheckDisplayText = "Camping Check"
	CampingCheckDescriptionText = "Players are checked for camping behavior."

	bCampingCheckSendLocation = True
	CampingCheckSendLocationPropertyText = "bCampingCheckSendLocation"
	CampingCheckSendLocationDisplayText = "Send Camper Location"
	CampingCheckSendLocationDescriptionText = "Players are temporarily tagged with a camper beacon giving away their location."

	CampingCheckVolume = 400.0
	CampingCheckVolume_Min = 64.0
	CampingCheckVolume_Max = 2048.0

	CampingCheckVolumePropertyText = "CampingCheckVolume"
	CampingCheckVolumeDisplayText = "Camping Detection Volume"
	CampingCheckVolumeDescriptionText = "The volume a player must stay within to be considered camping."

	bCampingCheckKickRepeatedOffenders = False
	CampingCheckKickRepeatedOffendersPropertyText = "bCampingCheckKickRepeatedOffenders"
	CampingCheckKickRepeatedOffendersDisplayText = "Kick Excessive Campers"
	CampingCheckKickRepeatedOffendersDescriptionText = "Players who receive four consecutive warnings get kicked."

	//

	bLogMatchStatistics = False
	LogMatchStatisticsPropertyText = "bLogMatchStatistics"
	LogMatchStatisticsDisplayText = "Log Match Statistics"
	LogMatchStatisticsDescriptionText = "Detailed statistics are logged at the end of each round and match."

	bIncludePlayerGUID = False
	LogPlayerGUIDPropertyText = "bIncludePlayerGUID"
	LogPlayerGUIDDisplayText = "Include Player GUID"
	LogPlayerGUIDDescriptionText = "Match Statistics logged on the server include the player GUID."

	bPublicMatchStatistics = False
	PublicMatchStatisticsPropertyText = "bPublicMatchStatistics"
	PublicMatchStatisticsDisplayText = "Public Match Statistics"
	PublicMatchStatisticsDescriptionText = "Players can request a copy of the match statistics at the end of the match."

	//

	bDamageBonus = False
	DamageBonusPropertyText = "bDamageBonus"
	DamageBonusDisplayText = "Damage Bonus"
	DamageBonusDescriptionText = "Players are awarded adrenaline and points for dealing damage to enemies."


	bDenySuicide = False
	DenySuicidePropertyText = "bDenySuicide"
	DenySuicideDisplayText = "Deny Suicide"
	DenySuicideDescriptionText = "Prevent players from using the Suicide command."


	AdrenalineComboControlPropertyText(0) = "bDisableBerserkCombo"
	AdrenalineComboControlDisplayText(0) = "Disable Berserk! Adrenaline Combo"
	AdrenalineComboControlPropertyText(1) = "bDisableBoosterCombo"
	AdrenalineComboControlDisplayText(1) = "Disable Booster! Adrenaline Combo"
	AdrenalineComboControlPropertyText(2) = "bDisableInvisibleCombo"
	AdrenalineComboControlDisplayText(2) = "Disable Invisible! Adrenaline Combo"
	AdrenalineComboControlPropertyText(3) = "bDisableSpeedCombo"
	AdrenalineComboControlDisplayText(3) = "Disable Speed! Adrenaline Combo"
	AdrenalineComboControlPropertyText(4) = "bDisablePintsizedCombo"
	AdrenalineComboControlDisplayText(4) = "Disable Pint-sized! Adrenaline Combo"
	AdrenalineComboControlPropertyText(5) = "bDisableCamouflageCombo"
	AdrenalineComboControlDisplayText(5) = "Disable Camouflage! Adrenaline Combo"


	bTeamCombos = False
	TeamCombosPropertyText = "bTeamCombos"
	TeamCombosDisplayText = "Team Combos"
	TeamCombosDescriptionText = "Adrenaline combos are shared between teammates."


	bRageQuitAnnouncement = True


	SpawnBot_Class_Pawn = Class'RePawn'

	// (Limit)

	MinNetPlayers_Min = 0
	MinNetPlayers_Max = 999
	NetWait_Min = 5
	NetWait_Max = 999
	RestartWait_Min = 0
	RestartWait_Max = 999
	SpawnProtectionTime_Min = 0.0
	SpawnProtectionTime_Max = 10.0
	GoalScore_Min = 0
	GoalScore_Max = 999
	MaxLives_Min = 0
	MaxLives_Max = 999
	MaxPlayers_Min = 0
	MaxPlayers_Max = 32
	MaxSpectators_Min = 0
	MaxSpectators_Max = 32
	ResetTimeDelay_Min = 0
	ResetTimeDelay_Max = 999
	TimeLimit_Min = 0
	TimeLimit_Max = 999

	// DeathMatch

	LoginMenuClass = "Re.UT2K4Menu_PlayerLogin"

	// GameInfo

	bGarbageCollection = False

	bCanChangeSkin = True // Not sure what effect this has.  Might be native handling.

	GameMessageClass = Class'LocalMessage_Game'

	DeathMessageClass = Class'LocalMessage_Deaths'
	GameReplicationInfoClass = Class'ReGRI'
	DefaultPlayerClassName = "Re.RePawn"
	HUDType = "Re.ReTeamGameHUD"
	PlayerControllerClassName = "Re.RePlayer"
	ScoreboardType = "Re.ReTeamGameScoreboard"

	MaxLives = 0
}
