// *F* Developed 2013 rejecht <rejecht at outlook dot com>
// - Adds separate damage type for arcing lightning.

class Statistics_SniperFire extends SniperFire
	;

// #

var protected const Vector NullVector;

var protected const Class<WeaponDamageType> ArcedLightningDamageTypeClass;

// #Statistics

var(Cached) protected int Statistics_PrimaryFireModeIndex;
var(Cached) protected int Statistics_HeadShotFireModeIndex;

// #Lifetime

simulated event PostBeginPlay ()
{
	if (Level.NetMode != NM_Client)
	{
		Statistics_PostBeginPlay ();
    }


	Super.PostBeginPlay ();
}

// #Statistics

protected function Statistics_PostBeginPlay ()
{
	local AwardPRI AwardPRI;

	AwardPRI = AwardPRI(Instigator.PlayerReplicationInfo);

	Statistics_PrimaryFireModeIndex = AwardPRI.Statistics_RegisterWeaponDamageType (Class'XWeapons.DamTypeSniperShot', 0);

	Statistics_HeadShotFireModeIndex = AwardPRI.Statistics_RegisterWeaponDamageType (Class'XWeapons.DamTypeSniperHeadShot', 2);

	AwardPRI.Statistics_RegisterWeaponDamageType (ArcedLightningDamageTypeClass, 3);
}

// To add the damage type for arced lightning.

event DoTrace (Vector Start, Rotator Dir)
{
	local AwardPRI AwardPRI;
    local Vector X,Y,Z, End, HitLocation, HitNormal, RefNormal;
    local Actor Other, mainArcHitTarget;
    local int Damage, ReflectNum, arcsRemaining;
    local bool bDoReflect;
    local xEmitter hitEmitter;
    local class<Actor> tmpHitEmitClass;
    local float tmpTraceRange;
    local vector arcEnd, mainArcHit;
    local Pawn HeadShotPawn;
	local vector EffectOffset;
// 	local bool bPrimaryCharge;
	local bool bHeadShot;

	// ++New
	AwardPRI = AwardPRI(Instigator.PlayerReplicationInfo);
	// New--

	// All about offsets.

	if (Class'PlayerController'.Default.bSmallWeapons)
	{
		EffectOffset = Weapon.SmallEffectOffset;
    }
	else
	{
		EffectOffset = Weapon.EffectOffset;
    }


    Weapon.GetViewAxes (X, Y, Z);


    if
	(
		Weapon.WeaponCentered ()
		||
		SniperRifle(Weapon).Zoomed
	)
	{
        arcEnd = (Instigator.Location + EffectOffset.Z * Z);
    }
	else if (Weapon.Hand == 0)
	{
		if (Class'PlayerController'.Default.bSmallWeapons)
		{
			arcEnd = (Instigator.Location + EffectOffset.X * X);
        }
		else
		{
			arcEnd = (Instigator.Location + EffectOffset.X * X - 0.5 * EffectOffset.Z * Z);
        }
	}
	else
	{
        arcEnd = (Instigator.Location + Instigator.CalcDrawOffset(Weapon) + EffectOffset.X * X + Weapon.Hand * EffectOffset.Y * Y + EffectOffset.Z * Z);
    }

    arcsRemaining = NumArcs;

    tmpHitEmitClass = HitEmitterClass;
    tmpTraceRange = TraceRange;

    ReflectNum = 0;

    while (True)
    {
        bDoReflect = false;

        X = Vector(Dir);

        End = Start + tmpTraceRange * X;

        Other = Weapon.Trace (HitLocation, HitNormal, End, Start, True);

        if
		(
			(Other != None)
			&&
			(
				(Other != Instigator)
				||
				(ReflectNum > 0)
			)
		)
        {
            if
			(
				bReflective
				&&
				Other.IsA ('xPawn')
				&&
				xPawn(Other).CheckReflect (HitLocation, RefNormal, DamageMin * 0.25f)
			)
            {
                bDoReflect = true;
            }
            else if (Other != mainArcHitTarget)
            {
                if (!Other.bWorldGeometry)
                {
                    Damage = (DamageMin + Rand(DamageMax - DamageMin)) * DamageAtten;


                    if (Vehicle(Other) != None)
                    {
                        HeadShotPawn = Vehicle(Other).CheckForHeadShot (HitLocation, X, 1.0);
                    }


                    if (HeadShotPawn != None) // Inside a vehicle.
                    {
						bHeadShot = True;


                        HeadShotPawn.TakeDamage (Damage * HeadShotDamageMult, Instigator, HitLocation, Momentum * X, DamageTypeHeadShot);
                    }
					else if
					(
						(Pawn(Other) != None)
						&&
						(arcsRemaining == NumArcs) // This is the first shot.
						&&
						Pawn(Other).IsHeadShot (HitLocation, X, 1.0)
					)
                    {
						bHeadShot = True;


                        Other.TakeDamage (Damage * HeadShotDamageMult, Instigator, HitLocation, Momentum * X, DamageTypeHeadShot);
                    }
                    else
                    {
						if (arcsRemaining < NumArcs)
						{
							Damage *= SecDamageMult;
                        }

						// ++New
						if (arcsRemaining < NumArcs)
                        {
	                        Other.TakeDamage (Damage, Instigator, HitLocation, Momentum * X, ArcedLightningDamageTypeClass);
                        }
                        else
                        {
	                        Other.TakeDamage (Damage, Instigator, HitLocation, Momentum * X, DamageType);
                        }
						// New--
					}
                }
                else
                {
					HitLocation = HitLocation + 2.0 * HitNormal;
                }
            }
        }
        else
        {
            HitLocation = End;

            HitNormal = Normal(Start - End);
        }

		//

        if
        (
			!bHeadShot
			&&
			(arcsRemaining == NumArcs)
			&&
			(AwardPRI != None)
		)
        {
			AwardPRI.Statistics_LogWeaponFiredByIndex (Statistics_PrimaryFireModeIndex, 1);
		}

		//

        if ( Weapon == None )
        {
			return;
        }


        hitEmitter = xEmitter(Weapon.Spawn (tmpHitEmitClass,,, arcEnd, Rotator(HitNormal)));

        if (hitEmitter != None)
        {
			hitEmitter.mSpawnVecA = HitLocation;
        }

        //

		if (HitScanBlockingVolume(Other) != None)
        {
			return;
        }

        //

        if (arcsRemaining == NumArcs)
        {
            mainArcHit = HitLocation + (HitNormal * 2.0);

            if
			(
				(Other != None)
				&&
				!Other.bWorldGeometry
			)
			{
                mainArcHitTarget = Other;
            }
        }

        //

        if (bDoReflect && ++ReflectNum < 4)
        {
            Start = HitLocation;

            Dir = Rotator( X - 2.0*RefNormal*(X dot RefNormal) );
        }
        else if ( arcsRemaining > 0 )
        {
            --arcsRemaining;

            Start = mainArcHit;

            Dir = Rotator(VRand());

            tmpHitEmitClass = SecHitEmitterClass;
            tmpTraceRange = SecTraceDist;

            arcEnd = mainArcHit;
        }
        else
        {
            break;
        }
    }
}

// #

DefaultProperties
{
	ArcedLightningDamageTypeClass = Class'DamTypeArcedLightning'
}
