// Developed 2011-2013 rejecht <rejecht at outlook dot com>

class UT2K4Tab_HitSounds extends UT2K4Tab_MidGameOptions
	HideDropDown
	;

var automated GUIMultiOptionListBox lb_Box;
var GUIMultiOptionList li_Options;

var protected int OptionTabOrder;

var protected automated GUISectionBackground sb_Background;

var protected array<SOption> Options_HitSounds;

// #

protected function Option_OnCreate (GUIMenuOption moNew, name Property)
{
	switch (Property)
	{
		case 'HitSoundsbUse':
			moNew.SetComponentValue (Class'PlayerSettings'.default.bUseHitSounds);
			return;

		case 'HitSoundsVolume':
			moSlider(moNew).Setup (0.1f, 2.0f);
			moNew.SetComponentValue (FClamp (Class'PlayerSettings'.default.HitSoundVolume, 0.1f, 2.0f));
			return;


		default:
			Warn ("Unhandled: Property"@Property);
			break;
	}
}

protected function Option_OnChange (GUIMenuOption moChanged)
{
	switch (Option_OutText_PropertyName (moChanged.Tag))
	{
		case 'HitSoundsbUse':
			Class'PlayerSettings'.default.bUseHitSounds = bool(moChanged.GetComponentValue ());
			Class'PlayerSettings'.static.StaticSaveConfig ();
			return;

		case 'HitSoundsVolume':
			Class'PlayerSettings'.default.HitSoundVolume = float(moChanged.GetComponentValue ());
			Class'PlayerSettings'.static.StaticSaveConfig ();
			AuditionHitSound ();
			return;


		default:
			Warn ("Unhandled: "@Option_OutText_PropertyName (moChanged.Tag));
			break;
	}
}

protected function AuditionHitSound ()
{
	local RePlayer RePC;

	RePC = RePlayer(PlayerOwner ());

	RePC.PlaySound (RePC.GetEnemyHitSound (), SLOT_None, Class'PlayerSettings'.Default.HitSoundVolume);
}

event InitComponent (GUIController MyController, GUIComponent MyOwner)
{
	Super.InitComponent (MyController, MyOwner);

	li_Options = lb_Box.List;
	li_Options.OnCreateComponent = ListOnCreateComponent;

	Option_Reset ();

	Option_AddGroup (li_Options, Options_HitSounds);
}

event InternalOnChange (GUIComponent Sender)
{
	if (GUIMultiOptionList(Sender) != None)
	{
		Option_OnChange (GUIMultiOptionList(Sender).Get ());
	}
}

function ListOnCreateComponent (GUIMenuOption NewComp, GUIMultiOptionList Sender)
{
	NewComp.LabelJustification = TXTA_Left;
	NewComp.ComponentJustification = TXTA_Right;

	if (NewComp.IsA ('GUIListHeader'))
	{
		NewComp.StyleName = "NoBackground";
		NewComp.LabelStyleName = "TextLabel";
		NewComp.FontScale = FNS_Medium;
	}
}

event InternalOnCreateComponent (GUIComponent NewComp, GUIComponent Sender)
{
	if (GUIMultiOptionList(NewComp) != None)
	{
		GUIMultiOptionList(NewComp).bDrawSelectionBorder = False;
		GUIMultiOptionList(NewComp).ItemPadding = 0.3f;

		if (Sender == lb_Box) // Pass it on to GUIMultiOptionListBox, since intercepted.
		{
			lb_Box.InternalOnCreateComponent (NewComp, Sender);
		}
	}
}

protected function Option_Reset ()
{
	Option_IndexTag = 0;
	Option_Group.Length = 0;
}

// - To add the options to the list.
// - To call Option_OnCreate for every option added.

protected function Option_AddGroup (GUIMultiOptionList moList, array<SOption> OptionGroup)
{
	local GUIMenuOption moNew;
	local int Index;
	local int Count;

	Count = OptionGroup.Length;

	for (Index = 0; Index < Count; ++Index)
	{
		moNew
		=
		moList.AddItem
		(
			Text_ClassName_Component[OptionGroup[Index].Component],
			,
			OptionGroup[Index].Caption
		);

		if (moNew != None)
		{
			moNew.bAutoSizeCaption = True;

			if (OptionGroup[Index].Component != C_Section)
			{
				moNew.TabOrder = OptionTabOrder++;

				Option_Append (Option_IndexTag, OptionGroup[Index].Property);

				if (OptionGroup[Index].Description != "")
				{
					moNew.SetHint (OptionGroup[Index].Description);
				}

				Option_OnCreate (moNew, OptionGroup[Index].Property);

				moNew.Tag = Option_IndexTag++;
			}
		}
		else
		{
			Warn ("(moNew == None) -- moList:"@moList@"moNew:"@moNew@"Index:"@Index);
		}
	}
}

// ####

DefaultProperties
{
    Begin Object Class=AltSectionBackground Name=sbBackground
        Caption="Hit Sounds"
        LeftPadding=0.0
        RightPadding=0.0
        TopPadding=0.0
        BottomPadding=0.0
		WinLeft=0.023625
		WinTop=0.025
		WinWidth=0.944875
		WinHeight=0.275
        bBoundToParent = true
        bScaleToParent = true
    End Object

    sb_Background = sbBackground

	Begin Object Class=GUIMultiOptionListBox Name=ConfigList
		bVisibleWhenEmpty = True
		bBoundToParent = True
		bScaleToParent = True
		NumColumns = 1
		OnCreateComponent = InternalOnCreateComponent
		OnChange = InternalOnChange
		RenderWeight = 0.9
		StyleName = "NoBackground"
		TabOrder = 1
		WinLeft = 0.1
		WinTop = 0.1
		WinWidth = 0.8
		WinHeight = 0.8125
	End Object

	lb_Box = ConfigList

	// Options

	OptionTabOrder = 2

	Options_HitSounds(0) = (Caption="Play Hit Sounds",Description="Lower pitch sounds indicate heavier damage done.",Property=HitSoundsbUse,Component=C_Boolean)
	Options_HitSounds(1) = (Caption="Hit Sound Volume",Property=HitSoundsVolume,Component=C_Range)
}
